;+
; NAME:
;       autolrisdefaults
;
; PURPOSE:
;       Set up the lpar file which contains the lris configuration based
;       on default values (either from a configuration file or from the 
;       internal defaults hard-coded in the program.)
;
; CALLING SEQUENCE:
;       autolrisdefaults, [/reset]
;
; EXPLANATION:
;       Creates and stores the lpipe common blocks need to keep and store
;       data about the pipeline configuration and status needed for 
;       numerous programs/subroutines.  Uses the 'default' settings from
;       lpipe.par (which can vary between installations and directores).
;       If the user overrides this with command-line options those settings
;       are specified within the lpipe routine itself.
;
; OPTIONAL INPUTS:
;       /reset - Set this to clear all user-specified settings back to
;                defaults.  (By default most pre-existing inputs are not
;                overwritten.)
;
;-

pro autolrisdefaults, reset=reset

   ; Load the configuration file that sets the external commands, directory locations, etc.
   ; If no local file is found it will look in the code directory.
   ; If no configuration file is found at all, internal defaults are available.
   ; (Neither of these two features are strictly necessary with the new setup implementation.
   ; Also creates the imredux/ and spredux/ subdirectories if necessary.

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail
   common landoltpathinfo, landoltpath

   paths = strsplit(!path,':',/extract)
   for p = -1, n_elements(paths)-1 do begin   ; find the directory this file is in
      if p ge 0 then path = paths[p] $
                     else path = ''
      sfile = path+'/lpipe.pro'
      n = countfiles(sfile)
      if n gt 0 then lrisautopath = path
      if n gt 0 then break
   endfor
   if n eq 0 then begin
      print, 'Problem finding lpipe directory.  This error should not occur.'
   endif

   if n_elements(lpar) eq 0 or keyword_set(reset) then begin
      lpar = {lrisautopath:lrisautopath, $
              refdatapath:lrisautopath+'/refdata', $
              defaultspath:lrisautopath+'/defaults', $
              autoastrocommand:'autoastrometry', $
              swarpcommand:'swarp', $
              datadir:'', $
              flatdir:'', $
              arcdir:'', $
              responsedir:'', $
              diagnosticdir:'', $
              imworkingdir:'', $
              spworkingdir2d:'', spworkingdir1d:'', $
              imfinaldir:'', spfinaldir:'', $
              lrisversion:0, $
              koa:0b, $
              wildchar:'??????_*[0-9]', $
              overwrite:0b, $
              fileseq:'', filterseq:'', gratingseq:'', targetseq:'', dichroicseq:'', $
              defaulttraceyred:-1, defaulttraceyblue:-1, $
              displayoff:0b, $
              debug:0, $
              reducer:'', $
              lpversion:'' $
             }
   endif
   lpar.fileseq = ''
   lpar.filterseq = ''
   lpar.gratingseq = '' ; these return to defaults and must be specified every time
   lpar.targetseq = ''
   lpar.dichroicseq = ''

   for p = -1, 1 do begin
      if p eq -1 then path = '.'
      if p eq  0 then path = lpar.lrisautopath
      if p eq  1 then path = lpar.lrisautopath + '/defaults'
      sfile = path+'/lpipe.par'       ; get the configuration file
      n = countfiles(sfile)
      if n gt 0 then begin
         openr, 5, sfile
         iline = ''
         while not eof(5) do begin
            readf, 5, iline
            hashpos = strpos(iline,'#')
            if hashpos ge 0 then iline = strmid(iline,0,hashpos)
            colonpos = strpos(iline,':')
            if colonpos lt 1 or colonpos ge strlen(iline) then continue
            varname = clip(strmid(iline,0,colonpos))
            varset = clip(strmid(iline,colonpos+1))
            if varname eq 'refdatadir' then lpar.refdatapath = varset
            if varname eq 'defaultsdir' then lpar.defaultspath = varset
            if varname eq 'autoastrocommand' then lpar.autoastrocommand = varset
            if varname eq 'swarpcommand' then lpar.swarpcommand = varset
            if varname eq 'sexcommand' then sexcommand = varset ; not in structure, uses sexsetpath
            if varname eq 'datadir' and lpar.datadir eq '' then lpar.datadir = varset ; datadir is 'sticky'
            if varname eq 'flatdir' then lpar.flatdir = varset
            if varname eq 'arcdir' then lpar.arcdir = varset
            if varname eq 'responsedir' then lpar.responsedir = varset
            if varname eq 'diagnosticdir' then lpar.diagnosticdir = varset
            if varname eq 'imworkingdir' then lpar.imworkingdir = varset
            if varname eq 'spworkingdir' then lpar.spworkingdir2d = varset
            if varname eq 'spworkingdir' then lpar.spworkingdir1d = varset
            if varname eq 'spworkingdir2d' then lpar.spworkingdir2d = varset
            if varname eq 'spworkingdir1d' then lpar.spworkingdir1d = varset
            if varname eq 'imoutputdir'  then lpar.imfinaldir = varset
            if varname eq 'spoutputdir'  then lpar.spfinaldir = varset
            if varname eq 'bluetracepos' then lpar.defaulttraceyblue = float(varset)
            if varname eq 'redtracepos' then lpar.defaulttraceyred = float(varset)
         endwhile
         close, 5
         break ; found the file: stop looping
      endif
   endfor

   ; this is not in the structure so special handling for now
   if n_elements(sexcommand) eq 0 then sexcommand = getsexpath() ; this will search some common locations
   if sexcommand eq '' then sexcommand = getsexpath()

   setswarppath, lpar.swarpcommand   ;extractpath(lpar.swarpcommand)
   setsexpath, sexcommand       ;extractpath(sexcommand)
   landoltpath = lpar.refdatapath+'/'

   ; these must always start blank
   prepfail = ['']
   flatfail = ['']
   catastrofail = ['']
   relastrofail = ['']
   fullastrofail = ['']
   extractfail = ['']
   wavsolfail = ['']
   wavsolwarn = ['']
   fluxcalfail = ['']

  lpar.displayoff = 0
end

