;+
; NAME:
;       blueshuttersubtract
;
; PURPOSE:
;       Automatically use all on-sky shutter correction frames to correct 
;       frames affected by a broken blue shutter (for runs in which the 
;       trapdoor was not used as a substitute shutter.)
;
; CALLING SEQUENCE:
;       blueshuttersubtract
;
; EXPLANATION:
;       The LRIS-B shutter frequently malfunctions and gets stuck open
;       until a manual intervention can fix it.  After this happens observers
;       can use the trapdoor as a substitute shutter (which forces red and blue
;       to be exactly synchronized) or they can take an extra 0-second blue
;       frame before and/or after each exposure to correct for signal
;       accumulated during readout.  If the observer does the latter, this 
;       program automatically matches all such correction pairs to science
;       images.  Should be run after the prepare step.
;
; INPUTS:
;       None.  The program expects files to be in sequential format and in
;       either the current working directory or a standard reduction 
;       subdirectory.
;
; OUTPUTS:
;       The files are edited to incorporate the correction.  
;       No new files are written.
; 
; NOTES:
;       Almost all observers use the trapdoor mode so this is rarely needed.
;       Its implementation is somewhat crude and pairing is not carefully 
;       checked.
;
;-
;------------------------------------------------------------------------------

pro blueshuttersubtract

  ; Look for paired exposures and subtract the zero-second on-sky "bias".

  files = [findfile('pb??????_????.fits'), findfile('spredux2d/pb??????_????.fits'), findfile('imredux/pb??????_????.fits')]
  w = where(files ne '', ct)
  if ct eq 0 then begin
    print, 'Found no processed files.'
    return
  endif
  files = files[w]
  rfiles = removepath(files)
  files = files[sort(rfiles)]

  filenums = fltarr(n_elements(files))
  exptimes = fltarr(n_elements(files))
  filters  = fltarr(n_elements(files))
  grisms   = fltarr(n_elements(files)) ; should check targname, as long as it's not blank
  trapdoor = strarr(n_elements(files))
  skybias  = strarr(n_elements(files))      ; should also check binning and windowing and everything else
  for f = 0, n_elements(files)-1 do begin
     h = headfits(files[f])
     filenums[f] = sxpar(h,'FRAMENO')
     exptimes[f] = sxpar(h,'EXPTIME')
     trapdoor[f] = clip(sxpar(h,'TRAPDOOR'))
     skybias[f]  = clip(sxpar(h,'SKYBIAS'))
  endfor

  for f = 0, n_elements(files)-2 do begin
     ;;if trapdoor[f] ne 'open' then continue
     if exptimes[f] lt 1 then continue
     if skybias[f] ne '' and skybias[f] ne '0' then begin
        print, removepath(files[f]), ' already subtracted (', skybias[f], ')'
        continue
     endif

     for b = f+1, (f+3 < n_elements(files)-1) do begin   ; formerly up to (f+10 < n_elements(files)-1) 
        if filenums[b] ge filenums[f]+3 then continue ; new
        if trapdoor[b] ne trapdoor[f] then continue ; formerly required both be open
        if exptimes[b] gt 0.5 then continue
        if filters[b] ne filters[f] then break

        print, 'Subtracting: ', removepath(files[f]), '-', removepath(files[b])
        im = readfits(files[f], h, /silent)
        bias = readfits(files[b], hb, /silent)
        im -= bias
        sxaddpar, h, 'SKYBIAS', removepath(files[b]), ' Blue shutter broken; subtracted frame'
        writefits, files[f], im, h

        break
     endfor
  endfor

end

