;+
; NAME: 
;   adjustzpt
;
; PURPOSE: 
;   Change the zeropoint of a photometric catalog, explicitly or by calibrating to
;   a comparison catalog (using getzptoffset).
;
; CALLING SEQUENCE: 
;   adjustzpt, catalog, reference, outcat=outcat, /rename, refunc=refunc, /nir, 
;              filter=filter, sys=sys, adjusted=adjusted
;       
; INPUTS:
;   catalog - Input catalog file.  (File will be overwritten if outcat not specified.)
;   reference - New zeropoint number, OR a reference catalog file with calibrated 
;               photometry of stars in the same field.
;   filter - Filter employed.
;      
; OPTIONAL INPUT KEYWORD PARAMETERS:
;   outcat - New catalog file to write.  The catalog file is overwritten if unset.
;   /rename - Option to write out new file to outcat and delete catalog file.
;   refunc  - Specify the systematic uncertainty (in mag) of the zeropoint of the 
;             reference catalog to propogate the error into the ZPTUNC keyword of output.
;   /nir - Specifies this is a near-IR catalog (forwarded to getzptoffset).
;   sys - Forwarded to getzptoffset and on to printcatalog...
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;   adjusted - Set to 1 if the adjustment succeeded, zero otherwise
;       
; OUTPUTS: 
;   The catalog is written to disk (or modified on disk).
;
; Last modified 4 Sep 2020.  Documentation added 3 July 2022.
;-

pro adjustzpt, catalog, reference, outcat=outcat, rename=rename, refunc=refunc, nir=nir, filter=filter, sys=sys, adjusted=adjusted
 
  ; Re-set the zeropoint of catalog based on input number (if int/float) or by
  ;   comparison to magnitudes in reference (if specified, otherwise SDSS/Landolt)
  ; Overwrites catalog by default, else prints to outcat (rename deletes input file on success)

  if n_elements(reference) eq 0 then zpt = getzptoffset(catalog, rms=rms, unc=unc, /zpt, getref=catname, nir=nir, filter=filter, sys=sys)  ; this works correctly.
  if n_elements(reference) eq 1 then begin
        if size(reference,/tn) eq 'STRING' then begin
            zpt = getzptoffset(catalog, reference, rms=rms, unc=unc, /zpt, getref=catname, nir=nir, filter=filter, sys=sys)       ; this does NOT.
        endif else begin
            zpt = reference
            catname = 'zeropoint'
        endelse
  endif
 
  adjusted = 0
  if finite(zpt) eq 0 then return  ; unsuccessful [no catalog to compare to, usually]
  if n_elements(refunc) eq 0 then refunc = 0

  cat = readcat(catalog, h)
  oldzeropt = sxpar(h,'ZEROPT', comment=zptcmt)
  sxaddpar, h, 'OLDZPT', oldzeropt, 'Zeropoint used before calibration', before='ZEROPT'
  dzpt = zpt - oldzeropt
  if n_elements(zptcmt) gt 0 then sxaddpar, h, 'ZEROPT', oldzeropt + dzpt, zptcmt $ 
                             else sxaddpar, h, 'ZEROPT', oldzeropt + dzpt  ; this shouldn't be necessary but sxaddpar is buggy
  if n_elements(rms) gt 0 then begin
    sxaddpar, h, 'ZPTRMS', rms, 'RMS of match to catalog stars'
    sxaddpar, h, 'ZPTUNC', sqrt(unc^2 + sxpar(h,'APUNC')^2 + refunc^2), 'Uncertainty of zeropoint'
    sxaddpar, h, 'NSTARZPT', n_elements(refunc), 'Number of stars used in zeropoint'
  endif
  cat.mag += dzpt
  if n_elements(catname) gt 0 then sxaddpar, h, 'PHOTCAL', removepath(catname), 'Catalog used to determine zeropoint'
  if n_elements(filter) gt 0 then calfilt = filter[0] else calfilt = sxpar(h,'FILTER')
  sxaddpar, h, 'CALFILT', calfilt, 'Catalog filter used to compute zeropt'
  if n_elements(outcat) eq 0 then outcatfile = catalog else outcatfile = outcat

  writecat, outcatfile, cat, h
  if keyword_set(rename) and file_test(outcatfile) and (catalog ne outcatfile) then file_delete, catalog
  adjusted = 1

end

