;+
; NAME: 
;   asinh2pogson
;
; PURPOSE: 
;   Convert SDSS asinh magnitudes to standard Pogson (logarithmic) magnitudes.   
;
; CALLING SEQUENCE: 
;   pmag = asinh2pogson(sdssmag, filt)
;       
; INPUTS:
;   sdssmag - Magnitude(s) in the SDSS asinh magnitude system.
;   filt    - The corresponding SDSS filter (u,g,r,i,z)    
;       
; OUTPUTS: 
;   pmag    - The magnitude converted to Pogson.
;
; COMMENTS:
;   The conversion formula is taken from the SDSS webpage.
;
; Last modified 2 Nov 2016.  Documentation added 3 July 2022.
;-

function asinh2pogson, mag, filt
  ; convert weird SDSS magnitude to a normal one

  ;   m = -2.5/ln(10) * [asinh((f/f0)/(2b)) + ln(b)]
  ;   m*(-ln(10)/2.5) = asinh((f/f0)/(2b)) + ln(b)
  ;   m*(-ln(10)/2.5) - ln(b) = asinh((f/f0)/(2b))
  ;   sinh( m*(-ln(10)/2.5) - ln(b) ) = (f/f0)/(2b)
  ;   2b * sinh( m*(-ln(10)/2.5) - ln(b) ) = f/f0
  ;   f = f0 * 2b * sinh( m*(-ln(10)/2.5) - ln(b) )

  ; Filter 	b 	Zero-flux Magnitude [m(f/f0 = 0)] 	m(f/f0 = 10b)
  ; u 	1.4×10-10 	24.63 	22.12
  ; g 	0.9×10-10 	25.11 	22.60
  ; r 	1.2×10-10 	24.80 	22.29
  ; i 	1.8×10-10 	24.36 	21.85
  ; z 	7.4×10-10 	22.83 	20.32

  rfilt = ['u','g','r','i','z']
  rb = [1.4,0.9,1.2,1.8,7.4]*1e-10
  rf0 = [3.46737,3.63078,3.63078,3.63078,3.71535]*1e+9
  if n_elements(filt) eq 1 then begin
    w = (where(filt eq rfilt)) [0]
  endif else begin
    w = fltarr(n_elements(filt))
    for i = 0, n_elements(filt)-1 do begin
       w[i] = (where(filt[i] eq rfilt)) [0]
    endfor
  endelse

  f = rf0[w] * 2*rb[w] * sinh( mag*(-alog(10)/2.5) - alog(rb[w]) )
  regularmag = ujy2mag(f,filt)
  
  return, regularmag
end

