;+
; NAME: atm_attenuate.pro
;
; PURPOSE: 
;       Calculate the expected atmospheric attenuation as a function of wavelength using
;       a simple polynomial model fit to a measured curve from Maunakea.
;
; CALLING SEQUENCE: 
;       atm_attenuate, wav, air, elevation=elevation, mag=mag, flux=flux
;
; INPUTS:
;       wav  = Wavelength in Angstroms.  Can be an array.
;       air  = Geometric airmass, i.e., sec(zenith angle)
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       elevation = Elevation of the observatory in meters.  (NOT the elevation angle.)
;                   Defaults to 4200 meters, which is the elevation of Maunakea.
;       mag  = Set to get the attenuation in magnitudes instead of as a flux factor.
;       flux = Set to get the attenuation as a flux factor.  (This is the default.)
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS: none.
; 
; OUTPUTS: An array containing the attenuation at each wavelength.
;
; REQUIREMENTS: none
;
; Last modified 27 Jun 2018.  Documentation added 2022.
;-


function atm_attenuate, wav, air, elevation=elevation, mag=mag, flux=flux

     if n_elements(elevation) eq 0 then  elev = 4200. else elev = elevation ; default Maunakea

      airp = 2^((4200-elev)/4570.)*[0.12947983, -0.13458371, 0.21833111, -0.48222307, $
                                    0.41800934, -0.11447838, 0.0032198187, -0.010178335, 0.0047375054]
      airtermarr = poly((wav-5000.)/2500.,airp)  ; magnitudes of extinction per airmass
      
      magatten = (air)*airtermarr ; air can be a differential

      fluxatten = 2.51^(-magatten)

    if keyword_set(mag) then return, magatten
    if keyword_set(flux) then return, fluxatten
    return, fluxatten

end

