;+
; NAME: combinesum.pro
;
; PURPOSE: 
;      Coadds a series of images (or more typically, 2D spectra) together.
;
; CALLING SEQUENCE: 
;       combinesum, filelist, outfile, masklist=masklist, /nomask,  
;                   expkey=expkey, yshift=yshift, /crclean, /noweight, 
;                   /silent, /verbose, idkey=idkey, addext=addext
;       
; INPUTS:
;      filelist - An array containing the images to be coadded.
;
; OPTIONAL INPUTS:
;      outfile  - The name of the output file to be written to disk.   If unspecified,
;                 a name ending in _sum.fits is created based on the input files.   
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;      masklist - Corresponding list of bad-pixel mask images.
;      /nomask - If the filelist has a BPM keyword pointing to a bad pixel mask, setting
;                this option causes the bad pixel masks to be ignored.
;      expkey -  If the exposure time is not stored as EXPTIME, the exposure keyword.
;      yshift -  Amount in pixels to offset each image vertically before combining.
;                Non-integer values are rounded to integers before applying.
;      /crclean - Attempt to detect and excise cosmic rays.
;      /noweight - Do not write a weight file that contains the number of overlapping
;                images for each pixel.
;      /silent - Suppress all output to the console.
;      /verbose - Write more output to the console.
;      idkey - Keyword to include in the header of the combined file.  Read from each
;               input image and output with the new keyword entries being the first five
;               letters of idkey followed by the number in the image sequence.
;               (Otherwise, the header is largely identical to the first image in the
;                sequence.)
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS: none
;       
; OUTPUTS: 
;     The new file is written to disk, along with the weight file (unless /noweight set).
;
; REQUIREMENTS:  IDLAstro and some other ccd files.
;
; Last modified 25 Feb 2019.  Documentation added 2 July 2022.
;-

pro combinesum, filelist, outfile, masklist=masklist, nomask=nomask, verbose=verbose, expkey=expkey, yshift=yshift, crclean=crclean, noweight=noweight, silent=silent, idkey=idkey, addext=addext

; Additive combination, mainly for sky-subtracted 2D spectra.  Does some outlier rejection.

; nomask causes the BPM keyword to be ignored.

if n_elements(outfile) eq 0 then begin
  filearr = strsplit(filelist[0],'.', /extract)
  fileroot = filearr[0]
  outfile = fileroot + 'sum.fits'
endif

if n_elements(filelist) eq 1 and strpos(filelist[0],'.fit') eq -1 then begin
  files = grabcolumn(filelist,0,/str)
  files = files[where(files ne '')]
endif else begin
  files = filelist
endelse

nfiles = n_elements(files)
nmid = nfiles/2

if nfiles eq 1 then if keyword_set(silent) eq 0 then print, ' Only one image; trivial sum.'

if n_elements(expkey) eq 0 then expkey = 'EXPTIME'

refdata = $;readfits(files[nmid], h, /silent)
          mrdfits(files[nmid], 0, h, /silent)
s = size(refdata)
nx = s[1]
ny = s[2]

skydata = fltarr(nx, ny, nfiles) ; with sky, not usually used
maskdata = bytarr(nx, ny, nfiles)
remaskdata = bytarr(nx, ny, nfiles)
exptime = fltarr(nfiles)
elaptime = fltarr(nfiles)
sunelev = fltarr(nfiles)
moonelev = fltarr(nfiles)
jdmid = dblarr(nfiles)
mjdmid = dblarr(nfiles)
airmass = fltarr(nfiles)
inmeds = fltarr(nfiles)
ut = strarr(nfiles)
utend = strarr(nfiles)
id = strarr(nfiles)

fits_info, files[0], /silent, n_ext=nextensions

noutextensions = nextensions + n_elements(addext)

for e = 0, noutextensions do begin

  for f = 0, nfiles-1 do begin
    if keyword_set(verbose) then print, files[f]
    if e le nextensions then begin
      indata = $ ;readfits(files[f], h, /silent)
              mrdfits(files[f], e, h, /silent)
    endif else begin
      indata = readfits(addext[e-nextensions-1], h, /silent)
    endelse
    if keyword_set(nomask) eq 0 then begin
      if n_elements(masklist) ge nfiles then inmask = readfits(masklist[f], hmask, /silent)
      maskname = clip(sxpar(h,'BPM'))
      if e eq 0 and maskname ne '0' and file_test(maskname) then inmask = mrdfits(maskname,0, hmask, /silent)
    endif
    if e eq 0 and nextensions ge 2 and keyword_set(crclean) then insky = mrdfits(files[f],1, hmask, /silent) ; not usually used

    if n_elements(yshift) gt 0 then begin
      ysh = round(yshift[f])
      if ysh ge ny then begin
        print, '  ERROR: yshift > ny; no overlap between frames.'
        return
      endif
      if ysh ne 0 then begin
        ;print, files[f], ysh
        indata = shift(indata,0,ysh)
        if ysh gt 0 then indata[*,0:ysh-1]     = !values.f_nan   ; these appear to "stick" into the output if other frame pixel is bad...
        if ysh lt 0 then indata[*,ny+ysh:ny-1] = !values.f_nan
      endif
    endif else begin
      if keyword_set(silent) eq 0 then print, 'No spatial shift applied.'
      yshift = 0+fltarr(nfiles)
    endelse
    if f eq 0 then begin
       h0 = h
       bitpix = sxpar(h0,'BITPIX')
       case bitpix of
         8: data = bytarr(nx,ny,nfiles)
         16: data = intarr(nx,ny,nfiles)
         32: data = lonarr(nx,ny,nfiles)
         -32: data = fltarr(nx,ny,nfiles)
         -64: data = dblarr(nx,ny,nfiles)
         else: print, 'unrecognized bitpix'
       endcase
    endif
    sin = size(indata)
    if total(sin[1:2] ne s[1:2]) gt 0 then begin
       ; cropping does NOT recognize physical coordinates here.
       print, '  WARNING - Trying to combine images of intrinsically different sizes'

       if sin[1] gt s[1] then indata = indata[0:s[1]-1,*]
       if sin[2] gt s[2] then indata = indata[*,0:s[2]-1]
       if sin[1] lt s[1] or sin[2] lt s[2] then begin
          newdata = refdata*!values.f_nan
          newdata[0:sin[1]-1,0:sin[2]-1] = indata
          indata = newdata
       endif
    endif
    data[*,*,f] = float(indata)
    if e eq 0 and n_elements(inmask) gt 0 then maskdata[*,*,f] = inmask
    if e eq 0 and n_elements(insky) gt 0 then skydata[*,*,f] = insky + indata
    airmass[f] = sxpar(h,'AIRMASS')
    exptime[f] = sxpar(h,expkey)
    ut[f] = sxpar(h,'UT')
    utend[f] = clip(sxpar(h,'UTC-END'))
    elaptime[f] = sxpar(h,'ELAPTIME')
    sunelev[f] = sxpar(h,'SUNELEV')
    moonelev[f] = sxpar(h,'MOONELEV')
    jdmid[f] = sxpar(h,'JDMID')
    mjdmid[f] = sxpar(h,'MJDMID')
    if keyword_set(idkey) then id[f] = sxpar(h,idkey)
  endfor

  relexptime = exptime/median(exptime)
  bitpix = sxpar(h0,'BITPIX')
  case bitpix of
    8: comb = bytarr(nx,ny)
    16: comb = intarr(nx,ny)
    32: comb = lonarr(nx,ny)
    -32: comb = fltarr(nx,ny)
    -64: comb = dblarr(nx,ny)
    else: print, 'unrecognized bitpix'
  endcase

  if e eq 0 and keyword_set(silent) eq 0 then print, '  Sum-combining...'
  i = 0l
  for y = 0, ny-1 do begin    
    for x = 0, nx-1 do begin  ; x is the fast array and should be iterated inside
      i = i + 1 ; needed only for the verbose option

      ; if /crclean set, does a little bit of CR filtering on its own
      if keyword_set(crclean) and e eq 0 and nextensions gt 1 and e eq 0 then begin     
         ww = where(maskdata[x,y,*] eq 0 and finite(data[x,y,*]), ctww) ; nearly everything!
         if ctww gt 0 then begin
            low = (-median(-data[x,y,ww]/relexptime[ww])) > 40  ; equivalent to min(data) for n=2
            dev = sqrt(max(skydata[x,y,ww]/min(relexptime[ww])) > 40)
            cr  = where(data[x,y,*] gt (low+8*dev)*relexptime+20 and maskdata[x,y,*] eq 0, ct)
            if ct gt 0 then begin
                 if keyword_set(verbose) then begin
                 print, '-----'
                 print, 'low = ', low
                 print, 'dev = ', dev
                 print, (data[x,y,*]) [*]
                 print, (data[x,y,*]/relexptime) [*]
                 print, (skydata[x,y,*]) [*]
                 print, ' Flagged ', clip(x)+','+clip(y)+','+clip(cr[0]), ' = ', skydata[x,y,cr[0]]
               endif
               remaskdata[x, y, cr] = 1
               if y lt ny-1 then remaskdata[x, y+1, cr] = 1  ; also flag neighbors
               if y gt 0    then remaskdata[x, y-1, cr] = 1
               if x lt nx-1 then remaskdata[x+1, y, cr] = 1
               if x gt 0    then remaskdata[x-1, y, cr] = 1
            endif
         endif
      endif

      comb[x, y] = total(data[x, y, *]) ; mediansigclip?
      if keyword_set(verbose) then begin
        if i mod 157557 eq 0 then begin
          st = stddev(data[x, y, *])
          val = comb[x, y]
          print, clip(x,5), clip(y,5), 'stdev=',clip(st,6), '  val=',clip(val,6);, '  stdev/sqrt(val)=',fpr(st/sqrt(val),3.2) 
        endif
      endif
    endfor
  endfor
  
  maskdata = maskdata > remaskdata
  
  if (size(maskdata)) [0] eq 3 then $  ; number of dimensions; normally 3 but 2 for trivial stack of 1 image
    addmask = total(maskdata, 3) $
  else $
    addmask = maskdata
  
  w = where(addmask gt 0 or finite(comb) eq 0)
  for i = 0L, n_elements(w)-1 do begin
    if w[i] eq -1 then continue
    val = 0.
    itime = 0.
    x = w[i] mod nx
    y = w[i] / nx
    for f = 0, nfiles-1 do begin
       if finite(data[x, y, f]) and maskdata[x,y,f] eq 0 then begin
          val += data[x, y, f]
          itime += exptime[f]
       endif
    endfor
    if itime gt 0 and itime lt total(exptime) then begin
       val *= total(exptime)/itime  ; scale
       comb[x,y] = val
    endif
  endfor
  
  hout = h0 ; Most keywords are copied from the FIRST file.
  if e eq 0 then begin
    for f = 0, nfiles-1 do begin
      sxaddpar, hout, 'SFILE'+strtrim(string(f+1),2), files[f], f eq 0 ? 'summed 2D frame file name' : ''
    endfor
    for f = 0, nfiles-1 do begin
      sxaddpar, hout, 'YSHIFT'+strtrim(string(f+1),2), yshift[f]
    endfor
    sxaddpar, hout, 'BZERO', 0
    sxaddpar, hout, 'BSCALE', 1
    sxaddpar, hout, 'NSUM', nfiles
    sxaddpar, hout, 'EXPTIME',  total(exptime)
    sxaddpar, hout, 'TOTALEXP', total(exptime)
    if total(elaptime) gt 0. then sxaddpar, hout, 'TOTALELP', total(elaptime)
    sxaddpar, hout, 'AIRMASS', mean(airmass)
    if total(sunelev) ne 0. then sxaddpar, hout, 'SUNELEV', max(sunelev) ; max ?
    if total(moonelev) ne 0. then sxaddpar, hout, 'MOONELEV', max(moonelev) ; max ?
    if total(jdmid) gt 0. then sxaddpar, hout, 'JDMID', mean(jdmid)
    if total(mjdmid) gt 0. then sxaddpar, hout, 'MJDMID', mean(mjdmid) ; would be better to weight by EXPTIME, but
                                                                       ; this introduces large numerical inaccuracy
    endkeys = ['DATE_END','UTC-END','DATE-END','UTEND','JDEND','MJDEND','AIRMEND'] ; These come from the LAST file.
    for ek = 0, n_elements(endkeys)-1 do begin
       comment = ''
       val = sxpar(h,endkeys[ek], count=cte, comment=comment, /silent) ; look in the last file
       if cte gt 0 then sxaddpar, hout, endkeys[ek], val, comment
    endfor
    mjd = sxpar(h0,'MJD',count=ctm,/silent) ; could loop over all possible time keys
    if ctm gt 0 then begin
       ;sxaddpar, hout, 'ELAPTIME', (sxpar(h,'MJD')-sxpar(h0,'MJD'))*24*3600. + sxpar(h,'EXPTIME') ; might redefine ELAPTIME
       sxaddpar, hout, 'DIFFTIME', (sxpar(h,'MJD')-sxpar(h0,'MJD'))*24*3600. + sxpar(h,'EXPTIME')
    endif

    for f = 0, nfiles-1 do sxaddpar, hout, 'EXPT'+strtrim(string(f+1),2), exptime[f]
    for f = 0, nfiles-1 do sxaddpar, hout, 'AIRMASS'+strtrim(string(f+1),2), airmass[f]
    for f = 0, nfiles-1 do sxaddpar, hout, 'UT'+strtrim(string(f+1),2), ut[f]
    for f = 0, nfiles-1 do if utend[f] ne '0' then sxaddpar, hout, 'UTEND'+strtrim(string(f+1),2), utend[f]
    if n_elements(idkey) gt 0 then $
    for f = 0, nfiles-1 do sxaddpar, hout, clip(clip(idkey,5))+strtrim(string(f+1),2), id[f]

  endif
  
  hist = 'Processed by combinesum '+tstamp(/ut)+' UT'
  sxaddhist, hist, hout

  hout = hout[where(strpos(hout,'END') ne 0)]
  hwh = where(strpos(hout,'HISTORY') eq 0, complement=hwnh) ; Sort header so history is at end
  hout = hout[[hwnh, hwh]]
  hout = [hout, 'END                                                                             ']  ; note the whitespace

  if e eq 0 then mwrfits, comb, outfile, hout, /create, /silent $
            else mwrfits, comb, outfile, hout, /silent
  ;          else writefits, outfile, comb

endfor

if keyword_set(noweight) eq 0 then begin
  outweightfile = repstr(outfile,'.fits','.weight.fits')
  if outweightfile eq outfile then outweightfile += '.weight'
  writefits, outweightfile, addmask, hmask
endif

if keyword_set(silent) eq 0 then print, '  Written to ', outfile

end

