;+
; NAME: 
;   connectspec
;
; PURPOSE: 
;   Connect two spectra with different but (ideally) slightly-overlapping wavelength 
;   ranges (e.g., LRIS R and B).  Performs no rescaling (scale before using this routine)
;
; CALLING SEQUENCE: 
;   s = connectspec(spec1, spec2, attachwav=attachwav)
;       
; INPUTS:
;   spec1 - First spectrum (spec structure - not a file)
;   spec2 - Second spectrum
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;   attachwav - Wavelength to attach spectra at.  (Uses midpoint of overlap region if unset.)
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS: none
;       
; OUTPUTS:  The connected spectrum.
;
; Last modified 18 Sep 2019.  Documentation added 3 July 2022.
;-

function connectspec, spec1in, spec2in, attachwav=attachwav

   ; connect two spectra with different wavelength ranges at the attachment point.
    
   spec1 = spec1in
   spec2 = spec2in
   if median(spec1.wav) gt median(spec2.wav) then begin
      spec2hold = spec2
      spec2 = spec1
      spec1 = spec2hold
   endif

   if n_elements(attachwav) eq 0 then attachwav = median([max(spec1.wav),min(spec2.wav)],/even)
   w1below = where(spec1.wav le attachwav, ct1)
   w2above = where(spec2.wav gt attachwav, ct2)
   n = ct1 + ct2

   spec = replicate({wav:0.0d, flux:0.0, sky:0.0, unc:0.0, pix:0L, ypix:0.0, resp:0.0, flag:0.0, dwav:0.0},n)

   wav = [spec1[w1below].wav , spec2[w2above].wav]
   wsbelow = where(wav le attachwav)
   wsabove = where(wav gt attachwav)

   spec[wsbelow] = spec1[w1below]
   spec[wsabove] = spec2[w2above]

   return, spec

end

