;+
; NAME: editheader.pro
;
; PURPOSE: 
;       Tool to change one or more FITS keyword values in a header.
;       Now also works on .spec files (FITS-like header in a text file)
;
; CALLING SEQUENCE: 
;       editheader, files, keys, newval, remove=remove
;
; INPUTS:
;       files  = name of file to edit.  Can also be an array of files.
;       keys   = keyword(s) to modify.
;       newval = the new value(s) of the keywords designated by "keys".
;       remove = set this to remove the keyword(s) instead of changing them.
;
; OPTIONAL INPUT KEYWORD PARAMETERS: none.
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS: none.
; 
; OUTPUTS: none; edits file on disk.
;
; REQUIREMENTS:
;       - Requires the IDL Astronomy User's Library routines 
;       (http://idlastro.gsfc.nasa.gov/) 
;       - Processing spec files requires readspec/writespec.
;-

pro editheader, file, keys, newval, silent=silent

  if n_elements(file) eq 0 then print, ' No file specified.'
  if n_elements(keys) eq 0 then print, ' No keyword specified.'
  if keyword_set(remove) eq 0 and $
       n_elements(newval) eq 0 then print, ' No value specified.'
  if n_elements(file) eq 0 or n_elements(keys) eq 0  or $
     (keyword_set(remove) eq 0 and n_elements(newval) eq 0) then return

  files = ['']
  for f = 0, n_elements(file)-1 do begin
     if  strpos(file[f],'*') ge 0 or strpos(file[f],'?') ge 0 or $
        (strpos(file[f],'[') ge 0 and strpos(file[f],']') ge 0) then begin
          expfiles = file_search(file[f], count=ct)
          if ct eq 0 then print, ' ', file[f], ' not found'
          if ct gt 0 then files = [files, expfiles]
     endif else begin
        if file_test(file[f]) eq 0 then print, ' ', file[f], ' not found' $
        else files = [files, file[f]]
     endelse
  endfor
  files = files[where(files ne '', ctf)]
  if ctf eq 0 then return

  for f = 0, n_elements(files)-1 do begin
     if strpos(files[f],'.spec') eq -1 then begin
        h = headfits(files[f])
        typ = 'fits'
     endif else begin
        sp = readspec(files[f], h)
        typ = 'spec'
     endelse

     for k = 0, n_elements(keys)-1 do begin
       oldval = sxpar(h, keys[k], count=ct, /silent)
       if ct eq 0 then oldval = '[absent]'
       if keyword_set(remove) then begin
         if ct gt 0 then begin
           sxdelpar, h, keys[k]
           if keyword_set(silent) eq 0 then print, ' ', files[f], '  ', keys[k], ' removed.'
         endif else begin
           if keyword_set(silent) eq 0 then print, ' ', files[f], '  ', keys[k], ' not in header.'
         endelse
       endif else begin
         if keyword_set(silent) eq 0 then print, ' ', files[f], '  ', keys[k], ': ', oldval, ' -> ', newval[k]
         if ct gt 1 then sxdelpar, h, keys[k] ; if multiple entries, sxaddpar only edits one
         if ct gt 1 and keys[k] eq 'OBJECT' then before='OBSERVER' else delvarx, before  ; common situation for LRIS
         sxaddpar, h, keys[k], newval[k], ' (edited)', before=before
       endelse
     endfor

     if typ eq 'fits' then modfits, files[f], 0, h, exten=0
     if typ eq 'spec' then writespec, files[f], sp, h
  endfor

end

