;+
; NAME: 
;   fieldcat
;
; PURPOSE: 
;   Extract a photometric catalog from an image and write to disk.
;   (Mostly a wrapper for fieldphot.)    
;
; CALLING SEQUENCE: 
;       
; INPUTS:
;       
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;       
; OUTPUTS: 
;
; Last modified 4 Sep 2020.  Documentation added 3 July 2022.
;-

; Basically a wrapper for fieldphot.  Find all stars in an image 
; (automatically excludes galaxies and extended sources!) and
; write photometry to a text file.
; Photometry is UNCALIBRATED aside from the zeropoint information given.

pro fieldcat, imagename, zpt1=zpt1, airtermi=airtermi, rinner=rinner, router=router, seeing=seeing, relative=relative, errthresh=errthresh, fwhmthreshfactor=fwhmthreshfactor, distthresh=distthresh, $
    outfile=outfile, copykey=copykey, verbose=verbose, outfilter=outfilter

 zptset = n_elements(zpt1)
 if zptset gt 0 then zpt1s = zpt1 - 25 ; for now have to make this silly adjustment, will fix fieldphot later?

 phot = fieldphot(imagename, zpt1=zpt1s, airtermi=airtermi, rinner=rinner, router=router, seeing=seeing, relative=relative, errthresh=errthresh, fwhmthreshfactor=fwhmthreshfactor, distthresh=distthresh, getinfo=info, verbose=verbose)

 if n_elements(info) eq 0 then begin
    logprint, '   Could not extract photometric catalog for '+imagename+'; returning...'
    return
 endif

if n_elements(outfile) eq 0 then outfile = repstr(imagename, '.fits', '.cat')
if outfile eq imagename then outfile = imagename + '.cat'

defaultcopykey = ['OBJECT','TARGNAME','TELESCOP','INSTRUME','INSTRMNT','FILTER','DICHNAME','AIRMASS','RA','DEC','OBSERVER','EXPOSURE','EXPTIME','DATE-OBS','UT','UTC','JD','MJD','MJD-OBS','COUNTS','SKYBKG','CTRATE']
if n_elements(copykey) gt 0 then copykey = unique([defaultcopykey, copykey]) $
                            else copykey = defaultcopykey

openw, 1, outfile


   h = headfits(imagename)
   for k = 0, n_elements(copykey)-1 do begin
     val = sxpar(h,copykey[k],comment=comment,count=ct,/silent)
     if ct eq 0 then continue
     outl =  '# '+clip(copykey[k],8)+'= '+strpad(val,21)
     if strlen(comment) gt 0 then outl += ' / ' + comment
     printf, 1, outl
   endfor

   ;if n_Elements(zpt1) eq 0

   ; strpad actually trims and doesn't just pad, why is IDL so awful :(
   ; furthermore it causes double precision information to be lost - relevant to JD...
   ; most likely will have to rewrite strpad to solve this.  add text to do a full-accuracy representation of the doluble.

   printf,1, "# IMFILE  = " + "'"+removepath(imagename)+"'" + ' / FITS image analyzed'
   printf,1, "# PHOTCAL = " + strpad("'NONE'",21) + ' / Origin of calibration'
   printf,1, "# SEEING  = " + strpad(info.seeing,21) + ' / Median stellar FWHM (pixels)'
   printf,1, "# RAPER   = " + strpad(info.r1,21) + ' / Aperture radius employed (pixels)'
   printf,1, "# RZPT    = " + strpad(info.r2,21) + ' / Aperture radius for zeropoint (pixels)'
   printf,1, "# ZEROPT  = " + strpad(25+info.zpteff,21)  + ' / Zeropoint used to convert flux->mag'
   if keyword_set(relative) eq 0 then begin
   printf,1, "# APCORR  = " + strpad(info.apcorr,21) + ' / Aperture correction (in mag)'
   printf,1, "# APUNC   = " + strpad(info.apunc,21) + ' / Uncetainty of APCORR'
   printf,1, "# APDEV   = " + strpad(info.apdev,21) + ' / RMS of APCORR'
   printf,1, "# NSTARSAP= " + strpad(info.napstars,21) + ' / Number of stars used to determine APCORR'
   endif
   if keyword_set(outfilter) then $
   printf,1, "# FILTER  = " + strpad(outfilter,21) + ' / Filter name (set by fieldcat)'
   if zptset then begin
   printf,1, "# ZEROPT1 = " + strpad(25+info.zpt1,21) + ' / Input zeropoint at zenith'   ; was .zpt1s, but this doesn't exist...
   printf,1, "# AIRTERM = " + strpad(info.airterm,21) + ' / Airmass dependence of zeropoint'
   endif
   printf,1, "# MINVAL  = " + strpad(info.minval,21) + ' / Minimum pixel value for photometry'
   printf,1, "# MAXVAL  = " + strpad(info.satval,21) + ' / Maximum pixel value for photometry'
   printf,1, "# SKYRAD  = " + strpad(info.skyr,21) + ' / Inner radius of sky annulus'
   printf,1, "# SKYWIDTH= " + strpad(info.skywidth,21) + ' / Width of sky annulus'
   printf,1, "# DIST_TH = " + strpad(info.distthresh,21) + ' / Distance threshold used'
   printf,1, "# FWHM_TH = " + strpad(info.fwhmthresh,21) + ' / FWHM threshold used'
   printf,1, "# NSTARS  = " + strpad(info.nstars,21) + ' / Number of stars in catalog'

   printf,1, "#ra          dec         mag     emag   x         y"
   for s = 0, n_elements(phot)-1 do begin
      printf, 1, fpr(phot[s].ra,3.7), ' ', fpr(phot[s].dec,3.7), ' ', fpr(phot[s].mag,3.3), ' ', fpr(phot[s].magerr,2.3), ' ', fpr(phot[s].x,5.2), ' ', fpr(phot[s].y,5.2)
   endfor

close, 1
  


end

