;+
; NAME: flatcombine.pro
;
; PURPOSE: 
;     Median-combine multiple individual flat-field images together to make a combined 
;     flat field.  Various features to aid in the rejection of stars and cosmic rays,
;     to normalize a 2D spectrum, and other options are available.
;
; CALLING SEQUENCE: 
;     flatcombine, filelist, outfile, /removeobjects, /response, /indresp 
;                  yresp=yresp,  objthresh=objthresh, objbuffer=objbuffer, 
;                  verbose=verbose, /median, /mean
;                  trimlo=trimlo, trimhi=trimhi, mincounts=mincounts, maxcounts=maxcounts,
;                  type=type, normzonein=normzonein, linewidth=linewidth, 
;                  lineblocks=lineblocks, satlevel=satlevel, satradius=satradius, 
;                  xnormkey=xnormkey, ynormkey=ynormkey, slitnorm=slitnorm, 
;                  xdiscontinuity=xdiscontinuity, smoothscale=smoothscale, 
;                  outobject=outobject
;                  lines=lines
;       
; INPUTS:
;             filelist  - List of files (an array, a file containing the list, or wildcard)
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;             outfile   - Filename to write to.  Created from filelist if unspecified.
;             normzone  - X region over which to calculate normalization factor
;                         (Two-element array, [xmin:xmax]) 
;             ynormzone - Y region over which to calculate normalization factor
;             /verbose  - Print more information (mostly debug info) to command line
;             /median   - Normalize by median (default)
;             /mean     - Normalize by trimmed mean
;             trimlo    - For trimmed mean, the trimming number on low side; default (N+1)/4
;             trimhi    - For trimmed mean, the trimming number on high side; default =trimlo
;             outobject - Set the OBJECT keyword of the output image.
;       Relevant for imaging:
;             /removeobjects - Search for and attempt to remove objects (for image flats)
;             satlevel  - Maximum pixel value for treating pixels as saturated for object
;                         detection purposes (bright star; very large exclusion radius)
;             satradius - Exclusion radius around saturated objects
;       Relevant for spectroscopy:
;             /response - Divide columns by median after combination (for spec flats)
;             /indresp  - Divide columns by median before combination (for spec flats:
;                         set if colour of the source is not stable.)
;             yresp     - Array of y pixel values over which to calculate the response
;             objthresh - Sigma threshold for object detection (default 6)
;             objbuffer - Grow region around detected objects (in pixels) to remove
;             mincounts - Count level below which columns are excluded and an average taken
;             maxcounts - Count level above which columns are excluded and an average taken
;             type      - Description of the type of flat-field to write to header as FLATTYPE
;             xnormkey  - Header keyword containing X region over which to calc norm factor
;             ynormkey  - Header keyword containing Y region over which to calc norm factor
;             slitnorm  - Apply a normalization in the Y (spatial) direction as well
;             smoothscale - Response paramater; TBD
;       Relevant for spectroscopy; not commonly used and not well tested:
;             lines     - List of pixel locations including lines to apply special 
;                         correction for.
;             linewidth - Number of pixels (full-width) to apply around lines
;             lineblocks - Number of blocks up and down the line to model/remove it
;             xdiscontinuity - Mark a discontinuity in the X direction
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS: none
;       
; OUTPUTS: 
;            The combined flat-field is written to disk.
;
; REQUIREMENTS: 
;
; Last modified 25 Jun 2020.  Documentation added 2 July 2022.
;-

pro flatcombine, filelist, outfile, removeobjects=removeobjects, response=response, indresp=indresp, yrespin=yrespin, lines=lines, objthresh=objthresh, objbuffer=objbuffer, verbose=verbose, mediani=mediani, meani=meani, trimlo=trimlo, trimhi=trimhi, mincounts=mincounts, maxcounts=maxcounts, type=type, normzonein=normzonein, linewidth=linewidth, lineblocks=lineblocks, satlevel=satlevel, satradius=satradius, xnormkey=xnormkey, ynormkey=ynormkey, slitnorm=slitnorm, xdiscontinuity=xdiscontinuity, smoothscale=smoothscale, outobject=outobject


; combine a bunch of flats with optional response division

; Now can accept flats of different sizes and align by physical coordinates.  Use with caution.

algorithm = 'median' ; default
if keyword_set(mediani) then algorithm = 'median'
if keyword_set(meani)   then algorithm = 'mean'  ; the 'i' is because otherwise it overwrites mean()?
if n_elements(trimlo) eq 0 then trimlo = (n_elements(filelist)+1)/4
if n_elements(trimhi) eq 0 then trimhi = trimlo
if n_elements(mincounts) eq 0 then mincounts = 100
if n_elements(maxcounts) eq 0 then maxcounts = 55000
if n_elements(satradius) eq 0 then satradius = 50
if n_elements(smoothscale) eq 0 then smoothscale = 7
if n_elements(linewidth) eq 0 then linewidth = 12
if n_elements(lineblocks) eq 0 then lineblocks = 5


if n_elements(outfile) eq 0 then begin
  filearr = strsplit(filelist[0],'.', /extract)
  fileroot = filearr[0]
  fileroot = repstr(fileroot,'*','')
  fileroot = repstr(fileroot,'?','')
  if strpos(fileroot,'flat') eq -1 then  outfile = fileroot + '.flat.fits' else $
                                         outfile = fileroot + '.fits'  ; could overwrite an input file...
endif


if n_elements(filelist) eq 1 and strpos(filelist[0],'.fits') eq -1 then begin
  if n_elements(file_search(filelist)) eq 0 then begin
     logprint, 'No matching file(s): ', filelist
  endif
  if n_elements(file_search(filelist)) eq 1 then begin
     files = grabcolumn(filelist,0,/str)
     files = files[where(files ne '')]
  endif
  if n_elements(files) eq 1 and n_elements(filelist) eq 1 then files = file_search(filelist)
endif else begin
  files = filelist
endelse

nfiles = n_elements(files)
;nmid = nfiles/2  now see below

fxo = intarr(nfiles)
fyo = intarr(nfiles)
fnx = intarr(nfiles)
fny = intarr(nfiles)
for f = 0, nfiles-1 do begin
   h = headfits(files[f])
   fxo[f] = -fix(sxpar(h, 'LTV1')) ; x origin in physical coordinate system
   fyo[f] = -fix(sxpar(h, 'LTV2')) ; y origin in physical coordinate system
   fnx[f] = sxpar(h, 'NAXIS1')
   fny[f] = sxpar(h, 'NAXIS2')
   ;print, sxpar(h,'PANE'), '    ', fnx[f], fny[f]
endfor

nx = min(fnx) ; could have an option to make these max, also.
ny = min(fny)
nmid = (where(fnx eq min(fnx) and fny eq min(fny))) [0]
xo = fxo[nmid]
yo = fyo[nmid]
hout = headfits(files[nmid])
;print, 'out: ', sxpar(hout,'PANE'), '    ', nx, ny


if n_elements(yrespin) eq 0 then yresp = indgen(ny) $
                            else yresp = yrespin[where(yrespin ge 0 and yrespin le ny-1)]

;refdata = mrdfits(files[nmid], 0, h, /silent, /fscale)
;s = size(refdata)
;nx = s[1]
;ny = s[2]

data = fltarr(nx, ny, nfiles) + !values.f_nan
inmeds = fltarr(nfiles)
usefiles = strarr(nfiles)

sumcounts = 0.
z = 0
for f = 0, nfiles-1 do begin
   if fxo[f] eq xo and fyo[f] eq yo and fnx[f] eq nx and fny[f] eq ny then begin
      indata = mrdfits(files[f], 0, h, /silent, /fscale)
      section = ''
   endif else begin
      readdata = mrdfits(files[f], 0, h, /silent, /fscale)
      indata = readdata[xo-fxo[f]:xo-fxo[f]+nx-1,yo-fyo[f]:yo-fyo[f]+ny-1] 
      section = '['+clip(xo-fxo[f])+'-'+clip(xo-fxo[f]+nx-1)+','+clip(yo-fyo[f])+'-'+clip(yo-fyo[f]+ny-1)+']'
   endelse

   ins = size(indata)
   inx = ins[1]
   iny = ins[2]
   if (inx ne nx) or (iny ne ny) then begin
      ; should no longer trigger.
      logprint, 'File ', files[f], ' has wrong dimensions ('+clip(inx)+'x'+clip(iny)+'; should have '+clip(nx)+'x'+clip(ny)+')'
      continue
   endif

   ;if f eq 0 then window, 0, xsize=1600, ysize=300
   ;if f eq 0 then plot, indata[*,700]/median(indata[*,700]), xrange=[500,2048], /xsty, /ysty
   ;if f gt 0 then oplot, indata[*,700]/median(indata[*,700])

   ;indata = indata + sxpar(h,'BZERO')
   if n_elements(normzonein) gt 0 then normzone = normzonein
   if n_elements(xnormkey) gt 0 then begin
     xnormstr = sxpar(h,xnormkey)
     xnormzone = fix(strsplit(xnormstr,',',/extract))
     normzone = [xnormzone[0], 0, xnormzone[1], -1]
   endif
   if n_elements(ynormkey) gt 0 then begin
     for k = 0, n_elements(ynormkey)-1 do begin
        ynormstr = sxpar(h,ynormkey[k], count=ct)
        if ct gt 0 then break
     endfor
     ynormzone = fix(strsplit(ynormstr,',',/extract))
     normzone = [0, ynormzone[0], -1, ynormzone[1]]
   endif
   if n_elements(normzone) eq 4 then begin
      if normzone[2] lt 0 then normzone[2] = nx+normzone[2]
      if normzone[3] lt 0 then normzone[3] = ny+normzone[3]
      if normzone[0] lt 0 or normzone[1] lt 0 or normzone[2] ge nx or normzone[3] ge ny then logprint, '  WARNING - normzone extends outside image'
      inmed = sigclipmedian(indata[normzone[0]>0:normzone[2]<(nx-1),normzone[1]>0:normzone[3]<(ny-1)])
   endif else begin
      inmed = sigclipmedian(indata)
   endelse
   inmeds[f] = inmed
   if inmed ge mincounts and inmed le maxcounts then begin
      if keyword_set(indresp) then begin
         iresp = median(indata[*,yresp], dimension=2) ; if each flat has a different lamp spectrum (was heating up, etc.)...
         ; it would probably be desirable to smooth this.
         for x = 0, nx-1 do indata[x,*] = indata[x,*] * inmed / iresp[x] ; inmed will immediately get normalized out in the next step, which is what we want
         if n_elements(avresp) eq 0 then avresp = iresp else avresp += iresp
      endif
      logprint, '  ' + removepath(files[f]) + section + ' (' + strtrim(long(inmed),2) + ' counts/pix)'
      data[*,*,z] = float(indata) / inmed
      ;oplot, data[*,700,z], color=rgb(255,255,0)
      usefiles[z] = files[f]
      sxaddpar, hout, 'COUNTS'+strtrim(string(z+1),2), sxpar(h,'COUNTS'), ' '+strtrim(long(inmed),2) 
      sumcounts += inmed

      z = z + 1
   endif else begin
      if inmed lt mincounts then $
      logprint, '  ' + removepath(files[f]) + section + ' (' + strtrim(long(inmed),2) + ' counts/pix) - too few counts; excluding'
      if inmed gt maxcounts then $
      logprint, '  ' + removepath(files[f]) + section + ' (' + strtrim(long(inmed),2) + ' counts/pix) - too many counts; excluding'
   endelse
endfor

if z lt 1 then begin ; formerly 3
    logprint, 'ERROR - Not enough counts to make a flat with these data!'
    return
endif

if z ne nfiles then data = data[*,*,0:z-1]

if keyword_set(indresp) then avresp  = avresp / z

flat = fltarr(nx, ny)

if algorithm eq 'median' then begin
   print, '  Median-combining...'
   i = 0l
   for y = 0, ny-1 do begin    
     for x = 0, nx-1 do begin  ; x is the fast array and should be iterated inside
       i = i + 1
       flat[x, y] = median(data[x, y, *], /even) ; mediansigclip?
       if i mod 157557 eq 0 then begin
          st = stddev(data[x, y, *])
          val = flat[x, y]
          if keyword_set(verbose) then print, clip(x,5), clip(y,5), 'stdev=',clip(st,6), '  val=',clip(val,6);, '  stdev/sqrt(val)=',clip(st/sqrt(val),6) 
       endif
     endfor
   endfor
endif
if algorithm eq 'mean' then begin
   print,  '  Combining via trimmed mean...'
   i = 0l
   nslice = (size(data)) [3]
   for y = 0, ny-1 do begin    
     for x = 0, nx-1 do begin  ; x is the fast array and should be iterated inside
       slice = data[x,y,*]
       slice = slice[sort(slice)]
       slice = slice[trimlo:nslice-1-trimhi]
       flat[x,y] = mean(slice)
     endfor
   endfor
endif

;if keyword_set(indresp) then begin
   ; Remove this again so the more refined response method can be applied later
   ; for x = 0, nx-1 do data[x,*,*] = data[x,*,*] * iresp[x]   ;  data not used again for spec
   ; for x = 0, nx-1 do flat[x,*]   = flat[x,*]   * iresp[x]   ; ?????
   ; This was re-introducing noise AND variations from the last flat of the sequence.
;endif

;plot, flat[*,400]


if keyword_set(removeobjects) then begin
   print, '  Identifying objects...'

   if n_elements(objthresh) eq 0 then objthresh = 6
   if n_elements(objbuffer) eq 0 then objbuffer = 12
   for f = 0, z-1 do begin
      if keyword_set(verbose) then print, '  ', removepath(files[f])
      indata = data[*,*,f] / flat
      mask = fltarr(nx, ny)
      satmask = fltarr(nx, ny)

      sigclipstats, indata, sigmahi=5, sigmalo=5, median=datamed, stdev=datastdev
      ;datamed = sigclipmedian(indata, sigmahi = 5, sigmalo = 5)
      ;datastdev = sigclipstdev(indata, sigmahi = 5, sigmalo = 5)

      sourcepixels = where(indata ge datamed + objthresh * datastdev, ctsourcepix)
      if ctsourcepix gt 0 then begin
         mask[sourcepixels] = 1
         smoothmask = smooth(mask, 1+2*objbuffer, /edge_truncate)
      endif
      ctsatpix = 0
      if n_elements(satlevel) gt 0 then begin
         satpixels = where(indata ge satlevel, ctsatpix)
         if ctsatpix gt 0 then begin
            satmask[satpixels] = 1
            smoothsatmask = smooth(mask, 1+2*satradius, /edge_truncate)
         endif
      endif

      if ctsourcepix gt 0 then begin
         sourcepixels = where(smoothmask gt 1.0E-6, ct)
         indata[sourcepixels] = !Values.F_NAN
      endif
      if ctsatpix gt 0 then begin
         nearsatpixels = where(smoothsatmask gt 1.0E-6, ct)
         indata[nearsatpixels] = !Values.F_NAN
      endif


      data[*,*,f] = indata  ; inefficient to copy over unaffected elements, but oh well
   endfor


   reflat = fltarr(nx, ny)

   if algorithm eq 'median' then begin
     print, '  Median-combining...'
     i = 0l
     for y = 0, ny-1 do begin
       for x = 0, nx-1 do begin
         i = i + 1
         reflat[x, y] = median(data[x, y, *], /even)   ;sigclipping this is REALLY slow...
         if i mod 157557 eq 0 and keyword_set(verbose) then begin
            slice = data[x, y, *]
            good = where(finite(slice), ct)
            if ct gt 3 then st = stddev(slice[good]) else st = 0
            val = flat[x,y] * reflat[x, y]
            print, '  ', clip(x,5), clip(y,5), 'n=', clip(ct,2), ' stdev=',fpr(st,1.4), '  val=',fpr(val,1.4);, '  stdev/sqrt(val)=',fpr(st/sqrt(val),3.2) 
         endif
       endfor
     endfor
     bad = where(finite(reflat) eq 0, ct)
     if ct gt 0 then reflat[bad] = 1
   endif

   if algorithm eq 'mean' then begin
     print,  '  Combining via trimmed mean...'
     i = 0l
     for y = 0, ny-1 do begin
       for x = 0, nx-1 do begin
         slice = data[x,y,*]
         good = where(finite(slice) eq 1, ctgood)
         if ctgood eq 0 then begin
             reflat[x,y] = 1
         endif else begin
             slice = slice[good]
             itrimlo = trimlo
             itrimhi = trimhi
             while ctgood-itrimlo-itrimhi lt 1 do begin
               itrimlo = (itrimlo - 1) > 0
               itrimhi = (itrimhi - 1) > 0
             endwhile
             slice = slice[sort(slice)]
             slice = slice[itrimlo:ctgood-1-itrimhi]
             reflat[x,y] = mean(slice)
         endelse
       endfor
     endfor
   endif

   flat = flat * reflat    ; want to divide by flat, then divide by reflat, so divide by (flat * reflat)

endif


if keyword_set(response) then begin
  if n_elements(lines) eq 55 then begin ; this is not used
    if lines eq 'find' then begin
        lines = [-1]
        nyr = n_elements(yresp)
        lineyresp = yresp[(nyr/2-20) > 0 : (nyr/2+20) < (nyr-1)]
        medcolx = fltarr(nx)
        for x = 0, nx-1 do begin
           medcolx[x] = median(flat[x,lineyresp], /even)
        endfor
        for x = 50, nx-50-1, 100 do begin
           xb     = (indgen(nx)) [(x-75)>0:(x+75)<(nx-1)]
           par = poly_fit(xb, medcolx[xb], 3, yfit=yfit)
           ydiff = medcolx[xb]-yfit
           posslines = where(ydiff gt 3.5*stddev(ydiff), ct) ; positive deviation only
           if ct gt 0 then print, xb[posslines]
           if ct gt 0 then lines = [lines,xb[posslines]]
        endfor
    endif
  endif

  ; the response procedure is unnecessarily repeated (for the already-corrected combined flat) when indrep is set.
  medcolx = fltarr(nx)+!values.f_nan
  for x = 0, nx-1 do begin
     if total(finite(flat[x,yresp])) gt 0.75*n_elements(yresp) then $
     medcolx[x] = median(flat[x,yresp], /even)
  endfor

  ;oplot, medcolx, col=rgb(200,0,0)
  
  ; could run a line detection algorithm here by doing a fit with a broad coverage
  ; and look for sharp deviances.

  dbound = [0, nx]
  if n_elements(xdiscontinuity) gt 0 then if xdiscontinuity[0] gt 0 then $
                                          dbound = [0, xdiscontinuity, nx]

  responsex = medcolx
  for x = 0, nx-1 do begin 
     if finite(medcolx[x]) eq 0 then continue
     d = max(where(x ge dbound))
     xi = (findgen(nx)) [(x-smoothscale)>dbound[d]:(x+smoothscale)<dbound[d+1]-1]  ; formerly +/- 10
     xi = xi[where(finite(medcolx[xi]))]
     ;par = polyiterfit(xi, medcolx[xi], order=2, yfit=yfit)
     par = poly_fit(xi-x, medcolx[xi], 3, yfit=yfit)
     ;delvarx, fpar
     ;for o = 1, 3 do begin
     ;  fpar = polyfitg(xi, medcolx[xi], o, yfit=yfit, guess=fpar)
     ;endfor
     ;if x eq 1660 then stop
     responsex[x] = yfit[where(xi eq x)]
  endfor
  for d = 1, n_elements(dbound)-2 do begin
     x = indgen(nx)
     responsex[where(x ge dbound[d])] += medcolx[dbound[d]-1] - medcolx[dbound[d]]
  endfor
  for x = 0, nx-1 do begin
     flat[x,*] = flat[x,*] / responsex[x]
  endfor
  ;oplot, responsex, col=rgb(0,40,255)

  ; mwrfits, flat, 'flatresp1.fits', /create

  ; Line filtering - not normally used
  ystart = fix(ny*(findgen(lineblocks)/lineblocks))
  ystop =  fix(ny*((1+findgen(lineblocks))/lineblocks))-1
  for l = 0, n_elements(lines)-1 do begin
     linestart = lines[l] - linewidth/2
     linestop = lines[l] + linewidth/2
     for b = 0, lineblocks-1 do begin
        medleft = median(flat[linestart-9:linestart-3,ystart[b]:ystop[b]], /even)
        medright = median(flat[linestop+3:linestop+9,ystart[b]:ystop[b]], /even)
        medtarget = (medleft + medright)/2.
        for x = linestart, linestop do begin
           medzone = median(flat[x,ystart[b]:ystop[b]], /even)
           flat[x,ystart[b]:ystop[b]] = medtarget * flat[x,ystart[b]:ystop[b]] / medzone
        endfor
     endfor
     medleft = median(flat[linestart-9:linestart-3,*], /even)
     medright = median(flat[linestop+3:linestop+9,*], /even)
     medtarget = (medleft + medright)/2.
     for x = linestart, linestop do begin
        flat[x,*] = medtarget * flat[x,*] / median(flat[x,*])
     endfor
  endfor

  ;oplot, median(flat[*,700:1200],dimension=2), color=rgb(0,255,0)

  ; don't even bother to flatfield for counts < ~1000*gain: even with 20 flatfields
  ; error introduced is 1% which exceeds typical flat variation


  if keyword_set(indresp) eq 0 then begin
    reqcounts = 1000 < (max(medcolx)/10.)
    lowcounts = where(medcolx lt reqcounts, ctlow, complement=goodcounts)
  endif else begin
    reqcounts = 1000 < (max(avresp)/10.)
    lowcounts = where(avresp lt reqcounts, ctlow, complement=goodcounts) ; formerly medcolx*median(inmeds)
  endelse

  ;window, 1
  ;plot, avresp
  ;oplot, [0, 4000], [reqcounts, reqcounts]

  ;maxx = (where(medcolx eq max(medcolx))) [0] ; ???
  if ctlow gt 0 then begin
    ;if min(lowcounts) lt maxx then $
    ;   logprint, '  Low counts for x=', $
    ;       clip(min(lowcounts[where(lowcounts lt maxx)])),'-',clip(max(lowcounts[where(lowcounts lt maxx)]))
    ;if max(lowcounts) gt maxx then $
    ;   logprint, '  Low counts for x=', $
    ;       clip(min(lowcounts[where(lowcounts gt maxx)])),'-',clip(max(lowcounts[where(lowcounts gt maxx)]))
    outxstr = clip(lowcounts[0])
    lastx = -1
    for xx = 1, n_elements(lowcounts)-1 do begin
       if lowcounts[xx] gt lowcounts[xx-1]+1 then begin
          if lowcounts[xx-1] ne lastx then $ 
            outxstr = outxstr+'-'+clip(lowcounts[xx-1])
          outxstr = outxstr+','+clip(lowcounts[xx])
          lastx = lowcounts[xx]
       endif
    endfor
    if lowcounts[n_elements(lowcounts)-1] ne lastx then $ 
    outxstr = outxstr + '-'+clip(lowcounts[n_elements(lowcounts)-1])
    print, '   Low counts for '+outxstr
    logprint, '   Using only a slit-averaged flat in this region.'
    for y = 0, ny-1 do begin
      wleft = where(lowcounts lt median(goodcounts), ctlowleft)
      wright = where(lowcounts gt median(goodcounts), ctlowright)
      if ctlowleft gt 0 then $
         flat[lowcounts[wleft],y] = median(flat[goodcounts[0:100<(n_elements(goodcounts)-1)],y])   
      if ctlowright gt 0 then $
         flat[lowcounts[wright],y] = median(flat[goodcounts[0>(n_elements(goodcounts)-100):n_elements(goodcounts)-1],y])   
    endfor
  endif

  ; in case there are unilluminated pixels, do a final overall division
  unilluminated = where(flat gt 0.5, ct)
  if ct gt 0 then begin
    med = median(flat[unilluminated], /even)
    flat /= med
  endif

endif

if keyword_set(slitnorm) then begin  ; halogen flats have an artificial slit response superimposed; remove it here.
  if n_elements(goodcounts) eq 0 then goodcounts = indgen(nx)
  for y= 0, ny-1 do begin
     medrowy = median(flat[goodcounts,y], /even)
     if medrowy gt 0.25 then flat[*,y] /= medrowy ; less than this is assumed to not be a false slit variation but genuinely 
                                                  ; dimmed part of the chip (off slit), which should be kept for later cropping
  endfor
  
endif


badval = where(flat le 0.01, ctbad)  ; zero values will create problems (nans) later and negative values are nonsense.
                                     ; also flagging very low values in advance.
if ctbad gt 0 then begin
  flat[badval] = 0.01  ; Prevents noise from being amplified too ridiculously if this flat is actually used.  Could set to Nan...
endif

sxaddpar, hout, 'BZERO', 0

sxaddpar, hout,'SUMCOUNTS',sumcounts, ' Sum of flat field count values'
for f = 0, z-1 do begin
  sxaddpar, hout, 'FLAT'+strtrim(string(f),2), removepath(usefiles[f])
endfor

if keyword_set(type) then sxaddpar,  hout, 'FLATTYPE', type, ' Type of flat-field'
sxaddpar, hout, 'NCOMBINE', nfiles, ' Number of flats averaged'

get_date, now
hist = 'Processed by flatcombine '+now
sxaddhist, hist, hout

if n_elements(outobject) eq 1 then sxaddpar, hout, 'OBJECT', outobject

sxaddpar, hout, 'BZERO', 0
sxaddpar, hout, 'BSCALE', 1
mwrfits, flat, outfile, hout, /create

logprint, '  Written to ', outfile

end
