;+
; NAME: flatcombine3.pro
;
; PURPOSE: 
;    Alternative procedure for combining multiple spectroscopic flats together in the 
;    presence of strong artificial spatial variations along the slit (LRIS halogen flats)
;
; CALLING SEQUENCE: 
;    flatcombine3, files, outfile, yresp
;       
; INPUTS:
;    files - List of filenames for input flats (array or wildcard string)
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;    outfile - Filename to write to.  Created from filelist if unspecified.
;    yresp - Array of y pixels indicating the area to use for the normalization.
;            Defaults to (ny/2+50):(ny-100)
;    outobject - Value of the OBJECT keyword to write to the output file.
;    type      - Value of the FLATTYPE keyword to write to the output file
;    mincounts - Count level below which columns are excluded and an average taken
;    maxcounts - Count level above which columns are excluded and an average taken
;    nblocks   - Number of spatial median blocks to employ.
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS: none
;       
; OUTPUTS:  The combined flat
;
; REQUIREMENTS: 
;
; Last modified 10 Apr 2020.  Documentation added 2 July 2022.
;-

pro flatcombine3, files, outfile, yresp=yresp, outobject=outobject, type=type, $
   mincounts=mincounts, maxcounts=maxcounts, nblocks=nblocks

   ; flat combination with strong spatial median filtering to remove flat-lamp spatial irregularities and slit irregularities.
   ; for LRIS halogen flats.

   nf = n_elements(files)
   for f = 0, nf-1 do begin
     if file_test(files[f]) eq 0 then begin
       print, '  ERROR: ', files[f], ' not found'
       return
     endif
   endfor
   if n_elements(outfile) eq 0 then begin
     filearr = strsplit(files[0],'.', /extract)
     fileroot = legalfilestr(filearr[0])
     if strpos(fileroot,'flat') eq -1 then  outfile = fileroot + '.flat.fits' else $
                                            outfile = fileroot + '.fits'  ; could overwrite an input file...
   endif

   if n_elements(mincounts) eq 0 then mincounts = 100
   if n_elements(maxcounts) eq 0 then maxcounts = 55000

   h0 = headfits(files[0])
   nx = sxpar(h0,'NAXIS1')
   ny = sxpar(h0,'NAXIS2')
   outblock = fltarr(nx,ny,nf) + !values.f_nan
   usefiles = ''
   hout = h0
   sumcounts = 0.

   minx = 50 ; These describe 'absolute' bad regions of the chip to never use in median calculations.
   maxx = nx-50
   if n_elements(yresp) eq 0 then ynorm = [ny/2+50, ny-100]
   if n_elements(yresp) eq 0 then yresp = indgen(ynorm[1]-ynorm[0]) + ynorm[0]
   if n_elements(nblocks) eq 0 then nblocks = 12

   z = 0
   for f = 0, nf-1 do begin
      inim = readfits(files[f], hf, /silent)
      inmed = median(inim)
      if inmed lt mincounts or inmed gt maxcounts then begin
        if inmed lt mincounts then $
        logprint, '  ' + removepath(files[f]) + '' + ' (' + strtrim(long(inmed),2) + ' counts/pix) - too few counts; excluding'
        if inmed gt maxcounts then $
        logprint, '  ' + removepath(files[f]) + '' + ' (' + strtrim(long(inmed),2) + ' counts/pix) - too many counts; excluding'
        continue
      endif

      xprofile = median(inim[*,yresp], dimension=2) ; spectral profile
      ;plot, xprofile, xrange=[3500,4000], psym=10, /ynoz
      xprofile = median(xprofile, 5) ; edge issue?? also a bit crude given gradients, but performs OK.
      ;oplot, xprofile, color=rgb(255,0,0), psym=10
      ;oplot, xprofile+sqrt(xprofile), color=rgb(155,0,0), psym=10
      ;oplot, xprofile-sqrt(xprofile), color=rgb(155,0,0), psym=10
      for x = 0, n_elements(xprofile)-1 do inim[x,*] /= xprofile[x]
      minbx = min(where(xprofile gt 0.1*max(xprofile)))+50
      maxbx = max(where(xprofile gt 0.1*max(xprofile)))-50
      nb = nblocks
      bxc = minbx + (maxbx-minbx)*findgen(nb)/(nb-1)
      db = bxc[1]-bxc[0]
      yprofileb = fltarr(nb, ny)
      xweightb  = fltarr(nx, nb)
      xx = findgen(nx)
      for b = 0, nb-1 do begin
        xc = bxc[b]
        yprofileb[b,*] = median(inim[(xc-50)>minx:(xc+50)<(maxx),*], dimension=1) ; slit profile
        xweightb[*,b] = (db-abs(xx-xc))/(db) > 0
      endfor
      left = where(xx lt bxc[0], ctleft)
      right = where(xx gt bxc[nb-1], ctright)
      if ctleft gt 0 then xweightb[left,0] = 1.
      if ctright gt 0 then xweightb[right,nb-1] = 1.
      for x = 0, nx-1 do begin
        xweightb[x,*] /= total(xweightb[x,*]) ; normalize
      endfor
      ;plot, xweightb[*,0], /nodata
      ;for b = 0, nb-1 do begin
      ;  oplot, xweightb[*,b], color=rgb(255-20*b,0,20*b)
      ;endfor
      ydiv = fltarr(nx,ny)
      for b = 0, nb-1 do begin
         ydiv += yprofileb[b,*] ## xweightb[*,b]
      endfor

      bad = where(ydiv lt 0.02*median(ydiv), ctbad)
      if ctbad gt 0 then ydiv[bad] = !values.f_nan

      ;yprofile1 = median(inim[300:500,*], dimension=1) ; slit profile
      ;yprofile2 = median(inim[(nx/2-50):(nx/2+50),*], dimension=1) ; slit profile
      ;yprofile3 = median(inim[nx-201:nx-101,*], dimension=1) ; slit profile
      ;yweight1 = 1-(findgen(nx)/((nx-1)/2.)) > 0
      ;yweight3 = -1+findgen(nx)/((nx-1)/2.) > 0
      ;yweight2 = 1-yweight1-yweight3
      ;ydiv = yprofile1##yweight1 + yprofile2##yweight2 + yprofile3##yweight3
      ;writefits, 'ydiv.fits', ydiv
      inim /= ydiv

      ;cyprofile = yprofile-median(yprofile,30) ; edge issue??
      ;for y = 0, n_elements(yprofile)-1 do inim[*,y] /= yprofile[y]

      if z eq 0 then begin
         usefiles = files[f]
      endif else begin
         usefiles = [usefiles, files[f]]
      endelse
      sxaddpar, hout, 'COUNTS'+strtrim(string(z+1),2), sxpar(hf,'COUNTS')
      sumcounts += sxpar(hf,'COUNTS')
      logprint, '  ' + removepath(files[f]) + '' + ' (' + strtrim(long(inmed),2) + ' counts/pix)'
      
      outblock[*,*,z] = inim
      z += 1
   endfor
   if z eq 0 then begin
      logprint, '  ERROR - no useable flats!'
      return
   endif

   outblock = outblock[*,*,0:(z-1)]

   if z gt 1 then $
     flat = median(outblock, dimension=3, /even) $
   else $
     flat = outblock

   reqcounts = 1000 < (max(xprofile)/10.) ; based on a single one only
   lowcounts = where(xprofile lt reqcounts, ctlow, complement=goodcounts)
   if ctlow gt 0 then begin
     outxstr = clip(lowcounts[0])
     lastx = -1
     for xx = 1, n_elements(lowcounts)-1 do begin
        if lowcounts[xx] gt lowcounts[xx-1]+1 then begin
           if lowcounts[xx-1] ne lastx then $ 
             outxstr = outxstr+'-'+clip(lowcounts[xx-1])
           outxstr = outxstr+','+clip(lowcounts[xx])
           lastx = lowcounts[xx]
        endif
     endfor
     if lowcounts[n_elements(lowcounts)-1] ne lastx then $ 
     outxstr = outxstr + '-'+clip(lowcounts[n_elements(lowcounts)-1])
     print, '   Low counts for '+outxstr
     logprint, '   Using only a slit-averaged flat in this region.'
     for y = 0, ny-1 do begin
       wleft = where(lowcounts lt median(goodcounts), ctlowleft)
       wright = where(lowcounts gt median(goodcounts), ctlowright)
       if ctlowleft gt 0 then $
          flat[lowcounts[wleft],y] = median(flat[goodcounts[0:100<(n_elements(goodcounts)-1)],y])   
       if ctlowright gt 0 then $
          flat[lowcounts[wright],y] = median(flat[goodcounts[0>(n_elements(goodcounts)-100):n_elements(goodcounts)-1],y])   
     endfor
   endif

  if keyword_set(slitnorm) then begin 
    if n_elements(goodcounts) eq 0 then goodcounts = indgen(nx)
    for y= 0, ny-1 do begin
       medrowy = median(flat[goodcounts,y], /even)
       if medrowy gt 0.25 then flat[*,y] /= medrowy
          ; less than this is assumed to not be a false slit variation but genuinely 
          ; dimmed part of the chip (off slit), which should be kept for later cropping
    endfor
  
  endif


  nans = where(finite(flat) eq 0)
   flat = flat > 0.01
  flat[nans] = !values.f_nan

   sxaddpar, hout,'SUMCOUNTS',sumcounts, ' Sum of flat field count values'
   for f = 0, z-1 do begin
     sxaddpar, hout, 'FLAT'+strtrim(string(f+1),2), removepath(usefiles[f])
   endfor

   if keyword_set(type) then sxaddpar,  hout, 'FLATTYPE', type, ' Type of flat-field'
   sxaddpar, hout, 'NCOMBINE', z, ' Number of flats averaged'

   get_date, now
   hist = 'Processed by lineflatcombine '+now
   sxaddhist, hist, hout

   if n_elements(outobject) eq 1 then sxaddpar, hout, 'OBJECT', outobject

   sxaddpar, hout, 'BZERO', 0
   sxaddpar, hout, 'BSCALE', 1

   writefits, outfile, flat, hout

   logprint, '  Written to ', outfile

end

