;+
; NAME: lineflatcombine.pro
;
; PURPOSE: 
;    Alternative procedure for combining multiple spectroscopic flats together in the 
;    presence of strong narrow line features in the light source (sky or deuterium
;    flats). 
;
; CALLING SEQUENCE: 
;    lineflatcombine, files, outfile, ydiscontinuity=ydiscontinuity, outobject=outobject, 
;         type=type, mincounts=mincounts, maxcounts=maxcounts, slitnorm=slitnorm       
;
; INPUTS:
;    files - List of filenames for input flats (array or wildcard string)
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;    outfile - Filename to write to.  Created from filelist if unspecified.
;    ydiscontinuity - Location of a y-discontinuity (e.g. a chip gap).
;    outobject - Value of the OBJECT keyword to write to the output file.
;    type      - Value of the FLATTYPE keyword to write to the output file
;    mincounts - Count level below which columns are excluded and an average taken
;    maxcounts - Count level above which columns are excluded and an average taken
;    slitnorm  - Apply a normalization along the y-direction as well.
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;       
; OUTPUTS: 
;
; REQUIREMENTS: 
;
; Last modified 10 Apr 2020.  Documentation added 2 July 2022.
;-

pro lineflatcombine, files, outfile, ydiscontinuity=ydiscontinuity, outobject=outobject, type=type, $
   mincounts=mincounts, maxcounts=maxcounts, slitnorm=slitnorm

   nf = n_elements(files)
   for f = 0, nf-1 do begin
     if file_test(files[f]) eq 0 then begin
       print, '  ERROR: ', files[f], ' not found'
       return
     endif
   endfor
   if n_elements(outfile) eq 0 then begin
     filearr = strsplit(files[0],'.', /extract)
     fileroot = legalfilestr(filearr[0])
     if strpos(fileroot,'flat') eq -1 then  outfile = fileroot + '.flat.fits' else $
                                            outfile = fileroot + '.fits'  ; could overwrite an input file...
   endif

   if n_elements(mincounts) eq 0 then mincounts = 100
   if n_elements(maxcounts) eq 0 then maxcounts = 55000

   h0 = headfits(files[0])
   nx = sxpar(h0,'NAXIS1')
   ny = sxpar(h0,'NAXIS2')
   outblock = fltarr(nx,ny,nf) + !values.f_nan
   usefiles = ''
   hout = h0
   sumcounts = 0.

   z = 0
   for f = 0, nf-1 do begin
      inim = readfits(files[f], hf, /silent)
      inmed = median(inim)
      if inmed lt mincounts or inmed gt maxcounts then begin
        if inmed lt mincounts then $
        logprint, '  ' + removepath(files[f]) + '' + ' (' + strtrim(long(inmed),2) + ' counts/pix) - too few counts; excluding'
        if inmed gt maxcounts then $
        logprint, '  ' + removepath(files[f]) + '' + ' (' + strtrim(long(inmed),2) + ' counts/pix) - too many counts; excluding'
        continue
      endif

      if z eq 0 then begin
         profile = median(inim, dimension=2)
         cprofile = profile-median(profile,30)
         gradient = abs(shift(cprofile,-1)-shift(cprofile,+1))
         smgradient = convol(gradient,[1,1,1,1,1]/5.)
         best = median(where(smgradient eq max(smgradient)))
         xcheck = [(best-200)>0, (best+200)<(nx-1)]
         ;print, xcheck ;[864,1164]
         usefiles = files[f]
      endif else begin
         usefiles = [usefiles, files[f]]
      endelse
      sxaddpar, hout, 'COUNTS'+strtrim(string(z+1),2), sxpar(hf,'COUNTS') ; also get it from inside removeflatlines?
      sumcounts += sxpar(hf,'COUNTS')
      logprint, '  ' + removepath(files[f]) + '' + ' (' + strtrim(long(inmed),2) + ' counts/pix)'
      linim = removeflatlines(inim, ydisc=ydiscontinuity, xcheck=xcheck)
      outblock[*,*,z] = linim
      z += 1
   endfor
   if z eq 0 then begin
      logprint, '  ERROR - no useable flats!'
      return
   endif

   outblock = outblock[*,*,0:(z-1)]

   flat = median(outblock, dimension=3, /even)

   reqcounts = 1000 < (max(profile)/10.)
   lowcounts = where(profile lt reqcounts, ctlow, complement=goodcounts)
   if ctlow gt 0 then begin
     outxstr = clip(lowcounts[0])
     lastx = -1
     for xx = 1, n_elements(lowcounts)-1 do begin
        if lowcounts[xx] gt lowcounts[xx-1]+1 then begin
           if lowcounts[xx-1] ne lastx then $ 
             outxstr = outxstr+'-'+clip(lowcounts[xx-1])
           outxstr = outxstr+','+clip(lowcounts[xx])
           lastx = lowcounts[xx]
        endif
     endfor
     if lowcounts[n_elements(lowcounts)-1] ne lastx then $ 
     outxstr = outxstr + '-'+clip(lowcounts[n_elements(lowcounts)-1])
     print, '   Low counts for '+outxstr
     logprint, '   Using only a slit-averaged flat in this region.'
     for y = 0, ny-1 do begin
       wleft = where(lowcounts lt median(goodcounts), ctlowleft)
       wright = where(lowcounts gt median(goodcounts), ctlowright)
       if ctlowleft gt 0 then $
          flat[lowcounts[wleft],y] = median(flat[goodcounts[0:100<(n_elements(goodcounts)-1)],y])   
       if ctlowright gt 0 then $
          flat[lowcounts[wright],y] = median(flat[goodcounts[0>(n_elements(goodcounts)-100):n_elements(goodcounts)-1],y])   
     endfor
   endif

  if keyword_set(slitnorm) then begin 
    if n_elements(goodcounts) eq 0 then goodcounts = indgen(nx)
    for y= 0, ny-1 do begin
       medrowy = median(flat[goodcounts,y], /even)
       if medrowy gt 0.25 then flat[*,y] /= medrowy
          ; less than this is assumed to not be a false slit variation but genuinely 
          ; dimmed part of the chip (off slit), which should be kept for later cropping
    endfor
  
  endif


   flat = flat > 0.01

   sxaddpar, hout,'SUMCOUNTS',sumcounts, ' Sum of flat field count values'
   for f = 0, z-1 do begin
     sxaddpar, hout, 'FLAT'+strtrim(string(f+1),2), removepath(usefiles[f])
   endfor

   if keyword_set(type) then sxaddpar,  hout, 'FLATTYPE', type, ' Type of flat-field'
   sxaddpar, hout, 'NCOMBINE', z, ' Number of flats averaged'

   get_date, now
   hist = 'Processed by lineflatcombine '+now
   sxaddhist, hist, hout

   if n_elements(outobject) eq 1 then sxaddpar, hout, 'OBJECT', outobject

   sxaddpar, hout, 'BZERO', 0
   sxaddpar, hout, 'BSCALE', 1

   writefits, outfile, flat, hout

   logprint, '  Written to ', outfile

end

