;+
; NAME: maskobjects2.pro
;
; PURPOSE: 
;     Detect sources in an image and produce a mask identifying source-affected pixels.
;     Used for excluding sources from a sky flat-field or similar.
;
; CALLING SEQUENCE: 
;     maskobjects2, img, smoothsize=smoothsize, objthresh=objthresh, objbuffer=objbuffer,
;                   niter=niter, nan=nan, mask=mask, halothresh=halothresh, halobuffer=halobuffer
;       
; INPUTS:
;     img = filename or 2D array to use for object detection 
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;     smoothsize - Smoothing kernel size to apply prior to object detection
;     niter      - Number of iterations to perform (default 1)
;     /nan       - Alter input array to set source values to Nan. 
;     objthresh  - Sigma threshold for object detection (default calculated by formula)
;     objbuffer  - Grow region around detected objects (in pixels) to remove (default 5)
;     halothresh - Threshold to identify bright stars
;     halobuffer - Approx. exclusion region around bright stars
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;     mask       - Boolean 2D array where 1 = source pixel.
;       
; OUTPUTS: 
;     img is altered to set the source values to NaN if given as an array and /nan set.
;     Masked image is written to 'out.fits' otherwise (for now).
;
; REQUIREMENTS: 
;
; Last modified 11 Jun 2013.  Documentation added 3 July 2022.
;-


pro maskobjects2, img, smoothsize=smoothsize, objthresh=objthresh, objbuffer=objbuffer, niter=niter, nan=nan, mask=mask, $
   halothresh=halothresh, halobuffer=halobuffer

   ; Currently the most modern version of the masking code
   ; returns a mask where 1=object, 0=blank sky
   ; if nan set, actually modifies the img to set objects to NaN
   
   if n_elements(smoothsize) eq 0 then smoothsize = 5
   if n_elements(objthresh) eq 0 then objthresh = 6-(smoothsize*0.6 < 4)
   if n_elements(objbuffer) eq 0 then objbuffer = 5
   if n_elements(niter) eq 0 then niter = 1 ; usually fine
   if n_elements(halothresh) eq 0 then halothresh = 500.
   if n_elements(halobuffer) eq 0 then halobuffer = 60.

   if n_elements(img) eq 1 then begin
      img = readfits(img)
      outfile = 'out.fits'
   endif

   nx = (size(img)) [1] 
   ny = (size(img)) [2]
   mask = intarr(nx, ny)
   smimg = smooth(img,smoothsize, /nan)
   sky = smimg

   sigclipstats, sky, sigmahi=5, sigmalo=5, median=skymed, stdev=skystdev
   smimg -= skymed

   halocheck = 1
   if halocheck eq 1 then begin
     ;writefits, 'smimg.fits', smimg
     brightpixels = where(smimg ge halothresh, ctbrightpix)
     if ctbrightpix gt 0 then begin
        sourceimg = fltarr(nx,ny) ; this is an image with only sources included
        sourceimg[brightpixels] = smimg[brightpixels]   
        sourceimg = smooth(sourceimg,1+halobuffer*0.8, /edge_truncate, /nan)
        sourceimg = smooth(sourceimg,1+halobuffer*0.8, /edge_truncate, /nan)
        thr = 5. ; formerly 10
        ;writefits, 'brimg.fits', sourceimg
        masked = where(sourceimg gt thr, ctmasked)
        mask[masked] = 1
        sky[masked] = !values.f_nan
     endif
   endif

   for i = 0, niter-1 do begin
     sigclipstats, sky, sigmahi=5, sigmalo=5, median=skymed, stdev=skystdev
     smimg -= skymed
     ;writefits, 'smimg.fits', smimg
     sourcepixels = where(smimg ge objthresh * skystdev, ctsourcepix)
     ;print, objthresh * skystdev
     ;print, ctsourcepix
     if ctsourcepix gt 0 then begin
       sourceimg = fltarr(nx,ny)
       sourceimg[sourcepixels] = 1
       sourceimg = smooth(sourceimg,  1+objbuffer, /edge_truncate, /nan) 
       sourceimg = smooth(sourceimg, 1+objbuffer, /edge_truncate, /nan) 
       ;writefits, 'sourceimg.fits', sourceimg
       masked = where(sourceimg gt 0.05, ctmasked)
       mask[masked] = 1
       if i lt niter-1 then sky[masked] = !values.f_nan
     endif

   endfor

   allmasked = where(mask, ctm)
   if keyword_set(nan) and ctm gt 0 then img[allmasked] = !values.f_nan

   if keyword_set(outfile) then writefits, outfile, img

end

