;+
; NAME: 
;   readcat
;
; PURPOSE: 
;   Read a catalog file from disk.  See writecat for more information.  
;
; CALLING SEQUENCE: 
;   cat = readcat(filename, h, header=header, addunc=addunc, magcolnum=magcolnum)
;       
; INPUTS:
;   filename - Input catalog filename.
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;   addunc    - Add a systematic uncertainty in quadrature to the magnitude uncertainty
;   magcolnum - Specify the column containing the magnitude, if reading a nonstandard file
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;   header   - Alternate means of obtaining header (see h below)
;
; OUTPUTS: 
;   cat      - Catalog structure.    
;   h        - Header string array containing FITS-like header cards
;
; Last modified 17 Feb 2016.  Documentation added 3 July 2022.
;-


function readcat, filename, h, header=header, addunc=addunc, magcolnum=magcolnum

  ; reads an ASCII star catalog (single-filter format, usually associated with extraction of a single image)
  ; see also readcalibration.pro

  if n_elements(filename) eq 0 then begin
     print, 'No file specified.'
     return, ['']
  endif
  if file_test(filename) eq 0 then begin
     print, 'File '+filename+' not found!'
     return, ['']
  endif
  if strpos(filename,'.fit') ge 0  and strpos(filename,'.cat') lt 0 then begin
     print, 'Cannot process fits files.'
     return, ['']
  endif

  if n_elements(magcolnum) eq 0 then magcol = 2 else magcol = magcolnum

  nl = countlines(filename)  
  if nl eq 0 then begin
     print, 'Blank file: ', filename
     return, -1
  endif
  cat = replicate({ra:0.0d, dec:0.0d, x:0., y:0., mag:0.0, unc:0.},nl) ; mimics readcalibration
  header = ['']

  close, 1
  openr, 1, filename
  inline = ''
  i = 0
  for l = 0l, nl-1 do begin
     readf, 1, inline
     inline = clip(inline)
     if strlen(inline) eq 0 then continue
     if strmid(inline,0,1) eq '#' then begin
        if strpos(inline,'=') ge 0 then begin
           inline = clip(strmid(inline,1))
           hline = processheaderline(inline)
           if hline ne '' then begin
             header = [header, clip(hline,78)]  ; might not be strictly valid
           endif
        endif
     endif else begin
        splitline = strsplit(inline, /extract)
        ni = n_elements(splitline)
        if ni eq 0 then continue
        if ni lt 2 then begin
           print, 'Problem reading this line:', inline
        endif else begin
           ncol = n_elements(splitline)
           cat[i].ra = double(splitline[0])
           cat[i].dec = double(splitline[1])
           cat[i].mag = float(splitline[magcol])
           cat[i].unc = float(splitline[magcol+1])
           if n_elements(addunc) gt 0 then cat[i].unc = sqrt(cat[i].unc^2 + addunc^2)
           if ncol gt 5 then begin
             cat[i].x = float(splitline[4])
             cat[i].y = float(splitline[5])
           endif
        endelse
        i += 1
     endelse
  endfor

  cat = cat[0:i-1]

  close, 1
  h = header
  return, cat


end

