;+
; NAME: 
;   readimheaders
;
; PURPOSE: 
;   Read in header metadata from multiple image files into a structure array. 
;   Used to help filter on specific objects, filters, and similar. (See also readcatheaders)
;
; CALLING SEQUENCE:
;   info = readimheaders(files, expkey=expkey, instkey=instkey, 
;         filtkey=filtkey, filt2key=filt2key, skykey=skykey, objkey=objkey, seekey=seekey) 
;       
; INPUTS:
;   files - Array of files (FITS images) to read in
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;   expkey   - Specify header keyword containing exposure time
;   instkey  - Specify header keyword containing instrument name
;   filtkey  - Specify header keyword containing filter name (default FILTER)
;   filt2key - Specify header keyword containing second filter name (default FILTER2)
;   skykey   - Specify header keyword containing sky level
;   timekey  - Specify header keyword containing UT/time (default UTC)
;   objkey   - Specify header keyword containing object name (default OBJECT)
;   seekey   - Specify header keyword containing seeing estimate
;       
; OUTPUTS:
;   info  - Structure array containing information about the images.
;
; COMMENTS:
;   Used at the start of calibratenight only; subsequent operations use readcatheaders.
;
; Last modified 4 Sep  2020.  Documentation added 3 July 2022.
;-


function readimheaders, files, expkey=expkey, instkey=instkey, filtkey=filtkey, filt2key=filt2key, skykey=skykey, objkey=objkey, seekey=seekey

   imh = replicate({filename:'', date:'', ut:'', jd:0.d, mjd:0.d, instrument:'', object:'', filter:'', filter2:'', rastr:'', decstr:'', radeg:!values.f_nan, decdeg:!values.f_nan, exptime:!values.f_nan, airmass:!values.f_nan, skycounts:!values.f_nan, ctrate:!values.f_nan, nx:0, ny:0, pxscale:!values.f_nan, satval:!values.f_nan, seeing:!values.f_nan}, n_elements(files))
 
  ; get relevant header information for a block of files.

   for f = 0, n_elements(files)-1 do begin
     if files[f] eq '' then continue
     if file_test(files[f]) eq 0 then begin
        print, ' Cannot find file ', files[f]
        continue 
     endif

     imh[f].filename = files[f]
     h = headfits(files[f], /silent)

       ; Date/time
     imh[f].date = clip(sxpar(h,'DATE'))   ; date string, e.g. "2013-01-13T08:48:37"
     imh[f].ut   = clip(sxpar(h,'UT', /silent, count=ct))
        if ct eq 0 then imh[f].ut   = clip(sxpar(h,'UTC', /silent, count=ct))
     imh[f].jd   = sxpar(h,'JD')
        if ct eq 0 then imh[f].jd   = double(sxpar(h,'JD-OBS', /silent, count=ct))
     imh[f].mjd   = sxpar(h,'MJD')
        if ct eq 0 then imh[f].mjd   = double(sxpar(h,'MJD-OBS', /silent, count=ct))

       ; Exposure
     imh[f].exptime = float(sxpar(h,expkey))

       ; Target and sky position
     imh[f].object = clip(sxpar(h,objkey))
     imh[f].rastr  = clip(sxpar(h,'RA'))
     imh[f].decstr = clip(sxpar(h,'DEC'))
     imh[f].radeg  =  ten(sxpar(h,'RA'))*15.
     imh[f].decdeg =  ten(sxpar(h,'DEC'))

       ; Telescope position
     imh[f].airmass = float(sxpar(h,'AIRMASS'))

       ; Instrument state
     if n_elements(instkey) gt 0 then imh[f].instrument = clip(sxpar(h, instkey)) 
     if n_elements(dichkey) gt 0 then imh[f].dichroic   = clip(string(sxpar(h,dichkey)))
     imh[f].filter = clip(sxpar(h,filtkey))
     if n_elements(filt2key) gt 0 then imh[f].filter2 = clip(sxpar(h,filt2key))

       ; Background-related keywords
     imh[f].skycounts = float(sxpar(h,skykey))
     imh[f].ctrate  = imh[f].skycounts/imh[f].exptime

       ; CCD keywords
     imh[f].nx = sxpar(h,'NAXIS1')
     imh[f].ny = sxpar(h,'NAXIS2')

      ; Other
     cd11 = sxpar(h,'CD1_1')   &    cd12 = sxpar(h,'CD1_2')
     cd21 = sxpar(h,'CD2_1')   &    cd22 = sxpar(h,'CD2_2')
     imh[f].pxscale = 3600. * sqrt(sqrt( (cd11^2 + cd12^2)*(cd21^2 + cd22^2) ))
           
     imh[f].satval =  sxpar(h,'SATURATE', /silent, count=ct)
        if ct eq 0 then imh[f].satval = !values.f_nan
     if n_elements(seekey) gt 0 then imh[f].seeing = sxpar(h,seekey)

   endfor

   return, imh

end

