;+
; NAME: readsdss2.pro
;
; PURPOSE: 
;       Read the complete SDSS calibration telescope bright star catalog.
;
; CALLING SEQUENCE: 
;       c = readsdss2(filename=filename)
;
; INPUTS:
;       No required inputs.
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       liverpool : Set to tranform the magnitudes to the Liverpool Telescope system.
;       filename = Specify the filename explicitly; otherwise the IDL path tree is 
;                  is searched for landolt2009.txt.
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS: none.
; 
; OUTPUTS: 
;      A structure containing the magnitude of SDSS standards in the catalog.
;      Transformed magnitudes in UBVRI system are also included.
;
; REQUIREMENTS:  none
;
; Last modified 18 Apr 2020.  Documentation added 2 Jul 2022
;-


function readsdss2, filename=filename, liverpool=liverpool

; Read in the (complete) SDSS secondary standards photometric catalogs.
; Generally both keywords are left blank; the code will find the catalog file on its own.

common sdss2info, sdss2path, sdss2out

if n_elements(sdss2out) gt 1 and n_elements(filename) eq 0 and keyword_set(liverpool) eq 0 then begin
  ; should have 1659 elements
  return, sdss2out
endif

if n_elements(sdss2path) eq 0 then begin
      paths = strsplit(!path,':',/extract)
      for p = -1, n_elements(paths)-1 do begin   ; find the directory this file is in
         if p ge 0 then path = paths[p] $
                        else path = '.'
         sfile = path+'/readsdss2.pro'
         if file_test(sfile) then begin
            sdss2path = path+'/'
         endif
      endfor
endif

;if n_elements(landoltpath) eq 0 then landoltpath = '~/progs/idl/phot/'

if n_elements(filename) eq 0 then filename = sdss2path+'sdss2.txt'
if n_elements(filenamex) eq 0 then filenamex = sdss2path+'sdss2extended.txt'

;    StarName      RA (J2000.0)    DEC (J2000.0)          r'        u'-g'      g'-r'      r'-i'      i'-z'         rms r'   rms u'-g'  rms g'-r'  rms r'-i'  rms i'-z'       n u'    n g'    n r'    n i'    n z' 

;     Hilt 31     00:28:11.15      +64:07:51.8         10.996      0.999      0.587      0.294      0.147         0.001      0.003      0.001      0.001      0.002          16      16      16      15      15  


n = 158
sdss2 = replicate({name:'', ra:0.D, dec:0.D, rastr:'', decstr:'', Uj:0., Bj:0., Vj:0., Rc:0., Ic:0., us:0., gs:0., rs:0., is:0., zs:0., $
                                                                 eUj:0.,eBj:0.,eVj:0.,eRc:0.,eIc:0.,eus:0.,egs:0.,ers:0.,eis:0.,ezs:0., n:0, m:0}, n)

test = file_test(filename)
if test eq 0 then begin
  logprint, ' WARNING - ', filename, ' not found.'
  return, -1
endif
 
openr, 1, filename

iline = ''
l = 0l
while not eof(1) do begin
   readf, 1, iline
   if strmid(iline,0,1) eq '#' then continue
   if strlen(iline) lt 120 then continue

   dat = ['',strsplit(strmid(iline,17),/extract)]
   sdss2[l].name = clip(strmid(iline,0,16))
   sdss2[l].rastr = dat[1]
   sdss2[l].decstr = dat[2]
   sdss2[l].ra = 15*ten(sdss2[l].rastr)
   sdss2[l].dec = ten(sdss2[l].decstr)
   r = float(dat[3])
   g = r+float(dat[5])
   u = g+float(dat[4])
   i = r-float(dat[6])
   z = i-float(dat[7])
   sdss2[l].n  = float(dat[15])
  ; sdss2[l].m  = fix(strmid(iline,84,3))
   r_unc = float(dat[8])
   g_unc = sqrt(r_unc^2 + float(dat[10])^2)
   u_unc = sqrt(g_unc^2 + float(dat[9])^2)  
   i_unc = sqrt(r_unc^2 + float(dat[11])^2)
   z_unc = sqrt(i_unc^2 + float(dat[12])^2)

   sdss2[l].us = u 
   sdss2[l].gs = g
   sdss2[l].rs = r
   sdss2[l].is = i
   sdss2[l].zs = z

   if keyword_set(liverpool) then begin
      sdss2[l].us -= 0.031*((u-g)-(1.156)) ; Smith & Steele 
      sdss2[l].gs += 0.057*((g-r)-(0.596))
      sdss2[l].rs -= 0.024*((g-r)-(0.596))
      sdss2[l].is += 0
      sdss2[l].zs -= 0.237*((i-z)-(0.124))
   endif

   sdss2[l].eus = u_unc
   sdss2[l].egs = g_unc
   sdss2[l].ers = r_unc
   sdss2[l].eis = i_unc
   sdss2[l].ezs = z_unc

   sdss2[l].Uj =  u - 0.0216*(u - g) - 0.799 ; combination of 2 equations, fortunately not needed

   sdss2[l].Bj = (u - 0.8116*(u - g) + 0.1313)*0+$;  sigma = 0.0095
                 (g + 0.3130*(g - r) + 0.2271)*1;    sigma = 0.0107

   sdss2[l].Vj = (g - 0.2906*(u - g) + 0.0885)*0+$;  sigma = 0.0129
                 (g - 0.5784*(g - r) - 0.0038)*1;    sigma = 0.0054

   sdss2[l].Rc = (r - 0.1837*(g - r) - 0.0971)/2+$;  sigma = 0.0106
                 (r - 0.2936*(r - i) - 0.1439)/2;    sigma = 0.0072

   sdss2[l].Ic = (r - 1.2444*(r - i) - 0.3820)*1+$;  sigma = 0.0078
                 (i - 0.3780*(i - z) - 0.3974)*0;    sigma = 0.0063

   sdss2[l].eUj =  u_unc
   sdss2[l].eBj = sqrt((g_unc * 1.3130)^2 + (r_unc*0.2271)^2 + 0.0107^2)
   sdss2[l].eVj = sqrt((g_unc*(1-0.5784))^2 + (r_unc*0.5784)^2 + 0.0054^2)
   sdss2[l].eRc = sqrt(r_unc^2 + 0.0106^2)
   sdss2[l].eIc = sqrt((r_unc*0.2444)^2 + (i_unc*1.2444)^2 + 0.0078^2)

   l = l + 1

endwhile
close, 1

; --------- extended star list --------

nx = 1627
sdss2x =  replicate({name:'', ra:0.D, dec:0.D, rastr:'', decstr:'', Uj:0., Bj:0., Vj:0., Rc:0., Ic:0., us:0., gs:0., rs:0., is:0., zs:0., $
                                                                 euj:0.,eBj:0.,eVj:0.,eRc:0.,eIc:0.,eus:0.,egs:0.,ers:0.,eis:0.,ezs:0., n:0, m:0}, nx)

test = file_test(filename)
if test eq 0 then begin
  logprint, ' WARNING - SDSS extended standards catalog (', filename, ') not found.'
  return, sdss2  
endif else begin
 
 openr, 1, filenamex
 
 iline = ''
 l = 0l
 while not eof(1) do begin
   readf, 1, iline
   if strmid(iline,0,1) eq '#' then continue
   if strlen(iline) lt 120 then continue

   dat = strsplit(iline,/extract)

   ra = float(dat[1])
   dec = float(dat[2])
   cosdec = cos(!pi*dec/180.)
   dist = 3600.*( abs(ra-sdss2.ra)*cosdec + abs(dec-sdss2.dec))  ; approximate
   if min(dist) lt 3. then continue ; avoid duplicates

   sdss2x[l].name = clip(dat[0])
   sdss2x[l].rastr = clip(dat[1])
   sdss2x[l].decstr = clip(dat[2])
   sdss2x[l].ra = ra
   sdss2x[l].dec = dec
  

  ; print, sdss2x[l].ra,sdss2x[l].dec
   u = float(dat[3])
   g = float(dat[6])
   r = float(dat[9])
   i = float(dat[12])
   z = float(dat[15])
  ; sdss2[l].n  = float(dat[15])
  ; sdss2[l].m  = fix(strmid(iline,84,3))
   u_unc = float(dat[4])
   g_unc = float(dat[7])
   r_unc = float(dat[10])
   i_unc = float(dat[13])
   z_unc = float(dat[16])

   sdss2x[l].us = u
   sdss2x[l].gs = g
   sdss2x[l].rs = r
   sdss2x[l].is = i
   sdss2x[l].zs = z

   if keyword_set(liverpool) then begin
      sdss2x[l].us -= 0.031*((u-g)-(1.156)) ; Smith & Steele 
      sdss2x[l].gs += 0.057*((g-r)-(0.596))
      sdss2x[l].rs -= 0.024*((g-r)-(0.596))
      sdss2x[l].is += 0
      sdss2x[l].zs -= 0.237*((i-z)-(0.124))
   endif

   sdss2x[l].eus = u_unc
   sdss2x[l].egs = g_unc
   sdss2x[l].ers = r_unc
   sdss2x[l].eis = i_unc
   sdss2x[l].ezs = z_unc

   sdss2x[l].Uj =  u - 0.0216*(u - g) - 0.799 ; combination of 2 equations, fortunately not needed

   sdss2x[l].Bj = (u - 0.8116*(u - g) + 0.1313)*0+$;  sigma = 0.0095
                 (g + 0.3130*(g - r) + 0.2271)*1;    sigma = 0.0107

   sdss2x[l].Vj = (g - 0.2906*(u - g) + 0.0885)*0+$;  sigma = 0.0129
                 (g - 0.5784*(g - r) - 0.0038)*1;    sigma = 0.0054

   sdss2x[l].Rc = (r - 0.1837*(g - r) - 0.0971)/2+$;  sigma = 0.0106
                 (r - 0.2936*(r - i) - 0.1439)/2;    sigma = 0.0072

   sdss2x[l].Ic = (r - 1.2444*(r - i) - 0.3820)*1+$;  sigma = 0.0078
                 (i - 0.3780*(i - z) - 0.3974)*0;    sigma = 0.0063

   sdss2x[l].eUj =  u_unc
   sdss2x[l].eBj = sqrt((g_unc * 1.3130)^2 + (r_unc*0.2271)^2 + 0.0107^2)
   sdss2x[l].eVj = sqrt((g_unc*(1-0.5784))^2 + (r_unc*0.5784)^2 + 0.0054^2)
   sdss2x[l].eRc = sqrt(r_unc^2 + 0.0106^2)
   sdss2x[l].eIc = sqrt((r_unc*0.2444)^2 + (i_unc*1.2444)^2 + 0.0078^2)

   l = l + 1
 endwhile
 sdss2x = sdss2x[0:l-1]
   
 close, 1

 sdss2out = [sdss2, sdss2x]
 return, sdss2out
  
endelse


end

