; +
; NAME:
;   solvearcspec
;
; PURPOSE:
;   Fully-automated solution of an arc spectrum.
;
; CALLING SEQUENCE:
;
; INPUTS:
;   farc               The arc/sky spectrum flux(x)
;   reflinewav         List of reference line wavelengths (array)
;   brightrefwav       List of bright reference line wavelengths (array)
;   x0                 Central x value for polynomial fitting, usually nx/2
;   cwav               Estimated wavelength at x0
;                       (Return value will be improved estimate)
;   unccwav            Uncertainty in estimate; scalar or 2-element vector
;   disp               Estimated dispersion
;                       (Return value will be improved estimate)
;   uncdisp            Uncertainty in estimate; scalar or 2-element vector
;                       WARNING: unccwav and uncdisp include physical variation
;                       across chip, so don't make too small!
;   maxsep             Maximum separation between wavelength triplets during
;                       automated solving
;   uncrefwav          Uncertainty in listed line positions.  Does not take
;                       centroiding error in account, that is dealt with 
;                       internally right now (potentially not well?)
;   initsol            Initial solution.  Will skip the automated line search
;                       if this is set and identify lines from this solution
;                       plus a wavelength shift. Set only if you are confident
;                       the >=1st order terms are close to accurate.
;   nofit              Only solve the offset term (e.g. in sky line refinement)
;   nbrightline        Number of lines to use in automated matching.
;   checkplotfile      File to print out verification plots.  Plots to active
;                       device if this is not set.
;   leaveplotopen      Don't close the checkplot, allowing future runs to
;                       add new pages.
;   checkwav           Wavelength range to check for confirming bright lines
;                       are matched (for validation purposes); 2-element array
;   errorcode          Returns a negative number of solution fails entirely.
;
; OUTPUTS:
;   wav                Solved wavelength of each pixel.
;   ndet               Number of lines detected
;   nmatch             Number of "likely" line matches in final solution.
;   ngoodmatch         Number of these matches that meet the uncrefwav tolerance
;   medabsresidual     Median absolute residual on the matches
;
; Last modified 23 Sep 2020
;-

function solvearcspec, farc, wav, refwav=refwav, brightrefwav=brightrefwav, refid=refid, brightrefid=brightrefid, $
       x0=x0, cwav=cwav, unccwav=unccwav, dispersion=dispersion, uncdisp=uncdisp, $
       maxsep=maxsep, uncrefwav=uncrefwav, initsol=initsol, nofit=nofit, nbrightline=nbrightline, $
       ndet=ndet, nmatch=nmatch, ngoodmatch=ngoodmatch, nbrightunmatched=nbrightunmatched, $
       checkplotfile=checkplotfile, medabsresidual=medabsresidual, $
       leaveplotopen=leaveplotopen, minorder=minorder, maxorder=maxorder, maxnudge=maxnudge, $
       outarclistfile=outarclistfile, centroidrad=centroidrad, quiet=quiet, xmin=xmin, checkwav=checkwav, $
       separate=separate, errorcode=errorcode
 
      nx = n_elements(farc)
      pix = findgen(nx)

      if n_elements(x0) eq 0 then x0 = nx/2. ;
      if n_elements(uncdisp) eq 0 then uncdisp  = dispersion*0.1
      if n_elements(unccwav) eq 0 then unccwav = cwav*0.15
      if n_elements(uncrefwav) eq 0 then uncrefwav = 0.3

      disprange = dispersion+[-1.,1.]*abs(uncdisp)
      cwavrange  = cwav + [-1.,1.]*abs(unccwav)     ; to pass to auto-solver
      minwav = cwavrange[0]-(x0- 0)*max(disprange)  ; to crop linelist
      maxwav = cwavrange[1]+(nx-x0)*max(disprange)  ;  "

      wavinrange = where(brightrefwav gt minwav and brightrefwav lt maxwav, ct)
      brightrefwav = brightrefwav[wavinrange]
      if n_elements(brightrefid) gt 0 then brightrefid = brightrefid[wavinrange]
      wavinrange = where(refwav gt minwav and refwav lt maxwav, ct)
      refwav = refwav[wavinrange]
      if n_elements(refid) gt 0 then refid = refid[wavinrange]

      ; default null initial values for return info, in case program returns early
      nmatch = 0
      ngoodmatch = 0
      nbrightunmatched = [0,0]
      medabsresidual = !values.f_nan

      if n_elements(checkplotfile) gt 0 then begin
      if checkplotfile ne '' then begin
         if !d.name eq 'PS' then psclose ; in case a plot was open
         !p.position = 0
         !p.multi = [0, 2, 1]
         if keyword_set(separate) then psopen, repstr(checkplotfile,'.ps','.0.eps'), xsize=7, ysize=5, /inch, /encaps else $
                                       psopen, checkplotfile, xsize=7, ysize=5, /inch

         device, /color
         colors = transpose([[0,0,0],$
                       [170,170,200],$  ; 1 = grey          (reference line)
                       [210,210,220],$  ; 2 = light grey    (faint reference line)
                       [64, 158, 64],$  ; 3 = green         (faint detected line)
                       [112,112, 64],$ 
                       [0,    0,192],$  ; 5 = blue          (strong detected line)
                       [192, 64, 64],$  ; 6 = reddish       (strong reference line)
                       [128, 64, 128]$
                    ])
         tvlct, colors
      endif
      endif


      linex = findlines(farc, farcsub, linestrength=linestrength, centroidrad=centroidrad)
      if n_elements(linex) lt 3 then begin
          ; not many lines -> sky in twilight?  try again.
          linex = findlines(farc, farcsub, linestrength=linestrength, thresh=5.5)
      endif
      if n_elements(linex) eq 1 and linex[0] eq -1 then begin
         ; try again if no lines found
         linex = findlines(farc, farcsub, linestrength=linestrength, thresh=4)
         if n_elements(linex) lt 3 and n_elements(initsol) le 1 then begin
            logprint, '  ERROR: Insufficient lines to solve the arc.'  
            if n_elements(checkplotfile) gt 0 and keyword_set(leaveplotopen) eq 0 and !d.name eq 'PS' then psclose
            errorcode = -1
            return, [0]
         endif
         if n_elements(linex) eq 1 and linex[0] eq -1 then begin
            logprint, '  WARNING: Spectrum has no lines.  Cannot refine initial wavelength solution.'
            arcsol = initsol
         endif
      endif
      if n_elements(xmin) gt 0 then begin
        wxok = where(linex gt xmin, ctok)
        if ctok ge 3 then linex = linex[wxok]
        if ctok ge 3 then linestrength = linestrength[wxok]
      endif

      ndet = n_elements(linex)
      if n_elements(linex) gt 1 then begin
         brightlinex = findbrightest2(linex, linestrength, nbrightline) ; formerly findbrightest
         ;print, brightlinex
         !p.position = 0
         !p.multi = [0, 1, 4]
         nbdat = n_elements(linex) ; this should probably trigger a warning
         if n_elements(minorder) gt 0 then uminorder = minorder < nbdat else delvarx, uminorder
         if n_elements(maxorder) gt 0 then umaxorder = maxorder < nbdat else delvarx, umaxorder
         arcsol = solvearc(linex, brightlinex, refwav, brightrefwav, $ ;brightrefid=brightrefid, 
                            cwavrange=cwavrange, disprange=disprange, maxsep=maxsep, initsol=initsol, nx=nx, uncrefwav=uncrefwav, $
                            nmatch=nmatch, ngoodmatch=ngoodmatch, medabsresidual=medabsresidual, nbrightunmatched=nbrightunmatched, $
                            nofit=nofit, minorder=uminorder, maxorder=umaxorder, maxnudge=maxnudge, silent=quiet, checkwav=checkwav)
         ; initsol = arcsol ; why ...?
      endif else begin
         arcsol = initsol
         xl = 0.03
         xr = 0.99
         !p.multi = 0
         !p.position = [xl, 0.1, xr, 0.5]
      endelse

      plot, pix, farcsub, /xstyle, /ylog, yrange=[10<max(farc)*0.9,max(farc)*1.02], /ystyle
      for l = 0, n_elements(brightlinex)-1 do begin
        ;oplot, [brightlinex[l],brightlinex[l]], [10,1000], color=5, thick=3
        oplot, [brightlinex[l]], [farcsub[fix(brightlinex[l])]], color=5, psym=5, symsize=0.7
      endfor
      for l = 0, n_elements(linex)-1 do begin
        ;oplot, [linex[l],linex[l]], [10,50], color=3
        oplot, [linex[l]], [farcsub[fix(linex[l])]], color=3, psym=5, symsize=0.5
      endfor

      wav = poly(pix-x0,arcsol)
      cwav = arcsol[0]
      disp = arcsol[1]

      ; temporary kludge to prevent awkward double-valued or negative solution in far-UV
      if min(wav) lt 2950 and max(wav) gt 3000 then begin
        dwav3000 = 0
        ; Freeze dispersion at 3000 and extrapolate linearly
        for i = n_elements(wav)-1, 0, -1 do begin
           if wav[i] lt 3000 and dwav3000 eq 0 and i gt 0 then begin
             dwav3000 = wav[i]-wav[i-1]
           endif
           if dwav3000 gt 0 then begin
             wav[i] = wav[i+1] - dwav3000
           endif
        endfor
      endif

      if n_elements(outarclistfile) gt 0 then begin
        openw, 3, outarclistfile
        printf, 3, '# x     wav(x)    strength'
        for i = 0, n_elements(linex)-1 do begin
           printf, 3, linex[i], poly(linex[i]-x0, arcsol), linestrength[i]
        endfor
        close, 3
      endif

      plotarccheck, pix, wav, arcsol, farcsub, $
                    linex, brightlinex, refwav, brightrefwav, $
                    refid=refid, brightrefid=brightrefid, nx=nx, $
                    separate=separate, plotfile=repstr(checkplotfile,'.ps','.eps')

      if n_elements(checkplotfile) gt 0 and keyword_set(leaveplotopen) eq 0 then begin
        psclose
        retitleplot, checkplotfile, removepath(checkplotfile)
      endif

      errorcode = 0
      return, arcsol

end

