;+
; NAME:
;   specwavcal
;
; PURPOSE: 
;   Wavelength-calibrate a 1D spectrum previously extracted by specextract.  
;
; CALLING SEQUENCE: 
;   specwavcal, infile, inh, outfile=outfile, 
;      solvearc=solvearc, arcfile=arcfile, arclist=arclist, brightarclist=brightarclist,
;      /skyrefine, skylist=skylist, skyfile=skyfile,
;      /telrefine, telspec=telspec,
;      cwav=cwav, dispersion=dispersion, unccwav=unccwav, uncdisp=uncdisp, uncrefwav=uncrefwav,
;      initsol=initsol, minorder=minorder, maxorder=maxorder,  maxnudge=maxnudge,
;      success=success, quiet=quiet, setkey=setkey, setval=setval, setcom=setcom
;       
; INPUTS:
;   infile - Filename of spectrum to calibrate, or the spec structure of the spectrum
;   inh - If a spec structure is input, the header
;
; INPUT KEYWORDS:
;   outfile   - Filename to write calibrated spectrum to.  Default 'w'+infile
;   initsol    - Guess at the solution polynomial (for a science spectrum, the arc solution)
;
;      Solution-finding keywords (mostly passed to solvearcspec; see documentation there):
;   /solvearc - (Re)solve the arc solution.  If unset, only refinement will be performed
;   arcfile   -  Filename of arc to use to obtain initial wavelength solution, if infile
;                itself is not the arc.
;   arclist   -  Filename or array containing reference wavelengths of expected arc lines
;   brightarclist - Filename or array containing reference wavelengths of bright arc lines
;   cwav       - Guess at central wavelength
;   dispersion - Guess at the dispersion of the spectrum (Ang/pix)
;   unccwav    - Estimated uncertainty in cwav estimate
;   uncdisp    - Uncertainty in dispersion estimate
;   uncrefwav  - Uncertainty in wavelength measurement accuracy
;   minorder   - Minimum order for wavelength solution
;   maxorder   - Maximum order for wavelength solution
;   maxnudge   - Maximum change in central wavelength possible from initsol
;
;      Solution-refinement keywords:
;   /skyrefine - Use sky lines (from the sky column or sky file) to perform 0th-order
;                refinement to the existing wavelength solution from initsol (or arc sol)
;   skylist   -  Filename containing reference wavelengths of expected sky lines 
;   /telrefine - Use telluric absorption region to refine wavelength solution via crosscorrelation
;   telspec    - Filename containing a reference spectrum of the telluric region(s)
;       
;     Other keywords:
;   /quiet   - Suppress output to console
;   setkey   -   Set additional keywords (keyword names)
;   setval   -   Set additional keywords (values)
;   setcom   -   Set additional keywords (comments)
;
; OUTPUT KEYWORDS:
;   success  -   Returns a negative flag if procedure failed, or 1 on success
;
; OUTPUTS: 
;   The calibrated spectrum is written to disk.
;
; COMMENTS:
;  This is a multi-purpose routine to obtain a wavelength calibration given an input
;  spectrum that is uncalibrated or only approximately calibrated.
; 
;  If given an arc spectrum and /solvearc is set, will (first) obtain a solution 
;  "from scratch" using solvearcspec.  This functionality is not currently used by 
;  LPipe (solvearcspec is called directly without this wrapper).  
;
;  If /skyrefine or /telrefine is set, the higher-order terms of the solution will be
;  taken from initsol (or from the solved arc solution if /solvearc is also set).
;  Refinement of the central wavelength will be performed using the sky spectrum or
;  telluric absorption pattern.
;
;  If none of these options are set, initsol is assumed to be the final solution and it
;  is applied to the spectrum with no further adjustments.
; 
;
; Last modified 18 Sep 2019.  Documentation added 3 July 2022.
;-

pro specwavcal, infile, inh, outfile=outfile,  $ 
   solvearc=solvearc, arcfile=arcfile, arclist=arclist, brightarclist=brightarclist, $
   skyrefine=skyrefine, skylist=skylist, skyfile=skyfile, $
   telrefine=telrefine, telspec=telspec, $
   cwav=cwav, dispersion=dispersion, unccwav=unccwav, uncdisp=uncdisp, uncrefwav=uncrefwav, $
   initsol=initsol, minorder=minorder, maxorder=maxorder,  maxnudge=maxnudge, $
   success=success, quiet=quiet, setkey=setkey, setval=setval, setcom=setcom

 
  ; Wavelength calibrate a 1D spectrum.

  if n_elements(infile) eq 0 then return
  if n_elements(skylist) eq 0 then skylist = ''

  ; Allows spectrum to be passed as an array, for some reason.
  if n_elements(infile) gt 1 then begin
    ; Input as spectrum object
    sp = infile
    if n_elements(inh) gt 0 then h = inh
  endif else begin
    ; Input as file to read
    if file_test(infile) eq 0 then begin
       print, ' File ', infile, ' not found.'
    endif
    path = extractpath(infile)
    if n_elements(outfile) eq 0 then outfile = path+'w'+removepath(infile)
    sp = readspec(infile, h)
    nx = n_elements(sp)
  endelse

  pix = findgen(nx) ; should make cropping adjustment...

  ;  nx = n_elements(inspec)
  ;  sp = replicate({x:0., wav:0., flux:0.}, nx) 
  ;  if keyword_set(solve) then sp.x = inspe
  ;  if keyword_set(skyrefine) or keyword_set(telrefine) then sp.wav = infile
  ;  sp.flux = outfile

  checkplotfile = repstr(outfile,'.spec','.wavsol.ps')

  flextype = 'none'

   ; ------------------ Calibrate Wavelength -------------------

   if keyword_set(solvearc) then begin

      ; Solve from scratch.

      if n_elements(arcfile) gt 0 then begin
         ; arcfile is a spectrum from a separate arc extracted at this CCD position.
         arc = readspec(arcfile)
         farc = arc.flux
      endif else begin
         ; generally we solve the arc alone
         farc = sp.flux
      endelse
    
      if n_elements(brightarclist) eq 1 then $
        brightreflinewav = grabcolumn(brightarclist, 0) $
      else $
        brightreflinewav = brightarclist    

      if n_elements(arclist) eq 1 then $
        morereflinewav = grabcolumn(arclist, 0) $
      else $
        morereflinewav = arclist    

      arcsol = solvearcspec(farc, wav, refwav=morereflinewav, brightrefwav=brightreflinewav, $
       cwav=cwav, unccwav=unccwav, dispersion=dispersion, uncdisp=uncdisp, $
       maxsep=maxsep, minorder=minorder, maxorder=maxorder, uncrefwav=uncrefwav, initsol=initsol, nbrightline=20, $
       nmatch=nmatch, ngoodmatch=ngoodmatch, medabsresidual=medabsresidual, $
       checkplotfile=checkplotfile, maxnudge=maxnudge)
      ; could set /leaveplotopen to keep plot open for sky solution writing.
      ; see the old version of specextract.  Probably better not to.      

      sol = arcsol
      initsol = arcsol
      wav = poly(pix-nx/2., sol)
   endif 


   if keyword_set(skyrefine) then begin

      ; should explore if cross-calibration method works better for this.

      if n_elements(skyfile) gt 0 then begin
         sk = readspec(skyfile)
         fsky = sk.flux
         sp.sky = fsky
      endif else begin
         fsky = sp.sky
      endelse

      if n_elements(fsky) gt 0 then begin
      if skylist ne '' then begin
        if keyword_set(oldskymethod) then begin
          readcol, skylist, idn, allreflinewav, allreflinestr, format='i,f,f', /silent
          thresh = 100.
          if n_elements(cwav) eq 0 then cwav = poly(nx/2,initsol)
          if n_elements(dispersion) eq 0 then dispersion = poly(nx/2,initsol)-poly(nx/2-1,initsol)
          if n_elements(uncdisp) eq 0 then uncdisp = dispersion*0.1
          minwav = cwav-unccwav[0] - (dispersion+max(uncdisp))*nx/2. ; estimation only
          maxwav = cwav+unccwav[0] + (dispersion+max(uncdisp))*nx/2. ; estimation only
          w = where(allreflinestr gt thresh and allreflinewav gt minwav and allreflinewav lt maxwav)
          skylinewav = allreflinewav[w] ; ultimately should edit the file?
          skylinestr = allreflinestr[w]

          nline = 100
          if maxwav lt 7000. then nline = 20
          if maxwav lt 6000. then nline = 5

          brightskywav = findbrightest(skylinewav, skylinestr, (nline/5)>3)   
          moreskywav = findbrightest(skylinewav, skylinestr, nline)

          maxsep = nx*dispersion/2.
          ;logprint, '  Correcting wavelengths using sky lines...'
          nofitsky = 1

          ; Note that solvearcspec doesn't actually do refinement: wav is an OUTPUT and the 
          ; wavelength solution is solved 'from scratch' like an arc.

          skysol = solvearcspec(fsky, wav, refwav=moreskywav, brightrefwav=brightskywav, $
           cwav=cwav, unccwav=unccwav, dispersion=dispersion, uncdisp=uncdisp, $
           maxsep=maxsep, uncrefwav=uncrefwav, initsol=initsol, nofit=nofitsky, nbrightline=20, $
           nmatch=nmatch, ngoodmatch=ngoodmatch, medabsresidual=medabsresidual, $
           checkplotfile=checkplotfile, maxnudge=maxnudge, quiet=quiet)
          nudge = skysol[0]-initsol[0]  ; solely for logging purposes.
          sol = skysol
          if nudge eq 0. then logprint, '  No correction applied.' $
          else flextype = 'skylines'
         
          if n_elements(wav) eq 0 then wav = poly(pix-nx/2., sol) ; only needed if solvearcspec failed?
          ;                        else wav = wav + nudge ; Bad! This would double-correct!  wav already corrected in solvearcspec
        endif else begin ; new simpler algorithm (brute force)
          if n_elements(wav) eq 0 then wav = poly(pix-nx/2., initsol)
          readcol, skylist, idn, allreflinewav, allreflinestr, format='i,f,f', /silent
          thresh = 100. ; for sky lines in the reference catalog
          maxshift = maxnudge
          w = where(allreflinestr gt thresh and allreflinewav gt (min(wav)-maxnudge) and allreflinewav lt max(wav)+maxnudge, nsky)
          skylinewav = allreflinewav[w] ; ultimately should edit the file?
          skylinestr = allreflinestr[w]
          ; maybe should only ever use the 5577 line on the blue as long as it's available
          ; maybe match independently on bright and faint lines and make sure they're consistent

          linex = findlines(fsky, fskysub, linestrength=linestrength, centroidrad=centroidrad) ;, /debugplot
          linewav = poly(linex-nx/2,initsol)
          killline = intarr(n_elements(linex))
          for i = 0, n_elements(linex)-1 do begin
             if i ge 1 then $
                if linestrength[i] gt linestrength[i-1]*10 and abs(linewav[i]-linewav[i-1]) lt 20 then $
                  killline[i-1] = 1
             if i le n_elements(linex)-2 then $
                if linestrength[i] gt linestrength[i+1]*10 and abs(linewav[i]-linewav[i-1]) lt 20 then $
                  killline[i+1] = 1
          endfor
          keep = where(killline eq 0 and linewav gt min(skylinewav)-200, ndet)
          linex = linex[keep]
          linewav = linewav[keep]
          ;ndet = n_elements(linewav)

          if ndet gt 0 and nsky gt 0 then begin
            meddisp = median(wav-shift(wav,1))
            matchthresh = 1.0*meddisp+uncrefwav ; formerly 1.5*meddesp
            wavstep = 0.2*meddisp
            nstep = fix(round((2*maxshift)/wavstep))
            shiftgrid = (indgen(nstep)-nstep/2)*wavstep
            measshifts = fltarr(nstep) + !values.f_nan
            madshifts = fltarr(nstep) + !values.f_nan
            nmatchs = intarr(nstep)
            for s = 0, n_elements(shiftgrid)-1 do begin
              shift = shiftgrid[s]
              linewavshift = linewav + shift
              diff = -1.
              for l = 0, n_elements(linewav)-1 do begin
                 dist = linewavshift[l]-skylinewav
                 wmindist = (where(abs(dist) eq min(abs(dist)), ctm)) [0]
                 if ctm eq 0 then continue
                 if abs(linewavshift[l]-skylinewav[wmindist]) lt matchthresh then begin
                   nmatchs[s] = nmatchs[s] + 1
                   diff = [diff, skylinewav[wmindist]-linewavshift[l]] ; sign convention: additional shift needed.
                 endif
              endfor
              if n_elements(diff) gt 1 then begin
                  diff = diff[1:*]
                  measshifts[s] = shift+median(diff)
                  madshifts[s] = median(abs(diff-median(diff))) ; need an (n-1)/n adjustment
                  ;print, 'sh = '+fpr(shift,4.2)+': '+clip(nmatchs[s],3)+' '+clip(max(nmatchs),3)+' med='+$
                  ;       fpr(median(diff),4.3)+' newsh='+fpr(shift+median(diff),4.3)+' mad='+fpr(median(abs(diff-median(diff))),4.3)
              endif
            endfor

            if max(nmatchs) gt 0 then begin ; a single match is good enough (OK for the blue)
              wgood = where(nmatchs ge (max(nmatchs)*0.9))
              wbest = (where(nmatchs ge max(nmatchs), ct))
              wbest = fix(median(wbest))
              bestmad = madshifts[wbest]
              if max(nmatchs) lt fix(0.70*min([ndet,nsky]-2)) then warnstr = ' - possibly incorrect?' else warnstr = ''
              logprint, '  Matched '+clip(max(nmatchs))+' / ['+clip(n_elements(linewav))+','+clip(n_elements(skylinewav))+'] sky lines  (med.abs.dev. = '+clip(fpr(madshifts[wbest],3.3))+')'+warnstr
              if max(nmatchs) ge fix(0.55*min([ndet,nsky]-2)) then begin
                nudge = median(measshifts[wgood])
                sol = initsol
                sol[0] += nudge
                wav += nudge
                logprint, '  Nudging solution by '+clip(fpr(nudge,5.3))
                flextype = 'skylines'
              endif else begin
                logprint, '  Sky-line adjustment of '+clip(fpr(median(measshifts[wgood]),5.2))+' is questionable; making no change.'
              endelse
            endif else begin
               britelinelist = ' (Brightest lines:'
               for bi = 0, ((n_elements(linewav)-1) < 2) do britelinelist += ' '+clip(fix(round(linewav[bi])))
               if n_elements(linewav) eq 0 then brightlinelist = ' (No bright lines' 
               britelinelist += ')'
               logprint, '  Found no sky line matches. '+britelinelist
            endelse
           
          endif ; lines were found
        endelse ; new method
      endif ; skylist not ''
      endif ; fsky
   endif ; skyrefine

   if n_elements(arcfile) gt 0 or n_elements(skyfile) gt 0 and skylist ne '' then psclose ; close it here

   if keyword_set(telrefine) then begin
      if n_elements(wav) eq 0 then wav = poly(pix-nx/2., initsol)
      ; kind of silly to keep recreating wav from the equation?

      ; should check for wavelength coverage and sufficient signal

      aband = where(wav gt 7530-100 and wav lt 7740+100, cta)
      bband = where(wav gt 6840-100 and wav lt 6910+100, ctb)
      if min(wav) lt 6840 and max(wav) gt 6910 then tel = bband
      if min(wav) lt 7530 and max(wav) gt 7740 then tel = aband
      ;if cta gt 10 then tel = aband
      ;if cta lt 15 and ctb gt 25 then tel = bband
  
      ;set_plot, 'X'
      spnorm = sp[tel].flux / median(sp[tel].flux)
      ;plot, wav[tel], spnorm, /xsty

      if min(convol(spnorm,[1,1,1]/3.,/edge_mirror)) gt 0.75 then begin
         logprint, '  ERROR - telluric region does not seem to contain a telluric line?'
         ;logprint, '  Performing no wavelength flexure correction.'
      endif else begin

        readcol, telspec, telrefwav, telrefabs, format='f,f'

        catch, shifterror

        if shifterror eq 0 then begin
          nudge = ccshift(wav[tel], spnorm, telrefwav, telrefabs, /normdivide, max=maxnudge)
          sol = initsol
          sol[0] += nudge
          wav += nudge
          ;print, nudge
          ;oplot, telrefwav, telrefabs, col=rgb(255,255,0)
          ;oplot, wav[tel], sp[tel].flux/median(sp[tel].flux), col=rgb(0,255,0)
          flextype = 'telluric'
 
          logprint, '  Nudging solution by '+clip(fpr(nudge,5.3))
 
        endif else begin
          logprint, '  ERROR - failure in telluric-based wavelength correction.'
        endelse

        catch, /cancel
     endelse
   endif


   if file_test(checkplotfile) then $
      retitleplot, checkplotfile, removepath(checkplotfile)+' ('+clip(sxpar(h,'TARGNAME'))+')'
   !p.multi = 0

   ; Refinement failed, or wasn't even attempted: just use initsol
   if n_elements(sol) eq 0 and n_elements(initsol) gt 1 then begin
      logprint, '  Performing no refinements to the input wavelength calibration.'
      sol = initsol
      wav = poly(pix-nx/2., initsol)
      success = 3
   endif else begin 
      success = 1
   endelse

   if n_elements(wav) eq 0 then begin
      if n_elements(arcfile) gt 0 and n_elements(arclist) gt 0 or $
         n_elements(skyfile) gt 0 and skylist ne '' then begin
        logprint, '  Unable to calculate wavelength solution.'
        success = -4  ; failed wavelength solution
        return
      endif else begin
        logprint, '  No wavelength calibration information provided; expressing in pixels.'
      endelse
   endif
   if n_elements(nmatch) gt 0 then if ngoodmatch lt fix(0.3*nmatch)-2 then $
      success = 2   ; possible bad wavelength solution

   outh = h
   ;if n_elements(cwav) gt 0 then sxaddpar, outh, 'CWAV', cwav
   ;if n_elements(dispersion) gt 0 then sxaddpar, outh, 'DISPAV', dispersion
   if n_elements(wav) gt 0 then begin ; should always be true now
     sxaddpar, outh, 'CWAV', wav[nx/2]
     sxaddpar, outh, 'DISPAV', wav[nx/2]-wav[nx/2-1]
     sxaddpar, outh, 'NWAV', n_elements(wav)
     sxaddpar, outh, 'MINWAV', min(wav)
     sxaddpar, outh, 'MAXWAV', max(wav)
     sxaddpar, outh, 'WAVORDER', n_elements(sol)
     if n_elements(nmatch) gt 0 then sxaddpar, outh, 'WAVMATCH', nmatch
     if n_elements(nmatchgood) gt 0 then sxaddpar, outh, 'WAVMATCG', nmatchgood
     if n_elements(medabsresidual) gt 0 then sxaddpar, outh, 'WAVRESID', medabsresidual
     if n_elements(nudge) gt 0 then sxaddpar, outh, 'WAVNUDGE', nudge, ' Flexure correction (Ang)'
     if flextype ne 'none' then sxaddpar, outh, 'FLEXMETH', flextype, ' Flexure correction method'
   endif

   for k = 0, n_elements(setkey)-1 do begin
      if setkey[k] ne '' then begin
        if n_elements(setcom) ge k+1 then sxaddpar, outh, setkey[k], setval[k], setcom[k] else $
                                          sxaddpar, outh, setkey[k], setval[k]
      endif
   endfor

   now = tstamp(/ut)
   ;hist = sxpar(outh,'HISTORY')
   ;for hi = 0, n_elements(hist)-1 do begin
   ;  if clip(hist[hi]) ne '' then sxaddhist, hist[hi], outh
   ;endfor
   newhist = 'Processed by specwavcal '+now+' UT'
   sxaddhist, newhist, outh
   ;if success eq 3 then sxaddhist, 'No flexure correction applied!!', outh, /comment

   sp.wav = wav

   logprint, '  Wavelength-calibrated spectrum written to ', removepath(outfile)
   writespec, outfile, sp, outh

end
