;+
; NAME: subtractsky2.pro
;
; PURPOSE: 
;    Subtract sky lines (and continuum) from a 2D spectrum to leave only signal from 
;    objects outside the atmosphere.   Uses blocked medians (along y direction)
;    and interpolates.  Assumes that the sky lines are close to vertical.
;
; CALLING SEQUENCE: 
;    s = subtractsky2(spec, ysamplestep=ysamplestep, discontinuity=discontinuity, sky=sky, 
;                  order=order, debug=debug, sourceprofile=sourceprofile, plotx=plotx,
;                  removegradients=removegradients, outname=outname)
;       
; INPUTS:
;      spec - Image or filename of 2D spectrum to skysubtract
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;      outname - Name of output file.  (default: 's'+filename)
;      ysamplestep - Y-direction size to perform block median subtraction.
;      discontinuity - Location of a discontinuity (chip gap) in the y direction.
;      order - Polynomial order to use for interpolation
;      /plotx - Make a plot of the image profile at a specific x pixel (for debugging)
;      removegradients - Experimental option, do not use   
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;      sky - Image (2D array) containing sky model subtracted from data.
;      sourceprofile - Output profile of the "slit" (y-axis)
;       
; OUTPUTS: 
;      If spec is a filename, written to disk as 's'+filename.  The output file contains
;        an extension with the sky model used.
;      If spec is a 2D array, s is the sky-subtracted spectrum
;
; REQUIREMENTS: 
;
; Last modified 18 Sep 2019.  Documentation added 3 July 2022.
;-

function subtractsky2, spec, ysamplestep=ysamplestep, discontinuity=discontinuity, sky=sky, order=order, debug=debug, sourceprofile=sourceprofile, plotx=plotx, removegradients=removegradients, outname=outname

   ; greatly simplified sky subtractor.
  
   removegrad = keyword_set(removegradients)
 
   inmode =  size(spec, /tn)
   if inmode eq 'STRING' then begin
      filename = spec
      fits_info, filename, /silent, n_ext=n_ext
      img=mrdfits(filename, 0, header, /silent)
      ;if n_ext gt 0 then begin  ; preserve extensions (for now they are still deleted)
      ;   s = size(img)
      ;   f = fltarr(s[1],s[2],n)
      ;   for e = 1, n_ext do begin
      ;      f[*,*,e] = mrdfits(filename, e, he, /silent)
      ;   endfor
      ;endif
      sxaddpar, header, 'BZERO', 0.
   endif else begin
      img = spec
   endelse

   debug = keyword_set(debug)

   dims = size(img, /dimens)
   nx = dims[0]
   ny = dims[1]

   yref1 = [ny/6, ny/2]
   yref2 = [ny/2, 5*ny/6]
   refspec = fltarr(nx)
   for x = 0, nx-1 do begin
        refspec[x] = 0.5 * sigclipmedian(img[x,yref1[0]:yref1[1]]) + $
                     0.5 * sigclipmedian(img[x,yref2[0]:yref2[1]]) 
   endfor

   xblocksizeapprox = 256
   nxspecblock = round(nx*1.0/xblocksizeapprox)
   xspecblock = ceil(nx*1.0/nxspecblock)
   xspecstart =  findgen(nxspecblock)*xspecblock
   xspecend   = (findgen(nxspecblock)*xspecblock+xspecblock-1) < (nx-1)
   continuum = fltarr(nx) + !values.f_nan
   for xi = 0, nxspecblock-1 do begin
      xrefspec = refspec[xspecstart[xi]:xspecend[xi]]
      cut = 4*abs(median(xrefspec)) + 10
      notlines = where(xrefspec lt cut)
      cut = 3*abs(median(xrefspec[notlines]))  + 10
      notlines = where(xrefspec lt cut)
      cut = 1.5*abs(median(xrefspec[notlines])) + 10
      notlines = where(xrefspec lt cut)
      x = xspecstart[xi]
        if refspec[x] lt cut                         and refspec[x+1] lt cut then continuum[x] = median(img[x,*])
      for x = xspecstart[xi]+1, xspecend[xi]-2 do begin
        if refspec[x] lt cut and refspec[x-1] lt cut and refspec[x+1] lt cut then continuum[x] = median(img[x,*])
      endfor
      x =  xspecend[xi]-1
        if refspec[x] lt cut and refspec[x-1] lt cut                         then continuum[x] = median(img[x,*])

      ;xx = indgen(nx)
      ;fit =  where(finite(continuum) and xx ge xspecstart[xi] and xx le xspecend[xi])
      ;contpar = poly_fit(xx[fit], continuum[fit], 2)
      ;nans = where(finite(continuum) eq 0 and xx ge xspecstart[xi] and xx le xspecend[xi])
      ;continuum[nans] = (poly(xx[nans], contpar))
   endfor
;   newcontinuum = continuum ; newcontinuum includes estimates for continuum over the sky lines.  Probably not robust.
;   for x = 0, nx-1 do begin
;       newcontinuum[x] = median(continuum[x - 11 > 0 : x+11 < nx-1])
;       if finite(newcontinuum[x]) eq 0 then newcontinuum[x] = median(continuum[x - 25 > 0 : x+25 < nx-1])
;       if finite(newcontinuum[x]) eq 0 then newcontinuum[x] = median(continuum[x - 55 > 0 : x+55 < nx-1])
;   endfor 

   ;;print, total(finite(continuum))
   contimg = continuum # (1+fltarr(ny))
   contsubimg = img - contimg
   if debug then writefits, 'continuum.fits', contimg, header
   if debug then writefits, 'contsubimg.fits', contsubimg, header

   sourceimg = img * 0.
   slitfluxx = fltarr(nxspecblock,ny)
   for xi = 0, nxspecblock-1 do begin
      sfxi = median( ( (contsubimg[xspecstart[xi]:xspecend[xi],*])), dimension=1) 
                          ;* (xspecend[xi]-xspecstart[xi]+1)
      if removegrad then begin
         ; new in 2016: remove gradients from source image
         ; This was a bad idea - don't use it.
         std = stdev(sfxi[where(finite(sfxi))])
         w = where(finite(sfxi) and sfxi lt median(sfxi, /even)+3.5*std, ct)
         yy = findgen(ny)
         cpar = poly_fit(yy[w],sfxi[w],3) ; fit a polynomial to the full trace
         slitfluxx[xi,*] = sfxi - poly(yy, cpar)
      endif else begin
         slitfluxx[xi,*] = sfxi
      endelse
      ; plot, sfxi[w]
      ; oplot, poly(yy, cpar), color=rgb(255,0,0)
      for x = xspecstart[xi], xspecend[xi] do $
         sourceimg[x,*] = slitfluxx[xi,*]
   endfor
   kernel = (1+fltarr(xspecblock))
   sourceimg = convol(sourceimg,kernel/total(kernel), /EDGE_TRUNCATE)
   if debug then writefits, 'sourceimg.fits', sourceimg, header


   ;plot, slitfluxx[xi,*]
   ;print, mean(slitfluxx[xi,*])
   slitflux = total(slitfluxx, 1)*1.0/nxspecblock    ; slitflux is a profile of the avg (over wavelengths) flux under the slit.
                                                     ; only exists as a convenience to the user for identifying the trace.
   ;plot, slitflux, psym=10  
   ;for xi = 0, nxspecblock-1 do begin
   ;    oplot, slitfluxx[xi,*], linestyle=1
   ;endfor
   ;oplot, slitflux, psym=10  

   if max(slitflux) eq min(slitflux) then begin
     logprint, ' The spectrum has completely blank columns!  Not sky-subtracting...'
     if inmode eq 'STRING' then return, 0 else return, spec
   endif


   stdslitflux = sigclipstdev(slitflux)
   medslitflux = sigclipmedian(slitflux)
   
   sx = findgen(n_elements(slitflux))
   ok = where(slitflux gt medslitflux-4*stdslitflux and slitflux lt medslitflux+4*stdslitflux, ct)
   if ct ge 5 then begin
      trend = poly_fit(sx[ok], slitflux[ok], 2)
      slitflux = slitflux - poly(sx, trend)      ; remove the linear trend
   endif else begin
      slitflux -= median(slitflux, /even)
   endelse

   stdslitflux = sigclipstdev(slitflux)
   medslitflux = sigclipmedian(slitflux)

   if n_elements(filename) eq 0 then filename = ''
   if stdslitflux eq 0 then begin
      logprint, clip(filename)+': Found large blank regions; cannot calculate image statistics.  Not sky subtracting...'
      if inmode eq 'STRING' then return, 0 else return, spec
   endif

   sourcethreshold = medslitflux+3*stdslitflux
   ;print, sourcethreshold
   sourcezone = where(convol(slitflux,[1,1,1]/3.) ge sourcethreshold)
   nonsourcezone = where(convol(slitflux,[1,1,1]/3.) lt sourcethreshold)
   ;plot, slitflux, psym=10, yrange=[-50,150]
   ;oplot, convol(slitflux,[1,1,1]/3.), linestyle=1
   ;oplot, [0.,10000], [1,1]*medslitflux+3.5*stdslitflux

   ; preliminaries done, now proceed with sky subtraction.


   img = img - sourceimg ; note that this is temporary; we'll add it back at the end.
                         ; not subtracting the continuum right now - it shouldn't matter.


   if debug then writefits, 'sourcesubimg.fits', img, header

   sky = fltarr(nx, ny)

   blocksizeapprox = 36 ;50 ;36
   buffer =  30 ;16  ;12

   if n_elements(order) eq 0 then order = 2;3

   if n_elements(discontinuity) eq 0 then discontinuity = [0, ny] $
                                     else discontinuity = [0, discontinuity, ny]

   if n_elements(plotx) gt 0 then begin
       psopen, 'plotx.ps', xsize=11.5, ysize=8, /inches
   endif

   for d = 0, n_elements(discontinuity)-2 do begin ; change this to big b
     nyd = discontinuity[d+1] - discontinuity[d]
     nyblockapprox = round(nyd * 1.0 / blocksizeapprox)
     ysamplestep = ceil(ny*1.0 / nyblockapprox)
     nyblock = ceil(nyd*1.0/ysamplestep)
     offy   =  discontinuity[d] + ysamplestep*indgen(nyblock)

     for x = 0, nx-1 do begin
       for i = 0, nyblock-1 do begin
         ystart = (offy[i] - buffer ) > 0
         ystop = (offy[i]+ysamplestep-1 + buffer)  < (discontinuity[d+1]-1)

         ystartedit = offy[i]
         ystopedit  = (offy[i]+ysamplestep-1)  < (discontinuity[d+1]-1)

         ;if x eq 0 then begin
         ;   print, clip(ystartedit,4), '-', clip(ystopedit,4), '  (',clip(ystart,4),'-', clip(ystop,4),')'
         ;endif

         minsize = ystop-ystart-1
         nssec = where(nonsourcezone ge ystart and nonsourcezone le ystop, ct)
         iter = 0
         while ct lt minsize do begin
           if ct lt minsize then begin ; expand to make sure we bridge the gap
              if total(nonsourcezone eq ystart) gt 0 and total(nonsourcezone eq ystop) gt 0 then begin
                  ystart = (ystart - 6) > 0
                  ystop  = (ystop + 6) < ny-1
              endif
              if total(nonsourcezone eq ystart) eq 0 then ystart = (ystart - 3) > 0
              if total(nonsourcezone eq ystop)  eq 0 then ystop  = (ystop + 3) < ny-1
           endif
           nssec = where(nonsourcezone ge ystart and nonsourcezone le ystop, ct)
           iter +=1
           if iter gt 1000 then break
         endwhile
         if iter gt 1000 then break
         ysec = nonsourcezone[nssec]
         ;if x eq 1000 then print, clip(i,4), ': ', ysec, '(,',clip(ct),')'

         section = (img[x,ysec]) [*] ;ystart:ystop  ; the (photon) values
         indices = (indgen(ystop+1)) [ysec] ;  actual y pixel indices
 
         sigthresh = [5.0, 3.5, 3.2, 3.2, 3.0]
         orders = [0, 0, (1 < order), (2 < order), order]
         good = where(indices ge 0 and finite(section), ctgood) ; all
         if ctgood le 1 then continue  ; new
         for iter = 0, n_elements(sigthresh)-1 do begin
            par = poly_fit(indices[good], section[good], orders[iter], yfit=yfit)
            residual = section[good] - yfit
 
            if n_elements(residual) le 1 then break
            sig = stdev(residual) 

            model = poly(indices, par) ; model includes even flagged indices

            possgood = where(finite(section) and abs(section-model) lt sigthresh[iter]*sig, ctgood)
            if ctgood gt 3 then good = possgood ; else print, 'Warning: flagged too many points.'

            ;if x eq 3819 then print, clip(ystart,3), '-', clip(ystop,3), '  ', clip(ystartedit,3), '-', clip(ystopedit,3), '  (', clip(iter),')', median(section-model), mean(section-model), clip(n_elements(section)-ctgood,4)
         endfor  
         par = poly_fit(indices[good], section[good], order, yfit=yfit)
         if total(finite(par) eq 0) gt 0 then begin 
              print, x, i, indices[good], section[good]
              stop
         endif
         editindices = (indgen(ystopedit+1)) [ystartedit:ystopedit]
         editmodel = poly(editindices, par)
         sky[x,ystartedit:ystopedit] = editmodel
       endfor
       xsky = sky[x,*]
       sky[x,*] = convol(sky[x,*],transpose(1+fltarr(ysamplestep/3))/(ysamplestep/3), /edge_truncate)

       if n_elements(plotx) gt 0 then if total(x eq plotx) gt 0 then begin
           plot, img[x,0:582], psym=10
           oplot, sky[x,*]
       endif
     endfor
   endfor



  ;plot, img[3819,*], psym=10
  ;oplot, sky[3819,*]

  if debug then writefits, 'sky.fits', sky, header

  sourceprofile=slitflux

  if n_elements(plotx) gt 0 then begin
     psclose
  endif

  outimg = img-sky+sourceimg

  if inmode eq 'STRING' then begin
     hist = 'Processed by subtractsky2 '+tstamp(/ut)+' UT' ;+systime()
     sxaddhist, hist, header

     sourcemax = max(sourceimg[where(finite(sourceimg))])
     sxaddpar, header, 'SOURCMAX', sourcemax, ' Maximum source continuum average'
     skymax = max(sky)
     sxaddpar, header, 'SKYMAX', skymax, ' Maximum source continuum average'
     skymed = median(sky, /even)
     sxaddpar, header, 'SKYMED', skymed, ' Median sky value; subtracted out'
     exptime = sxpar(header,'EXPTIME', count=ct)
     if ct gt 0. then sxaddpar, header, 'SKYRATE', skymed/exptime, ' Median sky rate'

     slashpos = strpos(spec,'/',/reverse_search)
     if n_elements(outname) eq 0 then $
       outname = strmid(spec,0,slashpos+1) + 's' + strmid(spec,slashpos+1)
     logprint, ' Writing subtracted/sky extensions to ', removepath(outname)
     mwrfits, outimg, outname, header, /create, /silent
     mwrfits, sky, outname, /silent               ; extension 1: sky
     ; mwrfits, sourceimg, outname, /silent         ; extension 2: source (useful for tracing later; NOT for any sort of data extraction!)
     return, 0
  endif else begin
     return, outimg
  endelse


end

