;+
; NAME: 
;    writespec
;
; PURPOSE: 
;    Write an IDL spec structure to a corresponding spec file (ASCII-formatted text file 
;    with wavelength, flux, and other columns plus a FITS-like header).
;
; CALLING SEQUENCE: 
;    writespec, filename, spec, h
;       
; INPUTS:
;    filename - Output filename 
;    spec     - The spec IDL structure (see below for notes on format).
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;    h        - Header; a string array of FITS-like header key-value cards.
;       
; OUTPUTS: 
;    The spectrum is written to disk as a file.
;
; COMMENTS:
;    The routines readspec and writespec are meant to replicate the functionality of
;    readfits and writefits but for ASCII data (containing a 1D spectrum) instead of
;    FITS data.  On disk, spectra are stored as text files beginning with a FITS-like
;    header (the first character of all header lines is a #) of keyword-value pairs,
;    followed by a column header line (also indicated by #), followed by the data
;    columns containing the spectrum and associated information.  When loaded into memory,
;    the columns are stored as elements of an IDL structure variable as indicated below:
;      wav  : Wavelength
;      flux : Flux (or counts if uncalibrated)
;      sky  : Sky flux (or counts)
;      unc  : Flux uncertainty estimate
;      pix  : Wavlength pixel location
;      ypix : Spatial centroid pixel
;      resp : Response function applied to convert counts to flux, if available
;      flag : Value between 0 and 1 indicating the fraction of pixel data that was bad
;    The header is stored in a separate variable as a string array or header cards that
;    can be manipulated using sxpar(), sxaddpar(), etc. in exactly the same way as a
;    normal FITS header.
;
; Last modified 18 Sep 2019.  Documentation added 3 July 2022.
;-

pro writespec, filename, spec_in, h

  if n_elements(spec_in) eq 0 then begin
    print, 'No spectral data; not writing file to disk.'
  endif

  openw, 1, filename
  lastkey = 'END'
  for i = 0, n_elements(h)-1 do begin
    printf, 1, '# '+h[i]
    if i ge n_elements(h)-1 then lastkey = (strsplit(h[i],/extract)) [0]
  endfor
  if n_elements(h) gt 0 and lastkey ne 'END' then printf, 1, '# END'

  if size(spec_in, /tn) eq 'STRUCT' then begin
     spec = spec_in
  endif else begin
     nl = n_elements(spec_in) 
     for i = 0, n_elements(spec_in)-1 do printf, 1, spec_in[i]
     close, 1
     return
  endelse

  ; ideally should see if these member variables even exist
  wavyes = total(spec.wav) ne 0 and total(finite(spec.wav)) gt 0
  skyyes = total(spec.sky) ne 0 and total(finite(spec.sky)) gt 0
  uncyes = total(spec.unc) ne 0 and total(finite(spec.unc)) gt 0
  pixyes = total(spec.pix) ne 0 and total(finite(spec.pix)) gt 0
  if tag_exist(spec,'YPIX') then ypixyes = total(spec.ypix) ne 0 and total(finite(spec.ypix)) gt 0 else ypixyes = 0
  if tag_exist(spec,'RESP') then respyes = total(spec.resp) ne 0 and total(finite(spec.resp)) gt 0 else respyes = 0
  if tag_exist(spec,'FLAG') then flagyes = total(spec.flag) ne 0 and total(finite(spec.flag)) gt 0 else flagyes = 0
  if wavyes eq 0 then begin
     ; pixel comes first if no wavelength
     cols=['xpixel', 'flux'] 
     if skyyes then cols = [cols, 'sky_flux']
     if uncyes then cols = [cols, 'flux_unc']
     if ypixyes then cols = [cols,'ypixel']
     if respyes then cols = [cols,'response']
     if flagyes then cols = [cols,'flag']
  endif else begin
     ; wavelength comes first otherwise
     cols=['wavelen','flux']
     if skyyes then cols = [cols, 'sky_flux']
     if uncyes then cols = [cols, 'flux_unc']
     if pixyes then cols = [cols, 'xpixel']
     if ypixyes then cols = [cols,'ypixel']
     if respyes then cols = [cols,'response']
     if flagyes then cols = [cols,'flag']
  endelse

  colheadstr = '## '
  collen = intarr(n_elements(cols))+12
  if pixyes then collen[where(cols eq 'xpixel')] = 7
  if ypixyes then collen[where(cols eq 'ypixel')] = 9
  if flagyes then collen[where(cols eq 'flag')] = 6
  collen[0] -= 3
  for c = 0, n_elements(cols)-1 do colheadstr += rclip(cols[c],collen[c])+' '
  printf, 1, colheadstr
  ; if respyes eq 0 then printf, 1, '## wavelength     flux         sky_flux     flux_unc     xpixel  ypixel'
  ; if respyes eq 1 then printf, 1, '## wavelength     flux         sky_flux     flux_unc     xpixel  ypixel  response'
  for i = 0, n_elements(spec)-1 do begin
    outstr = ''

    for c = 0, n_elements(cols)-1 do begin
       if cols[c] eq 'wavelen'  then outstr += rclip(spec[i].wav,12) + ' '
       if cols[c] eq 'flux'     then outstr += rclip(spec[i].flux,12) + ' '   ; if it's a double precision array this will produce a bad value
       if cols[c] eq 'sky_flux' then outstr += rclip(spec[i].sky,12)+' '
       if cols[c] eq 'flux_unc' then outstr += rclip(spec[i].unc,12)+' '
       if cols[c] eq 'xpixel'   then outstr += rclip(long(spec[i].pix),7)+' '
       if cols[c] eq 'ypixel'   then outstr += rclip(fpr(spec[i].ypix,5.3),9)+' '
       if cols[c] eq 'response' then outstr += rclip(spec[i].resp,12)+' '
       if cols[c] eq 'flag'     then outstr += ' '+clip(clipzero(spec[i].flag),5)+' '
    endfor
    printf, 1, outstr

    ; if respyes eq 0 then printf, 1, fpr(spec[i].wav,6.4) + string(spec[i].flux) + string(spec[i].sky) + string(spec[i].unc) + '  '+clip(long(spec[i].pix),7) + ' '+fpr(spec[i].ypix,5.3)
    ; if respyes eq 1 then printf, 1, fpr(spec[i].wav,6.4) + string(spec[i].flux) + string(spec[i].sky) + string(spec[i].unc) + '  '+clip(long(spec[i].pix),7) + ' '+fpr(spec[i].ypix,5.3) + string(spec[i].resp)
    ;if skyyes eq 0 and pixyes eq 0 then printf, 1, spec[i].wav, spec[i].flux
    ;if skyyes eq 1 and pixyes eq 0 then printf, 1, spec[i].wav, spec[i].flux, spec[i].sky
    ;if skyyes eq 1 and pixyes eq 1 then printf, 1, spec[i].wav, spec[i].flux, spec[i].sky, spec[i].unc
    ;if skyyes eq 1 and pixyes eq 1 then printf, 1, spec[i].wav, spec[i].flux, spec[i].sky, spec[i].pix
  endfor


  close, 1
end
