;+
; NAME:
;       downloadkeckfiles
;
; PURPOSE:
;       Download Keck data directly from a Keck computer.
;       For live reduction of new data as it is being acquired.
;
; CALLING SEQUENCE:
;       downloadkeckfiles, inst=inst, accountnum=accountnum, 
;                          scratchnum=scratchnum, outdir=outdir, 
;                          skipfirewall=skipfirewall
;
; EXPLANATION:
;       Convenience wrapper for opening the Keck firewall and rsync-ing
;       files to a local disk; helpful for performing at-the-telescope
;       reductions.  The user will need to manually enter the usernames
;       and passwords, which will need to be obtained from the support
;       astronomer.
;
; INPUT KEYWORDS:
;       inst - The instrument to download from.  Defaults to LRIS.
;       accountnum - The account number, usually a number from 1 to 25.
;       scratchnum - Scratch disk to download files from.  LRIS will tell
;                    you this when you log in at the start of the night.
;       outdir - Local directory to save to.  Defaults to current working
;                directory.
;       skipfirewall - Don't (re)open the firewall (e.g. because it is already
;                open or because you are reducing at Keck HQ.)
;                The program will skip opening the firewall automatically if 
;                you have already run this routine once within an IDL session. 
;
; OPTIONAL INPUT KEYWORDS:
;       gzip - Compress the files after downloading to save space.
;              Compression won't occur if the program is aborted.
;       camera - Download only red or blue files (specify 'r' or 'b')
;       /confirmed - Specify to skip user confirmation check.
;       destination - Subdirectory to download to; otherwise the current
;              working directory.  Use sortkoadata to sort into nightly
;              subdirectories after the fact.
;       generatedatabase - Update the local copy of the archive database
;              file to include newly publicly-released data.  Will only
;              affect downloads into the current working directory unless
;              the koadatabase.sav file is copied to the lpipe refdata/ 
;              subdirectory.
;
; OUTPUTS:
;       The requested files are downloaded to disk.
; 
; EXAMPLES:
;       downloadkeckfiles, inst='LRIS', account=11, scratch=17, outdir='raw'
;
; NOTES:
;       This is just a wrapper for rsync (+telnet for the fireall).  Existing
;       files will not be overwritten, so you can just run the same command 
;       repeatedly in an open IDL session to keep your local files synched to
;       the observatory.  Has not been tested since at least 2015: if Keck
;       security procedures or directory conventions change this will need to
;       be updated.
;       Only recognizes LRIS, DEIMOS, and MOSFIRE file locations and filenames.
;
;-
;------------------------------------------------------------------------------


pro downloadkeckfiles, inst=inst, accountnum=accountnum, scratchnum=scratchnum, outdir=outdir, skipfirewall=skipfirewall


   if n_elements(inst) eq 0 then inst = 'LRIS'
   ;if inst eq 'LRIS' and n_elements(autocheck) eq 0 then autocheck = 1

   common keckpath, firewallopen, account, scratch

   if n_elements(skipfirewall) gt 0 then firewallopen = 1

   if n_elements(accountnum) gt 0 then account = accountnum
   if n_elements(scratchnum) gt 0 then scratch = scratchnum

   if n_elements(account) eq 0 then print, 'Account number not specified.'
   if n_elements(scratch) eq 0 then print, 'Disk number not specified.'
   if n_elements(account) eq 0 or n_elements(scratch) eq 0 then return

   if n_elements(firewallopen) eq 0 then begin
      print, 'Opening firewall... enter the appropriate ssh username and password.'
      spawn, 'telnet 128.171.99.33 259'
      firewallopen = 1
      print
   endif

   user = strlowcase(inst) + clip(account) 
   server = 'sshserver1.keck.hawaii.edu'
   remotepath = '/h/scratch'+clip(scratch)+'/'+strlowcase(inst)+clip(account)+'/'
  
   if strlowcase(inst) eq 'lris' then filestr = '???????_????.fits'
   if strlowcase(inst) eq 'deimos' then filestr = 'd????_????.fits'
   if strlowcase(inst) eq 'mosfire' then filestr = 'm??????_????.fits'
   if n_elements(filestr) eq 0 then filestr = '*.fits'
   if n_elements(outdir) eq 0 then outdir = './'

   print, 'Enter instrument account password:'
   cmd = 'rsync --ignore-existing -v -e ssh ' + user+'@'+server+':"'+remotepath+filestr+'" '+outdir
   print, cmd
   spawn, cmd

   ;if keyword_set(autocheck) then begin
   ;    print
   ;    print, 'Automatically checking for missing header data...'
   ;    lriscorrectheaders
   ;endif

end

