;+
; NAME:
;      loadobjectposfile 
;
; PURPOSE:
;      Load from disk a file containing saved aperture and trace properties.
;
; CALLING SEQUENCE:
;      result = loadobjectposfile(objectfile, ct=ct)
;
; EXPLANATION:
;      Files like objectpos.txt and userobjectpos.txt are used by LPipe to 
;      store automatically and user-specified apertures for spectrum
;      extraction.  This program loads a file in that format from disk and
;      stores it in a special structure used to give instructions to the
;      spectrum extraction tool.
;      Some information about the format of this file is given in the LPipe
;      documentation.
;
; INPUTS:
;      objectfile - Name of the file to read.
;
; OPTIONAL OUTPUTS:
;      ct - An integer giving the number of apertures read in successfully.
;
; OUTPUTS:
;      Returns a trace/aperture structure.
; 
;-
;------------------------------------------------------------------------------


function loadobjectposfile, objectfile, ct=ct

   ; Needs options to specify the coordinate system.  Currently only handles image coordinates.
   coord = 'image'

   tt = 0
   if file_test(objectfile) gt 0 then begin 
      ntr = countlines(objectfile)
      if ntr eq 0 then begin
        ct = 0
        return, ''
      endif
      tr = replicate({file:'', ymin:0., ymax:0., yfix:-1., aprad:0., bakmin:[0.,0.], bakmax:[0.,0], $ 
                      tracefile:'', tracefileypos:0., apname:'', addsky:0, comment:'', coord:''},ntr)

      openr, 9, objectfile
      inline = ''
      for t = 0, ntr-1 do begin
          readf, 9, inline
          hashpos = strpos(inline,'#')
          if hashpos ge 0 then line = strmid(inline,0,hashpos) else line = inline
          if hashpos gt 0 then comment = strmid(inline,hashpos+1)
          line = clip(line)
          if strlen(line) eq 0 then continue
          if hashpos gt 0 then tr[tt].comment=comment
          tr[tt].coord = coord ; could use mode-change strings, or a special flag on the line
          line = repstr(repstr(line,'(',' '),')','')  ; makes compatible with old format
          inarr = strsplit(line,/extract)
          if n_elements(inarr) lt 3 then continue   ; the radius is now required information
          tr[tt].file = inarr[0]
          if strpos(tr[tt].file,'|') gt 0 then begin
              ininarr = strsplit(tr[tt].file,'|',/extract)
              tr[tt].file = ininarr[0]
              tr[tt].apname = ininarr[1]
          endif
          if strpos(tr[tt].file,'.fits') le 0 then tr[tt].file = tr[tt].file + '.fits'
          tr[tt].yfix = inarr[1]
          tr[tt].aprad = inarr[2]
          if n_elements(inarr) ge 4 then begin
            ; background region, which can be asymmetric:  e.g., 10-15|12-19
            if strpos(inarr[3],'*') gt 0 then begin    ; * means don't use the 2D sky subtraction
              tr[tt].addsky = 1                        ;   when extracting
              inarr[3] = repstr(inarr[3],'*','')
            endif
            if inarr[3] ne '' then begin
              barr = strsplit(inarr[3],'|',/extract)
              if n_elements(barr) eq 1 then barr = [barr[0],barr[0]]
              for b = 0, n_elements(barr)-1 do begin
                 bdarr = strsplit(barr[b],'-',/extract)
                 tr[tt].bakmin[b] = float(bdarr[0])
                 if n_elements(bdarr) gt 1 then tr[tt].bakmax[b] = float(bdarr[1])
              endfor
            endif
          endif
          if n_elements(inarr) ge 5 then begin
            tr[tt].tracefile = inarr[4]
          endif else begin
            tr[tt].tracefile = repstr(tr[tt].file, '.fits','.trace')
          endelse
          tt += 1
      endfor
      close, 9
      if tt ge 1 then tr = tr[0:tt-1] ; ignore unfilled lines
   endif else begin
       tr = ''
   endelse
   ct = tt

   return, tr
end

