;+
; NAME:
;       lrisapertures
;
; PURPOSE:
;       An interactive tool to visually specify spectrum extraction apertures 
;       in LPipe.
;
; CALLING SEQUENCE:
;       lrisapertures, [profilefile, /all] 
;
; EXPLANATION:
;       This is a graphical tool for selecting, changing, or adding apertures
;       along the longslit to extract within LPipe.  Brings up a window showing
;       an object profile plot with aperture locations marked, and allows the 
;       user to change all details of the extraction visually.
;
; OPTIONAL INPUTS:
;       profilefile - Specify to edit just one aperture.  Rarely used: generally
;              this program is called with no inputs.
;
; OPTIONAL INPUT KEYWORDS:
;       /all - Include all files, including bright standards (otherwise skipped
;              under the assumption that automated centering for them is always
;              reliable)
;
; OUTPUTS:
;       Edits (or creates) the userobjectpos.txt file, which stores all user
;       changed to the default apertures.  Note that the pipeline-chosen
;       automatic apertures are always preserved and can be reverted to with
;       the 'default' button (or by deleting userobjectpos.txt).
; 
; NOTES:
;       Designed to be fairly intuitive: most features are best learned by 
;       exploration, but the arrow buttons scroll between objects while the
;       center/width/background buttons control the aperture and background
;       placement. (There are many options for specifying the 
;       background.)
;       Note that it edits one file at a time and does not automatically 
;       modify the red/blue 'pair' associated with that file.  However,
;       the aperture width and background values can be passed from the 
;       active camera to the paired file (or vice versa) with the "copy to" or
;       "copy from" buttons.  These do not change the aperture centers.
;       Edited apertures are NOT saved until the save button is clicked.
;       Adding multiple extraction apertures uses the "new" button and the 
;       numerical for a source is not frequently 
;       tested.
;       The spectral reductions directory is hard-coded as spredux1d/.
;       Button interactivity is coded in an amateurish way that should be
;       improved, but works for now.
;
;  Last update 26 May 2021; very minor update 3 July 2022
;
;-
;------------------------------------------------------------------------------


@saveobjectposfile

pro tvsc2, im, minval, maxval, x0, y0, x1, y1
   ; should also rebin, if necessary
   imx = congrid(im, (x1-x0), (y1-y0))
   imx = (imx > minval) < maxval
   imx = imx-minval
   imx = 255.*imx/(maxval-minval)
   tvscl, imx, x0, y0, xsize=(x1-x0), ysize=(y1-y0), /nan
end

pro drawbox, coords, color=color, fillcolor=fillcolor, norm=norm, data=data, label=label, charsize=charsize, textcolor=textcolor, orient=orient
  ; draws a rectangle in the active window, optionally with a text caption.

  ; coords are: x0 y0 x1 y1, like p.position
  x0 = coords[0]
  y0 = coords[1]
  x1 = coords[2]
  y1 = coords[3]
  
  if n_elements(fillcolor) gt 0 then $
  polyfill, [x0, x1, x1, x0, x0], [y0, y0, y1, y1, y0], color=fillcolor, norm=norm, data=data

  if n_elements(color) gt 0 then $
  plots, [x0, x1, x1, x0, x0], [y0, y0, y1, y1, y0], color=color, norm=norm, data=data
  
  if !p.font eq 0 then xb = 0.003
  if !p.font eq 0 then xm = 0.1
  if !p.font eq -1 then if n_elements(orient) gt 0 then xb = 0.017 else xb =0.00
  if !p.font eq -1 then xm = 0.03
  if n_elements(label) gt 0 then xyouts, x0+(x1-x0)*xm+xb, y0+((y1-y0-0.01) > 0)*0.3*((n_elements(orient) eq 0)*0.7 + 0.3), label, charsize=charsize, norm=norm, data=data, color=textcolor, orient=orient
end

pro drawbutton, coords, state, astate, oncolors=oncolors, offcolors=offcolors, darkcolors=darkcolors, norm=norm, data=data, label=label, charsize=charsize, orient=orient
  ; draws a button.  Basically a wrapper around drawbox that allows all possible colors to be passed along with two binary flags (stats and astate)
  ; indicating how to color:  "on" (state=1), "off" (state=0, astate=1), "inactive"/dark (astate=0)
  ; astate is optional.

  if state then colors = oncolors else colors = offcolors
  if n_elements(astate) gt 0 then if astate eq 0 then colors = darkcolors
  drawbox, coords, color=colors[0], fillcolor=colors[1], textcolor=colors[2], norm=norm, data=data, label=label, charsize=charsize, orient=orient
end

function iswithin, cursorcoords, boxcoords
  x = cursorcoords[0]
  y = cursorcoords[1]
  x0 = boxcoords[0]
  y0 = boxcoords[1]
  x1 = boxcoords[2]
  y1 = boxcoords[3]
  return, (x ge x0 and x le x1 and y ge y0 and y le y1)
end

function isbothwithin, cursorcoords1, cursorcoords2, boxcoords
  return, iswithin(cursorcoords1, boxcoords) and iswithin(cursorcoords2, boxcoords)
end

pro msg, message, color=color
   common message, lastmessage

   if n_elements(lastmessage) gt 0 then $
      if message eq lastmessage then return
   msgpos =  [0.28,0.01, 0.78, 0.038]
   if n_elements(color) eq 0 then color=rgb(200,200,200)
   drawbox, msgpos+[-0.005,-0.005,0,0], /norm, color=rgb(0,0,0), fillcolor=rgb(0,0,0)  ; clear message window
   xyouts, msgpos[0], msgpos[1], /norm, message, color=color
   lastemessage=message
end

pro updatetr, tr, trf

  ; Update the user trace structure (tr) with the edited apertures corresponding to the current object (trf)

  trffile = trf[0].file

  insertpos = -1
  label = strarr(n_elements(tr))
  for i = 0, n_elements(tr)-1 do begin
     ; Identify which entries are from other objects before this one, which are this object, and which are
     ; from other objects after this one.
     if tr[i].file eq '' then continue
     if tr[i].file ne trffile and insertpos eq -1 then label[i] = 'b' ; before
     if tr[i].file eq trffile and insertpos eq -1 then insertpos = i
     if tr[i].file ne trffile and insertpos ne -1 then label[i] = 'a' ; after
  endfor

  before = where(label eq 'b', nbefore)
  after  = where(label eq 'a', nafter)
  if nbefore gt 0 and nafter eq 0 then newtr = [tr[before], trf]
  if nbefore gt 0 and nafter gt 0 then newtr = [tr[before], trf, tr[after]]
  if nbefore eq 0 and nafter gt 0 then newtr = [trf, tr[after]]
  if nbefore eq 0 and nafter eq 0 then newtr = trf

  tr = newtr

end




; -------------------------------------------------------------------------

pro lrisapertures, profilefile, all=all

; The main code - an interactive GUI-based user interfact to specify the extraction aperture(s) for a given target.
; Set profilefile to choose which object to edit.  Otherwise it will find the first file in the current directory.

xsize = 800
ysize = 580
window, 0, retain=2, xsize=xsize, ysize=ysize
viewmode = 'profile'
incmode = 'sci'
if keyword_set(all) then incmode = 'all'
!quiet = 1
lrisversion = getlrisversion(datadir='spredux1d/')


if n_elements(profilefile) gt 0 then begin
  if strpos(profilefile,'.profile') eq 0 then begin
    if strpos(profilefile,'.fits') eq 0 then profilefile = repstr(profilefile,'.fits','.profile')
  endif
  if file_test(profilefile) eq 0 then begin
     print, 'Cannot find profile file ', profilefile
     return
  endif
endif

; Find all the profile files (for << and >> buttons)
allprofilefiles = findfile('*.profile', count=nfiles)
if nfiles eq 0 then  allprofilefiles = findfile('spredux1d/*.profile', count=nfiles)
; if nfiles eq 0 then  allprofilefiles = findfile(path+'/*.profile', count=nfiles)
if nfiles eq 0 then begin
  print, 'No profiles in this directory or in spredux1d/.  Need to run trace step.
endif


; Drop anything that doesn't have an entry in objectpos.txt
objfile = 'objectpos.txt'
if file_test(objfile) eq 0 then begin
    objfile = 'spredux1d/objectpos.txt'
    if file_test(objfile) eq 0 then begin
        print, ' Cannot find objectpos.txt.  Must run trace step before lrisapertures.'
        return
    endif
endif
checktr = loadobjectposfile(objfile, ct=ntr)
checktrfileroot = repstr(checktr.file, '.fits','')
drop = intarr(n_elements(allprofilefiles))
for i = 0, (n_elements(allprofilefiles)-1) do begin
    fileroot = removepath(repstr(allprofilefiles[i],'.profile',''))
    ow = where(checktrfileroot eq fileroot, nobj)
    if nobj eq 0 then begin
       print, ' No objectpos entry for ', allprofilefiles[i]
       drop[i] = 1
    endif
endfor
if total(drop) gt 0 then allprofilefiles = allprofilefiles[where(drop eq 0)]


allfitsfiles = repstr(allprofilefiles,'.profile','.fits')
allhinfo = lrisheaderinfo(allfitsfiles, /silent)

isstd = allhinfo.exptime le 61. and allhinfo.maxsource gt 4000. 
wsci = where(isstd eq 0, nsci)
if nsci gt 0 then sciprofilefiles = allprofilefiles[wsci] else sciprofilefiles = allprofilefiles

if n_elements(profilefile) eq 0 then begin
   if nsci gt 0 then w = min(wsci) else w = 0
   profilefile = allprofilefiles[w]
endif

userobjfile = 'userobjectpos.txt'
autotr = loadobjectposfile(objfile, ct=ntr)
usertr = loadobjectposfile(userobjfile, ct=nutr)
if nutr eq 0 then begin
  usertr = autotr[0]
  usertr[0].file = ''
  usertr[0].tracefile = '' ; create a dummy line
endif

unsavedchangeselsewhere = 0b  ; For notifying the user to save on quitting.
rb = 0b

   ; tr is an array of all apertures from objectpos.txt and userobjectpos.txt
   ; trf is the subset corresponding to the currently-viewed profile
   ; trs is the current aperture (just a shorthand for trf[obj]


start:   ;;; returns to this point when the file is incremented or decremented
           ; (i.e. this is the 'loop' for different files.)
   if n_elements(profilefile) eq 0 then begin
       print, " No profile file specified at command line, and can't find any in this directory"
       return
   endif
   if file_test(profilefile) eq 0 then begin
      print, ' Profile file ', profilefile, ' does not exist!'
   endif
   ; print, 'Loading ', profilefile, ' in interactive window'

   ; get the position of this file in the sequence of all files
   thisfilen = (where(allprofilefiles eq profilefile, nmatchingfiles)) [0]
   thisfilewn = (where(sciprofilefiles eq profilefile, nmatchingfiles)) [0]
   
   ; see if there is a matching fits file with header data
   isfits = allfitsfiles[thisfilen] ne ''
   hinfo = allhinfo[thisfilen]
   rbmatchn = -1
   rbprofilefile = ''
   rbfitsfile = ''
   cam = ''

   if isfits then begin
        pxscale = clip(hinfo.pxscale)   ;sxpar(h,'PXSCALE')
        grname =  clip(hinfo.grname)    ;sxpar(h,'GRNAME')
        fileobjectname = clip(hinfo.target)
        exptime = clipzero(clip(fpr(hinfo.exptime,4.1)))   ;clipzero(clip(fpr(sxpar(h,'EXPTIME',4.1))))
        slitpa =  clipzero(clip(fpr(hinfo.slitpa,4.1)))    ;clipzero(clip(fpr(sxpar(h,'SLITPA',4.1))))
        slit   =  clip(hinfo.slit)    ;clipzero(clip(fpr(sxpar(h,'SLITPA',4.1))))
        fitsfilename = hinfo.filename
        cam = hinfo.camera

        ; also see if we can find a matching red/blue file (crude)
        othercam = where(allhinfo.camera ne hinfo.camera and $
                         ((allhinfo.mjd     gt hinfo.mjd and allhinfo.mjd lt hinfo.mjdstop) or $
                          (allhinfo.mjdstop ge hinfo.mjd and allhinfo.mjdstop le hinfo.mjdstop) or $
                          (allhinfo.mjd le hinfo.mjd and allhinfo.mjdstop ge hinfo.mjdstop)), ctrbmatch )
        if ctrbmatch gt 0 then begin
             hmatch = allhinfo[othercam]
             if ctrbmatch gt 0 then wbest = where(hmatch.exptime eq max(hmatch.exptime)) else wbest = 0
             rbmatchn = othercam[0] ;where(allhinfo[othercam].exptime eq max(allhinfo[othercam].exptime))] 
             rbfitsfile = allfitsfiles[rbmatchn]
             rbprofilefile = repstr(rbfitsfile,'.fits','.profile')
             rbcam = allhinfo[rbmatchn].camera
             if file_test(rbprofilefile) eq 0 then rbprofilefile = ''
        endif else begin 
             if cam eq 'blue' then rbcam = 'red'
             if cam eq 'red' then rbcam = 'blue' ; just for the plot
        endelse

        ;if rbmatch gt 0 then print, '  match is ', allfitsfiles[rbmatchn] else $
        ;                     print, '  no match...'
   endif



   ; Identify this object's entry/ies within the tr database.
   fileroot = removepath(repstr(profilefile,'.profile',''))
   usertrfileroot = repstr(usertr.file, '.fits','')
   autotrfileroot = repstr(autotr.file, '.fits','')

   ; See if this object is in the user database.
   userchanged = 0
   uw = where(usertrfileroot eq fileroot, nobj) 
   if nobj gt 0 then begin
      userchanged = 1
      trf = usertr[uw]
   endif else begin
   ; If not, get it from the auto database.
      aw = where(autotrfileroot eq fileroot, nobj) 
      if nobj eq 0 then begin
         print, 'Cannot find ', fileroot, ' in either objectpos.txt or userobjectpos.txt.  Need to rerun trace step.'
         return ; ???
      endif
      trf = autotr[aw]
   endelse
      
   inittrf = trf ; for undo purposes

   ; Find the RB match in the tr database, too.
   rbfileroot = removepath(repstr(rbprofilefile,'.profile',''))
   rbuw = where(usertrfileroot eq rbfileroot, nrbobj)
   if nrbobj gt 0 then begin
     rbtrf = usertr[rbuw]
   endif else begin
     rbaw = where(autotrfileroot eq rbfileroot, nrbobj)
     if nrbobj gt 0 then rbtrf = autotr[rbaw] else rbtrf = ''
   endelse

   ; Load in the profile file
   readcol, profilefile, yp, prof, comment='#', format='f,f', /silent
   if rbprofilefile ne '' then begin
      readcol, rbprofilefile, rbyp, rbprof, comment='#', format='f,f', /silent
   endif

   ; Set some data needed for plotting.
   !p.font = 0
   plotpos = [0.17,0.28,0.99,0.90]
   rbplotpos = [0.17,0.09,0.99,0.21]

   statpos = [0.01,0.01, 0.27, 0.032]

   ; The below has been updated for LRIS-R4 (and R1)
   np = n_elements(prof)
   maxpgood = fix(0.9*np)
   fullyrange = [0 < min(prof[0:maxpgood]), max(prof[0:maxpgood])*1.1]
   fullxrange = [min(yp), max(yp)]
   yrange = fullyrange
   xrange = fullxrange
   if lrisversion eq 1 or lrisversion eq 4 and cam eq 'red' then begin
       xrange[0] = xrange[1]/2
       yrange = [0 < min(prof[xrange[0]:xrange[1]]), max(prof[xrange[0]:xrange[1]])*1.1]
   endif   
   if rbprofilefile ne '' then begin
     nprb = n_elements(rbprof)
     rbfullxrange = [0, n_elements(rbprof)]
     rbxrange = xrange
     minpgoodrb = fix(0.05*nprb)
     if lrisversion eq 1 or lrisversion eq 4 and cam eq 'red' then begin
       rbxrange = rbfullxrange
     endif
     if lrisversion eq 1 or lrisversion eq 4 and cam eq 'blue' then begin
       rbxrange[0] = [max(rbyp)/2, max(rbyp)]
     endif
     maxpgoodrb = fix(0.9*nprb)
     rbfullyrange = [0 < min(rbprof[minpgoodrb:maxpgoodrb]), max(rbprof[minpgoodrb:maxpgoodrb])*1.1]
     rbyrange = rbfullyrange
   endif

   ; in the future, button state should be a global variable set that stores positions, colors, text, etc., so it can be tested anywhere.

   ; set the locations for all the buttons
   pl = 0.01 ; panel left
   pr = 0.12 ; panel right
   ph = (pl+pr)/2.
   pw = (pr-pl)
   pt1 = pl+pw*1./3
   pt2 = pl+pw*2./3
   pq1 = pl+pw*1./4
   pq3 = pl+pw*3./4
   pd2 = pl+pw*2./10
   pd3 = pl+pw*3./10
   pd7 = pl+pw*7./10
   pd8 = pl+pw*8./10
   ;yy = [0.90, 0.95]
   ;buttonfirst =[pl,  yy[0], pd2-0.001, yy[1]]
   ;buttonprev = [pd2, yy[0], ph-0.001,  yy[1]]
   ;buttonnext = [ph,  yy[0], pd8-0.001, yy[1]]
   ;buttonlast = [pd8, yy[0], pr,  yy[1]]
   ;buttonrb   = [pr+0.001,yy[0], pr+0.033,yy[1]]
   yy = [0.91, 0.95]
   buttonfirst = [pl+0.01,        yy[0], pl+0.01+0.04*1-0.001-0.01, yy[1]]
   buttonprev  = [pl+0.01+0.04*1-0.01, yy[0], pl+0.01+0.04*2-0.001, yy[1]]
   buttonnext  = [pl+0.01+0.04*2, yy[0], pl+0.01+0.04*3-0.001+0.01, yy[1]]
   buttonlast  = [pl+0.01+0.04*3+0.01, yy[0], pl+0.01+0.04*4-0.001, yy[1]]
   buttonrb    = [pl+0.01+0.04*4, yy[0], pl+0.01+0.04*5-0.001, yy[1]]

   ;yy = [0.82, 0.89] & yym = median(yy,/even)
   ;buttonobj0 = [0.025, yy[0], 0.045, yy[1]]
   ;buttonobj1 = [0.046, yy[0], 0.060, yy[1]]
   ;buttonobj2 = [0.061, yy[0], 0.075, yy[1]]
   ;buttonobj3 = [0.076, yy[0], 0.090, yy[1]]
   ;buttonobjnew = [0.091, yy[0], pr,  yym]
   ;buttonobjdel = [0.091, yym, pr,  yy[1]]

   yy = [0.86, 0.90] & yym = median(yy,/even)
   plo = 0.025
   poarr = plo+findgen(5)*(pr-plo)/4.
   buttonobj0 = [poarr[0], yy[0], poarr[1]-0.001, yy[1]]
   buttonobj1 = [poarr[1], yy[0], poarr[2]-0.001, yy[1]]
   buttonobj2 = [poarr[2], yy[0], poarr[3]-0.001, yy[1]]
   buttonobj3 = [poarr[3], yy[0], poarr[4],       yy[1]]
   yy = [0.82, yy[0]-0.001]
   buttonobjnew = [plo, yy[0], plo+(pr-plo)*0.5-0.001,  yy[1]]
   buttonobjdel = [plo+(pr-plo)*0.5, yy[0],   pr,  yy[1]]

   yy = [0.7615, 0.80]
   buttonu = [pl, yy[0], 0.055, yy[1]]
   buttondef = [0.056, yy[0], pr, yy[1]]

   yy = [0.721, 0.76]
   buttons = [pl, yy[0], ph-0.001, yy[1]]
   ;buttonlist = [pt2, yy[0], pr, yy[1]]
   buttonq = [ph, yy[0], pr, yy[1]]

   ;yy = [0.681, 0.72]
   ;buttonq = [pl, yy[0], pr, yy[1]]

   yy[1] = 0.681-0.01 + 0.04
   yy[0] = yy[1]-0.04
   buttonview = [pl, yy[0], pr, yy[1]]
   yy[1] = yy[0]-0.01
   yy[0] = yy[1]-0.08
   yym = median(yy,/even)
   buttonz = [pl, yy[0], 0.055, yy[1]]
   buttonzfull = [0.055, yy[0], pr, yym]
   buttonzauto = [0.055, yym,   pr, yy[1]]

   yy[1] = yy[0]-0.025+0.01
   yy[0] = yy[1]-0.04
   buttonr = [pl,yy[0], pr,yy[1]]
   yy[1] = yy[0]-0.01
   yy[0] = yy[1]-0.06
   button1 = [pl,yy[0], pr,yy[1]]
   yy[1] = yy[0]-0.01
   yy[0] = yy[1]-0.06
   button2 = [pl,yy[0], pr,yy[1]]
   yy[1] = yy[0]
   yy[0] = yy[1]-0.04
   button2l = [pl,  yy[0], ph, yy[1]]
   button2r = [ph,  yy[0], pr, yy[1]]
   yy[1] = yy[0]-0.01
   yy[0] = yy[1]-0.05
   button3   = [pl, yy[0], pr, yy[1]]
   yy[1] = yy[0]
   yy[0] = yy[1]-0.06
   button3i  = [pd3,yy[0], pd7-0.001,yy[1]]
   button3o1 = [pl, yy[0], pd3-0.001,yy[1]]
   button3o2 = [pd7,yy[0], pr, yy[1]]
   yy[1] = yy[0]
   yy[0] = yy[1]-0.06
   button3lo = [pl, yy[0], pq1-0.001,yy[1]]
   button3li = [pq1,yy[0], ph-0.001, yy[1]]
   button3ri = [ph, yy[0], pq3-0.001,yy[1]]
   button3ro = [pq3,yy[0], pr, yy[1]]
   yy[1] = yy[0]-0.0024
   yy[0] = yy[1]-0.03
   button3n  = [pl, yy[0], ph, yy[1]]
   button3sky = [ph, yy[0], pr, yy[1]]

   yy[1] = yy[0]-0.02
   yy[0] = yy[1]-0.035
   buttoncopyrb = [pl, yy[0], pr, yy[1]]
   yy[1] = yy[0]
   yy[0] = yy[1]-0.035
   buttoncopy2rb = [pl, yy[0], pr, yy[1]]
   ;buttoncopyrb = [rbplotpos[0]+0.025, rbplotpos[1]+0.065, rbplotpos[0]+0.15, rbplotpos[1]+0.10]
   ;buttoncopy2rb = [rbplotpos[0]+0.025, rbplotpos[1]+0.03, rbplotpos[0]+0.15, rbplotpos[1]+0.065]

   ; set the color scheme for all the buttons.
   white = rgb(255,255,255)
   grey = rgb(160,160,160)
   black = rgb(0,0,0)
   buttonscolors = [white,0,white]
   buttonsoncolors = [white,rgb(0,0,255),white]
   buttonsoffcolors = [rgb(100,100,100),0,rgb(100,100,100)]
   buttonqoncolors = [white,rgb(255,0,0),white]
   buttonzoncolors = [white, rgb(128,0,128), white]
   buttonzoffcolors = [white, rgb(40,0,40), white]
   button1color = rgb(255,0,0)
   button1fadecolor = rgb(130,0,0)
   button1offcolor = rgb(60,0,0)
   button1oncolors  = [button1color,button1fadecolor,white]
   button1offcolors = [grey,button1offcolor,grey]
   button2color = rgb(0,255,0)
   button2fadecolor = rgb(0,120,0)
   button2offcolor = rgb(0,60,0)
   button2oncolors  = [button2color,button2fadecolor,white]
   button2offcolors = [grey,button2offcolor,grey]
   button3color = rgb(255,255,0)
   button3fadecolor = rgb(100,100,0)
   button3fadecolor2 = rgb(90,90,0)
   button3offcolor = rgb(60,60,0)
   button3oncolors  = [button3color,button3fadecolor,white]
   button3offcolors = [grey,button3offcolor,grey]
   button3oncolors2  = [grey,button3fadecolor2,white]

   buttoncopyrboffcolors = [rgb(200,200,200),black,rgb(200,200,200)]
   if cam eq 'red' then buttoncopyrboncolors  = [white,rgb(0,0,128),white]
   if cam eq 'red' then buttoncopyrboncolors2  = [white,rgb(0,0,255),white]
   if cam eq 'blue' then buttoncopyrboncolors  = [white,rgb(128,0,0),white]
   if cam eq 'blue' then buttoncopyrboncolors2  = [white,rgb(255,0,0),white]

   ; set some default variables
   autoxrange = [-1,-1]
   autoyrange = [-1,-1]
   cursormode = ''
   infitsfile = ''
   if rb eq 0 then begin
     obj = 0
   endif else begin
     if obj gt nobj then obj = 0
   endelse
   newcen = 0b
   redraw = 1b
   quit = 0b
   unsavedchanges = 0b ; only to this object!  Flags that the usertr structure needs to be updated if we change objects.
   stickymsg = 0b
   deletionoccurred = 0

   while quit eq 0b do begin  ; this is the primary loop that iterates every time the user clicks

     if obj ge n_elements(trf) then obj = n_elements(trf)-1

     ; set some additional default variables
     list = 0b
     save = 0b
     prev = 0b
     next = 0b
     first = 0b
     last = 0b
     rb = 0b
     trs = trf[obj]

     ; if needed, completely redraw the active plot (the spatial profile or the 2D spectrum)
     !p.position = plotpos
     if redraw eq 1 then begin
       ; set up the plot
       if viewmode eq 'profile' then begin

          ; The other-camera plot (rb) has to be first.
          !p.position = rbplotpos
          if rbcam eq 'blue' then textcol = rgb(150,150,225)
          if rbcam eq 'red'  then textcol = rgb(215,150,150)
          if rbprofilefile ne '' then begin
            plot, [0], [0], xrange=rbxrange, yrange=rbyrange, xsty=1, /ysty, xtitle='Y pixel'
            oplot, rbyp, rbprof, psym=10, color=rgb(205,205,205)
            for pp = 0, n_elements(rbtrf)-1 do begin
               oplot, rbtrf[pp].yfix*[1,1]-rbtrf[pp].aprad,     [-1e7,1e7], color=button2fadecolor
               oplot, rbtrf[pp].yfix*[1,1]+rbtrf[pp].aprad,     [-1e7,1e7], color=button2fadecolor
               oplot, rbtrf[pp].yfix*[1,1]-rbtrf[pp].bakmax[0], [-1e7,1e7], color=button3fadecolor
               oplot, rbtrf[pp].yfix*[1,1]-rbtrf[pp].bakmin[0], [-1e7,1e7], color=button3fadecolor
               oplot, rbtrf[pp].yfix*[1,1]+rbtrf[pp].bakmin[1], [-1e7,1e7], color=button3fadecolor
               oplot, rbtrf[pp].yfix*[1,1]+rbtrf[pp].bakmax[1], [-1e7,1e7], color=button3fadecolor
               oplot, rbtrf[pp].yfix*[1,1], [-1e7,1e7],               color=button1fadecolor
            endfor
            xyouts, rbplotpos[0]+0.45, rbplotpos[3]-0.03, /norm, removepath(rbprofilefile), color=textcol
            oplot, xrange, yrange[[0,0]]
          endif else begin
            plot, [0], [0], xrange=xrange, yrange=[0,1], xsty=1, /ysty, xtitle=''
            xyouts, rbplotpos[0]+0.45, rbplotpos[3]-0.03, /norm, 'No '+rbcam+' match.', color=textcol
          endelse
          !p.position = plotpos

          ; Display the summed profile
          if isfits then begin ; header info
            plot, [0], [0], xrange=xrange, yrange=yrange, xsty=1+8, /ysty, xtitle='Y pixel', /noerase;, ytitle='Summed spatial profile'
            axis, xaxis=1, xrange=xrange*pxscale, /xsty, xtitle='Y pos (arcsec)     '
          endif else begin
            plot, [0], [0], xrange=xrange, yrange=yrange, xsty=1, /ysty, xtitle='Y pixel', /noerase;, ytitle='Summed spatial profile'
          endelse
          oplot, yp, prof, psym=10

       endif 
       if viewmode eq 'fits' then begin 
         ; Display the 2D spectrum
         if infitsfile ne fitsfilename then begin  ; scrolling between apertures doesn't trigger a reload of the sky.  bug or feature?
            infitsfile = fitsfilename
            fitsim = readfits(fitsfilename, /silent)
            if trs.addsky then skyim = mrdfits(fitsfilename,1, /silent)
            if trs.addsky then fitsim += skyim
            nx = (size(fitsim)) [1]
            ny = (size(fitsim)) [2]
            tracefile = extractpath(fitsfilename)+trs.tracefile
            if tracefile eq '' then tracefile = repstr(fitsfilename,'.fits','.trace')
            if file_test(tracefile) then begin
               openr, 3, tracefile
               inline = ''
               readf, 3, inline
               while eof(3) eq 0 and (strlen(inline) eq 0 or strmid(clip(inline),0,1) eq '#') do readf, 3, inline
               close, 3
               tracepar = float(strsplit(inline,/extract))
               tracey = poly(findgen(nx), tracepar)
               traceyoff = tracey-tracey[nx/2] ; different for B400?
               if grname eq '400/3400' then traceyoff = tracey-tracey[fix(round(nx*0.785))]
               for x = 0, nx-1 do begin
                  fitsim[x,*] = interpol(fitsim[x,*], findgen(ny), findgen(ny)+traceyoff[x])
               endfor
            endif else begin 
               print, ' Cannot find trace file ', tracefile, ' - no trace straightening; position aperture with caution'
               traceyoff = fltarr(nx)
            endelse
         endif
         if grname eq '400/3400' then y0 = ((size(fitsim)) [1])*1/4 else y0 = 0
         y1 = (size(fitsim)) [1]
         zfitsim = fitsim[y0:*,xrange[0]:xrange[1]]
         rebinfactor = 1
         while ysize lt ((y1-y0)/rebinfactor)/2 do rebinfactor *= 2
         zfitsim = rebin(zfitsim, ((size(zfitsim)) [1])/rebinfactor, (size(zfitsim)) [2]) ; bin along spectral direction 
         zfitsim = rotate(zfitsim,4)

         plot, [0], [0], xrange=xrange, xsty=1, /ysty, xtitle='Y pixel', color=0 ; erase the plot
         tvsc2, zfitsim, median(zfitsim)-10., median(zfitsim)+100, !p.position[0]*xsize, !p.position[1]*ysize, !p.position[2]*xsize, !p.position[3]*ysize
         axis, xaxis=0, xrange=xrange, xsty=1+8, xtitle='Y pixel'
         axis, xaxis=1, xrange=xrange*pxscale, /xsty, xtitle='Y pos (arcsec)     '
         axis, yaxis=0, yrange=[y0,y1], /ysty
         axis, yaxis=1, yrange=[y0,y1], /ysty, ytickformat='(A1)'
       endif
       ; Add the info on top of the plot
       ;if isfits then $
       ;xyouts, 0.02, 0.97, /norm, titlestr
       ;xyouts, 0.66, 0.94, /norm, 'tracefile='+trs.tracefile, color=rgb(160,160,160)
       if cam eq 'blue' then textcol = rgb(150,150,225)
       if cam eq 'red'  then textcol = rgb(215,150,150)
       if trs.tracefile ne repstr(removepath(fitsfilename),'.fits','.trace') then tracefilestr = trs.tracefile else tracefilestr = ''
       xyouts, 0.01, 0.97, /norm, 'LPipe aperture tool', color=rgb(200,200,200)
       xyouts, 0.011, 0.97, /norm, 'LPipe aperture tool', color=rgb(200,200,200)
       xyouts, 0.17, 0.97, /norm, fileobjectname, color=rgb(250,250,250)
       xyouts, 0.171, 0.97, /norm, fileobjectname, color=rgb(250,250,250)
       if tracefilestr ne '' then xyouts, 0.66, 0.97, /norm, 'tracefile='+tracefilestr, color=rgb(160,160,160)
       if isfits then begin
         info =                '   gr=' + grname  $
                                + '  exp=' + exptime+'s'  $
                                + '  PA='  + slitpa $ 
                                + '  slit=' +repstr(slit,'long_','')
         xyouts, 0.29, 0.97, /norm, info, color=textcol
       endif

       if viewmode eq 'profile' then $
            xyouts, plotpos[0]+0.45, plotpos[3]-0.04, /norm, removepath(profilefile), color=textcol
       if viewmode eq 'fits' then $
            xyouts, plotpos[0]+0.45, plotpos[3]-0.04, /norm, removepath(fitsfilename), color=textcol
   
       redraw = 0
     endif

     ; Display the message at the bottom, unless a special (sticky) message is already up there
     if stickymsg eq 0 then begin
       if cursormode eq '' then msg, '              '
       if cursormode eq 'c' and newcen eq 0 then msg, 'Select new center position', color=rgb(255,0,0)
       if newcen then msg, 'Select new center position (new aperture)', color=rgb(255,0,0)
       if cursormode eq 'e' then msg, 'Select new aperture edge position', color=rgb(0,255,0)
       if cursormode eq 'el' then msg, 'Select new aperture left edge position', color=rgb(0,255,0)
       if cursormode eq 'er' then msg, 'Select new aperture right edge position', color=rgb(0,255,0)
       if cursormode eq 'bi' then msg, 'Select new background inner edge position', color=rgb(255,255,0)
       if cursormode eq 'bo' then msg, 'Select new background outer edge position', color=rgb(255,255,0)
       if cursormode eq 'blo' then msg, 'Select new background left outer edge position', color=rgb(255,255,0)
       if cursormode eq 'bli' then msg, 'Select new background left inner edge position', color=rgb(255,255,0)
       if cursormode eq 'bri' then msg, 'Select new background right inner edge position', color=rgb(255,255,0)
       if cursormode eq 'bro' then msg, 'Select new background right outer edge position', color=rgb(255,255,0)
       if cursormode eq 'z' then msg, 'Select zoom corners', color=rgb(255,0,255)
     endif
     stickymsg = 0

     newcen = 0b


     mode = cursormode ; just a shorthand for lines below
     m0 = strmid(cursormode,0,1)
     ; Draw the bars for apertures which are not the current aperture
     for pp = 0, n_elements(trf)-1 do begin
        if pp eq obj then continue
        oplot, trf[pp].yfix*[1,1]-trf[pp].aprad,     [-1e7,1e7], color=button2fadecolor
        oplot, trf[pp].yfix*[1,1]+trf[pp].aprad,     [-1e7,1e7], color=button2fadecolor
        oplot, trf[pp].yfix*[1,1]-trf[pp].bakmax[0], [-1e7,1e7], color=button3fadecolor
        oplot, trf[pp].yfix*[1,1]-trf[pp].bakmin[0], [-1e7,1e7], color=button3fadecolor
        oplot, trf[pp].yfix*[1,1]+trf[pp].bakmin[1], [-1e7,1e7], color=button3fadecolor
        oplot, trf[pp].yfix*[1,1]+trf[pp].bakmax[1], [-1e7,1e7], color=button3fadecolor
        oplot, trf[pp].yfix*[1,1], [-1e7,1e7],               color=button1offcolor
     endfor
     ; Draw the bars corresponding to the location of the current aperture:  edges, background, and center
     oplot, trs.yfix*[1,1]-trs.aprad,     [-1e7,1e7], color=button2color+(button2fadecolor-button1color)*(mode eq 'c' or mode eq 'e' or mode eq 'el')
     oplot, trs.yfix*[1,1]+trs.aprad,     [-1e7,1e7], color=button2color+(button2fadecolor-button1color)*(mode eq 'c' or mode eq 'e' or mode eq 'er')
     oplot, trs.yfix*[1,1]-trs.bakmax[0], [-1e7,1e7], color=button3color+(button3fadecolor-button1color)*(mode eq 'c' or mode eq 'bo' or mode eq 'blo')
     oplot, trs.yfix*[1,1]-trs.bakmin[0], [-1e7,1e7], color=button3color+(button3fadecolor-button1color)*(mode eq 'c' or mode eq 'bi' or mode eq 'bli')
     oplot, trs.yfix*[1,1]+trs.bakmin[1], [-1e7,1e7], color=button3color+(button3fadecolor-button1color)*(mode eq 'c' or mode eq 'bi' or mode eq 'bri')
     oplot, trs.yfix*[1,1]+trs.bakmax[1], [-1e7,1e7], color=button3color+(button3fadecolor-button1color)*(mode eq 'c' or mode eq 'bo' or mode eq 'bro')
     oplot, trs.yfix*[1,1], [-1e7,1e7],               color=button1color+(button1fadecolor-button1color)*(mode eq 'c' or mode eq 'el' or mode eq 'er')
     ; Added dotted lines for whatever aperture is being controlled by the current mode
     if mode eq 'c'                 then oplot, trs.yfix*[1,1]         ,      [-1e7,1e7], color=button1color, linestyle=2
     if mode eq 'e' or mode eq 'el' then oplot, trs.yfix*[1,1]-trs.aprad,     [-1e7,1e7], color=button2color, linestyle=2
     if mode eq 'e' or mode eq 'er' then oplot, trs.yfix*[1,1]+trs.aprad,     [-1e7,1e7], color=button2color, linestyle=2
     if mode eq 'bo' or mode eq 'blo' and trs.bakmax[0] gt 0 then  oplot, trs.yfix*[1,1]-trs.bakmax[0], [-1e7,1e7], color=button3color, linestyle=2
     if mode eq 'bi' or mode eq 'bli' and trs.bakmin[0] gt 0 then  oplot, trs.yfix*[1,1]-trs.bakmin[0], [-1e7,1e7], color=button3color, linestyle=2
     if mode eq 'bi' or mode eq 'bri' and trs.bakmin[1] gt 0 then  oplot, trs.yfix*[1,1]+trs.bakmin[1], [-1e7,1e7], color=button3color, linestyle=2
     if mode eq 'bo' or mode eq 'bro' and trs.bakmax[1] gt 0 then  oplot, trs.yfix*[1,1]+trs.bakmax[1], [-1e7,1e7], color=button3color, linestyle=2

     ; Add the numbers for the current aperture stats at bottom left
     outstr = trsoutstring(trf[obj], /nocomment)
     strarr = strsplit(outstr,/extract)
     drawbox, statpos, /norm, color=rgb(0,0,0), fillcolor=rgb(0,0,0)
     xyouts, statpos[0], statpos[1], /norm, strarr[1], color=rgb(150,0,0)
     xyouts, statpos[0]+0.05, statpos[1], /norm, strarr[2], color=rgb(0,140,0)
     xyouts, statpos[0]+0.10, statpos[1], /norm, strarr[3], color=rgb(120,120,0)

     ; Get ready to redraw the buttons

     ; Calculate some variables to know which buttons are highlighted
     if incmode eq 'all' then begin
       firstactive =  thisfilen gt 0 and nfiles gt 1
       prevactive = thisfilen gt 0
       nextactive = thisfilen lt nfiles-1
       lastactive =  thisfilen lt nfiles-1 and nfiles gt 1
     endif else begin ; 'sci'
       firstactive =  thisfilewn gt 0 and nsci gt 1
       prevactive = thisfilewn gt 0
       nextactive = thisfilewn lt nsci-1
       lastactive =  thisfilewn lt nsci-1 and nsci gt 1
     endelse
     rbactive = rbmatchn gt -1
     iszfull = total([xrange,yrange] eq [fullxrange, fullyrange]) eq 4
     iszauto = total([xrange,yrange] eq [autoxrange, autoyrange]) eq 4

     ; Actually draw all the buttons on the left side of the screen (buttons are redrawn every single time...)

     ;drawbutton, buttonu, unsavedchanges, oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm, label='undo changes'
     drawbutton, buttonu, unsavedchanges, oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm, label='undo'
     drawbutton, buttondef, userchanged or unsavedchanges,  oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm, label='default'
     ;drawbutton, buttonlist, 1, oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm, label='list'
     drawbutton, buttons, unsavedchanges or unsavedchangeselsewhere, oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm, label='save'
     drawbutton, buttonq, 1, oncolors=buttonscolors, offcolors=buttonscolors, /norm, label='quit'    
     drawbutton, buttonfirst, firstactive, oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm, label='<<'
     drawbutton, buttonprev, prevactive,   oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm,  label=' <-'
     drawbutton, buttonnext, nextactive,   oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm,  label=' ->'
     drawbutton, buttonlast, lastactive,   oncolors=buttonscolors, offcolors=buttonsoffcolors, /norm, label='>>'
     drawbutton, buttonrb,   1, rbactive,  oncolors=buttonscolors, offcolors=button1colors, darkcolors=buttonsoffcolors, /norm, label='R/B'

     xyouts, 0.009, 0.855, 'ap', /norm
     drawbutton, buttonobj0, obj eq 0, 1,         oncolors=buttonsoncolors, offcolors=buttonscolors, darkcolors=buttonsoffcolors, /norm,  label='0'
     drawbutton, buttonobj1, obj eq 1, nobj gt 1, oncolors=buttonsoncolors, offcolors=buttonscolors, darkcolors=buttonsoffcolors, /norm,  label='1'
     drawbutton, buttonobj2, obj eq 2, nobj gt 2, oncolors=buttonsoncolors, offcolors=buttonscolors, darkcolors=buttonsoffcolors, /norm,  label='2'
     drawbutton, buttonobj3, obj eq 3, nobj gt 3, oncolors=buttonsoncolors, offcolors=buttonscolors, darkcolors=buttonsoffcolors, /norm,  label='3'
     drawbutton, buttonobjnew,      1, nobj le 3, oncolors=buttonscolors,   offcolors=buttonscolors, darkcolors=buttonsoffcolors, /norm,  label='new'
     drawbutton, buttonobjdel, nobj gt 1,         oncolors=buttonscolors,   offcolors=buttonsoffcolors, /norm,  label='del'

     drawbutton, buttonz, cursormode eq 'z',       oncolors=buttonzoncolors, offcolors=buttonzoffcolors, /norm,  label='zoom'
     drawbutton, buttonzfull, iszfull,  oncolors=buttonzoncolors, offcolors=buttonzoffcolors, /norm,  label='full'
     drawbutton, buttonzauto, iszauto,  oncolors=buttonzoncolors, offcolors=buttonzoffcolors, /norm,  label='auto'

     drawbutton, buttonview, 0, isfits,  oncolors=buttonsoncolors, offcolors=buttonscolors, darkcolors=buttonsoffcolors, /norm, label=((viewmode eq 'profile') ? 'view 2d img' : 'view profile')
     drawbutton, buttonr, cursormode eq '', oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label='inspect'

     drawbutton, button1, cursormode eq 'c', oncolors=button1oncolors, offcolors=button1offcolors, /norm, label='center', charsize=1.5
     drawbutton, button2, cursormode eq 'e', oncolors=button2oncolors, offcolors=button2offcolors, /norm, label='width',  charsize=1.5
     drawbutton, button2l, cursormode eq 'el', oncolors=button2oncolors, offcolors=button2offcolors, /norm, label='LW',  charsize=1.5
     drawbutton, button2r, cursormode eq 'er', oncolors=button2oncolors, offcolors=button2offcolors, /norm, label='RW',  charsize=1.5
     drawbutton, button3, strmid(cursormode,0,1) eq 'b', oncolors=button3oncolors, offcolors=button3offcolors, /norm, label='background',  charsize=1.5
     !p.font = -1
     drawbutton, button3i,  cursormode eq 'bi',  oncolors=button3oncolors, offcolors=button3offcolors, /norm, label='inner',charsize=1.1, orient=90
     drawbutton, button3o1, cursormode eq 'bo',  oncolors=button3oncolors, offcolors=button3offcolors, /norm, label='outer',  charsize=1.2, orient=90
     drawbutton, button3o2, cursormode eq 'bo', oncolors=button3oncolors, offcolors=button3offcolors, /norm, label='outer',  charsize=1.2, orient=90
     drawbutton, button3lo, cursormode eq 'blo', oncolors=button3oncolors, offcolors=button3offcolors, /norm, label='LO',   charsize=1.5, orient=90
     drawbutton, button3li, cursormode eq 'bli', oncolors=button3oncolors, offcolors=button3offcolors, /norm, label='LI',   charsize=1.5, orient=90
     drawbutton, button3ri, cursormode eq 'bri', oncolors=button3oncolors, offcolors=button3offcolors, /norm, label='RI',   charsize=1.5, orient=90
     drawbutton, button3ro, cursormode eq 'bro', oncolors=button3oncolors, offcolors=button3offcolors, /norm, label='RO',   charsize=1.5, orient=90
     !p.font = 0
     drawbutton, button3n,  max(trs.bakmin+trs.bakmax) le 0, oncolors=button3oncolors2, offcolors=button3offcolors, /norm, label='none'
     drawbutton, button3sky, trs.addsky eq 0, oncolors=button3oncolors2, offcolors=button3offcolors, /norm, label='2dsky'

     drawbutton, buttoncopyrb, rbprofilefile ne '', oncolors=buttoncopyrboffcolors, offcolors=buttoncopyrboffcolors, /norm, label='copy fr '+rbcam
     drawbutton, buttoncopy2rb, rbprofilefile ne '', oncolors=buttoncopyrboffcolors, offcolors=buttoncopyrboffcolors, /norm, label='copy to '+rbcam

     ; (Re)calculate the automatic ranges (probably could do this elsewhere and less often)
      axc = trs.yfix
      axr = 1.5*2*max([abs(trs.bakmax), 2*abs(trs.aprad), 10])
      autoxrange = [axc-axr, axc+axr]
      w = where(findgen(n_elements(prof)) gt autoxrange[0] and findgen(n_elements(prof)) lt autoxrange[1], ct)
      autoyrange = [min(prof[w]), max(prof[w])*1.05]
      autoxrange[0] = autoxrange[0] > fullxrange[0]
      autoxrange[1] = autoxrange[1] < fullxrange[1]
      autoyrange[0] = autoyrange[0] > fullyrange[0]
      autoyrange[1] = autoyrange[1] < fullyrange[1]


      wait, 0.1  ; this seems to be needed to properly catch inputs
      key = ''
      oldmouse = 0
      mousekey = 0
      simplecursor = 1
      if simplecursor then begin
         cursor, x, y, /norm  ; the program pauses here to collect a mouse click
         datax = xrange[0]+(xrange[1]-xrange[0])*(x-plotpos[0])/(plotpos[2]-plotpos[0])
         datay = yrange[0]+(yrange[1]-yrange[0])*(y-plotpos[1])/(plotpos[3]-plotpos[1])
         downdatax = datax
         downdatay = datay
         downx = x
         downy = y
         mousekey = !mouse.button
      endif else begin
        while key eq '' do begin
          ; This mode allows down/up inputs and also keyboard keypresses, but doesn't seem to work over ssh
          cursor, x, y, /norm, /nowait
          cursor, datax, datay, /data, /nowait
          key = get_kbrd(0)
          wait, 0.01
          if key eq '' then begin ; look for mouse activation
            if !mouse.button ne oldmouse then begin
               if !mouse.button gt 0 then begin ; mouse button is down
                  downx = x
                  downy = y
                  downdatax = datax
                  downdatay = datay
                  mousekey = !mouse.button   ; [1=left, 2=middle, 4=right]
                  if mousekey eq 2 then begin
                     plots, [datax], [datay], psym=1, /data, color=rgb(255,255,0), symsize=2
                  endif
                  ; ideally would colorize the button here - could make that step a function?
               endif else begin                ; mouse button went back up
                  upx = x
                  upy = y
                  updatax = datax
                  updatay = datay
                  break
               endelse
            endif
            oldmouse = !mouse.button
          endif
        endwhile
      endelse

      if key eq 'q' then quit = 1
    
      if mousekey eq 2 then begin
         ; middle key zooms (in simple mode just returns to default)
         if downdatax eq updatax and downdatay eq updatay then begin
           xrange = defaultxrange 
           yrange = defaultyrange
         endif else begin
           xrange = [updatax<downdatax, updatax>downdatax]
           yrange = [updatay<downdatay, updatay>downdatay]
         endelse
         redraw = 1
         continue
      endif
      if mousekey eq 4 then begin
        ; right key cancels
        cursormode = ''
      endif

      ;print, key, mousekey

      ; Keyboard actions, largely deprecated
      if key ne '' then begin
         if key eq 'r' then redraw = 1
         if key eq 'q' then quit = 1
         if key eq 's' then save = 1
      endif
      


      ; Mouse actions
      if mousekey eq 1 then begin
         ; Left keypress takes almost all major actions

         ; Actions off the plot (buttons)
         if iswithin([x,y], plotpos) eq 0 then begin
            ; Save, quit, move to a different file - set a flag and deal with the action later
            ; (many of these have overlapping actions and can be triggered by keyboard presses also)
            if iswithin([x,y], buttons) and iswithin([downx,downy], buttons) then save = 1
            ;if iswithin([x,y], buttonlist) and iswithin([downx,downy], buttonlist) then list = 1
            if iswithin([x,y], buttonq) and iswithin([downx,downy], buttonq) then quit = 1
            if iswithin([x,y], buttonfirst) and iswithin([downx,downy], buttonfirst) and firstactive then first = 1
            if iswithin([x,y], buttonprev) and iswithin([downx,downy], buttonprev) and prevactive then prev = 1
            if iswithin([x,y], buttonnext) and iswithin([downx,downy], buttonnext) and nextactive then next = 1
            if iswithin([x,y], buttonlast) and iswithin([downx,downy], buttonlast) and lastactive then last = 1
            if iswithin([x,y], buttonrb) and iswithin([downx,downy], buttonrb) and rbactive then rb = 1
 
            ; Select different object apertures on this same profile file
            oldobj = obj
            if iswithin([x,y], buttonobj0) then obj = 0
            if iswithin([x,y], buttonobj1) and nobj gt 1 then obj = 1
            if iswithin([x,y], buttonobj2) and nobj gt 2 then obj = 2
            if iswithin([x,y], buttonobj3) and nobj gt 3 then obj = 3
            if iswithin([x,y], buttonobjnew) and nobj le 3 then begin
              newcen = 1
              obj = nobj
              nobj += 1
              newtr = trs ; copy the structure
              newtr.yfix = median(xrange, /even)
              newtr.bakmin = max(newtr.aprad)+3
              newtr.bakmax = newtr.bakmin+10
              newtr.apname = clip(obj)
              trf = [trf, newtr]
              drawbutton, buttonobjnew, 1,   oncolors=buttonsoncolors, offcolors=buttonsoffcolors, /norm,  label='new'
              ; msg, 'Creating new aperture: please position centroid', color=rgb(255,0,0)
              wait, 0.2 
              unsavedchanges = 1
              cursormode = 'c'
            endif
            if iswithin([x,y], buttonobjdel) and nobj gt 1 then begin
              ; delete this aperture
              oo = indgen(n_elements(trf))
              trf = trf[where(oo ne obj)]
              nobj -= 1
              if nobj gt 1 then trf.apname = ['', clip(1+indgen(nobj-1))] else trf.apname = '' ; renumber (this also flushes all names...)
              obj = (obj - 1) > 0
              drawbutton, buttonobjdel, 1,   oncolors=buttonqoncolors, offcolors=buttonsoffcolors, /norm,  label='new'
              deletionoccurred = 1 ; no longer needed it seems
              wait, 0.2 
              unsavedchanges = 1
              redraw = 1
            endif

            if obj ne oldobj then redraw = 1
            ; if obj ne oldobj then trs = trf[obj]

            ; Zoom
            if iswithin([x,y], buttonz) and iswithin([downx,downy], buttonz) then cursormode = 'z0'
            if iswithin([x,y], buttonzfull) and iswithin([downx,downy], buttonzfull) then begin
               drawbutton, buttonzfull, 1,   oncolors=buttonzoncolors, offcolors=buttonzoffcolors, /norm,  label='full'
               drawbutton, buttonzauto, 0,   oncolors=buttonzoncolors, offcolors=buttonzoffcolors, /norm,  label='auto'
               xrange = fullxrange & yrange = fullyrange
               wait, 0.2
               redraw = 1
            endif
            if iswithin([x,y], buttonzauto) and iswithin([downx,downy], buttonzauto) then begin
               drawbutton, buttonzfull, 0,   oncolors=buttonzoncolors, offcolors=buttonzoffcolors, /norm,  label='full'
               drawbutton, buttonzauto, 1,   oncolors=buttonzoncolors, offcolors=buttonzoffcolors, /norm,  label='auto'
               xrange = autoxrange & yrange = autoyrange
               drawbox, [xrange[0], yrange[0], xrange[1], yrange[1]], /data, color=rgb(255,0,255)
               wait, 0.35 
               redraw = 1
            endif

            ; Change mode
            if iswithin([x,y], buttonr) and iswithin([downx,downy], buttonr) then begin
               cursormode = ''
               redraw = 1
            endif
            if iswithin([x,y], button1) then cursormode = 'c' ; center
            if iswithin([x,y], button2) then cursormode = 'e' ; aperture edges
            if iswithin([x,y], button2l) then cursormode = 'el' 
            if iswithin([x,y], button2r) then cursormode = 'er' 
            if iswithin([x,y], button3) and strmid(cursormode,0,1) ne 'b' then cursormode = 'bi' ; background
            if iswithin([x,y], button3o1) then cursormode = 'bo'
            if iswithin([x,y], button3o2) then cursormode = 'bo'
            if iswithin([x,y], button3i)  then cursormode = 'bi'
            if iswithin([x,y], button3lo) then cursormode = 'blo'
            if iswithin([x,y], button3li) then cursormode = 'bli'
            if iswithin([x,y], button3ri) then cursormode = 'bri'
            if iswithin([x,y], button3ro) then cursormode = 'bro'
  
            ; Undo changes - this is really more like 'restore' since it undoes ALL the changes since arriving at this object.
            if iswithin([x,y], buttonu) and iswithin([downx,downy], buttonu) and unsavedchanges then begin
               drawbutton, buttonu, 1,   oncolors=buttonsoncolors, offcolors=buttonsoffcolors, /norm,  label='undo'
               wait, 0.2
               trs = inittrf[obj]
               trf[obj] = trs
               redraw = 1
               unsavedchanges = 0
            endif

            ; Restore automatic defaults
            if iswithin([x,y], buttondef) and iswithin([downx,downy], buttondef) then begin
               drawbutton, buttondef, 1, oncolors=buttonsoncolors, offcolors=buttonsoffcolors, /norm,  label='default'
               wait, 0.2

               ; restore the defaults
               autotrfileroot = repstr(autotr.file, '.fits','')
               aow = where(autotrfileroot eq fileroot, nobj)
               trf = autotr[aow]
               trs = trf[0]
               obj = 0

               ; clear this object from the user data structure
               usertrfileroot = repstr(usertr.file, '.fits','')
               unow = where(usertrfileroot ne fileroot, nuobj) ; not the current file
               if nuobj gt 0 then begin
                  usertr = usertr[unow]
               endif else begin
                  usertr = autotr[0] ; this would empty the file so create a dummy line
                  usertr[0].file = ''
                  usertr[0].tracefile = ''
               endelse
               usertrfileroot = repstr(usertr.file, '.fits','')

               userchanged = 0
               unsavedchanges = 0 ; usertr structure is up to date.  
                                  ; This turns off unduing, although inittr still holds the old state.
               unsavedchangeselsewhere = 1 ; changes do still need to be saved to disk.
               redraw = 1
            endif


            ; Eliminate background
            if iswithin([x,y], button3n) and iswithin([downx,downy], button3n) then begin
               drawbutton, button3n, 1,   oncolors=button3oncolors, offcolors=button3offcolors, /norm,  label='none'
               if total(trs.bakmax+trs.bakmin) ne 0 then begin
                  msg, 'Background deleted.', color=rgb(255,255,0)
                  trs.bakmax = 0
                  trs.bakmin = 0
               endif else begin
                  msg, 'Background apertures created.', color=rgb(255,255,0)
                  trs.bakmin = trs.aprad*2
                  trs.bakmax = trs.aprad*4
               endelse
               wait, 0.45
               redraw = 1
               unsavedchanges = 1
               trf[obj] = trs
            endif

            ; Toggle readdition of sky
            if iswithin([x,y], button3sky) and iswithin([downx,downy], button3sky) then begin
               drawbutton, button3sky, 1,   oncolors=button3oncolors2, offcolors=button3offcolors, /norm,  label='2dsky'
               trs.addsky = ~ trs.addsky
               msg, 'Toggling use of 2D sky subtraction to '+(trs.addsky ? 'OFF' : 'ON'), color=rgb(255,255,0)
               wait, 0.45
               unsavedchanges = 1
               trf[obj] = trs
               infitsfile = ''  ; trigger redraw of sky if necessary
               if viewmode eq 'fits' then redraw = 1
            endif

            ; Copy apertures from the RB match
            if rbprofilefile ne '' then begin
            if iswithin([x,y], buttoncopyrb) and iswithin([downx,downy], buttoncopyrb) then begin
               drawbutton, buttoncopyrb, 1, oncolors=buttoncopyrboncolors2, offcolors=buttoncopyrboffcolors, /norm,  label='copy from '+rbcam
               msg, 'Copying aperture size info from '+rbcam+' pair', color=buttoncopyrboncolors2
               wait, 0.45
               unsavedchanges = 1

               if n_elements(rbtrf) gt nobj then begin
                 while n_elements(rbtrf) gt n_elements(trf) do trf = [trf, trs] ; grow the structure
               endif
               if n_elements(rbtrf) lt nobj then begin
                 trf = trf[0:n_elements(rbtrf)-1]
               endif
               nobj = n_elements(rbtrf)
               for pp = 0, nobj-1 do begin
                 trf[pp].yfix = trf[0].yfix+(rbtrf[pp].yfix-rbtrf[0].yfix) ; all these should be scaled by binning ratio
                 trf[pp].aprad  = rbtrf[pp].aprad
                 trf[pp].bakmin = rbtrf[pp].bakmin
                 trf[pp].bakmax = rbtrf[pp].bakmax
                 trf[pp].apname = rbtrf[pp].apname
               endfor
               if obj ge nobj then obj = 0
               trs = trf[obj]
               redraw = 1
            endif
            ; Copy apertures to the RB match
            if iswithin([x,y], buttoncopy2rb) and iswithin([downx,downy], buttoncopy2rb) then begin
               drawbutton, buttoncopy2rb, 1, oncolors=buttoncopyrboncolors2, offcolors=buttoncopyrboffcolors, /norm,  label='copy to '+rbcam
               msg, 'Copying aperture size info to '+rbcam+' pair', color=buttoncopyrboncolors2
               wait, 0.45
               unsavedchangeselsewhere = 1

               rbtrf = replicate(rbtrf[0], nobj)
               for pp = 0, nobj-1 do begin
                 rbtrf[pp].yfix = rbtrf[0].yfix+(trf[pp].yfix-trf[0].yfix) ; all these should be scaled by binning ratio
                 rbtrf[pp].aprad  = trf[pp].aprad
                 rbtrf[pp].bakmin = trf[pp].bakmin
                 rbtrf[pp].bakmax = trf[pp].bakmax
                 rbtrf[pp].apname = trf[pp].apname
               endfor

               updatetr, usertr, rbtrf

               redraw = 1
            endif
            endif           

            mode = cursormode
            m0 = strmid(cursormode,0,1)

            ; Switch to FITS view
            if iswithin([x,y], buttonview) and iswithin([downx,downy], buttonview) and isfits then begin
               drawbutton, buttonview, 1,   oncolors=buttonsoncolors, offcolors=buttonsoffcolors, /norm, label=((viewmode eq 'profile') ? 'view 2d img' : 'view profile')
               wait, 0.15
               if viewmode eq 'profile' then viewmode = 'fits' else viewmode = 'profile'
               redraw = 1
            endif

         endif

         ; Actions on the plot
         if iswithin([x,y], plotpos) then begin
           oldtrs = trs

           if cursormode eq '' then begin
            ; in principle the inspectmode could do something, but for now it doesn't.
            ; Maybe it could select the aperture or mode via the screen.
            ; The below just prints out the cursor location.
            ; plots, [x], [y], psym=1, /norm, color=rgb(0,0,255)
            ; drawbox, [0.87,0.87, 0.97,0.99], color=rgb(220,220,220), fillcolor=0, /norm
            ; xyouts, 0.88, 0.95, /norm, 'x='+clip(fpr(datax,5.2)), charsize=1.3, color=rgb(188,188,255)
            ; xyouts, 0.88, 0.91, /norm, 'y='+clip(fpr(datay,5.2)), charsize=1.3, color=rgb(188,188,255)
            ; need a better functionality here
           endif

           if cursormode ne '' and strmid(cursormode,0,1) ne 'z' then unsavedchanges = 1b

           if cursormode eq 'z' then begin
              ; Set zoom corners
              nc = n_elements(zoomcoords)
              if nc lt 2 then begin
                   zoomcoords = [datax, datay] 
                   plots, [datax], [datay], psym=1, /data, color=rgb(255,0,255), symsize=2
                   msg, 'Click the other corner of the zoom-in region', color=rgb(255,0,255)
                   stickymsg = 1
              endif
              if nc eq 2 then zoomcoords = [zoomcoords, datax, datay]
              if n_elements(zoomcoords) eq 4 then begin
                   drawbox, zoomcoords, /data, color=rgb(255,0,255)
                   wait, 0.35 
                   xrange = zoomcoords[[0,2]]
                   xrange = xrange[sort(xrange)]
                   yrange = zoomcoords[[1,3]]
                   yrange = yrange[sort(yrange)]
                   redraw = 1
                   zoomcoords = [0]
                   cursormode = ''
              endif
           endif

           ; Change aperture locations

           ; Center and profile edge are easy to calculate
           if cursormode eq 'c' then begin
             trs.yfix = datax
             msg, 'Setting centroid position to '+clipzero(trs.yfix), color=rgb(255,0,0)
             stickymsg = 1
           endif
           if cursormode eq 'e' then begin
             trs.aprad = abs(datax-trs.yfix)
             msg, 'Setting aperture half-width to '+clipzero(trs.aprad), color=rgb(0,255,0)
             stickymsg = 1
           endif
           ; Specific profile edge
           if cursormode eq 'el' or cursormode eq 'er' then begin             
             if datax lt trs.yfix then begin
               right = (trs.yfix+trs.aprad)
               left = datax
             endif else begin
               left = (trs.yfix-trs.aprad)
               right = datax
             endelse
             cshift = (right+left)/2. - trs.yfix
             trs.bakmin[0] += cshift
             trs.bakmax[0] += cshift
             trs.bakmin[1] -= cshift
             trs.bakmax[1] -= cshift
             trs.yfix = (right+left)/2.
             trs.aprad =(right-left)/2.
             msg, 'Setting aperture edges to '+clipzero(trs.yfix-trs.aprad)+','+clipzero(trs.yfix+trs.aprad), color=rgb(0,255,0)
             stickymsg = 1
           endif
           ; Background location
           if strmid(cursormode,0,1) eq 'b' then begin
             if cursormode eq 'bi'  then trs.bakmin = abs(datax-trs.yfix)
             if cursormode eq 'bo'  then trs.bakmax = abs(datax-trs.yfix)
             if cursormode eq 'blo' then trs.bakmax[0] = abs(datax-trs.yfix)
             if cursormode eq 'bli' then trs.bakmin[0] = abs(datax-trs.yfix)
             if cursormode eq 'bri' then trs.bakmin[1] = abs(datax-trs.yfix)
             if cursormode eq 'bro' then trs.bakmax[1] = abs(datax-trs.yfix)
             if cursormode eq 'bi' then msg, 'Setting inner background dist to '+clipzero(trs.bakmin[0]), color=rgb(255,255,0)
             if cursormode eq 'bo' then msg, 'Setting outer background dist to '+clipzero(trs.bakmax[0]), color=rgb(255,255,0)
             if cursormode eq 'blo' then msg, 'Setting outer left background dist to '+clipzero(trs.bakmax[0]), color=rgb(255,255,0)
             if cursormode eq 'bli' then msg, 'Setting inner left background dist to '+clipzero(trs.bakmin[0]), color=rgb(255,255,0)
             if cursormode eq 'bri' then msg, 'Setting inner right background dist to '+clipzero(trs.bakmin[1]), color=rgb(255,255,0)
             if cursormode eq 'bro' then msg, 'Setting outer right background dist to '+clipzero(trs.bakmax[1]), color=rgb(255,255,0)
             stickymsg = 1
           endif
           ; fix any inconsistencies
            ; radius extends off the image
           if trs.aprad gt trs.yfix-1      then trs.aprad = trs.yfix-1
           if trs.aprad gt (np-trs.yfix)-1 then trs.aprad = (np-trs.yfix)-1
            ; aperture overlaps background
           if trs.aprad gt trs.bakmin[0] and m0 eq 'e' and trs.bakmin[0] gt 0 then trs.bakmin[0] = trs.aprad+1
           if trs.aprad gt trs.bakmin[1] and m0 eq 'e' and trs.bakmin[1] gt 0 then trs.bakmin[1] = trs.aprad+1
           if trs.aprad gt trs.bakmin[0] and m0 eq 'b' and trs.bakmin[0] gt 0 then trs.aprad = trs.bakmin[0]-1
           if trs.aprad gt trs.bakmin[1] and m0 eq 'b' and trs.bakmin[1] gt 0 then trs.aprad = trs.bakmin[1]-1
            ; outer radius is within inner radius of background
           if trs.bakmax[0] lt trs.bakmin[0] then trs.bakmax[0] = trs.bakmin[0] + 10
           if trs.bakmax[1] lt trs.bakmin[1] then trs.bakmax[1] = trs.bakmin[1] + 10
           ;if strmid(cursormode,0,1) eq 'b' then begin
           ;    str = clipzero(trs.bakmin[0])+'-'+clipzero(trs.bakmax[0])+'|'+clipzero(trs.bakmin[1])+'-'+clipzero(trs.bakmax[1])
           ;    msg, 'Setting background to '+str, color=rgb(255,255,0)
           ;    stickymsg = 1
           ;endif
           if cursormode eq 'c' or strmid(cursormode,0,1) eq 'e' or strmid(cursormode,0,1) eq 'b' then begin
             ; Erase the old lines since we're going to redraw them in a new location in the next iteration
             oplot, oldtrs.yfix*[1,1], [-1e7,1e7], color=0*button1offcolor
             oplot, oldtrs.yfix*[1,1]-oldtrs.aprad, [-1e7,1e7], color=0*button2offcolor
             oplot, oldtrs.yfix*[1,1]+oldtrs.aprad, [-1e7,1e7], color=0*button2offcolor
             oplot, oldtrs.yfix*[1,1]-oldtrs.bakmin[0], [-1e7,1e7], color=0*button3offcolor
             oplot, oldtrs.yfix*[1,1]-oldtrs.bakmax[0], [-1e7,1e7], color=0*button3offcolor
             oplot, oldtrs.yfix*[1,1]+oldtrs.bakmin[1], [-1e7,1e7], color=0*button3offcolor
             oplot, oldtrs.yfix*[1,1]+oldtrs.bakmax[1], [-1e7,1e7], color=0*button3offcolor
             if viewmode eq 'profile' then oplot, yp, prof, psym=10, color=rgb(255,255,255)
           endif
           trf[obj] = trs
           ; fx = interpol(curvey, curvex, datax)
           ; xyouts, 0.02, 0.87, /norm, 'f(x)='+fpr(fx,3.2), charsize=1.2
         endif
      endif
      ; end mouse actions


      ; Take actions that can be triggered by either mouse or keyboard (prepare zoom, save, quit)

      if cursormode eq 'z0' then begin
         cursormode = 'z'
         zoomcoords = [0]
         ; msg, 'Click the two corners of the zoom-in region', color=rgb(255,0,255)
      endif

      if quit then begin
          ; Light up the button that was just pressed
          if first then drawbutton, buttonfirst, 1, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label='<<'          
          if last then drawbutton, buttonlast, 1, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label='>>'          
          if prev then drawbutton, buttonprev, 1, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label=' <-'          
          if next then drawbutton, buttonnext, 1, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label=' ->'          
          if rb   then drawbutton, buttonrb, 1, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label='R/B'          
          if quit then drawbutton, buttonq, 1, oncolors=buttonqoncolors, offcolors=buttonscolors, /norm, label='quit'          
          if unsavedchanges or unsavedchangeselsewhere then begin
           ; Check if the user really wants to proceed with unsaved changes
           if simplecursor eq 1 then begin
               msg, 'Unsaved changes have been made.  Continue anyway?', color=rgb(255,0,0)
               wait, 0.05
               pxc = (plotpos[0]+plotpos[2])/2.
               pxw = 0.18 
               pxh = 0.05
               buttony = [pxc-pxw/2, 0.1+0.43-pxh, pxc+pxw/2, 0.1+0.43]
               buttonn = [pxc-pxw/2, 0.1+0.49-pxh, pxc+pxw/2, 0.1+0.49]
               buttonsc = [pxc-pxw/2, 0.1+0.55-pxh, pxc+pxw/2, 0.1+0.55]
               drawbutton, buttony, 0, oncolors=buttonqoncolors, offcolors=buttonscolors, /norm, label='Quit w/o saving'
               drawbutton, buttonn, 0, oncolors=button2oncolors, offcolors=buttonscolors, /norm, label='Cancel and resume'
               drawbutton, buttonsc, 0, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label='Save and quit'
               wait, 0.05
               result = ''
               while result ne 'c' and result ne 'r' and result ne 's' do begin
                  cursor, x, y, /norm
                  if iswithin([x,y], buttony) then result = 'c'
                  if iswithin([x,y], buttonn) then result = 'r'
                  if iswithin([x,y], buttonsc) then result = 's'
               endwhile
               if result eq 'c' then drawbutton, buttony, 1, oncolors=buttonqoncolors, offcolors=buttonscolors, /norm, label='Quit w/o saving'
               if result eq 'r' then drawbutton, buttonn, 1, oncolors=button2oncolors, offcolors=buttonscolors, /norm, label='Cancel and resume'
               if result eq 's' then drawbutton, buttonsc, 1, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label='Save and quit'
               wait, 0.2
               if result eq 'r' then quit = 0
               if result eq 'r' then redraw = 1
               if result eq 's' then save = 1
           endif
           if simplecursor eq 0 then begin
             wait, 0.05
             msg, 'Unsaved changes have been made.  Continue anyway?  (C = continue, S = save+continue, any other key to cancel)', color=rgb(255,0,0)
             qkey = strupcase(get_kbrd())
             if qkey ne 'C' and qkey ne 'S' and qkey ne 'Q' then begin
                 quit = 0
                 first = 0
                 next = 0
                 prev = 0
                 last = 0
             endif 
             if qkey eq 'S' then save = 1 
           endif
          endif else begin
            wait, 0.2
          endelse
      endif

      if list then begin
           drawbutton, buttonlist, 1, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label='list'
           msg, 'Listing all apertures for this file to the terminal window...', color=rgb(200,200,200)
           wait, 0.05
           for i = 0, n_elements(trf)-1 do print, trsoutstring(trf[i], /nocomment)
           wait, 0.15
      endif
      if save then begin
          if unsavedchanges or unsavedchangeselsewhere then begin
             print, 'Saving to ', clip(userobjfile)
             drawbutton, buttons, 1, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label=' save'

             if unsavedchanges then updatetr, usertr, trf
             saveobjectposfile, userobjfile, usertr
             inittrf = trf

             msg, 'All changes saved.', color=rgb(120,120,255)
             wait, 0.6
             drawbutton, buttons, 0, oncolors=buttonsoncolors, offcolors=buttonscolors, /norm, label=' save'
             unsavedchanges = 0b
             unsavedchangeselsewhere = 0b
          endif
      endif

      if quit or prev or next or first or last or rb then break
  endwhile

  if prev or next or first or last or rb then begin  ; this line is probably superfluous since these are the only conditions that end the loop
     newfile = ''
     if incmode eq 'all' then begin
       if prev then newfile = allprofilefiles[thisfilen-1]
       if first then newfile = allprofilefiles[0]
       if last then newfile = allprofilefiles[nfiles-1]
       if next then newfile = allprofilefiles[thisfilen+1]
     endif else begin
       if prev then newfile = sciprofilefiles[thisfilewn-1]
       if first then newfile = sciprofilefiles[0]
       if last then newfile = sciprofilefiles[nsci-1]
       if next then newfile = sciprofilefiles[thisfilewn+1]
     endelse
     if rb then newfile = allprofilefiles[rbmatchn]
     if unsavedchanges then unsavedchangeselsewhere = 1b
     if unsavedchanges then updatetr, usertr, trf
     if newfile ne '' then begin ; should only not be true if quitting
        profilefile = newfile
        goto, start ; didn't want to put this in the loop but it's probably safe.
     endif
  endif

  print
  if unsavedchanges eq 0 and unsavedchangeselsewhere eq 0 then print, 'Closing...' $
                         else print, 'Closing without saving changes...'
  wait, 0.05
  wdelete, 0
  wait, 0.05
  !quiet = 0

end

