;+
; NAME:
;       lriscat
;
; PURPOSE:
;       Print a catalog of LRIS file information to the terminal (or to disk).
;
; CALLING SEQUENCE:
;       lriscat, str, [extrakeys=extrakeys, grep=, sortby=, outfile=]
;
; EXPLANATION:
;       Tool to print out a catalog of file information: times, configuration
;       details, etc.  Works on raw and on processed files, including ASCII
;       spectra.
;
; INPUTS:
;       str - Wildcard file string for files to check.
;
; OPTIONAL INPUT KEYWORDS:
;       extrakeys - Keyword (or list of additional keywords) to also include
;       grep - String to require in the filename to examing the file
;              (for fast filtering)
;       sortby - Keyword to sort the output catalog by.  
;       outfile - Print the catalog to a file instead of the string.
;       fkey/fval - Deprecated??
;
; OUTPUTS:
;
; 
; EXAMPLES:
;       lriscat, 'r170101_????.fits'
;       lriscat, 'spredux1d/cels*.spec', sortby='EXPTIME'
;
; NOTES:
;       See also lriscat.py for producing catalogues outside of IDL.
;       (That version has some additional features but doesn't recognize spectra.)
;
;-
;------------------------------------------------------------------------------

pro lriscat, str, extrakeys=extrakeys, grep=grep, sortby=sortby, outfile=outfile, fkey=fkey, fval=fval

   q = !quiet
   !quiet = 1

 
   if n_elements(str) eq 0 then str = '*.fits'
   ct = 1
   if n_elements(str) eq 1 then files = file_search(str, count=ct) else files = str
   if ct eq 0 then begin
      print, 'No matches.'
      return
   endif

   filestrlen = max(strlen(removepath(files))) > 22

   if n_elements(outfile) gt 0 then begin
     openw, 10, outfile
   endif

   head = clip('file',filestrlen)+' '+clip('UT',11)+' '+clip('target',16)+' '+clip('object',16)+' '+rclip('exp',6)+' '+$
          clip('ra',11)+' '+clip('dec',11)+' '+clip('filt/gr',9)+' '+clip('slit',4)+' '+clip('dic',3)+' '+clip('air',4)+' '+clip('PA',6)+' '+clip('bin',3)+' '+clip('wave',6)+'other'
   if n_elements(outfile) gt 0 then printf, 10, head else print, head

   outarr = ['']
   outsort = ['']
 
   for i = 0, n_elements(files)-1 do begin
      h = ''
      if strpos(files[i],'.fit') gt 0 then begin
         h = headfits(files[i])
      endif else begin
        if strpos(files[i],'.spec') gt 0 or strpos(files[i],'.resp') gt 0 or strpos(files[i],'.sol') then $
           ;dummy = readspec(files[i], h)
           h = headspec(files[i])
      endelse
      if n_elements(h) lt 2 then begin
         print, clip(removepath(files[i]))+': Unrecognized file type '
         continue
      endif

      if n_elements(fkey) gt 0 and n_elements(fval) eq n_elements(fkey) then begin
        for fi = 0, n_elements(fkey)-1 do begin
          fileval = sxpar(h,fkey[fi],count=ctc)
          if n_elements(ctc) eq 0 then begin
            if ctc eq 0 then begin
              print, ' '+fkey+' not recognized.' 
              return
            endif
          endif
          if clip(fileval) ne clip(fval[fi]) then break
        endfor
        if fi lt n_elements(fkey) then continue ; loop was broken, so skip this file
      endif

      out = ''
      out += clip(removepath(files[i]),filestrlen) + ' '
      
      out += clip(sxpar(h,'UTC'), 11) + ' '

      out += clip(sxpar(h,'TARGNAME'), 16) + ' '

      out += clip(sxpar(h,'OBJECT', /silent), 16) + ' '

      exptime = sxpar(h, 'EXPTIME', count=ct)
      if ct eq 0 then exptime = sxpar(h,'TTIME')
      out += rclip(fpr(exptime, 4.1), 6) + ' '

      out += clip(sxpar(h, 'RA'), 11) + ' '

      out += clip(sxpar(h, 'DEC'), 11) + ' '

      if clip(sxpar(h,'INSTRUME')) eq 'LRIS' then cam = 'r'
      if clip(sxpar(h,'INSTRUME')) eq 'LRISBLUE' then cam = 'b'
      if clip(sxpar(h,'INSTRUME')) eq 'LRIS+LRISBLUE' then cam = '2'
      if cam eq '2' then begin 
          out += clip(clip(sxpar(h,'GRANAME'))+'+'+clip(sxpar(h,'GRISNAME')),18)
          out += ' '
          ;out += clip(clip(sxpar(h,'REDFILT'))+'+'+clip(sxpar(h,'BLUFILT')),10)
      endif else begin
          if cam eq 'r' then grkey = 'GRANAME'
          if cam eq 'b' then grkey = 'GRISNAME'
          gr = clip(sxpar(h,grkey))
          if cam eq 'r' then filtkey = 'REDFILT'
          if cam eq 'b' then filtkey = 'BLUFILT'
          f = clip(sxpar(h,filtkey))
          if strpos(gr,'/') gt 0 then fgr = gr else fgr = f
          if strpos(gr,'/') gt 0 and strlowcase(f) ne 'clear' then fgr += '+'+f
          out += clip(fgr, 9) + ' '
      endelse

      slit = clip(sxpar(h,'SLITNAME'))
      if strpos(slit,'long') ge 0 or strpos(slit,'pol') ge 0 then begin
           if strpos(slit,'lon') ge 0 then slit = slit + '"'
           if strpos(slit,'pol') ge 0 then slit = slit + 'P'
           slit = strmid(slit,strpos(slit,'_')+1)
      endif else begin
           if slit eq 'direct' then slit = 'dir'
      endelse
      out += clip(slit,4) + ' '

      out += clip(sxpar(h, 'DICHNAME'), 3) + ' '

      airstr = fpr(sxpar(h, 'AIRMASS'), 1.2) + ' '
      if strmid(airstr,0,1) eq '*' then airstr = clip(clipzero(sxpar(h,'AIRMASS')),4)+' '
      out += airstr
      
      out += fpr(((90+sxpar(h, 'ROTPOSN') + 90 + 360) mod 360) - 90.,3.2) + ' ' ; PA=90+ROTPOSN; wrap about 270

      bin = repstr(repstr(clip(sxpar(h,'BINNING')),',','x'),' ','')
      if bin eq '0' then bin = ''
      out += bin + ' '

      wave = '-'
      if cam ne 'b' then wave = clip(sxpar(h,'WAVELEN')) else wave = '-'
      if strmid(wave,0,1) eq '0' then if float(wave) eq 0 then wave = '-'
      out += clip(wave,6) + ' '

      trapdoor = clip(sxpar(h,'TRAPDOOR', count=ct))
      if ct eq 0 then trapdoor = '-' ; 1D spec, not saved currently?
      if trapdoor eq '0' then trapdoor = '?'
      dooropen = strmid(trapdoor,0,1) eq 'o' or trapdoor eq '-'

      azimuth = sxpar(h,'AZ')
      elevation = sxpar(h,'EL')
      domeazimuth = sxpar(h,'DOMEPOSN')
      domerelaz = domeazimuth - azimuth
      while domerelaz lt -180.0 do domerelaz += 360.
      while domerelaz ge  180.0 do domerelaz -= 360.
      view = 'sky'
      if elevation lt 1.0 then view = 'horizon'
      if abs(elevation-45) lt 0.1 and abs(domerelaz) ge 2.0 and dooropen then view = 'dome?'
      if abs(elevation-45) lt 0.1 and abs(abs(domerelaz)-90) lt 1.0 and dooropen then view = 'dome'

      if view ne 'sky' and dooropen then out = out + view      
      if dooropen eq 0 then out = out + 'closed'

      Hg = 0 & Neon = 0 & Ar = 0 & Cd = 0 & Zn = 0 & hal = 0 & Kr = 0 & Xe = 0 & FeAr = 0 & De = 0
      lamps = clip(sxpar(h,'LAMPS'))
      if lamps eq '0' then begin
               if clip(sxpar(h,'MERCURY')) eq 'on' then Hg = 1
               if clip(sxpar(h,'NEON')) eq 'on' then Neon= 1
               if clip(sxpar(h,'ARGON')) eq 'on' then Ar = 1
               if clip(sxpar(h,'CADMIUM')) eq 'on' then Cd = 1
               if clip(sxpar(h,'ZINC')) eq 'on' then Zn = 1
               if clip(sxpar(h,'HALOGEN')) eq 'on' then hal = 1
               if clip(sxpar(h,'KRYPTON')) eq 'on' then Kr = 1                                            
               if clip(sxpar(h,'XENON')) eq 'on' then Xe = 1                                                  
               if clip(sxpar(h,'FEARGON')) eq 'on' then FeAr = 1
               if clip(sxpar(h,'DEUTERI')) eq 'on' then De = 1
               lamps = ''+clip(Hg)+','+clip(Neon)+','+clip(Ar)+','+clip(Cd)+','+clip(Zn)+','+clip(hal)+','+clip(Kr)+','+clip(Xe)+','+clip(Fear)+','+clip(De)
      endif else begin
         Hg = fix(strmid(lamps,0))
         Neon = fix(strmid(lamps,2))
         Ar = fix(strmid(lamps,4))
         Cd = fix(strmid(lamps,6))
         Zn = fix(strmid(lamps,8))
         hal = fix(strmid(lamps,10))
         Kr = fix(strmid(lamps,12))
         Xe = fix(strmid(lamps,14))
         Fe = fix(strmid(lamps,16))
         De = fix(strmid(lamps,18))
      endelse
      if Hg + Neon + Ar + Cd + Zn + hal + Kr + Xe + FeAr + De gt 0 and dooropen eq 0 then begin  ; trapdoor shuts off lamps if open
            out = out + ':'
            out = out + (Hg ? 'Hg' : '')
            out = out + (Neon ? 'Ne' : '')
            out = out + (Ar ? 'Ar' : '')
            out = out + (Cd ? 'Cd' : '')
            out = out + (Zn ? 'Zn' : '')
            out = out + (hal ? 'hal' : '')
            out = out + (Kr ? 'Kr' : '')
            out = out + (Xe ? 'Xe' : '')
            out = out + (FeAr ? 'Fe' : '')
            out = out + (De ? 'De' : '')
            out = out + ' '
      endif else begin
            if dooropen eq 0 then out = out + ':dark' + ' '
      endelse

      imlamp = clip(sxpar(h,'FLIMAGIN')) eq 'on'
      speclamp = clip(sxpar(h,'FLSPECTR')) eq 'on'
      lamp1 = clip(sxpar(h,'FLAMP1')) eq 'on'
      lamp2 = clip(sxpar(h,'FLAMP2')) eq 'on'
      if (imlamp or speclamp or lamp1 or lamp2) and dooropen      then out = out + ':flat'
      if (imlamp or speclamp or lamp1 or lamp2) and (dooropen eq 0 or view eq 'sky') then out = out + '(lamp on!)'


      for k = 0, n_elements(extrakeys)-1 do begin
          out += clip(sxpar(h,extrakeys[k])) + ' '
      endfor

      if keyword_set(grep) then begin
          if strpos(out, grep) eq -1 then continue
      endif

      if n_elements(sortby) gt 0 then begin
         outarr = [outarr, out]
         outsort = [outsort, sxpar(h,sortby, count=ct)]
         if ct eq 0 then begin
            print, 'Sort keyword not recognized: ', sortby
            return
         endif
      endif else begin
        if n_elements(outfile) gt 0 then printf, 10, out else print, out
      endelse

   endfor

   if n_elements(sortby) gt 0 and n_elements(outarr) gt 1 then begin
      outarr = outarr[1:*]
      outsort = outsort[1:*]
      outarr = outarr[sort(outsort)]
      for i = 0, n_elements(outarr)-1 do begin
        if n_elements(outfile) gt 0 then printf, 10, outarr[i] else print, outarr[i]
      endfor
   endif

   if n_elements(outfile) gt 0 then close, 10
   !quiet = q

end

