;+
; NAME:
;       lriscompletioncheck
;
; PURPOSE:
;       Tool to check if all files have been reduced.
;       Can be used to check multiple directories/nights at once and gather statistics
;         on the success rate of the pipeline in fully automatic mode.
;
; CALLING SEQUENCE:
;       lriscompletioncheck, prepfilesin, finalfilesin, 
;                            directories=directories, summaryfile=summaryfile
;
; EXPLANATION:
;       Compares the list of prepared files (prepfiles) and fully processed files
;       (finalfilesin) to see if any "raw" (prepared) files were not fully reduced.
;
; OPTIONAL INPUTS:
;       prepfilesin  - List of prepared files.  Checks spredux2d/p*.fits by default.
;       finalfilesin - List of final spectra.   Checks ./*.spec by default.
;       (Both of the above can be wildcard strings or lists.)
;
; OPTIONAL INPUT KEYWORDS:
;       directories - List of multiple directories to search.
;       summaryfile - Write out a summary file with completion/success statistics for
;                     each directory processed.
;
;
; EXAMPLES:
;       lriscrawler, '2017010[2-6]', mode='s'
;
; Last updated 2019 Mar 17, Documentation added 2022 Jul 03
;
;-
;------------------------------------------------------------------------------


pro lriscompletioncheck, prepfilesin, finalfilesin, directories=directories, summaryfile=summaryfile

 if n_elements(summaryfile) gt 0 then begin
   openw, 7, summaryfile
   printf, 7, '#directory  [by # of frames]     [by total exposure time]
 endif

 cd, current=initdir ; get starting directory
 if n_elements(directories) eq 0 then directories = '' else $
 if n_elements(directories) eq 1 then directories = file_search(directories, /mark_dir, /test_dir, count=ctd)


 nscialld = 0l
 expscialld = 0.
 nmatchedalld = 0l
 expmatchedalld = 0.

 for d = 0, n_elements(directories)-1 do begin
  if directories[d] ne '' then cd, directories[d]
  if directories[d] ne '' then print, directories[d]

  if n_elements(prepfilesin) eq 0 then prepfiles = file_search('spredux2d/p*.fits', count=ctp)
  if n_elements(prepfilesin) eq 1 then prepfiles = file_search(prepfilesin, count=ctp)
  if n_elements(prepfilesin) ge 2 then prepfiles = prepfilesin
  if n_elements(prepfilesin) ge 2 then ctp = n_elements(prepfiles)
  if n_elements(finalfilesin) eq 0 then finalfiles = file_search('*.spec', count=ctf)
  if n_elements(finalfilesin) eq 1 then finalfiles = file_search(finalfiles, count=ctf)
  if n_elements(finalfilesin) ge 2 then finalfiles = finalfilesin
  if n_elements(finalfilesin) ge 2 then ctf = n_elements(finalfiles)

  nsci = 0l
  expsci = 0.
  nmatched = 0l
  expmatched = 0.
  if ctp eq 0 then begin  
     if n_elements(prepfilesin) eq 0 then print, ' Cannot match '+directories[d]+'spredux2d/p*.fits''
     if n_elements(prepfilesin) eq 1 then print, ' Cannot match '+directories[d]+prepfilesin
     if directories[d] ne '' then cd, initdir
     continue
  endif
  if ctf eq 0 then begin  
     if n_elements(finalfilesin) eq 0 then print, ' No files matching '+directories[d]+'*.spec'
     if n_elements(finalfilesin) eq 1 then print, ' No files matching '+directories[d]+finalfilesin
  endif
  if ctp gt 0 and ctf gt 0 then begin

   hp = lrisheaderinfo(prepfiles)
   wsci = where(hp.trapdoor eq 'open' and hp.sunelev lt -12 and hp.exptime gt 45.)
   hp = hp[wsci]
   prepfiles = prepfiles[wsci]
   s = sort(hp.date)
   hp = hp[s]
   prepfiles = prepfiles[s]

   hf = replicate({filename:'', object:'', target:'', status:'', bexp:0., rexp:0., koaids:strarr(20)},n_elements(finalfiles))

   for i = 0, n_elements(hf)-1 do begin
     file = finalfiles[i]
     if i ge 0 then begin
       h = headspec(file)
       hf[i].status = clip(sxpar(h,'VALIDATE'))
       hf[i].object = clip(sxpar(h,'OBJECT'))
       hf[i].target = clip(sxpar(h,'TARGNAME'))
       hf[i].rexp = (sxpar(h,'REDEXPTM'))
       hf[i].bexp = (sxpar(h,'BLUEXPTM'))
       for j = 0, 19 do begin
          koaidj = sxpar(h,'KOAID'+clip(j), count=ct)
          if ct gt 0 then hf[i].koaids[j] = koaidj
       endfor
     endif
   endfor

   standardpath = '/home/aridperl/code/idl/lris/refdata'

   for i = 0, n_elements(hp)-1 do begin
     pfile = removepath(prepfiles[i])

     readstandard, hp[i].target, pos=[hp[i].radeg, hp[i].decdeg], $
                   stdname=stdname, /check, result=result, path=standardpath+'/'
     if result eq 1 then continue ; standard star

     if hp[i].koaid eq '' then begin 
        print, pfile+': no KOAID??'
        continue
     endif
     nsci += 1
     expsci += hp[i].exp
     w = where(hf.koaids eq hp[i].koaid, ct)
     targ = hp[i].target
     camera = strupcase(clip(hp[i].camera,1))
     cam = strlowcase(camera)
     gr = clip(hp[i].grname)
     exp = rclip(fix(round(hp[i].exptime)),6)
     out = pfile + ' '+ clip(targ,16) + ' '+camera+' '+clip(gr,9) + exp ;+ '  '+hp[i].koaid
     if ct eq 0 then begin
        flatf = repstr(repstr(prepfiles[i],'p'+cam+'1','fp'+cam+'1'),'.fits','r.fits')
        flatmsg = file_test(flatf) ? '' : ' [not flattened]'
        print, out + ' -> ??? '+flatmsg
     endif
     if ct eq 1 then begin 
        print, out + ' -> '+removepath(finalfiles[fix(w / 20)])
        nmatched += 1
        expmatched += hp[i].exp
     endif
     if ct ge 2 then print, out + ' -> multiple matches??'
     ;'['+hp[i].koaid+']'+'
   endfor
  endif

  if nsci gt 0 then begin
  print, ' '+clip(nmatched)+' / '+clip(nsci)+' matched ('+clip(fpr(100*nmatched*1./nsci,3.1),5)+'%)'
  print, ' '+clip(expmatched)+' / '+clip(expsci)+' seconds ('+clip(fpr(100*expmatched*1./expsci,3.1))+'%)'
  if n_elements(summaryfile) gt 0 then begin
    printf, 7, directories[d]+$
              ' '+rclip(nmatched,5)+' /'+rclip(nsci,5)+'  '+$
                    clip(fpr(100*nmatched*1./nsci,3.2),5)+'%'+$
              ' '+rclip(fpr(expmatched,8.1),10)+' / '+clip(fpr(expsci,7.1),9)+$
                    ' '+clip(fpr(100*expmatched*1./expsci,3.2))+'%'
  endif
  endif
  nscialld += nsci
  expscialld += expsci
  nmatchedalld += nmatched
  expmatchedalld += expmatched

  if directories[d] ne '' then cd, initdir
 endfor

 if n_elements(directories) gt 1 then begin
 print
 print, 'Combined statistics for '+clip(n_elements(directories))+' subdirectories:'
 print, ' '+clip(nmatchedalld)+' / '+clip(nscialld)+' matched ('+clip(fpr(100*nmatchedalld*1./nscialld,3.1),5)+'%)'
 print, ' '+clip(expmatchedalld)+' / '+clip(expscialld)+' seconds ('+clip(fpr(100*expmatchedalld*1./expscialld,3.1))+'%)'
 endif

 if n_elements(summaryfile) gt 0 then begin
    printf, 7, 'combined'+$
              ' '+rclip(nmatchedalld,5)+' /'+rclip(nscialld,5)+'  '+$
                    clip(fpr(100*nmatchedalld*1./nscialld,3.2),5)+'%'+$
              ' '+rclip(fpr(expmatchedalld,8.1),10)+' / '+clip(fpr(expscialld,7.1),9)+$
                    ' '+clip(fpr(100*expmatchedalld*1./expscialld,3.2))+'%'

   close, 7
 endif

end

