;+
; NAME:
;       lrisconnect
;
; PURPOSE:
;       Connect blue and red lris spectra into a single merged file.
;
; CALLING SEQUENCE:
;       lrisconnect, bluefilename, redfilename, [outfilename=, plotfilename=,
;                    /displayoff, setrescale=, rescaleside=]
;
; EXPLANATION:
;       Compares blue and red spectra in the region of overlap and rescales
;       the fluxes to make sure they 'match', then joins them at a junction
;       point chosen based on their respective response curves.
;
; INPUTS:
;      bluefilename - Name of the blue file.
;
; OPTIONAL INPUTS:
;      redfilename - Name of the red file.  If omitted, a 'trivial' junction
;                    is performed where the first file is unchanged, but some
;                    header data is added.
;
; OPTIONAL INPUT KEYWORDS:
;      outfilename - Output file.  If omitted, chosen automatically.
;      plotfilename - File for the check postscript plot.  
;                    (Also plotted to screen unless /displayoff is set).
;      /displayoff - Do not plot to screen.
;      setrescale - Force a specific rescaling value and thus bypass the 
;                   automatic rescaling calculation
;      rescaleside - A string indicating which side to rescale:
;        blue:     Rescale the blue side to match the red.
;        red:      Rescale the red side to match the blue.
;        both:     Rescale both sides by the same factor (one in inverse).
;        upward:   Rescale whichever side needs to be increased in flux. 
;        downward: Rescale whichever side needs to be decreased in flux. 
;
; OUTPUTS:
;     The rescaled spectrum is written to disk (see outfilename).
; 
; EXAMPLES:
;     lrisconnect, 'blue.spec', 'red.spec', outfile='combined.spec'
;
; NOTES:
;     Users may want to use this tool directly to connect up a spectrum
;     using a rescaling set 'by eye' if they do not trust the pipeline.
;
;-
;------------------------------------------------------------------------------

pro lrisconnect, bluefilename, redfilename, outfilename=outfilename, plotfilename=plotfilename, displayoff=displayoff, useskythresh=useskythresh, setrescale=setrescale, rescaleside=rescaleside

   if n_elements(displayoff) eq 0 then displayoff = 0

   ; combine blue and red spectra with automatic rescaling.

   if n_elements(rescaleside) eq 0 then rescaleside = 'upward' ; which half to actually adjust.  could allow geom. mean?
   ; rescale the fainter side to match the brighter one.

   trivial = 0
   if n_elements(redfilename) eq 0 then begin ; trivial connection
      trivial = 1
      redfilename = bluefilename
   endif
   if file_test(redfilename) eq 0 then logprint, ' Cannot find ', redfilename
   if file_test(bluefilename) eq 0 then logprint, ' Cannot find ', bluefilename

   if file_test(redfilename) eq 0 then return
   if file_test(bluefilename) eq 0 then return

   sb = readspec(bluefilename, hb)
   sr = readspec(redfilename, hr)
   
   if max(sb.wav) le 0 then logprint, ' Blue file is not wavelength-calibrated'
   if max(sr.wav) le 0 then logprint, ' Red file is not wavelength-calibrated'
   if max(sb.wav) le 0 or max(sr.wav) le 0 then return

   if max(sb.wav) gt max(sr.wav) then begin  ; if blue and red are backward, reverse them
       lrisconnect, redfilename, bluefilename, outfilename, plotfilename
   endif

   if trivial then only = clip(sxpar(hb,'CAMERA')) else only = ''

   if clip(sxpar(hb,'CAMERA')) eq 'blue' then begin
      gr = sxpar(hb,'GRISNAME')
      sb = sb[where(sb.wav lt 6700.)] ; anything beyond this is heavily contamined 
                                      ; (unless some long-pass filter is used - none available?)
   endif

   hout = hb
      ; could instead make a list of here of parallel keywords to keep both versions of...
   if ~(only eq  'red') then sxaddpar, hout, 'BLUFILE', removepath(bluefilename), ' Blue spectrum file'
   if ~(only eq 'blue') then sxaddpar, hout, 'REDFILE', removepath(redfilename), ' Red spectrum file'
   vv = (trivial?'':'; avg. of R&B')
   sxaddpar, hout, 'EXPTIME', mean([sxpar(hb,'EXPTIME'),sxpar(hr,'EXPTIME')]), ' Total exposure time (sec)'+vv
   sxaddpar, hout, 'TOTALEXP', mean([sxpar(hb,'TOTALEXP'),sxpar(hr,'TOTALEXP')]), ' Total exposure time (sec)'+vv
   sxaddpar, hout, 'OBSDATE', min([sxpar(hb,'DATE_BEG'),sxpar(hr,'DATE_BEG')]),' UT start of first exposure'
   sxaddpar, hout, 'DATE_BEG', min([sxpar(hb,'DATE_BEG'),sxpar(hr,'DATE_BEG')]),' UT start of first exposure'
   sxaddpar, hout, 'DATE_END', max([sxpar(hb,'DATE_END'),sxpar(hr,'DATE_END')]),' UT end of last exposure'
   sxaddpar, hout, 'JDSTART', min([sxpar(hb,'JD'),sxpar(hr,'JD')]), ' Julian date start of first exposure'
   sxaddpar, hout, 'JDEND',   max([sxpar(hb,'JDEND'),sxpar(hr,'JDEND')]), ' Julian date end of last exposure'
   sxaddpar, hout, 'MJDSTART',min([sxpar(hb,'MJD'),sxpar(hr,'MJD')]), ' Mean julian date start of first exposure'
   sxaddpar, hout, 'MJDEND',  max([sxpar(hb,'MJDEND'),sxpar(hr,'MJDEND')]), ' Mean julian date end of last exposure'
   sxaddpar, hout, 'MINWAV', min([sxpar(hb,'MINWAV'),sxpar(hr,'MINWAV')]), ' Minimum wavelength'
   sxaddpar, hout, 'MAXWAV', max([sxpar(hb,'MAXWAV'),sxpar(hr,'MAXWAV')]), ' Maximum wavelength'

   ; other userful keywords: flattype?

   target = repstr(sxpar(hb,'TARGNAME'),' ','')
   if strpos(target,'_S') gt 0 and strpos(target,'_S') ge strlen(target)-3 then $
      target = strmid(target,0,strpos(target,'_S'))  ; offset stars
   if n_elements(outfilename) eq 0 then outfilename = clip(target)+'.spec'


   if trivial eq 0 then begin ; actual connection steps below

      if n_elements(plotfilename) eq 0 then plotfilename = repstr(outfilename,'.spec','_rescale.eps')

      if n_elements(useskythresh) eq 0 then useskythresh = 150. ; counts/pix below which we also use the sky

      baprad = sxpar(hb,'APRAD')
      raprad = sxpar(hr,'APRAD')

      logprint, ' Connecting '+removepath(bluefilename)+' and '+removepath(redfilename), format='($,A)'
      pline = 0 ; says that the cursor is in the middle of a line.  Warnings will reset to a newline.
                ; Should consider moving this print to the end so that it can be formatted appropriately even if warnings.
      if clip(sxpar(hr,'INSTRUME')) eq 'LRIS' then begin 
          sr = sr[0:n_elements(sr)-2] ; last pixels on red seem to be bad. (LRIS-R3 only, technically)
      endif


      ; determine the full region of overlap
      overlaprange = [min(sr.wav), max(sb.wav)]

      nooverlap = 0
      if overlaprange[0] gt overlaprange[1] then begin 
         ; If no overlap we can't accurately scale anything, so skip the whole scaling section.
         logprint
         logprint, '   No red/blue overlap!', format='($,A)'
         pline = 1
         nooverlap = 1
      endif
      if clip(sxpar(hb,'GRNAME')) eq '300/5000' and clip(sxpar(hb,'DICHNAME')) eq '680' then begin
         logprint
         logprint, '   Cannot use rescaling algorithm for 300/5000 grism due to second-order contamination.', format='($,A)'
         pline = 1
         attachwav = 6500.
         nooverlap = 1
      endif
      if nooverlap eq 0 then begin

         ; Find the region of "optimum" overlap by calculating the product of response functions
         ;  and calculating the maximum +/- a bit

         woverblue = where(sb.wav ge overlaprange[0])
         bluerelrespover = sb[woverblue].resp / max(sb.resp)  ; blue response (normalized using the whole spec) for overlap pixels

         woverred =  where(sr.wav le overlaprange[1])
         redrelrespover =  sr[woverred].resp / max(sr[where(sr.wav lt overlaprange[1]+1000)].resp)
                                                 ; red response (normalized over the nearest 1000pix)

         ; interpolate sr[red] onto conv[blue] wavelengths
         wavover = sb[woverblue].wav
         redrelrespoveri = interpol(redrelrespover, sr[woverred].wav, wavover)

         ; heavily penalize pixels close to the red edge (where there are bias subtraction / vignetting issues)
         rxspecbin = (strsplit(sxpar(hr,'BINNING'),',',/extract)) [1]
         w = where(sr.pix lt 30/rxspecbin, ct)
         if ct gt 0 then redrelrespoveri[w] *= 0.2

         ; calculate product of the two response functions
         respxresp = redrelrespoveri * bluerelrespover

         ; try to find a region with relatively high response on both sides

         ;wgoodrespxresp = where(respxresp gt max(respxresp)*0.2 and $
         ;                                        bluerelrespover gt 0.15 and redrelrespoveri gt 0.15,  ct)
         wgoodrespxresp = where(respxresp gt max(respxresp)*0.2 and $
                                                 bluerelrespover gt 0.11 and redrelrespoveri gt 0.11,  ct)
         if ct lt 40 then wgoodrespxresp = where(respxresp gt max(respxresp)*0.2 and $
                                                 bluerelrespover gt 0.07 and redrelrespoveri gt 0.07,  ct)
         if ct lt 30 then wgoodrespxresp = where(respxresp gt max(respxresp)*0.2 and $
                                                 bluerelrespover gt 0.03 and redrelrespoveri gt 0.03,  ct)

         ; older method used a narrower region
         ;wgoodrespxresp = where(respxresp gt max(respxresp)*0.25 and $
         ;                                        bluerelrespover gt 0.15 and redrelrespoveri gt 0.15,  ct)
         ;if ct lt 45 then wgoodrespxresp = where(respxresp gt max(respxresp)*0.25 and $
         ;                                        bluerelrespover gt 0.11 and redrelrespoveri gt 0.11,  ct)
         ;if ct lt 40 then wgoodrespxresp = where(respxresp gt max(respxresp)*0.25 and $
         ;                                        bluerelrespover gt 0.07 and redrelrespoveri gt 0.07,  ct)
         ;if ct lt 30 then wgoodrespxresp = where(respxresp gt max(respxresp)*0.25 and $
         ;                                        bluerelrespover gt 0.03 and redrelrespoveri gt 0.03,  ct)

         if ct lt 10 then begin              
            logprint
            logprint, '   Insufficient blue/red overlap!  (Only '+$
                   fpr(overlaprange[1]-overlaprange[0],2.1)+' Angstrom / '+clip(ct)+' pix)', format='($,A)'
            nooverlap = 1
            pline = 1
         endif else begin
            useoverlaprange = [min(wavover[wgoodrespxresp]), max(wavover[wgoodrespxresp])] ; rescale calculation region

            ; useoverlapoptimum = (wavover[where(respxresp eq max(respxresp))]) [0]          ; connection point - old method
            ; new method - just check which is higher.
            higherresp = (bluerelrespover gt redrelrespoveri)  ; which relative response is higher?  1=blue, 0=red
            useoverlapoptimumw = where(higherresp ne shift(higherresp,-1), ct) ; will generally return two, first is the desired one
            if ct gt 0 then useoverlapoptimum = wavover(useoverlapoptimumw[0])
            if ct eq 0 and higherresp[0] eq 1 then useoverlapoptimum = useoverlaprange[1]
            if ct eq 0 and higherresp[0] eq 0 then useoverlapoptimum = useoverlaprange[0]
            useoverlapwidth = useoverlaprange[1]-useoverlaprange[0]
            nooverlap = 0
         endelse

         if nooverlap eq 0 and n_elements(setrescale) eq 0 then begin
            if useoverlapwidth lt 45. then begin
               logprint
               logprint, '   Caution: marginal blue/red overlap ('+fpr(useoverlapwidth,2.1)+' Ang):', format='($,A)'
               pline = 1
            endif
            ; print, ' '+clip(useoverlaprange[0])+'-'+clip(useoverlaprange[1])+' ', format='($,A)'

            ;window, 0
            ;!p.multi = 0
            ;!p.position = 0
            ;plot, wavover, wavover*0, /nodata, /xsty, yrange=[0,1]
            ; oplot, wavover, redrelrespoveri, color=rgb(255,0,0)
            ; oplot, wavover, bluerelrespover, color=rgb(0,0,255)
            ; oplot, wavover, respxresp/max(respxresp), color=rgb(255,255,0)
            ;stop

            srdeline = fltarr(n_elements(sr))
            sbdeline = fltarr(n_elements(sb))
            if sxpar(hb,'GRNAME') eq '600/4000' then begin
               wrline = where(sr.wav gt 5569 and sr.wav lt 5590, ct)  ; line is at 5577.346680
               if ct gt 0 then srdeline[wrline] = !values.f_nan ; avoid the bright night sky line
               wbline = where(sb.wav gt 5569 and sb.wav lt 5590, ct)  ; fixed a major bug here
               if ct gt 0 then sbdeline[wbline] = !values.f_nan ; avoid the bright night sky line
            endif

            bdwav = sb.wav - shift(sb.wav, 1)
            bdwav[0] = bdwav[1]            
            bozone = where(sb.wav ge useoverlaprange[0] and sb.wav le useoverlaprange[1] and $ 
                           finite(sb.flux+sb.sky+sbdeline) and sb.flag lt 0.5, ctb)

            rdwav = sr.wav - shift(sr.wav,1)
            rdwav[0] = rdwav[1]
            rozone = where(sr.wav ge useoverlaprange[0] and sr.wav le useoverlaprange[1] and $
                           finite(sr.flux+sr.sky+srdeline) and sr.flag lt 0.5, ctr)

            ; average S/N per pix in overlap region
            avovsnblue = median(sb[bozone].flux/sb[bozone].unc, /even) ; this can fail
            avovsnred  = median(sr[rozone].flux/sr[rozone].unc, /even)

            ; average counts in overlap region
            avovctblue = median(sb[bozone].flux*sb[bozone].resp)*sxpar(hb,'EXPTIME')
            avovctred  = median(sr[rozone].flux*sr[rozone].resp)*sxpar(hr,'EXPTIME')

            skyw = 0  ; can't be trusted as a general tool
            ;if avovctblue lt useskythresh or avovctred lt useskythresh then $
            ;   skyw = 1 else skyw = 0

            if (avovsnblue lt 3. or avovsnred lt 3.) or (ctb lt 8 or avovsnblue le 0) or (ctr lt 8 or avovsnred le 0) then begin
               if pline eq 0 then logprint
               pline = 1
            endif

            if avovsnblue lt 3. or avovsnred lt 3. then begin
               logprint, '  Warning: Low S/N in overlap region - blue:'+clip(fpr(avovsnblue,7.2))+', red:'+clip(fpr(avovsnred,7.2))
            endif

            if ctb lt 8 or avovsnblue le 0 then begin
               logprint, '   Insufficient useable blue pixels in overlap region ('+clip(ctb)+'); not rescaling.'
               bonorm = -1
            endif else begin
               ; +sb[bozone].sky*skyw ; add to flux below and set skyw to 1 to readd the sky
               bonorm = total((sb[bozone].flux) * bdwav[bozone]) / total(bdwav[bozone])
            endelse
  
            if ctr lt 8 or avovsnred le 0 then begin
               logprint, '   Insufficient useable red pixels in overlap region ('+clip(ctr)+'); not rescaling.'
               ronorm = -1
            endif else begin
               ; +sr[rozone].sky*skyw*baprad/raprad ; add to flux to readd the sky
               ronorm = total((sr[rozone].flux) * rdwav[rozone]) / total(rdwav[rozone])
            endelse

            if bonorm ne -1 and ronorm ne -1 then begin
               rescale = bonorm/ronorm
               if bonorm lt 0 then rescale = 0.
               if ronorm lt 0 then rescale = 0.
               if keyword_set(noscale) then rescale = 1.0 ; should automatically recognize really faint stuff

               if rescale lt 0.1 or rescale gt 10.0 or rescale le 0 then begin
                  logprint
                  if rescale gt 0 then logprint, '  WARNING - Red/blue rescaling factor is VERY large ('+clip(rescale)+').'
                  if rescale le 0 then logprint, '  ERROR   - Red/blue rescale value could not be calculated or is negative.'
                  print, '            This probably means traces of different objects are being matched or'
                  print, '            that there is a problem with tracing or flux calibration.'
                  print, '            Check the individual files.'
                  pline = 1
                  for s = 0, 1 do begin
                     if s eq 0 then hs = hb else hs = hr
                     ; sort of pointless now.
                     ;for cf = 0, 6 do begin
                     ;   filec = removepath(sxpar(hs, 'INFILE'+clip(cf), count=ct))
                     ;   if ct eq 0 then filec = removepath(sxpar(hs, 'COMB'+clip(cf), count=ct))
                     ;   expc = sxpar(hs, 'EXPTIME'+clip(cf), count=ct)
                     ;   if ct eq 0 then expc = sxpar(hs, 'EXPT'+clip(cf), count=ct)
                     ;   if ct gt 0 then print, '             ', filec, ' ', clip(expc)+'s'
                     ;endfor
                  endfor
                  if rescale lt 0.05 or rescale gt 20.0 then begin
                     logprint
                     logprint, '  ERROR - Red/blue rescaling factor is huge ('+clip(rescale)+')'
                     logprint, '  Not connecting.'
                     return  ; skip it entirely
                  endif
                  if rescale gt 0 then begin ; 0.05-0.1 or 10-20
                     logprint, '            Not performing any rescaling.'
                     rescale = 1.0
                  endif
               endif else begin  
                 if rescale lt 0.5 or rescale gt 2.0 then begin
                    logprint
                    logprint, '  WARNING - Red/blue rescaling factor is large ', format='($,A)'
                    pline = 1
                 endif               
                 if rescale lt 0.0 then begin
                    ; this cannot currently happen
                    logprint
                    logprint, '  WARNING - Rescaling factor is negative!!   Not rescaling at all.'
                    rescale = 1.0
                    pline = 1
                 endif
              endelse
            endif else begin ; rescale calculation bock
               logprint, '  Unable to measure rescale factor for blue/red spectra; not scaling.', format='($,A)'
               rescale = 1.0
               pline = 1
            endelse

         endif else begin ; not enough overlap
            if n_elements(setrescale) gt 0 then rescale = setrescale else $
            rescale = 1.0   ; if no overlap
         endelse

      endif else begin

         rescale = 1.0
      endelse

      ; actually do the rescaling

      rescaleb = 1.
      rescaler = 1.
      case rescaleside of
      'upward': begin
         if rescale lt 1.0  then rescaleb = 1./rescale
         if rescale gt 1.0  then rescaler = rescale
         end
      'downward': begin
         if rescale lt 1.0  then rescaleb = 1./rescale
         if rescale gt 1.0  then rescaler = rescale
         end
      'red': begin
         rescaler = rescale
         end
      'blue': begin
         rescaleb = 1./rescale
         end
      'both': begin
         rescaler = sqrt(rescale)
         rescaleb = 1/sqrt(rescale)
         end
      else: begin
         print, 'Invalid rescale type'
         return
         end
      endcase

      oldsr = sr
      oldsb = sb 
      if pline eq 0 then sp = '' else sp = '  ' ; spaces
      if rescaler ne 1.0 then begin
           logprint, sp+'(x'+clip(clip(rescaler,5))+') ', format='($,A)'
           sr.flux *= rescaler
           if tag_exist(sr, 'sky') then sr.sky *= rescaler ; should we aperture-scale this too?
           if tag_exist(sr, 'unc') then sr.unc *= rescaler
           if tag_exist(sr, 'resp') then sr.resp /= rescaler
      endif 
      if rescaleb ne 1.0 then begin
           logprint, sp+' (blue x'+clip(clip(rescaleb,5))+') ', format='($,A)'
           sb.flux *= rescaleb
           if tag_exist(sb, 'sky') then sb.sky *= rescaleb ; should we aperture-scale this too?
           if tag_exist(sb, 'unc') then sb.unc *= rescaleb
           if tag_exist(sb, 'resp') then sb.resp /= rescaleb
      endif

      makeplot = 1
      if makeplot and (trivial eq 0) and (nooverlap eq 0) then begin
        if n_elements(bozone) eq 0 then $
           bozone = where(sb.wav ge min(sr.wav) and sb.wav le max(sb.wav) and finite(sb.flux+sb.sky), ct)
        if n_elements(rozone) eq 0 then $
           rozone = where(sr.wav ge min(sr.wav) and sr.wav le max(sb.wav) and finite(sr.flux+sr.sky), ct)

        ; plot various things to the screen and to a ps file
        for output = displayoff, 1 do begin
           !p.multi = [0,2,2]
           if output eq 0 then begin
              errorstatus = 0
              catch, errorstatus
              if errorstatus ne 0 then begin
                 print, ' X-windows display not working: turning off display'
                 displayoff = 1
                 catch, /cancel
                 continue
              endif else begin
                 window, 8, xsize=950, ysize=400, title='LPipe: red-blue connection'
              endelse
              catch, /cancel
              
              title = ''
              oldbcol = rgb(55,45,128)
              newbcol = rgb(95,145,255)
              oldrcol = rgb(128,0,0)
              newrcol = rgb(255,128,0)
              gray = rgb(128,128,128)
              green = rgb(0,255,0)
              faintyellow = rgb(100,100,0)
           endif
           if output eq 1 then begin
              psopen, plotfilename, /encaps, xsize=9.5, ysize=4, /inches
              device, /color
              !p.font=0
              title = '!17'
              device, /times, font_index=17
              colors = transpose([[0,0,0],$
                                  [0,0,200],$ 
                                  [100,50,50],$ 
                                  [255,0,0],$
                                  [128,128,128],$
                                  [0,128,0],$
                                  [50,50,100],$
                                  [200,200,100]])
              tvlct, colors
              oldbcol = 6
              newbcol = 1
              oldrcol = 2
              newrcol = 3 
              gray = 4
              green = 5
              faintyellow = 7
           endif
           for px = 0, 3 do begin  ; plot-x
              if px eq 0 then !p.position = [0.09,0.4,0.39,0.98]
              if px eq 1 then !p.position = [0.39,0.4,0.69,0.98]
              if px eq 2 then !p.position = [0.69,0.4,0.99,0.98]
              if px eq 3 then !p.position = [0.09,0.1,0.99,0.4]
              if n_elements(useoverlaprange) eq 2 then $
                xrange=useoverlaprange else $
                continue  ; temporary solution
              yrange=[0,1.05*max([sb[bozone].flux+sb[bozone].sky,(sr[rozone].flux+sr[rozone].sky)*baprad/raprad])]
              if px eq 3 then xrange=useoverlaprange+[-1000,+1000]
              fluxmeds = [median(oldsr[rozone].flux),median(sr[rozone].flux),median(sb[bozone].flux)] ; for plot scaling
              if px eq 3 then yrange=[2*min([0,fluxmeds]), 2*max([fluxmeds,1e-18])] ;[,,
              if px eq 1 or px eq 2 then ytickformat='(A1)' else delvarx, ytickformat
              plot, [0], [0], xrange=xrange, yrange=yrange, /xsty, /ysty, charsize=0.75, title=title, ytickformat=ytickformat
                 if px eq 0 then xyouts, xrange[0]+5, yrange[1]*0.9, 'obj ('+repstr(clip(target),'!','')+')'
                 if px eq 1 then xyouts, xrange[0]+5, yrange[1]*0.9, 'sky'
                 if px eq 2 then xyouts, xrange[0]+5, yrange[1]*0.9, 'sum'
                 if px eq 3 then xyouts, xrange[0]+(xrange[1]-xrange[0])*0.05, yrange[1]*0.7, 'obj ('+clip(target)+')'
              ;oplot, [5577,5577], yrange, linestyle=1
              y0 = (px mod 3 ne 1)*oldsb.flux+(px mod 3 ne 0)*oldsb.sky*baprad/raprad ; sky is always matched to red, I guess
              y1 = (px mod 3 ne 1)*sb.flux   +(px mod 3 ne 0)*sb.sky*baprad/raprad
              y2 = (px mod 3 ne 1)*oldsr.flux+(px mod 3 ne 0)*oldsr.sky*baprad/raprad
              y3 = (px mod 3 ne 1)*sr.flux   +(px mod 3 ne 0)*sr.sky*baprad/raprad
              if px eq 3 then y0 = convol(y0,(1+fltarr(7))/7., /edge_truncate)
              if px eq 3 then y1 = convol(y1,(1+fltarr(7))/7., /edge_truncate)
              if px eq 3 then y2 = convol(y2,(1+fltarr(7))/7., /edge_truncate)
              if px eq 3 then y3 = convol(y3,(1+fltarr(7))/7., /edge_truncate)
              oplot, oldsb.wav, y0, color=oldbcol, psym=10
              oplot, sb.wav, y1, color=newbcol, psym=10
              oplot, oldsr.wav, y2, color=oldrcol, psym=10
              oplot, sr.wav, y3, color=newrcol, psym=10
              if px eq 0 then oplot, wavover, yrange[1]*bluerelrespover, color=gray
              if px eq 0 then oplot, wavover, yrange[1]*redrelrespoveri, color=gray
              if px eq 3 then oplot, sb.wav, yrange[1]*sb.resp/max(sb.resp), color=gray
              if px eq 3 then oplot, sr.wav, yrange[1]*sr.resp/max(sr[where(sr.wav lt useoverlaprange[1]+1000)].resp), color=gray
              if px eq 0 or px eq 3 then for i = 0, 1 do oplot, useoverlaprange[i]*[1,1], yrange, thick=2
              if n_elements(useoverlapoptimum) gt 0 then oplot, useoverlapoptimum+[0,0], yrange, color=green, linestyle=2, thick=2
              if px eq 1 then oplot, 5577.346680*[1,1], [-1,1], linestyle=1, color=faintyellow
              if px eq 1 then oplot, 6829.508301*[1,1], [-1,1], linestyle=1, color=faintyellow
              if px eq 1 then oplot, 6863.970703*[1,1], [-1,1], linestyle=1, color=faintyellow
           endfor ; plot panel loop
           if output eq 1 then psclose
        endfor ; plot output loop
      endif

      ;print, ' @ '+overlapcenter, format='($,A)'
      ;if n_elements(attachwav) lt 0 then delvarx, attachwav

      if n_elements(useoverlapoptimum) gt 0 then begin 
         attachwav = useoverlapoptimum 
      endif else begin
        bsig = where(sb.resp/max(sb.resp) gt 0.04, ctbsig)
        rsig = where(sr.resp/max(sr.resp) gt 0.04, ctrsig)
        if ctbsig gt 10 and ctrsig gt 10 then $
          attachwav = (max(sb[bsig].wav)+min(sr[rsig].wav))/2.
      endelse
      if n_elements(attachwav) eq 0 then attachwav = strnum(sxpar(hb,'DICHNAME'))*10.  

      smerge = connectspec(sb, sr, attachwav=attachwav)
      sxaddpar, hout, 'RESCALER', 1.0, ' (red not rescaled)'
      sxaddpar, hout, 'RESCALEB', 1.0, ' (blue not rescaled)'
      if rescaler ne 1.0 then sxaddpar, hout, 'RESCALER', rescaler, ' Rescaling factor applied to red' else $
      if rescaleb ne 1.0  then sxaddpar, hout, 'RESCALEB', rescaleb, ' Rescaling factor applied to blue'
      if n_elements(useoverlaprange) gt 0 then sxaddpar, hout, 'WAVOVER', clip(fpr(useoverlaprange[0],5.2))+'-'+clip(fpr(useoverlaprange[1],5.2)), ' Wavelength region of useable red-blue overlap'
      if n_elements(attachwav) gt 0 then sxaddpar, hout, 'ATCHWAV', attachwav, ' Red-blue connection wavelength'
   endif else begin
      smerge = sb
      logprint, ' No match available for '+bluefilename+' - trivial connection' ;, format='($,A)'
   endelse

   airs = [-1]
   outkoaid =['']

   ; rename some keywords to be instrument-specific
   if sxpar(hb,'INSTRUME') eq 'LRISBLUE' then begin
         blueexp = sxpar(hb,'TOTALEXP')
         if blueexp eq 0 then blueexp = sxpar(hb,'EXPTIME')
         sxaddpar, hout, 'BLUDET', sxpar(hb,'DETECTOR'), ' LRIS blue detector', before='DETECTOR'
         sxaddpar, hout, 'BLUDBEG', sxpar(hb,'DATE_BEG'), ' UT start of first blue exposure'
         sxaddpar, hout, 'BLUDEND', sxpar(hb,'DATE_END'), ' UT end of last blue exposure'
         sxaddpar, hout, 'BLUNEXP', sxpar(hb,'NSUM'), ' Num. blue exposures summed'
         sxaddpar, hout, 'BLUEXPTM', blueexp, ' Blue exposure time'
         for cf = 9, 0, -1 do begin
            expc = sxpar(hb, 'EXPTIME'+clip(cf), count=ct)
            if ct eq 0 then expc = sxpar(hb, 'EXPT'+clip(cf), count=ct)
            if ct gt 0 then sxaddpar, hout, 'BLUEXP'+clip(cf), expc, after='BLUEXPTM'
         endfor
         sxaddpar, hout, 'BLUFILT', sxpar(hb,'BLUFILT'), ' Blue filter'
         sxaddpar, hout, 'GRISNAME', sxpar(hb,'GRISNAME'), ' Blue grism'
         sxaddpar, hout, 'BLUBIN', sxpar(hb,'BINNING'), ' Blue binning (spat,spec)'
         for cf = 0, 9 do begin
            air = sxpar(hb, 'AIRMASS'+clip(cf), count=ct)
            if ct gt 0 then airs = [airs, air]
            if ct gt 0 then sxaddpar, hout, 'BLUAIR'+clip(cf), air, before='BLUBIN'
         endfor
         sxaddpar, hout, 'BLUXCENT', sxpar(hb,'XCENT'), ' Blue X reference pixel for extraction'
         sxaddpar, hout, 'BLUYCENT', sxpar(hb,'YCENT'), ' Blue Y extraction center position'
         sxaddpar, hout, 'BLUAPRAD', sxpar(hb,'APRAD'), ' Blue extraction aperture half-width'
         sxaddpar, hout, 'BLUBKDST', sxpar(hb,'BAKDIST'), ' Blue background aper. inner edge distance'
         sxaddpar, hout, 'BLUBKWID', sxpar(hb,'BAKWIDTH'), ' Blue background aperture width'
         if sxpar(hb,'BAKDISTM') then sxaddpar, hout, 'BLUBKDM', sxpar(hb,'BAKDISTM'), ' Blue back. aper. inner edge dist. (-)'
         if sxpar(hb,'BAKDISTP') then sxaddpar, hout, 'BLUBKDP', sxpar(hb,'BAKDISTP'), ' Blue back. aper. inner edge dist. (+)'
         if sxpar(hb,'BAKWIDTM') then sxaddpar, hout, 'BLUBKWM', sxpar(hb,'BAKWIDTM'), ' Blue back. aper. width (-)'
         if sxpar(hb,'BAKWIDTP') then sxaddpar, hout, 'BLUBKWP', sxpar(hb,'BAKWIDTP'), ' Blue back. aper. width (-)'
         sxaddpar, hout, 'BLUCWAV', sxpar(hb,'CWAV'), ' Blue central wavelength (Ang)'
         sxaddpar, hout, 'BLUDISP', sxpar(hb,'DISPAV'), ' Blue dispersion (Ang/pix)'
         sxaddpar, hout, 'BLU2DSPC', removepath(sxpar(hb,'2DSPEC')), ' Blue 2D spectrum file'
         for cf = 99, 0, -1 do begin
            sfile = sxpar(hb, 'SFILE'+clip(cf), count=ct)
            if ct gt 0 then sxaddpar, hout, 'BLU2DS'+clip(cf), removepath(sfile), after='BLU2DSPC', ' Individual blue 2D spectra'
         endfor
         sxaddpar, hout, 'BLUARCF', removepath(sxpar(hb,'ARCFILE')), ' Blue arc file'
         sxaddpar, hout, 'BLUSOL', removepath(sxpar(hb,'SOLFILE')), ' Blue arc solution file'
         sxaddpar, hout, 'BLUFLEX', sxpar(hb,'WAVNUDGE'), ' Blue flexure correction (Ang)'
         sxaddpar, hout, 'BLUTRAC', removepath(sxpar(hb,'TRACFILE')), ' Blue tracing file'
         sxaddpar, hout, 'BLUFCAL1', removepath(sxpar(hb,'FLUXCAL1')), ' Blue flux standard file'
         if clip(sxpar(hb,'FLUXCAL2')) ne '0' then sxaddpar, hout, 'BLUFCAL2', removepath(sxpar(hb,'FLUXCAL2')), ' Blue flux standard file'
         sxaddpar, hout, 'BLUFAIR1', sxpar(hb,'FLUXAIR1'), ' Blue flux standard airmass'
         if clip(sxpar(hb,'FLUXAIR2')) ne '0' then sxaddpar, hout, 'BLUFAIR2', sxpar(hb,'FLUXAIR2')
         sxaddpar, hout, 'BLUFNAM1', sxpar(hb,'FLUXNAM1'), ' Blue flux standard star name'
         if clip(sxpar(hb,'FLUXNAM2')) ne '0' then sxaddpar, hout, 'BLUFNAM2', sxpar(hb,'FLUXNAM2')
         sxaddpar, hout, 'BLUFLAT', sxpar(hb,'FLATFILE'), ' Blue flat field file name'
         for cf = 0, 99 do begin
            kid = sxpar(hb, 'KOAID'+clip(cf), count=ct)
            if ct gt 0 then outkoaid = [outkoaid, kid]
            ;if ct gt 0 then sxaddpar, hout, 'KOAIDB'+clip(cf), kid
         endfor
         maxifkeyb = ''
         ;for c = 0, n_elements(hb)-1 do begin
         ;   nam = clip(strmid(hb[c],0,strpos(hb[c],'=')))
         ;   if strmid(nam,0,6) eq 'INFILE' or strmid(nam,0,4) eq 'COMB' or $
         ;      strmid(nam,0,5) eq 'SFILE' or strmid(nam,0,6) eq 'CFILE' then begin
         ;       val = sxpar(hb, nam, /silent)
         ;       cf = clip(strnum(nam))
         ;       if strmid(nam,0,6) eq 'INFILE' or strmid(nam,0,5) eq 'SFILE' then sxaddpar, hout, 'SFILEB'+clip(cf), val
         ;       if strmid(nam,0,4) eq 'COMB' or strmid(nam,0,6) eq 'CFILE'   then sxaddpar, hout, 'CFILEB'+clip(cf), val
         ;   endif
         ;endfor
   endif
   if clip(sxpar(hr,'INSTRUME')) eq 'LRIS' then begin
         redexp = sxpar(hr,'TOTALEXP')
         if redexp eq 0 then redexp = sxpar(hr,'EXPTIME')
         sxaddpar, hout, 'REDEXPTM', redexp, after='BLUEXPTM', ' Red exposure time'
         for cf = 9, 0, -1 do begin
            expc = sxpar(hr, 'EXPTIME'+clip(cf), count=ct)
            if ct eq 0 then expc = sxpar(hr, 'EXPT'+clip(cf), count=ct)
            if ct gt 0 then sxaddpar, hout, 'REDEXP'+clip(cf), expc, after='REDEXPTM'
         endfor
         for cf = 0, 9 do begin
            air = sxpar(hr, 'AIRMASS'+clip(cf), count=ct)
            if ct gt 0 then airs = [airs, air]
            if ct gt 0 then sxaddpar, hout, 'REDAIR'+clip(cf), air, before='BLUBIN'
         endfor
         sxaddpar, hout, 'REDDET', sxpar(hr,'DETECTOR'), ' LRIS red detector', before='DETECTOR'
         sxaddpar, hout, 'REDDBEG', sxpar(hr,'DATE_BEG'), after='BLUDBEG', ' UT start of first red exposure'
         sxaddpar, hout, 'REDDEND', sxpar(hr,'DATE_END'), after='BLUDEND', ' UT end of last red exposure'
         sxaddpar, hout, 'REDNEXP', sxpar(hr,'NSUM'), after='BLUNEXP', ' Num. red exposures summed'
         sxaddpar, hout, 'REDFILT', sxpar(hr,'REDFILT'), after='BLUFILT', ' Red filter'
         sxaddpar, hout, 'GRANAME', sxpar(hr,'GRANAME'), after='GRISNAME', ' Red grating'
         sxaddpar, hout, 'REDBIN', sxpar(hr,'BINNING'), after='BLUBIN', ' Red binning (spat,spec)'
         sxaddpar, hout, 'REDXCENT', sxpar(hr,'XCENT'), after='BLUXCENT', ' Red X reference pixel for extraction'
         sxaddpar, hout, 'REDYCENT', sxpar(hr,'YCENT'), after='BLUYCENT', ' Red Y extraction center position'
         sxaddpar, hout, 'REDAPRAD', sxpar(hr,'APRAD'), after='BLUAPRAD', ' Red extraction aperture half-width'
         sxaddpar, hout, 'REDBKDST', sxpar(hr,'BAKDIST'), ' Red background aper. inner edge distance', after='BLUBKDST'
         sxaddpar, hout, 'REDBKWID', sxpar(hr,'BAKWIDTH'), ' Red background aperture width', after='BLUBKWID'
         if sxpar(hr,'BAKDISTM') gt 0 then sxaddpar, hout, 'REDBKDM', sxpar(hr,'BAKDISTM'), ' Red back. aper. inner edge dist. (-)', after='BLUBKDP'
         if sxpar(hr,'BAKDISTP') gt 0 then sxaddpar, hout, 'REDBKDP', sxpar(hr,'BAKDISTP'), ' Red back. aper. inner edge dist. (+)', after='REDBKDM'
         if sxpar(hr,'BAKWIDTM') gt 0 then sxaddpar, hout, 'REDBKWM', sxpar(hr,'BAKWIDTM'), ' Red back. aper. width (-)', after='BLUEBKWP'
         if sxpar(hr,'BAKWIDTP') gt 0 then sxaddpar, hout, 'REDBKWP', sxpar(hr,'BAKWIDTP'), ' Red back. aper. width (+)', after='REDBKWM'
         sxaddpar, hout, 'REDCWAV', sxpar(hr,'CWAV'), after='BLUCWAV', ' Red central wavelength (Ang)'
         sxaddpar, hout, 'REDDISP', sxpar(hr,'DISPAV'), after='BLUDISP', ' Red dispersion (Ang/pix)'
         sxaddpar, hout, 'RED2DSPC', removepath(sxpar(hr,'2DSPEC')), after='BLU2DSPC', ' Red 2D spectrum file'
         for cf = 99, 0, -1 do begin
            sfile = sxpar(hr, 'SFILE'+clip(cf), count=ct)
            if ct gt 0 then sxaddpar, hout, 'RED2DS'+clip(cf), removepath(sfile), before='BLUARCF', ' Individual red 2D spectra'
         endfor ; could also keep track of shifting...
         sxaddpar, hout, 'REDARCF', removepath(sxpar(hr,'ARCFILE')), after='BLUARCF', ' Red arc file'
         sxaddpar, hout, 'REDSOL', removepath(sxpar(hr,'SOLFILE')), after='BLUSOL', ' Red arc solution file'
         sxaddpar, hout, 'REDFLEX', sxpar(hr,'WAVNUDGE'), ' Red flexure correction (Ang)', after='BLUFLEX'
         sxaddpar, hout, 'REDTRAC', removepath(sxpar(hr,'TRACFILE')), after='BLUTRAC', ' Red tracing file'
         sxaddpar, hout, 'REDFCAL1', removepath(sxpar(hr,'FLUXCAL1')), after='BLUFCAL1', ' Red flux standard file'
         if clip(sxpar(hr,'FLUXCAL2')) ne '0' then sxaddpar, hout, 'REDFCAL2', removepath(sxpar(hr,'FLUXCAL2')), after='BLUFCAL2', ' Red flux standard file'
         sxaddpar, hout, 'REDFAIR1', sxpar(hr,'FLUXAIR1'), after='BLUFAIR1', ' Red flux standard airmass'
         if clip(sxpar(hr,'FLUXAIR2')) ne '0' then sxaddpar, hout, 'REDFAIR2', sxpar(hr,'FLUXAIR2'), after='BLUFAIR2'
         sxaddpar, hout, 'REDFNAM1', sxpar(hr,'FLUXNAM1'), after='BLUFNAM1', ' Red flux standard star name'
         if clip(sxpar(hr,'FLUXNAM2')) ne '0' then sxaddpar, hout, 'REDFNAM2', sxpar(hr,'FLUXNAM2'), after='BLUFNAM2'
         sxaddpar, hout, 'REDFLAT', sxpar(hr,'FLATFILE'), after='BLUFLAT', ' Red flat field file name'
         for cf = 0, 99 do begin
            kid = sxpar(hr, 'KOAID'+clip(cf), count=ct)
            if ct gt 0 then outkoaid = [outkoaid, kid]
            ;if ct gt 0 then sxaddpar, hout, 'KOAIDR'+clip(cf), kid
         endfor
         ;for c = 0, n_elements(hr)-1 do begin
         ;   nam = clip(strmid(hr[c],0,strpos(hr[c],'=')))
         ;   if clip(nam) eq '0' or clip(nam) eq '' or clip(nam) eq '#' then continue
         ;   val = sxpar(hr, nam, /silent)
         ;   if strmid(nam,0,6) eq 'INFILE' or strmid(nam,0,4) eq 'COMB' or $
         ;      strmid(nam,0,5) eq 'SFILE' or strmid(nam,0,6) eq 'CFILE' then begin
         ;       cf = clip(strnum(nam))
         ;       if strmid(nam,0,6) eq 'INFILE' or strmid(nam,0,5) eq 'SFILE' then sxaddpar, hout, 'SFILER'+clip(cf), val
         ;       if strmid(nam,0,4) eq 'COMB' or strmid(nam,0,6) eq 'CFILE'   then sxaddpar, hout, 'CFILER'+clip(cf), val
         ;   endif else begin
         ;     ; add any keywords in the input red header that aren't in the blue header
         ;     outval = sxpar(hout,nam,count=ct)
         ;     if ct eq 0 then $
         ;       sxaddpar, hout, nam, val
         ;  endelse
         ;endfor
   endif
 

   if trivial eq 0 then sxaddpar, hout, 'NSATURAT', sxpar(hr,'NSATURAT')+sxpar(hb,'NSATURAT'), ' Num of saturated pixels'
   sxaddpar, hout, 'NWAV', n_elements(smerge), ' Number of extracted wavelength columns'

   airs = airs[1:*]
   sxaddpar, hout, 'AIRMASS', mean(airs), ' Approximate av. airmass of combined spectrum'

   sxdelpar, hout, 'GRNAME' ; no longer relevant for combining both sides
   sxdelpar, hout, 'ELAPTIME'
   sxdelpar, hout, 'GAIN'
   sxdelpar, hout, 'FLATFILE'
   if trivial then begin
      if only eq 'red' then sxaddpar, hout, 'RESCALER', 1.0, ' (No blue data; no rescaling)'
      if only eq 'blue' then sxaddpar, hout, 'RESCALEB', 1.0, ' (No red data; no rescaling)'
   endif else begin
      sxaddpar, hout, 'INSTRUME', 'LRIS+LRISBLUE'
      sxdelpar, hout, 'CAMERA'
   endelse
   sxdelpar, hout, 'ARCFILE'
   sxdelpar, hout, 'CWAV'
   sxdelpar, hout, 'DISPAV'
   sxdelpar, hout, 'BINNING'
   sxdelpar, hout, 'FLATTYPE'
   sxdelpar, hout, '2DSPEC'
   sxdelpar, hout, 'YCENT'
   sxdelpar, hout, 'XCENT'
   for fi = 1, 99 do begin
     sxdelpar, hout, 'YSHIFT'+clip(fi)
     sxdelpar, hout, 'SFILE'+clip(fi)
     sxdelpar, hout, 'KOAID'+clip(fi)
   endfor
   sxdelpar, hout, 'BAKDIST'
   sxdelpar, hout, 'BAKWIDTH'
   sxdelpar, hout, 'DETECTOR'
   sxdelpar, hout, 'LAMPS'
   ; sxdelpar, hout, 'EXTDATE'

   sxaddpar, hout, 'DATE', sxpar(hout,'DATE_END'), ' UT end of last exposure'

   for i = 1, n_elements(outkoaid)-1 do sxaddpar, hout, 'KOAID'+clip(i), outkoaid[i]
  
   sxdelpar, hout, 'FLUXCAL1'
   sxdelpar, hout, 'FLUXCAL2'
   sxdelpar, hout, 'FLUXAIR1'
   sxdelpar, hout, 'FLUXAIR2'
   sxdelpar, hout, 'FLUXNAM1'
   sxdelpar, hout, 'FLUXNAM2'
   sxdelpar, hout, 'ARCFILE0'
   sxdelpar, hout, 'SOLFILE0'
   sxdelpar, hout, 'SOLFILE'
   sxdelpar, hout, 'FLUXWT1'
   sxdelpar, hout, 'FLUXWT2'
   sxdelpar, hout, 'FRAMENO'
   sxdelpar, hout, 'WINDET'
   sxdelpar, hout, 'FLEXMETH'

   if only ne 'blue' then sxaddpar, hout, 'WAVELEN', sxpar(hr,'WAVELEN'), ' Red specified central wavelength', after='GRANAME'

   for cf = 0, 9 do sxdelpar, hout, 'UT'+clip(cf) 
   for cf = 0, 9 do sxdelpar, hout, 'AIRMASS'+clip(cf) 
      ; could delete lots of other things associated with one file or another (and replace with averages / pair terms...)?

   delkeys = ['LTV1','LTV2','XBIN','YBIN','PXSCALE','DICHSEQF','RFLTSEQF','GRATSEQF','GRANSEQF', $
              'COMB'+clip(indgen(10)), 'EXPT'+clip(indgen(10)), 'WINDOW', 'PANE', 'NAXIS1','NAXIS2',$
              'YEXTRACT','CWAVE','NSUM','SATURATE','DISPERSN','APRAD','APSHARP','NSATURAT','NWAV',$
              'WAVORDER','WAVNUDGE','BAKDISTM','BAKDISTP','BAKWIDTM','BAKWIDTP','TRACFILE']
   for d = 0, n_elements(delkeys)-1 do sxdelpar, hout, delkeys[d]


   if trivial eq 0 then begin
      if n_elements(setrescale) gt 0 then logprint, '(x'+clip(clip(rescale,5))+') ', format='($,A)'
      logprint
      if nooverlap eq 0 and n_elements(skyw) gt 0 then if skyw then begin
          logprint, '    Faint source ('+clip(avovctblue)+'/'+clip(avovctred)+' ct/px in overlap region): using the sky. '
      endif
   endif
   logprint, '             -> ', removepath(outfilename)

   sxaddpar, hout, 'REDUCED',  repstr(tstamp(/ut),' ', 'T'), ' Timestamp of connect reduction step'
   hist = 'Processed by lrisconnect '+tstamp(/ut)+' UT'
   sxaddhist, hist, hout;, /history

   whist = where(strpos(hout,'HISTORY') eq 0 or strpos(hout,'END') eq 0, ct, complement=wnothist) ; put history keys at the end
   if ct gt 0 then hout = [hout[wnothist], hout[whist]]

   sxaddhist, 'Wavelengths are Angstroms in air', hout, /comment
   sxaddhist, 'Fluxes are erg/cm^2/s/Ang; absolute calibration is approximate', hout, /comment

   writespec, outfilename, smerge, hout
   ;q = get_kbrd()
end

