;+
; NAME:
;       lrisplot
;
; PURPOSE:
;       Quickly plot a fully- or partially-reduced LRIS spectrum to the screen
;       or to a file.
;
; CALLING SEQUENCE:
;       lrisplot, filename, outplot=
;
; EXPLANATION:
;       This is a basic (and older) plotting tool for examining a single 
;       spectrum.  
;       The plot is not interactive and there is no ability to control the
;       axis ranges, etc.  
;       lrisvalidate is a much more powerful tool.
;
; INPUTS:
;       filename = The filename to plot.
;
; OPTIONAL INPUT KEYWORDS:
;       outplot = The output file.  If it is not a string (e.g. if specified
;                 as /outplot) then it will plot to lris.eps.  If omitted,
;                 the plot will appear in a screen window.
;
; OUTPUTS:
;       The plot is written to disk if outplot is set.
; 
; EXAMPLES:
;       lrisplot, 'spectrum.spec'
;
; NOTES:
;       See also lrisplotall and lrisvalidate for more powerful inspection
;       tools.
;
;-
;------------------------------------------------------------------------------

pro lrisplot, filename, outplot=outplot

  if n_elements(filename) eq 0 then begin
    print, ' Need to specify spectrum file name to plot.'
    return
  endif
  if file_test(filename) eq 0 then begin
     print, ' '+filename+' not found'
     return
  endif

  if n_elements(outplot) gt 0 then begin
    if size(outplot,/tn) eq 'STRING' then outplotname = outplot else outplotname = 'lris.eps'
    psopen, outplotname, xsize=11, ysize=8.5, /inches, /landscape
    !p.multi = 0; [0,1,4]
    !p.position = 0
    !p.font = 0
    device, /helvetica, font_index=17
    device, /color
    colors = transpose([[0,0,0],$
                      [255,0,0], $
                      [100,100,100], $
                      [180,180,210],$ 
                      [64, 64, 192],$ 
                      [64, 136,128],$ 
                      [64, 158, 64],$  
                      [112,112, 64],$ 
                      [192, 64, 64],$
                      [128, 64, 128]$
                       ])
    tvlct, colors
    title = '!17'
  endif else begin
    window, 0, xsize=1800, ysize=500
    title = ''
  endelse

  ;filename = filestr
  s = readspec(filename, h)

  targname = sxpar(h,'TARGNAME')
  date = strmid(sxpar(h,'DATE'),0,strpos(sxpar(h,'DATE'),'T'))

  if strpos(targname,'_S') gt 0 then targname = strmid(targname,0,strpos(targname,'_S')) 
  print, targname, ' (', removepath(filename), ')'

  nx = n_elements(s)
  wn = where(s.wav gt 3300 and s.wav lt 10000 and s.wav lt max(s.wav)*0.95 and s.wav gt min(s.wav) * 1.05)
  xrange = [3100, 10350]
  yrange = [0,1.05*max(s[wn].flux)]
  plot, s.wav, s.flux, xrange=xrange, yrange=yrange, /xstyle, /ystyle, title=title
  junct = where(s.pix-shift(s.pix,1) lt 0) ; very crude; if points have been removed this will cause problems
  junct = [junct, n_elements(s)-1]
  for j = 0, n_elements(junct)-2 do begin
       midi = fix((junct[j+1]+junct[j])/2)
       nconv = round(10./(s[midi].wav-s[midi-1].wav)) ; convolve to 10 angstroms
       if nconv mod 2 eq 0 then nconv += 1 ; odd number
       convflux = convol(s.flux,(1.+fltarr(nconv))/nconv)
       oplot, s.wav, convflux, color=rgb(255,0,0)
       ;res = 3.  ; interpolation doesn't really make a difference; makes the file smaller presumably
       ;iwav = min(s.wav)+3*findgen((max(s.wav)-min(s.wav))/3.+1)
       ;iflux = interpol(convflux,s.wav,iwav)
       ;oplot, iwav, iflux, color=1
       if j gt 0 then oplot, s[junct[j]].wav+[0,0], [yrange[0],yrange[1]], color=rgb(0,0,255), linestyle=1
  endfor
  if yrange[0] lt 0 then oplot, xrange, [0,0], color=rgb(0,0,255)

  xyouts, 8500, yrange[1]*0.85, title+targname
 
  if n_elements(outplot) gt 0 then begin
    psclose
    title = 'LRIS spectrum'
    retitleplot, outplotname, title
  endif

end

