;+
; NAME:
;       lrisplotall
;
; PURPOSE:
;       Plot many processed LRIS spectra to a file.
;
; CALLING SEQUENCE:
;       lrisplotall, filestr=, order=
;
; EXPLANATION:
;       A plotting tool to quickly produce a many-page postscript plot with
;       a series of LRIS spectra.  Usually used for a quick inspection of
;       the output of LPipe (and is now run automatically by that package
;       following processing.)
;
; OPTIONAL INPUT KEYWORDS:
;       files = A wildcard file search string (or single file name) to 
;                 select which spectra to plot.
;       order = A keyword to sort the files in a specific way: usually 
;                 by time.  By default the plot order matches the file name
;                 alphabetical order.
;
; OUTPUTS:
;       The file lrisspectra.ps is written to disk.
; 
; EXAMPLES:
;       lrisplotall, order='time'
;
; NOTES:
;
;
;-
;------------------------------------------------------------------------------

pro lrisplotall, filestr=filestr, order=order  ;standards=standards

  if n_elements(filestr) eq 0 then $
    all = file_search('lris*.spec') $ 
  else $
    all = file_search(filestr)
  all = all[sort(all)]
  
  if n_elements(order) then begin 
     imh = lrisheaderinfo(all)
     if order eq 'ut' or order eq 'time' then all = all[sort(imh.ut)]
     if order eq 'config' then all = all[sort(imh.grname)]
  endif
 
  psopen, 'lrisspectra.ps', xsize=8.5, ysize=11, /inches ;, /seascape
  !p.multi = [0,1,5]
  !p.position = 0
  !p.font = 0
  device, /helvetica, font_index=17
  device, /color
  colors = transpose([[0,0,0],$
                      [255,0,0], $
                      [100,100,100], $
                      [180,180,210],$ 
                      [64, 64, 192],$ 
                      [64, 136,128],$ 
                      [64, 158, 64],$  
                      [112,112, 64],$ 
                      [192, 64, 64],$
                      [128, 64, 128]$
                       ])
  tvlct, colors

  daterange = ['','']
  for f = 0, n_elements(all)-1 do begin
    filename = all[f]
    if filename eq '' then continue
    if file_test(filename) eq 0 then continue ; this shouldn't occur
    s = readspec(filename, h)

    targname = sxpar(h,'TARGNAME')
    date = strmid(sxpar(h,'DATE'),0,strpos(sxpar(h,'DATE'),'T'))
    if daterange[0] eq '' then daterange = [date, date]
    if date lt daterange[0] then daterange[0] = date
    if date gt daterange[1] then daterange[1] = date

    if strpos(targname,'_S') gt 0 then targname = strmid(targname,0,strpos(targname,'_S')) 
    print, '  Plotting ', targname, ' (', removepath(filename), ')'
    if keyword_set(standards) eq 0 then begin
      if strpos(targname,'BD') ge 0 then continue
      if strpos(targname,'HD') ge 0 then continue
      if strpos(targname,'Feige') ge 0 then continue
      if strpos(targname,'Kopff') ge 0 then continue
    endif

    nx = n_elements(s)
    wn = where(s.wav gt 3300 and s.wav lt 10000 and s.wav lt max(s.wav)*0.95 and s.wav gt min(s.wav) * 1.05)
    xrange = [3100, 10350]
    yrange = [0,1.05*max(s[wn].flux)]
    plot, s.wav, s.flux, xrange=xrange, yrange=yrange, /xstyle, /ystyle, title='!17'
    junct = where(s.pix-shift(s.pix,1) lt 0)
    junct = [junct, n_elements(s)-1]
    for j = 0, n_elements(junct)-2 do begin
       midi = fix((junct[j+1]+junct[j])/2)
       nconv = round(10./(s[midi].wav-s[midi-1].wav)) ; convolve to 10 angstroms
       if nconv mod 2 eq 0 then nconv += 1 ; odd number
       convflux = convol(s.flux,(1.+fltarr(nconv))/nconv, /edge_truncate)
       oplot, s.wav, convflux, color=1
       ;res = 3.  ; interpolation doesn't really make a difference; makes the file smaller presumably
       ;iwav = min(s.wav)+3*findgen((max(s.wav)-min(s.wav))/3.+1)
       ;iflux = interpol(convflux,s.wav,iwav)
       ;oplot, iwav, iflux, color=1
       if j gt 0 then oplot, s[junct[j]].wav+[0,0], [yrange[0],yrange[1]], color=2, linestyle=1
    endfor
    if yrange[0] lt 0 then oplot, xrange, [0,0], color=2

    xyouts, 8500, yrange[1]*0.85, '!17'+targname
  endfor

  psclose
  if daterange[0] eq daterange[1] then title = 'LRIS spectra from '+daterange[0] else $
                                       title = 'LRIS spectra from '+daterange[0]+' to '+daterange[1]
  retitleplot, 'lrisspectra.ps', title

  !p.position = 0
  !p.multi = 0

end
