;+
; NAME:
;       lrisplotprofiles
;
; PURPOSE:
;       Plot all the spatial (slit) profiles for one camera.
;
; CALLING SEQUENCE:
;       lrisplotprofiles, camera=
;
; EXPLANATION:
;       Diagnostic tool for checking aperture placement/alignment.  
;
; INPUT KEYWORDS:
;       camera - 'r' or 'b' to indicate the camera to plot.
;
; OUTPUTS:
;       Writes a plot file to disk.
; 
; NOTES:
;       Largely deprecated by lrisapertures (and lrisvalidate).
;       See also lrisplotprofiles2, which pairs red and blue.
;
;-
;------------------------------------------------------------------------------


pro lrisplotprofiles, camera=camera

   common lrisauto, lpar

   if n_elements(lpar) eq 0 then begin
      print, " Must run autolrisdefaults,mode='s' to use lrisplotprofiles at command line."
      ; Could just run it...?
      return
   endif

   objfile = lpar.spworkingdir1d+'objectpos.txt'
   userobjfile = 'userobjectpos.txt'

   camchar = strlowcase(strmid(camera,0,1))

   if file_test(objfile) eq 0 then begin
      print, ' No object position file found.  Need to run tracing step.'
      return
   endif

   tr = loadobjectposfile(objfile, ct=ntr)
   usertr = loadobjectposfile(userobjfile, ct=nutr)
   for u = 0, nutr-1 do begin
      w = where(usertr[u].file eq tr.file, ctmatch) 
      if ctmatch gt 0 then tr[w[0]] = usertr[u]      ; should mark user-specified profiles on the plot
   endfor

   ofile = tr.file
   ocen = tr.yfix
   orad = tr.aprad

   ; Plot the profiles

   proffiles = file_search(lpar.spworkingdir1d+'ls'+camchar+'*.profile', count=nfiles) 
   fitsfiles = repstr(proffiles,'.profile','.fits')
   fileh = readlrisheaders(fitsfiles)
   profileplotfile = 'profiles'+camchar+'.ps'


     ; plot all the spatial profiles

      psopen, profileplotfile, xsize=7.5, ysize=10, /inches
      device, /color
      !p.font=0
      device, /helvetica, font_index=17 
      colors = transpose([[0,0,0],$
                      [255,0,0],$ 
                      [100,100,0],$ 
                      [0,200,0],$ 
                      [160,160,0],$  
                      [10,255,10],$
                      [100,100,100]])
      tvlct, colors
 
      !p.multi = [0,2,10]
      ybars = 0.04+0.95*findgen(11)/10
      py = 9
      dum = ''
      for f = 0, nfiles-1 do begin
         profilefile = proffiles[f]
         ; print, ' - ', profilefile
         fitsfilename = repstr(removepath(proffiles[f]), '.profile', '.fits')
         fileroot = repstr(fitsfilename,'.fits','')
         readcol, profilefile, y, prof, format='f,f', comment='#', /silent

         ow = where((ofile eq fitsfilename) or (ofile eq fileroot), ns)
         if ns eq 0 then begin
            ; print, 'No entry for ', fitsfilename, ' in '+objfile ; this is just a plotting routine
            continue
         endif

         for pi = 0, 1 do begin
           if pi eq 0 then !p.position = [0.04, ybars[py], 0.27, ybars[py+1]]
           if pi eq 1 then !p.position = [0.32, ybars[py], 0.99, ybars[py+1]]
           if pi eq 0 then begin
             xrange = ocen[ow[0]]+[-30,30]/fileh[f].spatbin
             closey = where(y ge xrange[0] and y le xrange[1])
             yrange = [1.1*min(prof[closey]) < (-1), 1.05*max(prof[closey])]
           endif
           if pi eq 1 then begin
             delvarx, xrange
             goody = where(y gt 5 and y lt max(y)-10)
             yrange = [1.1*min(prof[goody]) < (-1),  1.05*max(prof[goody])]
           endif
           plot, y, prof, /xsty, xrange=xrange, yrange=yrange, /ysty, ytickformat='(I)'
           oplot, [0,max(y)], [0,0], color=6
           for si = 0, ns-1 do begin
             s = ow[si]
             oplot, ocen[[s,s]], [-1e5,1e5], color=1
             xyouts, ocen[s]+1, max(yrange)-(max(yrange)-min(yrange))*0.1, clipzero(ocen[s]), orient=-90, charsize=0.6, color=1
             oplot, ocen[[s,s]]-orad[[s,s]], [-1e5,1e5], color=3, linestyle=2
             oplot, ocen[[s,s]]+orad[[s,s]], [-1e5,1e5], color=3, linestyle=2
             oplot, ocen[[s,s]]-tr[s].bakmin[0], [-1e5,1e5], color=2, linestyle=1
             oplot, ocen[[s,s]]-tr[s].bakmax[0], [-1e5,1e5], color=2, linestyle=1
             oplot, ocen[[s,s]]+tr[s].bakmin[1], [-1e5,1e5], color=2, linestyle=1
             oplot, ocen[[s,s]]+tr[s].bakmax[1], [-1e5,1e5], color=2, linestyle=1
           endfor
         endfor
         xyouts, 0.7, ybars[py+1]-0.02, fileroot, /norm, charsize=0.77
         ; xyouts, 0.7, ybars[py+1]-0.15, '@ x='+clip(xref), /norm, charsize=0.59

         py -= 1
         ;if py lt 0 then dum = get_kbrd()
         ;if dum eq 'q' then break
         if py lt 0 then py = 9
      endfor 
      psclose
      retitleplot, profileplotfile, camchar eq 'r' ? 'LRIS Red Profiles' : 'LRIS Blue Profiles'


end
