;+
; NAME:
;       lrisplotresponses
;
; PURPOSE:
;       Produce a combined response plot for all solved spectrophotometric
;       standard-stars processed during a run.
;
; CALLING SEQUENCE:
;       lrisplotresponses, camchar
;
; EXPLANATION:
;       Produce a final summary plot with all response curves.
;       Helpful for checking consistency in solutions and for checking
;       if the transmission/focus was good.
;
; INPUTS:
;       camchar - A one-character specifier ('b' for blue or 'r' for red)
;
; OUTPUTS:
;       A series of plots (one for each grism).
; 
; EXAMPLES:
;       lrisplotresponses, 'r'
;
;
;-
;------------------------------------------------------------------------------


pro lrisplotresponses, camchar

    common lrisauto, lpar

    respfiles = findfile(lpar.responsedir+'els'+camchar+'*.response')
    ;spfiles   = repstr(respfiles,'.response','.spec') ; no longer required; header is copied completely
    if n_elements(respfiles) eq 1 and respfiles[0] eq '' then return ; this assumes nothing comes after
    resph = lrisheaderinfo(respfiles)
    gr = unique(resph.grname)
    for g = 0, n_elements(gr)-1 do begin
      grstr = repstr(gr[g],'/','-')
      w = where(resph.grname eq gr[g], nsol)
      crespfiles = respfiles[w]
      crh = resph[w]

      psopen, lpar.spworkingdir1d+'response'+camchar+grstr+'.eps', xsize=11, ysize=6, /inches, /encaps
      device, /color
      !p.font=0
      !p.multi = 0
      device, /times, font_index=17                                  
      colors = transpose([[0,0,0],$
                      [255,50,50],$ 
                      [120,120,40],$ 
                      [50,200,50],$ 
                      [50,190,190],$
                      [60,60,255],$
                      [140,40,140],$
                      [200,200,200]])
      tvlct, colors
      ncols = 6; excludes black and gray
      minwav = 100000.
      maxwav = 0.
      maxresp = 0.
      for i = 0, n_elements(crh)-1 do begin
         readcol, respfiles[w[i]], rwav, rcounts, rflux, rresponse, rtelluricresponse, format='f,f,f,f,f', comment='#', /silent
         ; a blank response file will crash here
         minwav = minwav < min(rwav)
         maxwav = maxwav > max(rwav)
         sph = lrisheaderinfo(respfiles[w[i]])
         maxresp = maxresp > max(rresponse/sph.specbin)
      endfor
      ;print, minwav, maxwav, maxresp
      plot, [0], [0], xrange=[minwav, maxwav], yrange=[maxresp/50., maxresp*1.1], /xsty, /nodata, /ysty, /ylog, xtitle='Wavelength', ytitle='Response (counts / detector column)'
      for i = 0, n_elements(crh)-1 do begin
         ;print, respfiles[w[i]]
         ;r = readspec(respfiles[w[i]],h)
         readcol, respfiles[w[i]], rwav, rcounts, rflux, rresponse, rtelluricresponse, format='f,f,f,f,f', comment='#', /silent
         
         ;spf = spfiles[w[i]]
         sph = lrisheaderinfo(respfiles[w[i]])
         rresponseb = rresponse*(1./sph.specbin) ;*(1./sph.spatbin) only xbin matters, y is summed over.

         if camchar eq 'b' then ymax = 1e16
         if camchar eq 'r' then ymax = 8e16
         ;oplot, rwav, rresponse, color=1+i
         oplot, rwav, rresponseb*rtelluricresponse, color=1+(i mod ncols)
         wmax = where(rresponseb eq max(rresponseb))
         xyouts, rwav[wmax]-700, rresponse[wmax], strmid(repstr(repstr(removepath(respfiles[w[i]]), '.response', ''),'els',''),8), color=1+(i mod ncols)

         ; Effective area calculation below, not currently used for anything
         dwav = rwav-shift(rwav,1)
         dwav[0] = dwav[1]
         gain = 1
         if camchar eq 'r' and lpar.lrisversion eq 2 then gain = 1.9
         if camchar eq 'r' and lpar.lrisversion eq 2 then gain = 1.162 
         if camchar eq 'r' and lpar.lrisversion eq 3 then gain = 1.185  ; note independent of GAIN keyword...
         if camchar eq 'b' then gain = 1.63
         h = 6.6260755e-27
         nu = 2.998e18/rwav
         effectivearea_m2 = (rresponse * rtelluricresponse * gain * h * nu / dwav)/1e4
         fullsystemthoroughput = effectivearea_m2 / 73.3

           resparea20 = 20.*dwav*1e4/(gain * h * nu) ; differs a little for different wavelength solutions.
           oplot, rwav, resparea20, color=7 ; 20m^2 effective area curve (27% efficiency)
           resparea10 = 10.*dwav*1e4/(gain * h * nu)
           oplot, rwav, resparea10, color=7 ; 10m^2 effective area curve (13% efficiency)
           resparea5  =  5.*dwav*1e4/(gain * h * nu)
           oplot, rwav, resparea5,  color=7 ; 5m^2 effective area curve (6% efficiency)
         if i eq 0 then begin
           if resparea20[10] lt ymax then $
             xyouts, rwav[10], resparea20[10], /data, '27% efficiency', color=7, charsize=0.7
           if resparea10[10] lt ymax then $
             xyouts, rwav[10], resparea10[10], /data, '13% efficiency', color=7, charsize=0.7
           if resparea5[10] lt ymax then $
             xyouts, rwav[10], resparea5[10], /data, '6% efficiency', color=7, charsize=0.7
         endif
      endfor

      psclose
      retitleplot, lpar.spworkingdir1d+'response'+camchar+grstr+'.eps', strupcase(camchar)+gr[g]+' response curves'
    endfor 

end

