;+
; NAME: 
;       lrisprepare
;
; PURPOSE: 
;       All-purpose program to perform basic operations on an LRIS
;       image (or 2D spectrum) including bias subtraction, flagging
;       of bad pixels, header standardization, provisional astrometry,
;       etc.
;       Auto-detects whether it is red or blue, which LRIS version
;       it is, and whether it is imaging or spectroscopy.
;       This the latest, final, catch-all version.
;       It does not yet work with old-format images, but should work
;       with all modern LRIS imaging/spectroscopy regardless of amp
;       status, binning, etc.
;
; CALLING SEQUENCE: 
;       lrisprepare, filename, [options]
;
; INPUTS:
;	filename = name of the multi-HDU FITS file to prepare,
;                  or array of filenames, or file w/list of filenames
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       split       - write out left and right chips separately (default)
;       merge       - keep left and right chips together
;       multi       - like split but make a multi-extension file
;       rightonly 
;       leftonly
;       crops       - crop instruction ('auto' means use header information)
;       xcrop       - crop locations
;       lcrop       -
;       rcrop       -
;       ycrop       -
;       outname     - specify output file to write to disk
;       loutname
;       routname
;       prefix      - specify output prefix
;       lprefix     -
;       rprefix     -
;       suffix      - specify output suffix
;       lsuffix     -
;       rsuffix     -
;       nobias
;       scalarbias
;       gain        - gain correction for all four (two if LRISR1) amps
;       flag        - do not use?? implemented inconsistently
;       fixdefect   - replaces flagged values with interpolations
;       verbose
;       namefix     - correct names using a catalog
;       timer       - print out processing times
;       matchfile   - path+name of matching red/blue file (input)
;       setredwav   - replace central wavelength (for LRISR4)
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;       header (?)
; 
; OUTPUTS:
;       No formal outputs.
;       The prepared images are written to disk.
;
; REQUIREMENTS:
;       - Requires the IDL Astronomy User's Library routines 
;       (http://idlastro.gsfc.nasa.gov/) 
;       - Requires readmhdulp.pro to process LRIS images from 2009/later.

; Note - now attempts to add physical coordinates based on binned data.
; However, there is a genuine physical shift for binned observations - when binned 2x2,
; an extra row of dead pixels is read out along the edge of the gap (on the r chip and 
; looks like on the l chip too.)  This applies to R3; I don't know if it applies to R2.

;----------------------------------------------------------------------------------------
pro lrisprepare, filename, split=split, merge=merge, multi=multi, rightonly=rightonly, leftonly=leftonly, $
        crops=crops, xcrop=xcrop, lcrop=lcrop, rcrop=rcrop, ycrop=ycrop, notranspose=notranspose, $
        outname=outname, loutname=loutname, routname=routname, $
        prefix=prefix, lprefix=lprefix, rprefix=rprefix, $
        suffix=suffix, lsuffix=lsuffix, rsuffix=rsuffix, $
        nobias=nobias, scalarbias=scalarbias, gain=gain, verbose=verbose, $
        fixdefect=fixdefect, flag=flag, namefixfiles=namefixfiles, header=header, timer=timer, seq=seq, $
        requirecompleteheader=requirecompleteheader, newobj=newobj, $
        reducer=reducer, version=version, errcode=errcode, $
        matchfile=matchfile, keepmira=keepmira, $
        setredwav=setredwav
   ; arguments repeated once

common lrisauto, lpar
common lrisbpmblock, bpm1, bpm2, bpm3 
common lrisprepareprev, mask, prevcamera, prevversion, prevwindet, prevxbin, prevybin
common lrisbadarc, badarcdir, badarclist
common lrisdrop, dropcolumns
common lrishost, hostname, username
common lrishi, hicam, hikeys, hicomments, delkeys ; header keys to keep (with comments to add) and keys to drop


errcode = 0

if n_elements(lpar) gt 0 then begin
   bpmpath = lpar.refdatapath
endif else begin
   bpmpath = '/home/aridperl/progs/idl/lris/refdata/' ; allows lrisprepare to be run independently for testing
   if file_test(bpmpath) eq 0 then bpmpath = '/Users/dperley/code/idl/lris/refdata/'
endelse
if strlen(bpmpath) gt 0 then if strmid(bpmpath,strlen(bpmpath)-1,1) ne '/' then bpmpath += '/'

if keyword_set(merge) and keyword_set(split) then begin
   print, 'Cannot both split and merge.'
   errcode=-1
   return
endif
if keyword_set(merge) eq 0 and keyword_set(split) eq 0 then begin
   merge = 1   
endif
merge = keyword_set(merge)
split = keyword_set(split)
multi = keyword_set(multi)

if n_elements(crops) eq 0 and $
   n_elements(xcrop) eq 0 and n_elements(lcrop) eq 0 and n_elements(rcrop) eq 0 and n_elements(ycrop) eq 0 then crops = 'auto'

; ---------- Process input filename(s), call recursively if necessary----------

; Check for empty filename
if n_elements(filename) eq 0 then begin
   print, 'No filename specified.'
   errcode=-1
   return
endif

; Check for array input
if n_elements(filename) gt 1 then begin
   files = filename
endif

; Check for list input
if n_elements(filename) eq 1 then begin
   filearr = strsplit(filename,'.', /extract)
   if n_elements(filearr) lt 2 then begin
      print, '???'
      errcode=-2
      return
   endif
   fileroot = filearr[0]
   fileext = filearr[1]
   if fileext eq 'cat' or fileext eq 'lis' or fileext eq 'list' or fileext eq 'txt' then begin
     ;files = grabcolumn(filename,0,/str)
     readcol, filename, files, format='a'
   endif
endif


; Check for wildcard input
if n_elements(filename) eq 1 then begin
   starpos = strpos(filename,'*')
   qpos = strpos(filename,'?')
   if starpos ge 0 or qpos ge 0 then begin
      files = file_search(filename)
      if n_elements(files) eq 1 then begin
        if files[0] eq '' then print, 'Cannot find any files matching ', filename
        errcode=-1
        return
      endif
   endif
endif

; Check for array filename, loop recursively if necessary
if n_elements(files) gt 1 then begin
  for f = 0, n_elements(files)-1 do begin
    lrisprepare, files[f], split=split, merge=merge, rightonly=rightonly, leftonly=leftonly, $
        crops=crops, xcrop=xcrop, lcrop=lcrop, rcrop=rcrop, ycrop=ycrop, notranspose=notranspose, $
        outname=outname, loutname=loutname, routname=routname, $
        prefix=prefix, lprefix=lprefix, rprefix=rprefix, $
        suffix=suffix, lsuffix=lsuffix, rsuffix=rsuffix, $
        nobias=nobias, scalarbias=scalarbias, gain=gain, verbose=verbose, $
        fixdefect=fixdefect, flag=flag, namefixfiles=namefixfiles, timer=timer, header=header, seq=seq, $
        requirecompleteheader=requirecompleteheader, newobj=newobj, $
        reducer=reducer, version=version, errcode=errcode, $
        matchfile=matchfile, keepmira=keepmira  ; seq won't work for an array
  endfor
  return
endif


logprint, ' '+filename ;, format='($,A)'


; ---------- Read data and process header information ----------

; Read in the data and header

starttime = systime(/seconds) 
linebias = keyword_set(scalarbias) eq 0

check = file_test(filename)
if check eq 0 then begin
  logprint, ' not found!'
  errcode = -1
  return
endif

h = headfits(filename) ; re-read later as variable header...

; Notes on CCDs:

; LRIS-B:
;   AMP1
;  L1 (far left/bottom)  AMP1 00151-14-1L  Gain 
;  L2 (near left/bottom) AMP2 00151-14-1R
;  R1 (near right/top)   AMP3 00151-19-1L
;  R2 (far right/top)    AMP4 00152-19-1R

; LRIS-R3:
;  AMP=VidInp (order in FITS file).  AMPLOC (&CCDLOC) gives sequence from bottom.
;  L1 (far left/bottom):   AMP2, 19-3R  Gain 1.180   RN 4.76   401-1024
;  L2 (near left/bottom):  AMP1, 19-3L  Gain 1.255   RN 4.64  1025-2048
;  R1 (near right/top):    AMP4, 19-2R  Gain 1.162   RN 4.62  2049-3072 * sets the normalization *
;  R2 (far right/top):     AMP3, 19-2L  Gain 1.191   RN 4.54  3073-3800

; Note: LRIS has been on Keck I since 2000-02-24.
; It was on Keck II from 1996-10-01 through 2000-02-06.

; CCD information:
;   LRIS red CCD2: installed/first light 2009-06-06
;                  CTE problems begin                          ~ 2009-09
;                  CTE problems affect all but main amplifier  ~ 2010-01
;                  CTE problems affect main amplifier starting ~ 2010-10
;   LRIS red CCD3: installed 2010-11-29 (with new dewar)
;   LRIS red CCD4: installed 2021-05    (date unclear)

; Other notes:
;   831/8200 grating was replaced in Oct/Nov 2008
;   I-band filter shattered July 2007, was replaced July 2008
;   ADC installed 2007 July
;   Cd/Zn installed 2001-04, replaced Kr/Xe.

; Will need to account for shutter failures somehow

; Determine which red CCD was employed based on the date.
; Obviously this is not a very good way to do it.
; Some blue keywords seem to have changed with the red upgrade, too,
; so this is calculated for either.

;hdate = sxpar(header,'DATE', count=checkdate, /silent)
;if checkdate gt 0 and clip(hdate) ne '0' then begin
;   datestr = (strsplit(hdate,'T',/extract)) [0]
;   datearr = strsplit(datestr,'-',/extract)
;endif

vstr = sxpar(h,'INSTREV',count=ctv,/silent)
mjd = float(sxpar(h,'MJD-OBS', count=checkmjd, /silent))
if checkmjd eq 0 or clip(mjd) eq '0' then begin
   mjd = float(sxpar(h,'MJD', count=checkmjd, /silent))
endif
;if checkmjd eq 0 or clip(mjd) eq '0' and n_elements(datearr) eq 3 then begin
;   jdcnv, datearr[0], datearr[1], datearr[2], 0, jd ; just an estimate for now
;   mjd = jd-2400000.5d
;endif

; Determine the version (CCD) of LRIS being processed
; Use the header information directly if possible
if ctv gt 0 then begin
  if vstr eq 'REDMARK4' then lrisversion = 4
endif
; Otherwise, figure it out from the date
if n_elements(lrisversion) eq 0 then begin
  if mjd gt 0 and mjd lt 49000 then begin
    print
    logprint, ' WARNING: Observation date does not make sense!'
    logprint, '   MJD-OBS  '+clip(sxpar(h,'MJD-OBS'))
    logprint, '   DATE-OBS '+clip(sxpar(h,'DATE-OBS'))
    logprint, ' Assuming this is recent data.  If observations predate 2011 this will cause problems.'
    lrisversion = 3
  endif
  if mjd gt 49000 and mjd lt 54983 then lrisversion = 1 ; installation to 2009-05-31
  if mjd ge 54983 and mjd lt 55529 then lrisversion = 2 ; 2009-06-01 to 2010-11-28
  if mjd ge 55529 and mjd lt 59340 then lrisversion = 3 ; 2010-11-29 to 2021-05-06 (end date uncertain)
  if mjd ge 59340 then                  lrisversion = 4 ;
  if n_elements(lrisversion) eq 0 then begin
    year = strmid(sxpar(h,'DATE'),0,4)
    if fix(year) gt 2021 then lrisversion = 4
    if fix(year) gt 2010 then lrisversion = 3
    if fix(year) eq 2010 then lrisversion = 2
    if fix(year) lt 2010 and fix(year) gt 1993 then lrisversion = 1
    if n_elements(lrisversion) eq 0 then begin
      logprint, '  [Assuming LRIS3]' ;, format='($,A)'
      lrisversion = 3
    endif
  endif
endif

if sxpar(h, 'NAXIS') eq 2 then begin
   ;fits_info, filename, /SILENT, N_ext=n_ext, extname=extname  (an alternative way to do this)

   ; Single-extension LRIS file format

   if clip(sxpar(h,'INSTRUME')) eq 'LRIS' then begin
     if lrisversion eq 1 then begin
       ; Old LRIS red camera
       array = readlrisr1(filename, header=header, nobias=nobias, verbose=verbose) 
     endif
     if lrisversion eq 2 or lrisversion eq 3 then begin
        ; LRIS version 2 or 3, but old-style simple LRIS file format
        array = readsimplelris(filename, header=header, nobias=nobias, linebias=linebias, verbose=verbose) 
     endif
     if lrisversion eq 4 then begin
       array = readlrisr4(filename, header=header, nobias=nobias, verbose=verbose, setredwav=setredwav)
     endif
   endif
endif else begin

   ; Multi-extension LRIS file format
   
   array = readmhdufitslp(filename, header=header, nobias=nobias, linebias=linebias, verbose=verbose) ; do gain later

endelse

delvarx, h ; delete basic header: not really necessary, but useful for debugging

sin = size(array) ; input size
nxin = sin[1]
nyin = sin[2]

; for old-style LRIS, at this point we will have to indicate where the amps are manually.

if keyword_set(keepmira) eq 0 and nxin eq 4096 and nyin eq 512 and $
 sxpar(header,'INSTRUME') eq 'LRISBLUE' and clip(sxpar(header,'SLITNAME')) eq 'direct' then begin
   logprint, ' - MIRA file; skipping.'
   errcode = +1
   return
endif

readtime = systime(/seconds)  ; mark time after file read from disk with readmhdu

; Determine whether this is the red or blue camera
instrume = clip(sxpar(header,'INSTRUME'))
if instrume eq 'LRIS'     then camera = 'r'
if instrume eq 'LRISBLUE' then camera = 'b'
if camera eq '' then begin
   logprint, ': unrecognized camera (INSTRUME = ', instrume, ')'
   errcode = -4
   return
endif

; ---- Reorganize default header and add comments -----

sxdelpar, header, ''           ; remove blank lines
sxdelpar, header, 'COMMENT'    ; remove comment lines

; sxpar crashes on lines with multiple sequential apostrophes for some reason.  Delete these outright.
;keep = [-1]
;for i = 0, n_elements(header)-1 do begin
; print, header[i]
; napos = n_elements(strsplit(header[i],"'",/preserve))-1
; if napos le 2 then keep = [keep, i]
;endfor
;header = header[keep[1:*]]

sxrenamepar, header, 'OBJECT', 'TEMPOBJ' ; remove duplicate keyword
sxrenamepar, header, 'TEMPOBJ', 'OBJECT'
sxrenamepar, header, 'OBSERVER', 'TEMPOBS' ; remove duplicate keyword
sxrenamepar, header, 'TEMPOBS', 'OBSERVER'
sxrenamepar, header, 'DATE', 'TEMPDATE' ; remove duplicate keyword  (R1)
sxrenamepar, header, 'TEMPDATE', 'DATE'

; in principle should also check any other free-text field, including OBSERVER.
catch, objerror
if objerror ne 0 then begin
  print, '  [problematic OBJECT value] ;', format='($,A)'
  sxaddpar, header, 'OBJECT', '', 'Object name deleted due to invalid format'
endif
test = sxpar(header,'OBJECT')
catch, /cancel ; the top-level error handler is still intact

refreshhi = 0
if n_elements(hikeys) eq 0 then refreshhi = 1
if n_elements(hikeys) gt 0 then if hicam ne camera then refreshhi = 1
if refreshhi then begin
  hicam = camera
  hinfofile = bpmpath+'headerkeywords.txt'
  hikeys = ['']
  hicomments = ['']
  delkeys = ['']
  openr, 19, hinfofile
  inline = ''
  while not eof(19) do begin
    readf, 19, inline
    if strlen(inline) eq 0 then continue
    if strmid(inline,0,1) eq '#' then continue
    newkey = clip((strsplit(inline,'=', /extract)) [0])
    pipepos = strpos(inline,'|')
    if pipepos gt 0 then begin
       newcommentstr = strmid(inline,pipepos+1)
       if total(strmid(newcommentstr,0,1) eq ['X','R','B']) gt 0 then begin
         if strmid(newcommentstr,0,1) eq 'X' then delkeys = [delkeys, newkey]
         if strmid(newcommentstr,0,1) eq 'R' and camera eq 'b' then delkeys = [delkeys, newkey]
         if strmid(newcommentstr,0,1) eq 'B' and camera eq 'r' then delkeys = [delkeys, newkey]
         newcomment = clip(strmid(newcommentstr,1))
       endif else begin
         newcomment = clip(newcommentstr)
       endelse
       hikeys = [hikeys, newkey]
       hicomments = [hicomments, newcomment]
    endif 
  endwhile
  if n_elements(hikeys) gt 1 then hikeys = hikeys[1:*]
  if n_elements(hicomments) gt 1 then hicomments = hicomments[1:*]
  if n_elements(delkeys) gt 1 then delkeys = delkeys[1:*]
  close, 19
endif

if delkeys[0] ne '' then begin
  ;print, delkeys
  for k = 0, n_elements(delkeys)-1 do sxdelpar, header, delkeys[k]
endif
header = header[where(strmid(header,0,7) ne 'COMMENT')]
reorderheader, header, hikeys
annotateheader, header, hikeys, hicomments

; Get other essential info about the data structure: 
;   camera (already done), version, binning, windowing, lmode

if camera eq 'r' then sxaddpar, header, 'CAMERA', 'red', ' LRIS camera', after='INSTRUME'
if camera eq 'b' then sxaddpar, header, 'CAMERA', 'blue', ' LRIS camera', after='INSTRUME'
sxaddpar, header, 'LRISVER', lrisversion, ' LRIS red version'
if camera eq 'b' then sxaddpar, header, 'DETECTOR', 'LRISB', ' LRIS detector mosaic'
if camera eq 'r' then sxaddpar, header, 'DETECTOR', 'LRISR'+clip(lrisversion), ' LRIS detector mosaic'

; Determine the binning
binning = clip(sxpar(header,'BINNING',comment=comment))
bin = fix(strsplit(binning,',',/extract))  
xbin = bin[0]
ybin = bin[1]
spatbin = xbin
specbin = ybin
if camera eq 'r' and lrisversion eq 4 then begin
  xbin = bin[1]
  ybin = bin[0]
  specbin = xbin
  spatbin = ybin
endif
sxaddpar, header, 'XBIN', xbin, ' Binning along (LRIS detector) x-axis', after='BINNING'  ; this becomes YBIN for spectroscopy
sxaddpar, header, 'YBIN', ybin, ' Binning along (LRIS detector) y-axis', after='BINNING'  ; this becomes XBIN for spectroscopy
 ; These are reversed in ltranspose at the very end of this routine.
 ; So they refer to the detector frame throughout lrisprepare.pro, but refer to
 ; the image display frame in all codes that modify the output.

; Determine the windowing.
; This is complicated because there are many different ways this is represented
; in the header for different CCDs.

; We specifically want the windowing in original detector coordinates.

panestr = clip(sxpar(header,'PANE',count=ctpane))
windowstr = clip(sxpar(header,'WINDOW',count=ctwindow))
if ctpane eq 0 and ctwindow eq 0 then begin
stop
   logprint, 'ERROR:  No windowing keyword data?  Cannot determine amplifier boundaries.'
   errcode = -5
   return
endif

windet = intarr(4) ; this will store the (unbinned) x0, y0, nx, ny

; PANE is always in unbinned coordinates: [x0, y0, dx, dy]
; WINDOW is in binned coordinates:   [0/1, x0, y0, dx, dy]
;  for LRIS-B, dy is PER CHIP, so x2 to get value for combined detector.

if camera eq 'r' then begin
   if ctpane gt 0 then begin
       ; No adjustments required for PANE on red side, when available (R2 and R3).
       pane = fix(strsplit(panestr,',',/extract))
       windet = pane
   endif else begin
      if lrisversion eq 1 then begin
        windowv = fix(strsplit(windowstr,',',/extract))
        if n_elements(windowv) ge 5 then begin
          windet[0] = windowv[1] * xbin
          windet[1] = (2048/ybin - windowv[4] - windowv[2]) * ybin ; formerly /xbin... not well-tested...
          windet[2] = windowv[3] * xbin
          windet[3] = windowv[4] * ybin
        endif 
        ; old code - doesn't seem sensible since x y mixed
        ;windowvals = fix(strsplit(windowstr,',',/extract))
        ;windowvals[2] = 2048/xbin - windowvals[4] - windowvals[2] 
        ;panevals[0] = windowvals[1] * xbin
        ;panevals[1] = windowvals[2] * ybin
        ;panevals[2] = windowvals[3] * xbin
        ;panevals[3] = windowvals[4] * ybin
      endif else begin
        logprint, '  WARNING: Missing PANE data for new LRIS red?'
      endelse
   endelse
endif   
if camera eq 'b' and ctwindow gt 0 then begin
   windowv = fix(strsplit(windowstr,',',/extract))
   if n_elements(windowv) ge 5 then begin
     windet[0] = windowv[1] * xbin
     windet[1] = windowv[2] * ybin
     windet[2] = windowv[3] * xbin * 2
     windet[3] = windowv[4] * ybin
   endif
endif

if total(windet) eq 0 then begin
   logprint, '  ERROR: Unable to determine windowing from header info.'
   errcode = -5
   return
endif

windetstr = ''
for p = 0, 3 do windetstr += clip(windet[p]) + (p lt 3 ? ',' : '')
sxaddpar, header, 'WINDET', windetstr, ' Windowing in detector coordinates [X0,Y0,DX,DY]'

wintype = 'CUSTOM'
if windetstr eq '0,0,4096,4096' then wintype = 'FULL'
if windetstr eq '1500,0,1200,4096' then wintype = 'LONGSLIT'
if windetstr eq '400,625,3400,2500' then wintype = 'IMAGE'
if lrisversion eq 1 and camera eq 'r' then begin
  if windetstr eq '0,0,2048,2048' then wintype = 'FULL'
  if windetstr eq '0,550,2048,1000' then  wintype = 'LONGSLIT'
  ; if windetstr eq '0,0,550,1000' then wintype = 'LONGSLIT' ???
endif
if n_elements(wintype) gt 0 then sxaddpar, h, 'WINTYPE', wintype, ' Standard windowing type'

; validate windowing (except R4 where it is hopeless)
if camera ne 'r' or lrisversion ne 4 then begin
if (windet[2] ne nxin*xbin) or (windet[3] ne nyin*ybin) then begin
  if (windet[2] ne 1200 or nxin ne 602) or (windet[3] ne nyin*ybin) then begin 
     ; 1200 != 602*2: this happens every time in 2x1 red binning, related to CCD bug?

    if panestr eq '0,0,4096,4096' and windowstr eq '1,0,0,1024,2048' and camera eq 'b' then begin
       print, '   ERROR: Corrupted blue file; skipping...'
       errcode = -10
       return
    endif

    logprint, ' WARNING: WINDET ('+clip(windet[2])+'x'+clip(windet[3])+')/('+repstr(binning,',','x')+' bin)'+$
            ' inconsistent with data ('+clip(nxin)+'x'+clip(nyin)+')'
   ; if ctpane gt 0 then print, ' PANE = ', panestr
   ; if ctwindow gt 0 then print, ' WINDOW = ', windowstr
  endif
           ; X Y convention here is original data, not transposed data      
endif
endif

lmode = '' ; mode based on disperser
if camera eq 'b' and clip(sxpar(header,'GRISNAME'))  eq 'clear'  then lmode = 'img'
if camera eq 'r' and clip(sxpar(header,'GRANAME'))  eq 'mirror' then lmode = 'img'
if camera eq 'b' and strpos(sxpar(header,'GRISNAME'),'/') gt 0   then lmode = 'spec'
if camera eq 'r' and strpos(sxpar(header,'GRANAME'),'/') gt 0    then lmode = 'spec'

smode = '' ; mode based on slit
slit =  strlowcase(clip(sxpar(header, 'SLITNAME')))
if slit eq 'direct' then smode = 'img'
if strpos(slit,'_') gt 0 then smode = 'spec'

   ; or smode eq 'img' and lmode eq 'spec'  -- this is now OK, slitless spectrum 
; if lmode eq 'img' and smode eq 'spec' then begin  --- this is not OK, slitmask image
   ; Note that this (and other errors) are OK in principle for biases/darks, but still the pipeline doesn't process them
;   logprint
;   logprint, '     WARNING: Mixed-up grating/slit settings!  Ignoring file...'
;   errcode = -6
;   return
;endif
if lmode eq '' and smode eq '' then begin
   logprint
   logprint, '     WARNING: grating and slit keywords corrupt, absent, or unusual.'
  if camera eq 'b' then logprint, '      GRISNAME:', clip(sxpar(header,'GRISNAME')), '    SLITNAME:', clip(sxpar(header,'SLITNAME')), '  FILTER:', clip(sxpar(header,'BLUFILT')), '  :', lmode, '?'
  if camera eq 'r' then logprint, '       GRANAME:', clip(sxpar(header,'GRANAME')), '  SLITNAME:', clip(sxpar(header,'SLITNAME')), '  FILTER:', clip(sxpar(header,'REDFILT')), '  :', lmode, '?'
   logprint, '              Skipping...'
   errcode = -6
   return
endif
if lmode eq '' and smode ne '' then lmode = smode


   ; unnecessary indentation here

   hdate = sxpar(header,'DATE', count=checkdate, /silent)
   ; htime = sxpar(header,'UT', count=checktime)
   hdateobs = sxpar(header,'DATE-OBS', count=checkdateobs, /silent)
   if checkdate gt 0 and checkdateobs eq 0 and lrisversion eq 1 then sxaddpar, header, 'DATE-OBS', clip(hdate)


   ; fill in red central wavelength if manual override required

   if lmode eq 'spec' and n_elements(setredwav) gt 0 and camera eq 'r' then begin
      print, '  [WAVELEN: '+clip(sxpar(header,'WAVELEN'))+' -> '+clip(setredwav)+']'
      sxaddpar, header, 'WAVELEN', setredwav
   endif


   ; Check for randomly missing keywords that can be filled in using the matching (red/blue) file,
   ; if one was supplied.  (Also write the match filename to the header.)

   checkkeysall = ['TARGNAME','TELESCOP','TELFOCUS','TUBETEMP','DOMEPOSN','FLIMAGIN','FLSPECTR',$
      'DRA','DDEC','DTRACK','CURRINST','RA','ROTMODE','ROTPOSN','ROTPPOSN',$
      'AIRMASS','AZ','DEC','EL','EQUINOX','HA','PONAME','MJD-OBS','ST',$
      'DATE-OBS','SLITNAME']
   if lpar.lrisversion eq 4 then checkkeysall = [checkkeysall,'OBJECT','TRAPDOOR']
   if camera eq 'r' then checkkeysall = [checkkeysall, 'REDFILT'] 
   if camera eq 'r' and lmode eq 'spec' or smode eq 'spec' then checkkeysall = [checkkeysall, 'WAVELEN']
   if camera eq 'b' then checkkeysall = [checkkeysall, 'BLUFILT'] 
     ; domeposn ,flimagin, flspectr, dra, ddec, dtrack, adcneed, currinst, adcprena, 
     ; secfocus, secthetx, secthety, simulate, mjd-obs all missing from early LRIS
     ; 'SECFOCUS','SECTHETX','SECTHETY','SIMULATE','ADCNEED','ADCPRENA',
     ; blue/red specific keywords found only in the blue/red files
     ; lamps have a different convention; 4th(Cd) is halogen; 5th(Zn) might be 'Kr'; 6th(hal) is some other line lamp

   if n_elements(matchfile) gt 0 then begin
      sxaddpar, header, 'MATCH', removepath(matchfile), ' Matching R/B file' ; Add the name of the match.
      ; print, ' ... matches '+removepath(matchfile)
      if file_test(matchfile) gt 0 then begin
        matchheader = headfits(matchfile)
        nmissing = 0
        ncopied = 0
        listcopykey = ''
        for k = 0, n_elements(checkkeysall)-1 do begin ; Look for missing keywords.
          test = sxpar(header, checkkeysall[k], count=ct, /silent)
          if ct eq 0 then begin
             nmissing += 1
             matchval = sxpar(matchheader, checkkeysall[k], count=mct, comment=matchcomment, /silent)
             if mct gt 0 then begin
                sxaddpar, header, checkkeysall[k], matchval, matchcomment ; Replace missing keywords.
                misskey = checkkeysall[k]
                if ncopied gt 0 then listcopykey += ','
                listcopykey += misskey
                ncopied += 1
             endif
          endif
        endfor
        if ncopied eq 1 then logprint, '  ['+misskey+' copied from '+repstr(repstr(removepath(matchfile),'.fits',''),'.fits','')+']'
        if ncopied gt 1 and ncopied le 5 then logprint, '  ['+listcopykey+' copied from '+repstr(repstr(removepath(matchfile),'.fits',''),'.fits','')+']'
        if ncopied gt 5 then logprint, '  ['+clip(ncopied)+'/'+clip(nmissing)+' missing keywords copied from '+repstr(repstr(removepath(matchfile),'.fits',''),'.gz','')+']'
      endif
   endif
   


   ; Check for any (still-)missing keywords that are required for the pipeline to work

   checkkeys =  ['OBJECT','TARGNAME','DOMEPOSN','INSTRUME',$,
                 'RA','ROTMODE','ROTPOSN','ELAPTIME','SLITNAME',$
                 'AIRMASS','AZ','DEC','EL','HA','MJD-OBS','ST','DATE-OBS',$
                 'TRAPDOOR']
   ;checkkeysim = ['REDFILT','BLUFILT']
   checkkeysred = ['REDFILT','GRANAME']
   checkkeysblue = ['BLUFILT','GRISNAME']
   ;checkkeyssp = ['SLITNAME']
   checkkeysredsp = ['WAVELEN']
   ;checkkeysbluesp = ['GRISNAME'] ; 'mirror' etc still needed...

   ;if lmode eq 'img' or smode eq 'img' then checkkeys = [checkkeysboth, checkkeysim]
   if lmode eq 'spec' or smode eq 'spec' then begin
      ;checkkeys = [checkkeys, checkkeyssp]
      if camera eq 'r' then checkkeys = [checkkeys, checkkeysredsp]
      ;if camera eq 'b' then checkkeys = [checkkeys, checkkeysbluesp]
   endif
   if camera eq 'r' then checkkeys = [checkkeys, checkkeysred]
   if camera eq 'b' then checkkeys = [checkkeys, checkkeysblue]
                ; also important but have changed convention: UT/UTC, LAMPS/(lamp names), WINDOW/PANE
                ; dome lamp keywords? 
   missing = ['']
   for k = 0, n_Elements(checkkeys)-1 do begin
      test = sxpar(header, checkkeys[k], count=ct, /silent)
      if ct eq 0 then missing = [missing, checkkeys[k]]
   endfor
   if n_elements(missing) gt 1 then begin
       misstr = ''
       for km = 0, n_elements(missing)-1 do misstr += ','+missing[km]
       misstr = strmid(misstr,2)
       logprint, '  missing the following keywords: '+misstr, format='($,A)'
       if keyword_set(requirecompleteheader) then begin
          expstr = clip(fix(sxpar(header,'EXPTIME', count=ct)))
          if clip(sxpar(header,'OBJECT')) ne '' then objstr = 'of '+clip(sxpar(header,'OBJECT')) else objstr = ''
          if ct eq 0 then expstr = clip(fix(sxpar(header,'ELAPTIME', count=ct)))
          logprint
          logprint, ' - not processing '+expstr+'s exposure '+objstr
          errcode = -7
          return
       endif
   endif


if clip(sxpar(header, 'UTC')) eq '0' then sxaddpar, header, 'UTC', sxpar(header, 'UT'), after='UT'
if clip(sxpar(header, 'UT')) eq '0' then sxaddpar, header, 'UT', sxpar(header, 'UTC'), after='UTC'

sxaddpar, header, 'BZERO', 0 
sxaddpar, header, 'BSCALE', 1
sxaddpar, header, 'NAXIS1', 0, after='NAXIS' ; just to save the location
sxaddpar, header, 'NAXIS2', 0, after='NAXIS1'


; Compatibility of lamps format (must be done before namefix, which uses lamp info)
lamps = sxpar(header, 'LAMPS', count=lampct) ; need to also check for old lamp ordering
if lampct eq 0 then begin  ; compatibility with older lamp format
      hg = clip(sxpar(header, 'MERCURY')) eq 'on'
      neon = clip(sxpar(header, 'NEON')) eq 'on'
      ar = clip(sxpar(header, 'ARGON')) eq 'on'
      cd = clip(sxpar(header, 'CADMIUM')) eq 'on'
      zn = clip(sxpar(header, 'ZINC')) eq 'on'
      hal = clip(sxpar(header, 'HALOGEN')) eq 'on'
      kr = clip(sxpar(header, 'KRYPTON')) eq 'on'
      xe = clip(sxpar(header, 'XENON')) eq 'on'
      fear = clip(sxpar(header, 'FEARGON')) eq 'on'
      de = clip(sxpar(header, 'DEUTERI')) eq 'on'
      lamps = clip(fix(hg))+','+clip(fix(neon))+','+clip(fix(ar))+','+clip(fix(cd))+','+clip(fix(zn))+','+clip(fix(hal)) $
                        +','+clip(fix(kr))+','+clip(fix(xe))+','+clip(fix(fear))+','+clip(fix(de))
      sxaddpar, header, 'LAMPS', lamps, before='MERCURY'
endif

; lamps are automatically turned off if trapdoor was open; correct this in header
trapdoor = strlowcase(clip(sxpar(header,'TRAPDOOR', count=ct)))
if ct eq 0 then trapdoor = '-'
if trapdoor eq 'open' then if strpos(lamps, '1') ge 0 then $
  sxaddpar, header, 'LAMPS', repstr(lamps,'1','0')
  ; note: doesn't correct the individal lamp entries ('NEON', etc.) but LAMPS takes precedence


view = getlrisview(header, /update)

sxaddpar, header, 'ORIGOBJ', sxpar(header,'OBJECT'), ' Object name as entered by user'
result = 0
if n_elements(namefixfiles) gt 0 then begin
   namefix, header, namefixfiles, result=result
endif
if n_elements(newobj) gt 0 and result eq 0 then begin
   origobj = sxpar(header,'OBJECT')
     logprint, '  [OBJECT: '+clip(origobj)+' -> '+newobj+'] ' ;, format='($,A)'
   sxaddpar, header, 'OBJECT', newobj
endif

exptime = sxpar(header, 'EXPTIME', count=ctexptime)    ; "shutter start to shutter close" (usually not an integer)
elaptime = sxpar(header, 'ELAPTIME', count=ctelaptime) ; "actual exposure time" (not including pauses???), always? integer
ttime = sxpar(header, 'TTIME', count=ctttime)          ; requested time (does NOT include impact of aborts!)
utbeg = sxpar(header, 'DATE_BEG', count=ctb)
utend = sxpar(header, 'DATE_END', count=cte)
if ctb lt 1 or cte lt 1 then begin
   utbeg = sxpar(header, 'UTC', count=ctb)
   utend = sxpar(header, 'UTC-END', count=cte)
endif
if ctb gt 0 and cte gt 0 then begin
   ; lack of T should be harmless below.  Ignores midnight wrap.
   difftime = 3600*(ten(strmid(utend,1+strpos(utend,'T')))-ten(strmid(utbeg,1+strpos(utbeg,'T')))) ; ignores date wrap
   sxaddpar, header, 'DIFFTIME', difftime, ' Tend-Tstart (sec)'
endif


if ctexptime eq 0 then begin ; swarp needs this
    if ctelaptime gt 0 then begin
       sxaddpar, header, 'EXPTIME', elaptime, ' copied from ELAPTIME', before='ELAPTIME'
    endif else begin
       if ctttime gt 0 then $
          sxaddpar, header, 'EXPTIME', ttime, ' copied from TTIME', before='TTIME'
    endelse
endif 
if ctelaptime eq 0 and n_elements(difftime) gt 0 then begin
    sxaddpar, header, 'ELAPTIME', sxpar(header,'EXPTIME'), ' calculated from EXPTIME', after='EXPTIME'
endif

; ELAPTIME and EXPTIME have different values for red images??


; Saturation on the help page keeps changing.  Currently claims both are 60k.
if camera eq 'b' then sxaddpar, header, 'SATURATE', 60000, ' Saturation level', before='BINNING'
if camera eq 'r' and lrisversion eq 1 then sxaddpar, header, 'SATURATE', 60000, ' Saturation level', before='BINNING'
if camera eq 'r' and lrisversion ge 2 then sxaddpar, header, 'SATURATE', 60000, ' Saturation level', before='BINNING'

if camera eq 'b' then sxaddpar, header,'FILTER', sxpar(header,'BLUFILT'), ' Filter', before='BLUFILT'
if camera eq 'r' then sxaddpar, header,'FILTER', sxpar(header,'REDFILT'), ' Filter', before='BLUFILT'
if camera eq 'b' and clip(sxpar(header, 'GRISNAME'),2) eq 'NB' then begin
   ; narrowband blue filers are in grism carousel
  if clip(sxpar(header,'BLUFILT')) eq 'clear' then sxaddpar, header,'FILTER', sxpar(header,'GRISNAME'), ' Filter (NB)'  $
                                              else sxaddpar, header,'FILTER', sxpar(header,'GRISNAME')+sxpar(header,'BLUFILT'), ' Filter (NB+std)'
endif

if lmode eq 'img' then begin
   filtcwav = filtwv(sxpar(header,'FILTER'), fwhm=filtfwhm)
   if filtcwav gt 1 then sxaddpar, header, 'FILTCWAV', filtcwav, ' Approx. filter central wavelength', after='FILTER'
   if filtfwhm gt 1 then sxaddpar, header, 'FILTFWHM', filtfwhm, ' Approx. filter FWHM', after='FILTCWAV'
endif


if camera eq 'b' then sxdelpar, header, ['REDFILT','REDFNUM','REDFOCUS']
if camera eq 'r' then sxdelpar, header, ['BLUFILT','BLUFNUM','BLUFTRAN','BTRAN','BLUFOCUS']


if camera eq 'r' then begin
   if lrisversion eq 1 then pxscale = 0.2115*(xbin)   ; LRIS news says 0.2109, and apparently before 2000-09 was 0.215
   if lrisversion eq 2 then pxscale = 0.135*(xbin)
   if lrisversion eq 3 then pxscale = 0.135*(xbin)
   if lrisversion eq 4 then pxscale = 0.135*(xbin)
endif
if camera eq 'b' then pxscale = 0.1357*(xbin)

sxaddpar, header, 'PXSCALE', pxscale, ' Pixel scale (arcsec/pix), incl. binning', after='WAVELEN'
sxaddpar, header, 'OBSERVAT',"Keck", ' Observatory', before='TELESCOP'

if n_elements(hostname) eq 0 then hostname = gethostname()
if n_elements(username) eq 0 then username = getusername()
sxaddpar, header, 'HOSTNAME', hostname, ' LPipe host computer name'
sxaddpar, header, 'USERNAME', username, ' UNIX username running LPipe'
if n_elements(reducer) gt 0 then sxaddpar, header, 'REDUCER', reducer, ' Name of reducer'
if n_elements(version) gt 0 then sxaddpar, header, 'LPIPEVER', version, ' LPipe version number'
sxaddpar, header, 'PREPARED', repstr(tstamp(/ut),' ', 'T'), ' Timestamp of prepare reduction step'
sxaddpar, header, 'REDUCED',  repstr(tstamp(/ut),' ', 'T'), ' Timestamp of prepare reduction step'
sxaddpar, header, 'IDLVER', !version.release, ' IDL version number'

slitpa = (float(sxpar(header, 'ROTPOSN'))+90) mod 360.
sxaddpar, header, 'SLITPA', slitpa, ' Slit or right chip CCW from North (deg)', before='ROTPOSN'

sxaddpar, header, 'RDNOISE', 4, ' Read noise (??)', before='BINNING' ; default as placeholder
if camera eq 'b' then  sxaddpar, header, 'RDNOISE', 3.6, ' Read noise (??)'  ; accurate for right chip, left chip differs

; --- Component change tracking -----
if n_elements(seq) gt 0 then begin
  sxaddpar, header, 'DICHSEQF', repstr(removepath(seq.dichf),'.gz',''), ' First file after DICHNAME change'
  sxaddpar, header, 'RFLTSEQF', repstr(removepath(seq.rfiltf),'.gz',''), ' First file after REDFILT change' 
  sxaddpar, header, 'BFLTSEQF', repstr(removepath(seq.bfiltf),'.gz',''), ' First file after BLUFILT change' 
  sxaddpar, header, 'GRATSEQF', repstr(removepath(seq.gratf),'.gz',''), ' First file after GRANAME change' 
  sxaddpar, header, 'GRANSEQF', repstr(removepath(seq.granglef),'.gz',''), ' First file after GRANGLE change' 
  sxaddpar, header, 'SLITSEQF', repstr(removepath(seq.slitf),'.gz',''), ' First file after SLITNAME change' 
endif



; -- Exposure mid and end times --

datestr = sxpar(header,'DATE-OBS')  ; date only
uttstr = sxpar(header,'UT')         ; time only
tstart_sec = ten(uttstr)*3600. ; secs since midnight
tmid_sec   = tstart_sec + sxpar(header,'EXPTIME')/2. ; what about pausing...?
tend_sec   = tstart_sec + sxpar(header,'EXPTIME')
utmidarr = sixty(tmid_sec/3600. mod 24.)
utendarr = sixty(tend_sec/3600. mod 24.)
utstartstr = datestr + 'T' + uttstr
; need to check for wrapping around UT midnight, fortunately very rarely an issue on LRIS
utmidstr = datestr + 'T' + string(utmidarr[0],'(I02)')+':'+string(utmidarr[1],'(I02)')+':'+string(utmidarr[2],'(F05.2)')
utendstr = datestr + 'T' + string(utendarr[0],'(I02)')+':'+string(utendarr[1],'(I02)')+':'+string(utendarr[2],'(F05.2)')
sxaddpar, header, 'UTSTART', utstartstr, ' UTC datetime at start'
sxaddpar, header, 'UTMID', utmidstr, ' UTC datetime at midpoint'
sxaddpar, header, 'UTEND', utendstr, ' UTC datetime at end'
sxaddpar, header, 'OBSDATE', utstartstr, ' Calendar datetime (day and time)', after='DATE-OBS'

mjd = double(sxpar(header, 'MJD-OBS')) ; start time.  Technically already grabbed above
mjdmid = mjd + sxpar(header,'EXPTIME')/3600.d/24./2.
mjdend = mjd + sxpar(header,'EXPTIME')/3600.d/24.
sxaddpar, header, 'MJD', mjd, ' Mean julian date at start', after='MJD-OBS'
sxaddpar, header, 'MJDSTART', mjd, ' Mean julian date at start', after='UTEND'
sxaddpar, header, 'MJDMID', mjdmid, ' Mean julian date at midpoint', after='MJDSTART'
sxaddpar, header, 'MJDEND', mjdend, ' Mean julian date at end', after='MJDMID'
jd = double(mjd) + 2400000.5d
jdmid = jd + sxpar(header,'EXPTIME')/3600.d/24./2.
jdend = jd + sxpar(header,'EXPTIME')/3600.d/24.
sxaddpar, header, 'JD', jd, ' Julian date at start', after='MJD'
sxaddpar, header, 'JDSTART', jd, ' Julian date at start', after='MJDEND'
sxaddpar, header, 'JDMID', jdmid, ' Julian date at midpoint', after='JDSTART'
sxaddpar, header, 'JDEND', jdend, ' Julian date at end', after='JDMID'


; --- Sun/moon ephemeris ---

; ct2lst, lstcalc, -155.4747, dummy, jd ; very close to the below - this is LMST, not LAST, so needs a correction?
; Wikipedia states that nutation variations (LAST vs LMST) are of order arcseconds  (30x too small)

lst = sxpar(header, 'ST') ; assumed to be in current equinox, but could be vs 2000 equinox...?
; value is bad in some archive images.  What to do?
lststart = ten(lst) ; decimal hours
;lststart = lststart - (mjd - 51544.0)*0.0084/3600. adjustment for equinox precession seems to make it closer?
lstmid = (lststart + sxpar(header,'EXPTIME')/3600./2.) mod 24
lstend = (lststart + sxpar(header,'EXPTIME')/3600.) mod 24
zenithrastart = 15.*lststart ; decimal degrees
zenithra      = 15.*lstmid
zenithraend   = 15.*lstend
zenithdec = 19.82636 ; current equinox

; Moon/sun locations from IDLASTRO are in current equinox
moonpos, jdmid, moonra, moondec  ; needs isarray, which seems to have vanished from the GSFC library.
sunpos, jdmid, sunra, sundec    
mphase, jdmid, moonphase

radeg = 15.*ten(sxpar(header, 'RA'))
decdeg = ten(sxpar(header, 'DEC'))

radeg_eq = radeg
decdeg_eq = decdeg
eqyear = 2000. + (mjd - 51544.0)/365.242189  ; current equinox
precess, radeg_eq, decdeg_eq, 2000.0, eqyear ; precess equinox

gcirc, 2, zenithra, zenithdec, moonra, moondec, moonzdist
gcirc, 2, zenithra, zenithdec, sunra, sundec, sunzdist
gcirc, 2, radeg_eq, decdeg_eq, moonra, moondec, moondist
;gcirc, 2, zenithrastart, zenithdec,  radeg_eq, decdeg_eq, zdiststart
;gcirc, 2, zenithra,      zenithdec,  radeg_eq, decdeg_eq, zdistmid
;gcirc, 2, zenithraend,   zenithdec,  radeg_eq, decdeg_eq, zdistend

sunelev = 90. - sunzdist/3600.
moonelev = 90. - moonzdist/3600.
moonsep = moondist/3600.
;elevstart = 90. - zdiststart/3600.
;elevmid   = 90. - zdistmid/3600.
;elevend   = 90. - zdistend/3600.
;airstart = 1./sin(elevstart*!pi/180.)
;airmid   = 1./sin(elevmid*!pi/180.)
;airend   = 1./sin(elevend*!pi/180.)
; The above are close (0.1 deg) but not quite accurate, probably due to precession?
; May be safe to use a differential airmass calculation... header AIRMASS is probably at start time?
;print
;print, sxpar(header,'EL'), elevstart, elevmid, elevend
;print, sxpar(header,'AIRMASS'), airstart, airmid, airend

sxaddpar, header, 'SUNELEV', sunelev, ' Sun elevation '+(sunelev>0.?'(above horizon, deg)':'(deg)'), before='ROTMODE'
sxaddpar, header, 'MOONELEV', moonelev, ' Moon elevation '+(moonelev>0?'(above horizon, deg)':'(deg)'), before='ROTMODE'
sxaddpar, header, 'MOONDIST', moonsep, ' Distance to moon (deg)', before='ROTMODE'
sxaddpar, header, 'MOONPHAS', moonphase, ' Lunar phase (1 = full)', after='MOONDIST'

; --- Parallactic angle ---

hastr = clip(sxpar(header,'HA', count=ctha))
if ctha gt 0 then begin
  if hastr ne '+00:00:00.00' then begin
    hadeg = ten(hastr) ; in hours, as expected by parangle
    para = parangle(hadeg, decdeg, 19.8206)
    sxaddpar, header, 'PARANGLE', para, ' Parallactic angle (at start, deg)', after='AIRMASS'
    offpa = slitpa-para; might be able to get this from rotpposn, drive, etc.
    if offpa gt 360 then offpa -= 360.
    if offpa lt -180 then offpa += 360.
    sxaddpar, header, 'OFFPA', offpa, ' Slit rotation vs. parallactic angle (deg)', after='PARANGLE'
  endif
endif



; --- Dispersion and central wavelength ---

if lmode eq 'spec' then begin
  dispersion = 0.
  if camera eq 'b' then begin
    grism = clip(sxpar(header, 'GRISNAME'))
    if grism eq  '300/5000' then dispersion = 1.43
    if grism eq  '400/3400' then dispersion = 1.09
    if grism eq  '600/4000' then dispersion = 0.61 ; 0.63
    if grism eq '1200/3400' then dispersion = 0.24
    if grism eq  '300/5000' then cwav = 4495.
    if grism eq  '400/3400' then cwav = 3600. 
    if grism eq  '600/4000' then cwav = 4305.
    if grism eq '1200/3400' then cwav = 3400.
    if dispersion eq 0. then begin
       logprint, '   WARNING: Unrecognized grism GRISNAME=', grism 
    endif else begin
       dichroic = clip(sxpar(header,'DICHNAME'))
       if dichroic eq '460' and grism eq '600/4000' then cwav = cwav + 210.
       if dichroic eq '460' and grism eq '1200/3400' then cwav = cwav + 70.
       sxaddpar, header, 'DISPERSN', dispersion*specbin, ' Expected dispersion (Ang/pix)'  ; ybin (not rotated yet)
       sxaddpar, header, 'CWAVE',    cwav, ' Expected central wavelength'
    endelse
    sxaddpar, header, 'GRNAME', grism, ' Disperser (grism) name'
  endif
  if camera eq 'r' then begin
   grating = clip(sxpar(header,'GRANAME'))
   if grating eq  '150/7500' then dispersion = 3.0
   if grating eq  '300/5000' then dispersion = 1.60 ; 1.59
   if grating eq  '400/8500' then dispersion = 1.18 ; 1.16
   if grating eq  '600/5000' then dispersion = 0.80
   if grating eq  '600/7500' then dispersion = 0.80
   if grating eq  '600/10500' then dispersion = 0.80
   if grating eq  '831/8200' then dispersion = 0.588 ; 0.58
   if grating eq  '900/5500' then dispersion = 0.54  ; 0.53
   if grating eq '1200/7500' then dispersion = 0.40
   if grating eq  '600/10000' then dispersion = 0.80
   if grating eq '1200/9000' then dispersion = 0.40
   if lrisversion eq 1 then dispersion *= 0.210/0.135
   if lrisversion eq 4 then dispersion *= 1 ; 0.123/0.135 ; seems this is not necessary?
   if dispersion eq 0. then begin
      logprint, '   WARNING: Unrecognized grating GRANAME=', grating
   endif else begin     
      sxaddpar, header, 'DISPERSN', dispersion*specbin, ' Expected dispersion (Ang/pix)'  ; ybin (not rotated yet)
   endelse
   cwav = sxpar(header,'WAVELEN',count=ctcwav)
   if cwav le 0 and strpos(grating,'/') ge 0 then begin
      if ctcwav gt 0 then logprint, '  WARNING - central wavelength in header is set to '+clip(cwav)
      if ctcwav eq 0 then logprint, '  WARNING - central wavelength in header is absent'
      angle = float(sxpar(header,'GRANGLE'))
      if grating eq '400/8500' then begin
        cwav = 7828. + (angle-25.855530)*(580./0.73363304) ; Try to guess using the GRANGLE key?
        sxaddpar, header, 'WAVELEN', cwav, 'Artifically assigned central wavelength'
      endif else begin
         logprint, '          Inserting an arbitrary value, wavelength calibration WILL FAIL'
         sxaddpar, header, 'WAVELEN', 8000., 'Artifically assigned central wavelength'
      endelse
   endif
   if grating eq '600/10000' then cwav += 150. ; needed for 2013-12-29 and 2017-04-19, so seems quite general
   sxaddpar, header, 'GRNAME', grating, ' Disperser (grating) name'
   sxaddpar, header, 'CWAVE',  cwav, ' Expected central wavelength (Ang)'  ; + standard offset...?
  endif
endif


headtime = systime(/seconds)  ; mark time after header modifications, including epheremis



; -------- Fix CCD bug (?) -----------

; There is some bug in the binning procedure itself that produces a blank line 
; next to the chip gap in binned data.  On the red it seems to place two lines
; on the right and on the blue it seems to produce one line each on blue and red.
; Not certain if this is intrinsice or from readhmdufits.

; Fix it by shifting the image to remove this stripe.

; The "ground truth" standard here is the array produced in 1x1 binning by readhmdu.
; Images taken in 2x2 should be identical but resampled.

; correct for the different pixel readout pattern in binned data (some bug on the CCD?)
; Maybe this should be moved over to readmududap.

if lrisversion eq 3 and camera eq 'r' and xbin eq 2 then begin
   ampl2 = fix(strsplit(sxpar(header,'AMP1'),',',/extract))
   ampr1 = fix(strsplit(sxpar(header,'AMP4'),',',/extract))
   if n_elements(ampr1) ge 2 then $  ; not done in 1-amp upper amplifier mode, which may or may not be right
     array[ampr1[0]-1:nxin-3,*] = array[ampr1[0]+1:nxin-1,*] ; shift right half by 2 pixels
   
endif
if camera eq 'b' and xbin eq 2 then begin
   ampl2 = fix(strsplit(sxpar(header,'AMP2'),',',/extract)) 
   ampr1 = fix(strsplit(sxpar(header,'AMP3'),',',/extract))
   array[ampr1[0]-1:nxin-3,*] = array[ampr1[0]+1:nxin-1,*] ; shift right half by 2 (binned?) pixels

   array = shift(array,1,0)  ; shift EVERYTHING by 1 pixels because there is also an offset across the whole field...
endif



; New exciting bug discovered - diminished columns (5-10% lower response) seem to move around
; even within the same setting.


; -------------- Figure out the amp/chip boundaries ------------


if camera eq 'b' then begin
   ampl1 = fix(strsplit(sxpar(header,'AMP1'),',',/extract))
   ampl2 = fix(strsplit(sxpar(header,'AMP2'),',',/extract))
   ampr1 = fix(strsplit(sxpar(header,'AMP3'),',',/extract))
   ampr2 = fix(strsplit(sxpar(header,'AMP4'),',',/extract))
endif else begin
   if lrisversion eq 2 or lrisversion eq 3 then begin
     ampl1 = fix(strsplit(sxpar(header,'AMP2'),',',/extract))
     ampl2 = fix(strsplit(sxpar(header,'AMP1'),',',/extract))  ; red amps are not in sequence
     ampr1 = fix(strsplit(sxpar(header,'AMP4'),',',/extract))
     ampr2 = fix(strsplit(sxpar(header,'AMP3'),',',/extract))
   endif else begin
     amp1 = fix(strsplit(sxpar(header,'AMP1'),',',/extract))
     amp2 = fix(strsplit(sxpar(header,'AMP2'),',',/extract))
   endelse
endelse

if camera eq 'b' or lrisversion eq 2 or lrisversion eq 3 then begin
  if n_elements(ampl1) ge 2 and n_elements(ampl2) ge 2 then chipl = [min([ampl1,ampl2]), max([ampl1,ampl2])]
  if n_elements(ampl1) lt 2 and n_elements(ampl2) ge 2 then chipl = ampl2
  if n_elements(ampl1) ge 2 and n_elements(ampl2) lt 2 then chipl = ampl1
  if n_elements(ampr1) ge 2 and n_elements(ampr2) ge 2 then chipr = [min([ampr1,ampr2]), max([ampr1,ampr2])]
  if n_elements(ampr1) lt 2 and n_elements(ampr2) ge 2 then chipr = ampr2
  if n_elements(ampr1) ge 2 and n_elements(ampr2) lt 2 then chipr = ampr1
 
  chipboundaryl = chipl[1]
  chipboundaryr = chipr[0] ; should be separated by one...
endif else begin
  chip = [amp1[0] < amp2[0], amp1[1] > amp2[1]]
endelse


; ---------------- Flag/fix bad pixels -------------------


if camera eq 'r' then begin


  ; This block is not actually being used by the pipeline anymore.
  if keyword_set(flag) or keyword_set(fixdefect) then begin
    ; Load bad pixel mask and rebin/crop appropriately
   
    reusemask = 0
    if n_elements(mask) gt 0 and n_elements(prevversion) gt 0 and n_elements(prevwindet) eq n_elements(windet) then begin
      if lrisversion eq prevversion and xbin eq prevxbin and ybin eq prevybin and total(windet eq prevwindet) eq n_elements(windet) then $
         reusemask = 1      
    endif
    ; reuse the cropped mask if we can, although this saves only trivial processing time (aside from the time saved by not rereading the file)
 
    if reusemask eq 0 then begin ; load mask from disk
       ; should confirm that this works as planned....
       bpmfile = ''
       if camera eq 'r' and lrisversion eq 1 and n_elements(bpm1) lt 2 then bpmfile = bpmpath+'bpm_red1.fits'
       if camera eq 'r' and lrisversion eq 2 and n_elements(bpm2) lt 2 then bpmfile = bpmpath+'bpm_red2.fits'
       if camera eq 'r' and lrisversion eq 3 and n_elements(bpm3) lt 2 then bpmfile = bpmpath+'bpm_red3.fits'
 
       if bpmfile ne '' then if file_test(bpmfile) eq 0 then begin
          logprint
          logprint, 'Bad pixel mask '+bpmfile+' not found: check refdata directory'
       endif else begin
          if camera eq 'r' and lrisversion eq 1 and n_elements(bpm1) lt 2 then bpm1 = mrdfits(bpmfile,0,hbpm, /silent)
          if camera eq 'r' and lrisversion eq 2 and n_elements(bpm2) lt 2 then bpm2 = mrdfits(bpmfile,0,hbpm, /silent)
          if camera eq 'r' and lrisversion eq 3 and n_elements(bpm3) lt 2 then bpm3 = mrdfits(bpmfile,0,hbpm, /silent)
          if camera eq 'r' and lrisversion eq 1 then mask = bpm1[windet[0]:windet[0]+windet[2]-1,windet[1]:windet[1]+windet[3]-1]  ; windet is before binning
          if camera eq 'r' and lrisversion eq 2 then mask = bpm2[windet[0]:windet[0]+windet[2]-1,windet[1]:windet[1]+windet[3]-1]
          if camera eq 'r' and lrisversion eq 3 then mask = bpm3[windet[0]:windet[0]+windet[2]-1,windet[1]:windet[1]+windet[3]-1]
          sbpm = size(mask)
          if xbin gt 1 or ybin gt 1 then mask = rebin(mask, sbpm[1]/xbin, sbpm[2]/ybin)
          ;if xbin gt 1 or ybin gt 1 then mask = shift(mask, 1, 0) ; mysterious why this is necessary.  should be revisited with flats.
          ; should no longer be necessary with the problem now identified as a CCD bug, but revisit this to check someday
       endelse
    endif else begin
       ; just reuse the old mask that's already in there.
       sbpm = size(mask)
    endelse
    
    if keyword_set(flag) then array[where(mask gt 0)] = !values.f_nan
    if keyword_set(fixdefect) then begin
      array[where(mask gt 0)] = !values.f_nan
      ct = 1
      while ct gt 0 do begin 
        bad = where(finite(array) eq 0, ct)
        for m = 0, ct-1 do begin
          mm = bad[m]
          array[mm] = median(array[mm-2:mm+2])
        endfor
      endwhile
    endif

  endif

   ; would produce better results to do this after flat-fielding since there is horizontal structure in the response, but this is easier.
   ; Also, not every dropcolumn is affected every time.
   ; May want to consider running this for spectra also.
  if lrisversion eq 3 and lmode eq 'img' and xbin eq 1 then begin   ; does not support binning
      if n_elements(dropcolumns) eq 0 then begin
         readcol, bpmpath+'lrisred3dropcolumns.dat', dropcolumns, format='i', comment='#'
         dropcolumns -= 1  ; convert DS9 to array coordinate
      endif
      for i = 0, n_elements(dropcolumns)-1 do begin
          x = dropcolumns[i]
          for xm = x-1, x-5 do if total(xm eq dropcolumns) eq 0 then break ; find the nearest column that is not also a dropcolumn
          for xp = x+1, x+5 do if total(xp eq dropcolumns) eq 0 then break
          ax = x - windet[0]
          axm = xm - windet[0]
          axp = xp - windet[0]
          if axp gt (size(array)) [1] then continue
          coldiff = median((array[axm,*]+array[axp,*])/2 - array[ax, *],  /even) 
          array[ax,*] += coldiff ;;;;+ !values.f_nan
      endfor
 
      ; also correct the hot pixel cluster while we're at it.
      if (size(array)) [1] gt 1190-windet[0] and (size(array)) [2] gt 1710-windet[1] then $
         array[1185-windet[0]:1190-windet[0],1705-windet[1]:1710-windet[1]] = !values.f_nan   ; corners are probably ok...

      ; also flag some pixels flanking the chip gap.
      if (size(array)) [1] gt 2056-windet[0] then $
         array[2043-windet[0]:2056-windet[0],*] = !values.f_nan
  endif
 
  ; flag the bad columns close to the target position for spectroscopy
  if lrisversion eq 3 and lmode eq 'spec' then begin
     array[(2278-windet[0])/xbin,*] = !values.f_nan
     array[(2276-windet[0])/xbin,*] = !values.f_nan

     ; flag another bad column not close to the trace that messes up the profiles
     array[(2498-windet[0])/xbin,*] = !values.f_nan
  endif

  if lrisversion eq 1 and camera eq 'r' then begin
     s = size(array)
     nx = s[1]
     ny = s[2]
     badtrailx = ([863, 997, 1001]-1-40-windet[0])/xbin
     badtraily = ([723, 1484, 1486]-1-windet[1])/ybin
     for i = 0, n_elements(badtrailx)-1 do $
        if badtrailx[i] ge 0 and badtraily[i] ge 0 and badtrailx[i] lt nx and badtraily[i] lt ny then $
           array[badtrailx[i], 0:badtraily[i]] = !values.f_nan
     badx = ([995, 992]-1-40-windet[0])/xbin
     bady = ([1483, 1477]-1-windet[1])/ybin
     for i = 0, n_elements(badx)-1 do $
        if badx[i] ge 0 and bady[i] ge 0 and badx[i] lt nx and bady[i] lt ny then $
           array[badx[i], bady[i]] = !values.f_nan
  endif


   ; Recognize and replace hyper-saturated pixels (so saturated they register as dead)
   ; from neon or argon lamps
   ;  (could extend this to all files at the expense of processing time if it seems
   ;   necessary)
  if lrisversion eq 2 or lrisversion eq 3 then begin
    lamps = sxpar(header, 'LAMPS', count=lampct)
    lamparr = fix(strsplit(lamps,',',/extract))
    if n_elements(lamparr) ge 3 then begin ;  lately this keyword has been missing from some(all?) frames?
      if lamparr[1] or lamparr[2] then begin
        truesaturation = max(array) ; both chips should saturate at the same place?

        arrayl = array[chipl[0]:chipl[1]-4,*] 
        arrayr = array[chipr[0]+4:chipr[1],*] 
        arrayl = fixhypersaturation(arrayl, deadthresh=1500)
        arrayr = fixhypersaturation(arrayr, deadthresh=500)
        array[chipl[0]:chipl[1]-4,*] = arrayl
        array[chipr[0]+4:chipr[1],*] = arrayr

      endif
    endif
  endif

  prevversion = lrisversion
  prevxbin = xbin
  prevybin = ybin
  prevwindet = windet
   
endif

prevcamera = camera

flagtime = systime(/seconds)  ; mark time after flagging

; ---------- Raw counts estimate ------------------

; Identify a subset with expected signal to estimate the median counts to add to the header.

; as always, x and y are pre-transpose
ycount = [0, nyin-1] ; use the whole y (post-tranpose x) by default, will narrow if imaging
if camera eq 'r' and (lrisversion eq 1 or lrisversion eq 4) then begin
  if lrisversion eq 1 then xcount = [chipr[0], (1745-windet[0])/xbin-1 < chipr[1]] ; not sure if this works
  if lrisversion eq 4 then begin
     xcount = [chip[0], chip[1]]/xbin ; same as for cropping
     ycount = [1360,2700]/ybin  
  endif
  if lmode eq 'img' then ycount = [(116-windet[1])/ybin, (1745-windet[1])/ybin]
endif else begin
  xcount = [chipr[0], (2670-windet[0])/xbin-1 < chipr[1]] ; window both imaging and spec to slit zone (right chip)
  if lmode eq 'img' then begin
    if camera eq 'b' then ycount = [(760-windet[1])/ybin, (3200-windet[1])/ybin]
    if camera eq 'r' then ycount = [(638-windet[1])/ybin, (3100-windet[1])/ybin]
  endif else begin
    if camera eq 'b' and n_elements(grism) gt 0 then begin
      if grism eq '400/3400' then ycount = [(2048-windet[1])/ybin, (4096-windet[1])/ybin]
    endif
    if camera eq 'r' and n_elements(chipl) gt 1 then begin
       ; in case the right CCD is dead
       xcountl = [(1580-windet[0])/xbin, chipl[1]]
       xcountl = [xcountl[0] > 0, xcountl[1] < (nxin-1)]
    endif
  endelse
endelse
xcount = [xcount[0] > 0, xcount[1] < (nxin-1)]
ycount = [ycount[0] > 0, ycount[1] < (nyin-1)]
if xcount[1] le xcount[0] then xcount = [0, nxin-1]
if ycount[1] le ycount[0] then ycount = [0, nyin-1] ; if source region is cropped out use everything
counts =  median(array[xcount[0]:xcount[1],ycount[0]:ycount[1]])
;print, clip(xcount[0])+':'+clip(xcount[1])+','+clip(ycount[0])+':'+clip(ycount[1])+' = '+clip(counts)
countsr = counts
if camera eq 'r' and lrisversion eq 2 or lrisversion eq 3 and lmode eq 'spec' and n_elements(xcountl) gt 1 then begin
  if xcountl[1]-xcountl[0] gt 10 then begin
     ; in case the right CCD is dead
     countsl =  median(array[xcountl[0]:xcountl[1],ycount[0]:ycount[1]])
     if countsl gt counts*2 + 100 then counts = countsl
  endif 
endif
sxaddpar, header, 'COUNTS', counts, ' Median raw counts over expected signal area'
sxaddpar, header, 'MAXCOUNT', max(array), ' Maximum count value over detector'

; should really produce a conversion/dissecting function that turns detector-mosaic to array coordinates...



; ------------- Gain correction -------------------

if n_elements(gain) eq 0 then gain = [1.,1,1,1]

if camera eq 'r' and lrisversion eq 1 then sxaddpar, header, 'GAIN', 1.98, ' Inverse gain estimate', after='RDNOISE'
if camera eq 'r' and lrisversion eq 2 then sxaddpar, header, 'GAIN', 1.19/gain[2], ' Inverse gain estimate (??)', after='RDNOISE' ; unclear
if camera eq 'r' and lrisversion eq 3 then sxaddpar, header, 'GAIN', 1.162/gain[2], ' Inverse gain estimate (R1)', after='RDNOISE'
if camera eq 'r' and lrisversion eq 4 then sxaddpar, header, 'GAIN',  1., 'Guess of inverse gain.  Not actually known', after='RDNOISE'
if camera eq 'b' then sxaddpar, header, 'GAIN', 1.63/gain[2], ' Inverse gain estimate (R1)', after='RDNOISE'
if camera eq 'b' or lrisversion eq 2 or lrisversion eq 3 then begin
  if max(ampl1) gt 0 then array[ampl1[0]:ampl1[1],*] = array[ampl1[0]:ampl1[1],*] * gain[0]
  if max(ampl2) gt 0 then array[ampl2[0]:ampl2[1],*] = array[ampl2[0]:ampl2[1],*] * gain[1]
  if max(ampr1) gt 0 then array[ampr1[0]:ampr1[1],*] = array[ampr1[0]:ampr1[1],*] * gain[2]
  if max(ampr2) gt 0 then array[ampr2[0]:ampr2[1],*] = array[ampr2[0]:ampr2[1],*] * gain[3]
  if max(ampl1) gt 0 then sxaddpar, header, 'INGAINL1', gain[0], ' Gain adjustment factor'
  if max(ampl2) gt 0 then sxaddpar, header, 'INGAINL2', gain[1], ' Gain adjustment factor'
  if max(ampr1) gt 0 then sxaddpar, header, 'INGAINR1', gain[2], ' Gain adjustment factor'
  if max(ampr2) gt 0 then sxaddpar, header, 'INGAINR2', gain[3], ' Gain adjustment factor'
endif
if camera eq 'r' and lrisversion eq 1 then begin
  if max(amp1) gt 0 and amp1[1] gt amp1[0] then array[amp1[0]:amp1[1],*] = array[amp1[0]:amp1[1],*] * gain[0]
  if max(amp2) gt 0 and amp2[1] gt amp2[0] then array[amp2[0]:amp2[1],*] = array[amp2[0]:amp2[1],*] * gain[1]
  if max(amp1) gt 0 and amp1[1] gt amp1[0] then sxaddpar, header, 'INGAIN1', gain[0], ' Gain adjustment factor'
  if max(amp2) gt 0 and amp2[1] gt amp2[0] then sxaddpar, header, 'INGAIN2', gain[1], ' Gain adjustment factor'
endif

; ---------- Figure out crop boundaries ----------

if n_elements(ycrop) eq 0 then ycrop = [0,nyin-1]

if n_elements(crops) eq 1 then begin
  if crops eq 'none' then begin
     if camera eq 'blue' or lrisversion ge 2 then begin
       lcrop = chipl
       rcrop = chipr
       ycrop = [0, nyin-1]
       xcrop = [chipl[0], chipr[1]]
     endif else begin
       xcrop = chip
       ycrop = [0, nyin-1]
     endelse
  endif
  if crops eq 'auto' then begin
     ycrop = [0, nyin-1] ; default
     if camera eq 'b' then begin
        if lmode eq 'img' then begin
           lcrop = [(360-windet[0])/xbin > chipl[0], chipl[1]]
           rcrop = [chipr[0], (3790-windet[0])/xbin-1 < chipr[1]]
           ycrop = [(760-windet[1])/ybin > 0, (3200-windet[1])/ybin-1 < nyin-1]
           xcrop = [lcrop[0], rcrop[1]]
        endif
        if lmode eq 'spec' then begin
           lcrop = [(1440-windet[0])/xbin > chipl[0],      chipl[1]]
           rcrop = [chipr[0],      (2670-windet[0])/xbin-1 < chipr[1]]
           xcrop = [lcrop[0], rcrop[1]]
        endif
     endif
     if camera eq 'r' then begin
        if lrisversion eq 1 and lmode eq 'img' then begin
           xcrop = [(116-windet[0])/xbin > chip[0], (1745-windet[0])/xbin-1 < chip[1]]
           ycrop = [0, nyin-1]
        endif
        if lrisversion eq 1 and lmode eq 'spec' then begin
           xcrop = [chip[0], chip[1]] ; no cropping
           ycrop = [(580-windet[1])/ybin > 0, (1500-windet[1])/ybin-1 < nyin-1]
        endif
        if (lrisversion eq 2 or lrisversion eq 3) and lmode eq 'img' then begin
           lcrop = [(430-windet[0])/xbin > chipl[0], chipl[1]]
           rcrop = [chipr[0],  (3778-windet[0])/xbin-1 < chipr[1]]
           ycrop = [(638-windet[1])/ybin > 0, (3100-windet[1])/ybin-1 < nyin-1]
           xcrop = [lcrop[0], rcrop[1]]
        endif
        if (lrisversion eq 2 or lrisversion eq 3) and lmode eq 'spec' then begin
           lcrop = [(1520-windet[0])/xbin > chipl[0], chipl[1]]
           rcrop = [chipr[0],  (2660-windet[0])/xbin-1 < chipr[1]] ; 2660
           xcrop = [lcrop[0], rcrop[1]]
        endif
        if lrisversion eq 4 and lmode eq 'spec' then begin
       ;    xcrop = [chip[0], chip[1]]/xbin ; no cropping ; xbin wrong?? not sure why
           xcrop = [0, nxin-1] ; perform no cropping
       ;    print, xcrop
       ;    ycrop = [1360,2700]/ybin ; this is post-transpose y
           ycrop = [1340,2760]/ybin ; this is post-transpose y
        endif
        if lrisversion eq 4 and lmode eq 'img' then begin
           ycrop = [260,3863]/ybin  ; this is rough
           xcrop = [748,3287]/xbin           
        endif
     endif
  endif
endif

dotpos = strpos(filename,'.fits',/reverse_search)
slashpos = strpos(filename,'/',/reverse_search)
fileroot = strmid(filename,slashpos+1,dotpos-slashpos-1)
fileext = repstr(strmid(filename,dotpos+1),'.gz','')

if merge or multi then begin
   if n_elements(outname) gt 0 then outfilename = outname
   if n_elements(outfilename) eq 0 then begin
      if n_elements(prefix) eq 0 and n_elements(suffix) eq 0 then begin
         prefix = 'p'
      endif
      if n_elements(prefix) gt 0 then outfilename = prefix+fileroot+'.'+fileext
      if n_elements(suffix) gt 0 then outfilename = fileroot+suffix+'.fits'
   endif
endif


;get_date, now
hist = 'Processed by lrisprepare '+tstamp(/ut)+' UT'  ;+now

if ycrop[1] le ycrop[0] or xcrop[1] le xcrop[0] then begin
   ;logprint
   logprint, '  ERROR:  Invalid crop region.  Cannot prepare file.'
   errcode = -8
   return
   ; This seems to occur when processing MIRA files.
   ; investigate why, someday.  2012-04-27 run.
endif

; deal with crate crashes - doesn't seem to work though
if max(array) le 0 then begin
   logprint, '  File contains no positive data!  Not preparing.'
   errcode = -9
   return
endif



; ---- WCS keywords -----

; Generic keywords.  All of these are PRE transpose.

sxaddpar, header, 'LTM1_1', 1./xbin
sxaddpar, header, 'LTM2_2', 1./ybin

; CRPIX1 will generally be overwritten later.

if lrisversion eq 1 and camera eq 'r' then begin
   sxaddpar, header, 'CRPIX1', -windet[0]*1.0/xbin-xcrop[0]+1024./xbin
   sxaddpar, header, 'CRPIX2', -windet[1]*1.0/ybin-ycrop[0]+1024./ybin
   ;sxaddpar, header, 'CRPIX1', 1024/xbin - (windet[0]*1.0/xbin+xcrop[0])
   ;sxaddpar, header, 'CRPIX2', 1024/ybin - (windet[1]*1.0/ybin+ycrop[0])
endif else if lrisversion eq 4 and camera eq 'r' then begin
   sxaddpar, header, 'CRPIX1', -windet[0]*1.0/xbin-xcrop[0]+2000./xbin  ; not sure about x vs y
   sxaddpar, header, 'CRPIX2', -windet[1]*1.0/ybin-ycrop[0]+2000./ybin
endif else begin
   sxaddpar, header, 'CRPIX1', -windet[0]*1.0/xbin-xcrop[0]+2048./xbin ; midpoint
   sxaddpar, header, 'CRPIX2', -windet[1]*1.0/ybin-ycrop[0]+2048./ybin ; midpoint
   ;sxaddpar, header, 'CRPIX1', 2048/xbin - (windet[0]*1.0/xbin+xcrop[0]) , after='CRVAL2' ; effectively identical
   ;sxaddpar, header, 'CRPIX2', 2048/ybin - (windet[1]*1.0/ybin+ycrop[0]) , after='CRPIX1'
endelse

; --- Imaging WCS keywords ---

ra = ten(strtrim(sxpar(header,'RA'),2))*15  ; hours to degrees
dec = ten(strtrim(sxpar(header,'DEC'),2))
padeg = sxpar(header, 'ROTPOSN')
parad = padeg*3.1415927/180.
pxscaledeg = pxscale/3600.
parity = 1
if lrisversion eq 1 and camera eq 'r' then parity = -1 
if lrisversion eq 1 and camera eq 'r' then parad += !pi/2
if lrisversion eq 4 and camera eq 'r' then parity = -1 
if lrisversion eq 4 and camera eq 'r' then parad += !pi/2
; if lrisversion eq 1 and camera eq 'b' then parad += !pi 
if lmode eq 'img' then begin
   sxaddpar, header, 'CD1_1',  pxscaledeg * cos(parad)*parity
   sxaddpar, header, 'CD1_2',  pxscaledeg * sin(parad)
   sxaddpar, header, 'CD2_1', -pxscaledeg * sin(parad)*parity
   sxaddpar, header, 'CD2_2',  pxscaledeg * cos(parad)

   sxaddpar, header, 'CRVAL1', ra
   sxaddpar, header, 'CRVAL2', dec
   sxaddpar, header, 'CTYPE1', 'RA---TAN'
   sxaddpar, header, 'CTYPE2', 'DEC--TAN'
   ; crpix will wait until post-cropping
endif

; --- Spectroscopic WCS keywords ---

if lmode eq 'spec' then begin

   if camera eq 'r' and lrisversion eq 1 or lrisversion eq 4 then begin
     sxaddpar, header, 'CD1_1', sxpar(header, 'DISPERSN')
     sxaddpar, header, 'CD1_2', 0.
     sxaddpar, header, 'CD2_1', 0.
     sxaddpar, header, 'CD2_2', pxscaledeg*3600.
   endif else begin
     sxaddpar, header, 'CD1_1', 0.
     sxaddpar, header, 'CD1_2', sxpar(header, 'DISPERSN')
     sxaddpar, header, 'CD2_1', pxscaledeg*3600.  ; use arcseconds for convenience
     sxaddpar, header, 'CD2_2', 0.
   endelse

   sxaddpar, header, 'CTYPE1', 'WAVE-WAV' ; seems to not work anymore in DS9?
   sxaddpar, header, 'CTYPE2', 'DEC--TAN'
   sxaddpar, header, 'CRVAL1', sxpar(header, 'CWAVE')
   sxaddpar, header, 'CRVAL2', 0.   ; just give the relative offset
   sxaddpar, header, 'CNAME1', 'Wavelength'
   sxaddpar, header, 'CNAME2', 'Slit offset'
   sxaddpar, header,' CUNIT1', 'Angstrom'
   sxaddpar, header,' CUNIT2', 'deg'
   sxaddpar, header,' CRUNIT1', 'Angstrom', after='CUNIT1'
   sxaddpar, header, 'CRUNIT2', 'deg', after='CRUNIT1'

endif


; ------ Do the cropping and make final edits to the header, then write to disk ------

if lrisversion eq 1 or lrisversion eq 4 and camera eq 'r' then begin
   ; write the file out now for LRIS-R1/4 (which has only one chip) and stop.

   array  = array[xcrop[0]:xcrop[1],  ycrop[0]:ycrop[1]]
   sxaddpar, header, 'LTV1', -windet[0]*1.0/xbin-xcrop[0] ;
   sxaddpar, header, 'LTV2', -windet[1]*1.0/ybin-ycrop[0]+1024./ybin ;

     ; it's not really possible to align the two in DS9 because DS9 does not support fractional LTV offsets and the offsets are in
     ; binned pixels.  So with the different array origins and center vs. edge pixel definitions fractional pixel offsets can't be
     ; avoided.  This might also be going on with the other cameras?

   if n_elements(outname) gt 0 then outfilename = outname
   if n_elements(outfilename) eq 0 then begin
      if n_elements(prefix) eq 0 and n_elements(suffix) eq 0 then begin
         prefix = 'p'
      endif
      if n_elements(prefix) gt 0 then outfilename = prefix+fileroot+'.'+fileext
      if n_elements(suffix) gt 0 then outfilename = fileroot+suffix+'.fits'
   endif
   if lmode eq 'img' and clip(sxpar(header,'PONAME')) eq 'LRISB' then begin
     vshift = (30/0.135)
     xyad, header, sxpar(header,'CRPIX1'), sxpar(header,'CRPIX2')+vshift, va, vd
     sxaddpar, header, 'CRVAL1', va
     sxaddpar, header, 'CRVAL2', vd
   endif

   ;if camera eq 'b' and lmode eq 'spec' then countsval = percentile(array,80) $   ; could speed this up a bit...
   ;                                     else countsval = median(array)
   ;sxaddpar, header, 'COUNTS', countsval
   ;sxaddpar, header, 'MAXCOUNT', max(array) ; maximum value in the image

   mwrfits, array, outfilename, header, /create
   logprint, '  -> ', outfilename

endif else begin

if merge then begin
   array  = array[xcrop[0]:xcrop[1],  ycrop[0]:ycrop[1]]
   croptime = systime(/seconds)  ; mark time after cropping

   sxaddpar, header, 'LTV1', -windet[0]*1.0/xbin-xcrop[0]+2048./xbin ; output Y
   sxaddpar, header, 'LTV2', -windet[1]*1.0/ybin-ycrop[0]            ; output X
   ; LTV1 and LTV2 will be reversed later

   ; For consistency, we will always define CRPIX to be the center of the near-right amplifier.
   if lmode eq 'img' and lrisversion ge 2 then begin
     vshift = (30/0.135) ; the right CCD center is shifted relative to the blue (and off-center in general)
     xyad, header, sxpar(header,'CRPIX1'), sxpar(header,'CRPIX2')+vshift, va, vd
     sxaddpar, header, 'CRVAL1', va
     sxaddpar, header, 'CRVAL2', vd

     sxaddpar, header, 'CRPIX1', ampr1[0]+0.5  ; set the WCS y=0 to the chip gap (boundary of the edge pixels.) 
                                               ; R+0.5 instead of R-0.5 presumably because 1st pixel is "1" not 0?
                                               ; will be reversed later in transpose
   endif

   ampl1 -= xcrop[0]
   ampl2 -= xcrop[0]
   ampr1 -= xcrop[0]
   ampr2 -= xcrop[0]
   if n_elements(ampl1) ge 2 then sxaddpar, header, 'AMPL1', clip(ampl1[0])+','+clip(ampl1[1]), ' (image coordinates)'
   if n_elements(ampl2) ge 2 then sxaddpar, header, 'AMPL2', clip(ampl2[0])+','+clip(ampl2[1])
   if n_elements(ampr1) ge 2 then sxaddpar, header, 'AMPR1', clip(ampr1[0])+','+clip(ampr1[1])
   if n_elements(ampr2) ge 2 then sxaddpar, header, 'AMPR2', clip(ampr2[0])+','+clip(ampr2[1])
   ; needed for identifying the proper amplifier for later flat-field normalization

   ;if camera eq 'b' and lmode eq 'spec' then countsval = percentile(array,80) $   ; could speed this up a bit...
   ;                                     else countsval = median(array)
   ;sxaddpar, header, 'COUNTS', countsval
   ;sxaddpar, header, 'MAXCOUNT', max(array) ; maximum value in the image

   ; update gain and read noise
   if camera eq 'r' and lrisversion eq 3 then sxaddpar, header, 'RDNOISE', 4.62

   if keyword_set(notranspose) eq 0 then begin ; and lmode eq 'spec' 
      ltranspose, array, header
   endif

   logprint, '  -> ', outfilename

   sxaddhist, hist, header

   mwrfits, array, outfilename, header, /create

endif

if split then begin ; Not yet tested following February 2019 updates.
   rheader = header
   lheader = header

   if keyword_set(leftonly)  and n_elements(lcrop) eq 0 and n_elements(xcrop) eq 2 then lcrop = xcrop
   if keyword_set(rightonly) and n_elements(rcrop) eq 0 and n_elements(xcrop) eq 2 then rcrop = xcrop

   doleft  = keyword_set(rightonly) eq 0
   doright = keyword_set(leftonly)  eq 0
 
   if doleft  then arrayl = array[lcrop[0]:lcrop[1],ycrop[0]:ycrop[1]] 
   if doright then arrayr = array[rcrop[0]:rcrop[1],ycrop[0]:ycrop[1]] 

   croptime = systime(/seconds)  ; mark time after cropping


   if lmode eq 'img' then begin
     if camera eq 'r' then begin
        sxaddpar, rheader, 'CRPIX1', 2048/xbin - (windet[0]*1.0/xbin+rcrop[0]) , after='CRVAL2'
        sxaddpar, rheader, 'CRPIX2', 2048/ybin - (windet[1]*1.0/ybin+ycrop[0]) , after='CRPIX1'
 
        sxaddpar, lheader, 'CRPIX1', 2048/xbin - (windet[0]*1.0/xbin+lcrop[0]) , after='CRVAL2'
        sxaddpar, lheader, 'CRPIX2', 2048/ybin - (windet[1]*1.0/ybin+ycrop[0]) , after='CRPIX1'

        vshift = (30/0.135)
        xyad, rheader, sxpar(rheader,'CRPIX1'), sxpar(rheader,'CRPIX2')+vshift, va, vd
        sxaddpar, rheader, 'CRVAL1', va
        sxaddpar, rheader, 'CRVAL2', vd
        xyad, lheader, sxpar(lheader,'CRPIX1'), sxpar(lheader,'CRPIX2')+vshift, va, vd
        sxaddpar, lheader, 'CRVAL1', va
        sxaddpar, lheader, 'CRVAL2', vd

        leftshift = (-1024/xbin-30./0.135)
        xyad, rheader, sxpar(rheader,'CRPIX1')+leftshift, sxpar(rheader,'CRPIX2'), la, ld
        sxaddpar, lheader, 'CRVAL1', la
        sxaddpar, lheader, 'CRVAL2', ld

     endif
     if camera eq 'b' then begin
        sxaddpar, rheader, 'CRPIX1', 2048/xbin - (windet[0]*1.0/xbin+rcrop[0]), after='CRVAL2'
        sxaddpar, rheader, 'CRPIX2', 2048/ybin - (windet[1]*1.0/ybin+ycrop[0]), after='CRPIX1'

        sxaddpar, lheader, 'CRPIX1', 2048/xbin - (windet[0]*1.0/xbin+lcrop[0]), after='CRVAL2'
        sxaddpar, lheader, 'CRPIX2', 2048/ybin - (windet[1]*1.0/ybin+ycrop[0]), after='CRPIX1'

        leftshift = (-1024/xbin-14./0.135)
        xyad, rheader, sxpar(rheader,'CRPIX1')+leftshift, sxpar(rheader,'CRPIX2'), la, ld
        sxaddpar, lheader, 'CRVAL1', la
        sxaddpar, lheader, 'CRVAL2', ld
        ;leftrot = 0.5*!pi/180.
        ;sxaddpar, lheader, 'CD1_1',  pxscaledeg * cos(parad+leftrot)*parity
        ;sxaddpar, lheader, 'CD1_2',  pxscaledeg * sin(parad+leftrot)
        ;sxaddpar, lheader, 'CD2_1', -pxscaledeg * sin(parad+leftrot)*parity
        ;sxaddpar, lheader, 'CD2_2',  pxscaledeg * cos(parad+leftrot)
     endif
   endif
   ;if lmode eq 'spec' then begin
   ;   sxaddpar, header, 'CRPIX2', 0.  ; for now, set the WCS y=0 to y=0.
   ;endif

   if doleft  then sxaddpar, lheader, 'LTV1', -windet[0]/xbin-lcrop[0]  ; LTV1 and LTV2 will be reversed in transpose function
   if doright then sxaddpar, rheader, 'LTV1', -windet[0]/xbin-rcrop[0]

   if doleft  then sxaddpar, lheader, 'LTV2', -windet[1]/ybin-ycrop[0]
   if doright then sxaddpar, rheader, 'LTV2', -windet[1]/ybin-ycrop[0]

   if camera eq 'r' and lrisversion eq 3 then begin
      if doright then sxaddpar, rheader, 'RDNOISE', 4.64
      if doleft  then sxaddpar, lheader, 'RDNOISE', 4.62
   endif

   if doleft  then sxaddpar, lheader, 'CHIP', 'left'
   if doright then sxaddpar, rheader, 'CHIP', 'right'

   ;if doleft  then sxaddpar, lheader, 'COUNTS', median(arrayl)
   ;if doright then sxaddpar, rheader, 'COUNTS', median(arrayr)

   if multi and n_elements(outname) gt 0 then outfilename = outname
   if keyword_set(leftonly) and n_elements(loutfilename) eq 0 and n_elements(outfilename) gt 0 then loutfilename=outfilename
   if keyword_set(rightonly) and n_elements(routfilename) eq 0 and n_elements(outfilename) gt 0 then routfilename=outfilename

   if n_elements(prefix) gt 0 and n_elements(rprefix) eq 0 and keyword_set(rightonly) then rprefix = prefix
   if n_elements(prefix) gt 0 and n_elements(lprefix) eq 0 and keyword_set(rightonly) then lprefix = prefix

   if n_elements(loutname) gt 0 then loutfilename = loutname
   if n_elements(loutfilename) eq 0 then begin
      if n_elements(lprefix) eq 0 and n_elements(lsuffix) eq 0 then begin
        lprefix = 'p'
        lsuffix = 'l'
      endif
      if n_elements(lprefix) gt 0 then loutfilename = lprefix+fileroot+'.'+fileext
      if n_elements(lsuffix) gt 0 then loutfilename = fileroot+lsuffix+'.fits'
   endif
   if n_elements(routname) gt 0 then routfilename = routname
   if n_elements(routfilename) eq 0 then begin
      if n_elements(rprefix) eq 0 and n_elements(rsuffix) eq 0 then begin
         rprefix = 'p'
         rsuffix = 'r'
      endif
      if n_elements(rprefix) gt 0 then routfilename = rprefix+fileroot+'.'+fileext
      if n_elements(rsuffix) gt 0 then routfilename = fileroot+rsuffix+'.fits'
   endif

   if  keyword_set(notranspose) eq 0 then begin ; and lmode eq 'spec' 
      if doleft  then ltranspose, arrayl, lheader 
      if doright then ltranspose, arrayr, rheader 
   endif

   if multi then begin
      logprint, ' -> ', outfilename 
   endif else begin $
      if doleft and doright        then logprint, ' -> ', loutfilename, ',', routfilename
      if doleft and (doright eq 0) then logprint, ' -> ', loutfilename
      if doright and (doleft eq 0) then logprint, ' -> ', routfilename
   endelse

   ;if doleft  and file_test(loutfilename) then logprint, 'Overwriting ', loutfilename
   ;if doright and file_test(routfilename) then logprint, 'Overwriting ', routfilename

   sxaddhist, hist, lheader
   sxaddhist, hist, rheader

   if multi eq 0 then begin
     if doleft  then mwrfits, arrayl, loutfilename, lheader, /create
     if doright then mwrfits, arrayr, routfilename, rheader, /create
   endif else begin
     if doright then mwrfits, arrayr, outfilename, rheader, /create
     if doleft  then mwrfits, arrayl, outfilename, lheader     
   endelse

endif
endelse

finaltime = systime(/seconds)  ; mark time at end of file processing

if keyword_set(timer) then begin
print, 'Total duration:  ', finaltime-starttime
print, '  Read duration: ', readtime-starttime
print, '  Head duration: ', headtime-readtime
print, '  Flag duration: ', flagtime-headtime
print, '  Crop duration: ', croptime-flagtime
print, '  Write duration:', finaltime-croptime
endif

end

