;+
; NAME:
;       lrisselectfiles
;
; PURPOSE:
;       Get a list of files satisfying certain conditions.
;
; CALLING SEQUENCE:
;       lrisselectfiles, dirtype, prefix, suffix, filetype, [imh=, count=, /nofilter, /nofilefilter]
;
; EXPLANATION:
;       This is typically used by the pipeline to find the files that are
;       needed for the next processing step, e.g. all the flat-fielded blue files.
;
; INPUTS:
;       dirtype - String code for the directory to look.  Typically one of:
;           d: raw data directory
;           i: imaging working directory
;           s2: spectrum 2D working directory
;           s1: spectrum 1D working directory
;           f: processed flats directory
;           a: processed arcs directory
;       prefix - Standard processing history filename prefix.  e.g. "zfpr".
;       suffix - Chip suffix.  Generally 'l', 'r', or an empty string.
;       filetype - Filename suffix.  For spectra, '.fits' or '.spec'
;
; OPTIONAL INPUT KEYWORDS:
;       /nofilter - Ignore the lpar user-specified filter entirely.
;       /nofilefilter - Ignore the user-specified filename filter.
;
; OPTIONAL OUTPUT KEYWORDS:
;       imh = LRIS header structure (from lrisheaderinfo) to return.
;             Matches the output filename list.  
;             (Saves having to repeat the header read.)
;       count = Number of files.  Useful to check that matching files
;             actually exist and for loop condition.
;
; OUTPUTS:
;       Returns a list of filenames meeting the conditions.
; 
; EXAMPLES:
;       lrisselectfiles('s2','zfpb','r','.fits', count=ct)
;
; NOTES:
;
;
;-
;------------------------------------------------------------------------------

function lrisselectfiles, dirtype, prefix, suffix, filetype, imh=imh, count=count, $
   nofilter=nofilter, nofilefilter=nofilefilter
   ; unmatched=unmatched, matchstr=matchstr

   ; Find LRIS files on disk.  Can filter out by various user-specified keywords.
   ; nofilter turns off keyword filtering.  (note two different uses of word 'filter' here)
   ; nofilefilter turns off filtering by filenum and target ONLY (filters only by configuration, for calibrations)

   common lrisauto, lpar

   ; needs to able to deal with arrays for biases?
   if dirtype eq 'd' then workingdir = lpar.datadir
   if dirtype eq 'i' then workingdir = lpar.imworkingdir
   if dirtype eq 's1' then workingdir = lpar.spworkingdir1d
   if dirtype eq 's2' then workingdir = lpar.spworkingdir2d
   if dirtype eq 'if' then workingdir = lpar.imfinaldir
   if dirtype eq 'sf' then workingdir = lpar.spfinaldir
   if dirtype eq 'f' then workingdir = lpar.flatdir
   if dirtype eq 'a' then workingdir = lpar.arcdir
   if dirtype eq 'r' then workingdir = lpar.responsedir
   if n_elements(workingdir) eq 0 then begin
       print, 'Directory not recognized'
       count = 0
       return, ''
   endif

   ; would be nice to deal with badfiles.txt to veto at all stages

   ; files = file_search(workingdir+prefix+lpar.wildchar+suffix+'.fits', count=count)  
   selector = lpar.wildchar+suffix+filetype
   fullprefix = workingdir+prefix

   if n_elements(prefix) eq 1 then $
       files = choosefiles(selector,fullprefix, ct=count)
   if n_elements(prefix) eq 2 then $
       files = choosefiles(selector,fullprefix[0], fullprefix[1], ct=count)
   if n_elements(prefix) eq 3 then $
       files = choosefiles(selector,fullprefix[0], fullprefix[1], fullprefix[2], ct=count)
   if n_elements(prefix) eq 4 then $
       files = choosefiles(selector,fullprefix[0], fullprefix[1], fullprefix[2], fullprefix[3], ct=count)
   if n_elements(prefix) gt 4 then begin
       print, 'Selecting this many file types is not yet enabled...'
       files = choosefiles(selector,fullprefix[0], fullprefix[1], fullprefix[2], fullprefix[3], ct=count)
   endif

   if count eq 0 then return, ''

   imh = lrisheaderinfo(files)

   if keyword_set(nofilter) eq 0 then begin  

    if lpar.fileseq ne '' and keyword_set(nofilefilter) eq 0 then begin
      wlist = -1
      count = 0
      if lpar.koa eq 0 then begin 
         filen = filenum(files)
         if dirtype eq 's1' then filen = imh.framenum
      endif
      if lpar.koa eq 1 then filen = imh.koaframe
      fileseqarr = strseq(lpar.fileseq)
      for i = 0, n_elements(fileseqarr)-1 do begin
         w = where(filen eq fileseqarr[i], ct) ; imh.frameno
         if ct gt 0 then wlist = [wlist, w]
      endfor
      if n_elements(wlist) eq 1 then return, ''
      wlist = unique(wlist)
      count = n_elements(wlist)-1
      wlist = wlist[1:*]
      files = files[wlist]
      imh = imh[wlist]
    endif

    ; could replace all of the below with seqfilter now.

    if lpar.filterseq[0] ne '' then begin
      wlist = -1
      count = 0
      for i = 0, n_elements(lpar.filterseq)-1 do begin
         w = where(imh.filter eq lpar.filterseq[i], ct)
         if ct gt 0 then wlist = [wlist, w]
      endfor
      if n_elements(wlist) eq 1 then return, ''
      wlist = unique(wlist)
      count = n_elements(wlist)-1
      wlist = wlist[1:*]
      files = files[wlist]
      imh = imh[wlist]
    endif
   
    if lpar.gratingseq[0] ne '' then begin
      wlist = -1
      count = 0
      for i = 0, n_elements(lpar.gratingseq)-1 do begin
         w = where(imh.grname eq lpar.gratingseq[i], ct)
         if ct gt 0 then wlist = [wlist, w]
      endfor
      if n_elements(wlist) eq 1 then return, ''
      wlist = unique(wlist)
      count = n_elements(wlist)-1
      wlist = wlist[1:*]
      files = files[wlist]
      imh = imh[wlist]
    endif

    if lpar.targetseq[0] ne '' and keyword_set(nofilefilter) eq 0  then begin
      wlist = -1
      count = 0
      for i = 0, n_elements(lpar.targetseq)-1 do begin
         w = where(imh.target eq lpar.targetseq[i], ct)
         if ct gt 0 then wlist = [wlist, w]
      endfor
      wlist = unique(wlist)
      if n_elements(wlist) eq 1 then begin
        simptarg = imh.target
        for j = 0, n_elements(imh.target)-1 do simptarg[j] = (strsplit(imh[j].target,'_',/extract)) [0]
        for i = 0, n_elements(lpar.targetseq)-1 do begin
           w = where(simptarg eq lpar.targetseq[i], ct)
           if ct gt 0 then wlist = [wlist, w]
        endfor
      endif
      if n_elements(wlist) eq 1 then return, ''
      count = n_elements(wlist)-1
      wlist = wlist[1:*]
      files = files[wlist]
      imh = imh[wlist]
    endif

    if lpar.dichroicseq[0] ne '' then begin ; not actually an option yet
      wlist = -1
      count = 0
      for i = 0, n_elements(lpar.dichroicseq)-1 do begin
         w = where(imh.dichroic eq lpar.dichroicseq[i], ct)
         if ct gt 0 then wlist = [wlist, w]
      endfor
      if n_elements(wlist) eq 1 then return, ''
      wlist = unique(wlist)
      count = n_elements(wlist)-1
      wlist = wlist[1:*]
      files = files[wlist]
      imh = imh[wlist]
    endif

   endif

   count = n_elements(files)

   return, files
end


