;+
; NAME:
;       readlrisr4
;
; PURPOSE:
;       Reads and bias-corrects a new (2021+) LRIS red file.
;
; CALLING SEQUENCE:
;       readlrisr1, filename, [header=header, /nobias]
;
; EXPLANATION:
;       Reads, bias-subtracts, and adds some header information for
;       a post-replacement LRIS red ("LRIS-R4") file to mimic the output
;       format produced by readmhdufitslp for 2010-2020 LRIS files.
;
; INPUTS:
;       filename - File to read in.
;
; OPTIONAL INPUT KEYWORDS:
;       /nobias - Do not use the overscan to bias-correct.
;       /verbose - Does not do anything
;
; OPTIONAL OUTPUT KEYWORDS:
;       header - Contains the FITS header.
;
; OUTPUTS:
;       Returns a 2D array containing the FITS data.
; 
; EXAMPLES:
;       im = readlrisr4('r210701_00001.fits', header=h)
;
; NOTES:
;       Prototype version.  Not tested thoroughly.
;
;-
;------------------------------------------------------------------------------

function extractcoord, coordstr
  if size(coordstr,/tn) ne 'STRING' then return, [-1,-2,-1,-2]
  coordstr2 = repstr(repstr(repstr(coordstr,':',','),'[',''),']','')
  coord = fix(strnum(strsplit(coordstr2,',',/extract)))
  if n_elements(coord) eq 1 then return, [-1,-2,-1,-2]
  x = coord[0:1]
  y = coord[2:3]
  arr = [x[sort(x)], y[sort(y)]]
  return, arr
end  

function readlrisr4, filename, header=header, nobias=nobias, verbose=verbose, setredwav=setredwav

   ; setredwav is only for error message printing

   common warn4, msgprinted

   ; Note: compared to LRIS-R1 (previous single-extension CCD) the X and Y are transposed

   im = readfits(filename, header, /silent)*1.0
   sxaddpar, header, 'BZERO', 0

   s = size(im)
   nx = fix(s[1])
   ny = fix(s[2])
     
   ; overscan zones, in format x0 x1 y0 y1, zero-referenced
   bsec1 = extractcoord(sxpar(header, 'BSEC1'))-1
   bsec2 = extractcoord(sxpar(header, 'BSEC2'))-1
   bsec3 = extractcoord(sxpar(header, 'BSEC3'))-1
   bsec4 = extractcoord(sxpar(header, 'BSEC4'))-1
   ; science zones, in format x0 x1 y0 y1, zero-referenced
   dsec1 = extractcoord(sxpar(header, 'DSEC1'))-1
   dsec2 = extractcoord(sxpar(header, 'DSEC2'))-1
   dsec3 = extractcoord(sxpar(header, 'DSEC3'))-1
   dsec4 = extractcoord(sxpar(header, 'DSEC4'))-1

   namp = 4
   if bsec3[0] lt 0 then namp = 2 

   ; Check if these values actually make sense given the array dimensions
   secx = [ bsec1[[0,1]],bsec2[[0,1]] , dsec1[[0,1]],dsec2[[0,1]]]
   secy = [ bsec1[[2,3]],bsec2[[2,3]] , dsec1[[2,3]],dsec2[[2,3]]]
   if namp eq 4 then begin
     secx = [secx, bsec3[[0,1]],bsec4[[0,1]] , dsec3[[0,1]],dsec4[[0,1]]]
     secy = [secy, bsec3[[2,3]],bsec4[[2,3]] , dsec3[[2,3]],dsec4[[2,3]]]
   endif
   if min(secx) lt 0 or min(secy) lt 0 then begin
     logprint, '  Warning: SEC coordinate less than zero'     
   endif
   if max(secx) ge nx or max(secy) ge ny or $
      max(secx) lt nx*7/10 or max(secy) lt ny*7/10 then begin
     logprint, '  Warning: BSEC/DSEC coordinates inconsistent with array size' 
     modematch = ''
     if nx eq 2208 and ny eq 2124 then begin ; full-frame, 2 amps, binned 2x2 [9414720]
        modematch = 'full-frame, 2 amps, binned 2x2'
        sxaddpar, header, 'DSEC1', '[5:1032,1:2064]', ' Fixed by readlrisr4' 
        sxaddpar, header, 'BSEC1', '[1033:1103,1:2064]'
        sxaddpar, header, 'BSEC2', '[1106:1176,1:2064]'
        sxaddpar, header, 'DSEC2', '[1177:2204,1:2064]'
        sxdelpar, header, ['BSEC3','BSEC4','DSEC3','DSEC4']
     endif
     if nx eq 2208 and ny eq 4188 then begin ; full-frame, 2 amps, binned 2x1 [18529920]
        modematch = 'full-frame, 2 amps, binned 2x1'
        sxaddpar, header, 'DSEC1', '[5:1032,1:4128]', ' Fixed by readlrisr4'
        sxaddpar, header, 'BSEC1', '[1033:1103,1:4128]'
        sxaddpar, header, 'BSEC2', '[1106:1176,1:4128]'
        sxaddpar, header, 'DSEC2', '[1177:2204,1:4128]'
        sxdelpar, header, ['BSEC3','BSEC4','DSEC3','DSEC4']
     endif
     if nx eq 4234 and ny eq 4188 then begin ; full-frame, 2 amps, unbinned [35498880]
        modematch = 'full-frame, 2 amps, unbinned'
        sxaddpar, header, 'DSEC1', '[8:2064,1:4128]', ' Fixed by readlrisr4'
        sxaddpar, header, 'BSEC1', '[2065:2117,1:4128]'
        sxaddpar, header, 'BSEC2', '[2118:2168,1:4128]'
        sxaddpar, header, 'DSEC2', '[2171:4226,1:4128]'
        sxdelpar, header, ['BSEC3','BSEC4','DSEC3','DSEC4']
     endif
     if nx eq 4234 and ny eq 4248 then begin ; full-frame, 4 amps, unbinned [36008640]
        modematch = 'full-frame, 4 amps, unbinned'
        sxaddpar, header, 'DSEC3', '[8:2064,2185:4248]'
        sxaddpar, header, 'BSEC3', '[2065:2117,2185:4248]'
        sxaddpar, header, 'BSEC4', '[2118:2168,2185:4248]'
        sxaddpar, header, 'DSEC4', '[2171:4227,2185:4248]'
        sxaddpar, header, 'DSEC1', '[8:2064,1:2064]', ' Fixed by readlrisr4'
        sxaddpar, header, 'BSEC1', '[2065:2117,1:2064]'
        sxaddpar, header, 'BSEC2', '[2118:2168,1:2064]'
        sxaddpar, header, 'DSEC2', '[2171:4227,1:2064]'     
     endif
     if modematch ne '' then begin
        logprint, '  Image dimensions matched mode: ', modematch
        logprint, '  Overriding DSEC/BSEC values using reference data'
        bsec1 = extractcoord(sxpar(header, 'BSEC1'))-1
        bsec2 = extractcoord(sxpar(header, 'BSEC2'))-1
        bsec3 = extractcoord(sxpar(header, 'BSEC3'))-1
        bsec4 = extractcoord(sxpar(header, 'BSEC4'))-1
        dsec1 = extractcoord(sxpar(header, 'DSEC1'))-1
        dsec2 = extractcoord(sxpar(header, 'DSEC2'))-1
        dsec3 = extractcoord(sxpar(header, 'DSEC3'))-1
        dsec4 = extractcoord(sxpar(header, 'DSEC4'))-1
        namp = 4
        if bsec3[0] lt 0 then namp = 2 
     endif else begin
        logprint, ' Could not find a matching mode using image dimensions'
     endelse
   endif
   
   ; quadrants are (origin at bot-left):
   ; 3 4
   ; 1 2
   ; or in 2-amp mode:
   ; 1 2

   if nx gt 3000 then xbin = 1 else xbin = 2 ; spatial
   if ny gt 3000 then ybin = 1 else ybin = 2 ; spectral
   binning = clip(xbin)+','+clip(ybin) ; spatial x spectral.  Becomes Y x X after transpose.
   sxaddpar, header, 'BINNING', binning
   
   nxout = (dsec1[1]-dsec1[0]+1)+(dsec2[1]-dsec2[0]+1)
   if namp eq 2 then nyout = (dsec1[3]-dsec1[2]+1)
   if namp eq 4 then nyout = (dsec1[3]-dsec1[2]+1)+(dsec3[3]-dsec2[0]+1)
      
   outsec1 = dsec1 - [dsec1[0],dsec1[0], dsec1[2],dsec1[2]]
   outsec2 = dsec2 - [dsec2[0],dsec2[0], dsec2[2],dsec2[2]] + [outsec1[1]+1,outsec1[1]+1,0,0]
   outsec3 = dsec3 - [dsec3[0],dsec3[0], dsec3[2],dsec3[2]] + [0,0,outsec1[3]+1,outsec1[3]+1]
   outsec4 = dsec4 - [dsec4[0],dsec4[0], dsec4[2],dsec4[2]] + [outsec1[1]+1,outsec1[1]+1,outsec1[3]+1,outsec1[3]+1]
          
   sxaddpar, header, 'AMP1', clip(outsec1[0])+','+clip(outsec1[1]) ; x (->Y) only
   sxaddpar, header, 'AMP2', clip(outsec2[1])+','+clip(outsec2[1]) ; x (->Y) only
 
   if keyword_set(nobias) eq 0 then begin
     bias1 = fix(median(im[bsec1[0]:bsec1[1],bsec1[2]:bsec1[3]]))
     bias2 = fix(median(im[bsec2[0]:bsec2[1],bsec2[2]:bsec2[3]]))
     im[dsec1[0]:dsec1[1],dsec1[2]:dsec1[3]] -= bias1
     im[dsec2[0]:dsec2[1],dsec2[2]:dsec2[3]] -= bias2
 
     if namp eq 4 then begin
       bias3 = fix(median(im[bsec3[0]:bsec3[1],bsec3[2]:bsec3[3]]))
       bias4 = fix(median(im[bsec4[0]:bsec4[1],bsec4[2]:bsec4[3]]))
       im[dsec3[0]:dsec3[1],dsec3[2]:dsec3[3]] -= bias3
       im[dsec4[0]:dsec4[1],dsec4[2]:dsec4[3]] -= bias4
     endif

     if keyword_set(check) then begin
       im[bsec1[0]:bsec1[1],bsec1[2]:bsec1[3]] -= bias1
       im[bsec2[0]:bsec2[1],bsec2[2]:bsec2[3]] -= bias2
       if namp eq 4 then begin
         im[bsec3[0]:bsec3[1],bsec3[2]:bsec3[3]] -= bias3
         im[bsec4[0]:bsec4[1],bsec4[2]:bsec4[3]] -= bias4
       endif
     endif
   endif

   outim = fltarr(nxout, nyout)
   
   outim[outsec1[0]:outsec1[1],outsec1[2]:outsec1[3]] = im[dsec1[0]:dsec1[1],dsec1[2]:dsec1[3]]
   outim[outsec2[0]:outsec2[1],outsec2[2]:outsec2[3]] = im[dsec2[0]:dsec2[1],dsec2[2]:dsec2[3]]
   if namp eq 4 then begin
     outim[outsec3[0]:outsec3[1],outsec3[2]:outsec3[3]] = im[dsec3[0]:dsec3[1],dsec3[2]:dsec3[3]]
     outim[outsec4[0]:outsec4[1],outsec4[2]:outsec4[3]] = im[dsec4[0]:dsec4[1],dsec4[2]:dsec4[3]]
   endif
   
   outim = transpose(outim)
   outim = reverse(outim,1) ; obviously this can be simplified...
   outim = reverse(outim,2)
   
   sout = size(outim)
   sxaddpar, header, 'NAXIS1', fix(sout[1])
   sxaddpar, header, 'NAXIS2', fix(sout[2])
   sxaddpar, header, 'BITPIX', -32
   
   dum = sxpar(header,'PANE',count=ct)
   if ct eq 0 then sxaddpar, header, 'PANE', '0,0,4142,4114' ; assumes no windowing
   dum = sxpar(header,'MJD-OBS',count=ct)
   if ct eq 0 then sxaddpar, header, 'MJD-OBS', sxpar(header,'MJD')
   dum = sxpar(header,'ELAPTIME',count=ct)
   if ct eq 0 then sxaddpar, header, 'ELAPTIME', sxpar(header,'TELAPSE')
   dum = sxpar(header,'OBJECT',count=ct)
   if ct eq 0 then sxaddpar, header, 'OBJECT', ''
   dum = sxpar(header,'ST',count=ct)
   if ct eq 0 then sxaddpar, header, 'ST', sxpar(header,'LST')

   dum = sxpar(header,'FLIMAGIN',count=ct)
   if ct eq 0 then sxaddpar, header, 'FLIMAGIN', sxpar(header,'FLAMP1')
   dum = sxpar(header,'FLSPECTR',count=ct)
   if ct eq 0 then sxaddpar, header, 'FLSPECTR', sxpar(header,'FLAMP2') ; not certain about 1 vs 2 order
   
   dum = sxpar(header,'FRAMENO',count=ct)
   if ct eq 0 then begin
     ofile = sxpar(header,'OFNAME', count=ctofile)
     framenum = strmid(ofile,strpos(ofile,'_')+1,strpos(ofile,'.')-strpos(ofile,'_')-1)
     sxaddpar, header, 'FRAMENO', framenum
   endif
   
   checkwav = sxpar(header,'WAVELEN', count=ctw)
   checkgrating = clip(sxpar(header,'GRANAME'))
   checkversion = clip(sxpar(header,'INSTREV'))
   checktrap = clip(sxpar(header,'TRAPDOOR'))
   checkexp = fix(sxpar(header,'TELAPSE'))
   if strpos(checkgrating,'/') gt 0 and ctw gt 0 and checkwav eq 0. and checkversion eq 'REDMARK4' then begin
     if n_elements(msgprinted) eq 0 and n_elements(setredwav) eq 0 then begin
       logprint, ' WARNING: Key header information is written incorrectly in LRIS Mark 4 red files.'
       logprint, '       You may need to explicitly specify the central wavelength using the option'
       logprint, '       setredwav=[value] when calling LPipe, or you may need to edit the WAVELEN'
       logprint, '       and/or TRAPDOOR values in the headers in the raw files using editheader.' 
       logprint, '       LPipe will attempt to copy information from the blue file headers, but this'
       logprint, '       will usually miss some exposures.' 
       msgprinted = 1       
     endif
     sxdelpar, header, 'WAVELEN'   ; is inaccurate (always zero)
   endif
   
   if checktrap eq 'closed' then begin
        lampkeys = ['MERCURY','NEON','ARGON','CADMIUM','ZINC','HALOGEN',$
                    'KRYPTON','XENON','FEARGON','DEUTERI'] ; internal lamps
        nlamp = 0
        for k = 0, n_elements(lampkeys)-1 do begin
           nlamp += clip(sxpar(header,lampkeys[k])) eq 'on' 
        endfor      
        if nlamp eq 0 then sxaddpar, header, 'TRAPDOOR', 'open'
        ;if nlamp ge 1 then sxaddpar, header, 'TRAPDOOR', 'closed'
        if nlamp eq 0 then logprint, '  [TRAPDOOR: '+checktrap+' -> '+clip(sxpar(header,'TRAPDOOR'))+']'
   endif
   
  ; print
  ; print, sxpar(header, 'DSEC1')
  ; print, sxpar(header, 'DSEC2')
  ; print, sxpar(header, 'DSEC3')
  ; print, sxpar(header, 'DSEC4')   
  ; print
  ; print, dsec1
  ; print, dsec2
  ; print, dsec3
  ; print, dsec4
  ; print
  ; print, outsec1
  ; print, outsec2
  ; print, outsec3
  ; print, outsec4

   return, outim
end

