;+
; NAME:
;       rundatestr
;
; PURPOSE:
;       Determine the date a file was taken (wrapped appropriately for
;       Keck's geographical location) and encode it in a a six- or 
;       eight-character string.
;
; CALLING SEQUENCE:
;       result = rundatestr(hdate, /nocentury)
;
; EXPLANATION:
;       Used for file-renaming purposes to produce a date code for a file.
;       The wrap point is at 22:00 UT (=noon HST), to allow calibrations
;       taken between noon and 2 PM to be associated with the following
;       night rather than the previous night.
;
; INPUTS:
;       hdate - A standard date string, e.g. from a header. Generally expected
;               to be in yyyy-mm-ddThh:mm:ss format.
;
; OPTIONAL INPUT KEYWORDS:
;       /nocentury - Leave out the first to year digits from the output string.
;
; OUTPUTS:
;       Returns a six- or eight-character date string, e.g. 20070101.
; 
; EXAMPLES:
;       rundatestr('2006-12-31T23:01:44')
;
;
;-
;------------------------------------------------------------------------------

function rundatestr, hdate, nocentury=nocentury

   hdate = repstr(hdate,'/','-')
   hdt = strsplit(hdate,'T',/extract)
   utdate = hdt[0]
   utyear = fix((strsplit(utdate,'-',/extract)) [0])
   utmonth = fix((strsplit(utdate,'-',/extract)) [1])
   utday   = fix((strsplit(utdate,'-',/extract)) [2])
   if n_elements(hdt) ge 2 then begin
     uttime = hdt[1]
     uthour = fix((strsplit(uttime,':',/extract)) [0])
     if uthour ge 22 then utday += 1
       ; observations after 22:00 (=noon HST) are given the next day's file date code
     nday = [-1,31,28,31,30,31,30,31,31,30,31,30,31]
     if utyear mod 4 eq 0 then nday[1] = 29 ; we won't worry about 1900 or 2100 being wrongly treated as leap years...
     if utmonth ge 1 and utmonth le 12 then $
     if utday gt nday[utmonth] then begin
        utday = 1
        utmonth += 1
        if utmonth gt 12 then begin
          utmonth = 1
          utyear += 1
        endif
     endif    
   endif

   if keyword_set(nocentury) then datestr = string(utyear mod 100,'(I02)')+string(utmonth,'(I02)')+string(utday,'(I02)') $
              else datestr = string(utyear,'(I04)')+string(utmonth,'(I02)')+string(utday,'(I02)')

   return, datestr

end

