; +
; NAME:
;       unwrapkoa
;
; PURPOSE:
;       Un-tars a group of KOA tarballs and sorts the data by observation night.
;
; CALLING SEQUENCE:
;       unwrapkoa, [filestr, /deletetar, /rename, /destination]
;
; EXPLANATION:
;       A convenience tool for organizing large quantities of KOA data.
;       Once downloads complete, use this to sort them by night.  You should
;       generally delete the .tar files once this finishes.
;       Will also sort already-untarred files (becomes a wrapper to sortkoa).
;
; OPTIONAL INPUTS:
;       filestr - Wildcard string indicating which files.  Defaults to all
;                 .tar files in the current directory (e.g. to *.tar).
;
; OPTIONAL INPUT KEYWORDS:
;       /deletetar - Delete the tarballs automatically after unpacking/sorting.
;       /rename - Rename the files sequentially intead of using KOA filenames.
;       /destination - Specify a destination directory rather than sorting by
;                      night.
;
; EXAMPLES:
;       unwrapkoa, '*.tar', /rename
;
;-
;------------------------------------------------------------------------------

pro unwrapkoa, filestr, deletetar=deletetar, rename=rename, destination=destination

  ; Unpack a KOA tarball or many KOA tarballs.  
  ; By default puts all the data into directories associated with each night.
  ; Pool all the files together in the current directory by specifying destination = '.'

  if n_elements(destination) then begin
       deststr = destination
       if deststr eq '' then deststr = '.'
       if strmid(deststr,strlen(deststr)-1) eq '/' then deststr = strmid(deststr,0,strlen(deststr)-1)
       if file_test(deststr) eq 0 then begin
           print, deststr, ' not found'
           return
       endif
  endif

  if n_elements(filestr) eq 0 then filestr = '*.tar'

  if strpos(filestr,'.fits') gt 0 and strpos(filestr,'.tar') eq -1 then begin
     sortkoa, filestr, rename=rename
     return
  endif

  files = file_search(filestr)

  wtar = where(strpos(files,'.tar') gt 0, cttar)
  for i = 0, cttar-1 do begin
     arr = strsplit(removepath(files[wtar[i]]),'_',/extract)
     if n_elements(arr) ne 4 then begin
        print, ' This is not a standard KOA archive tar file: '+files[wtar[i]]
        continue
     endif
     jobid = arr[2]
     outdir = 'KOA_'+jobid   
     command = 'tar -xf '+files[wtar[i]]
     print, command
     spawn, command
     if file_test(outdir) eq 0 then begin
        print, ' Untar operation did not succeed?'
        continue
     endif
     for j = 0, 1 do begin 
       if j eq 0 then path = outdir+'/LRIS/raw/sci/'
       if j eq 1 then path = outdir+'/LRIS/raw/cal/'
       if file_test(path) then begin
          jpg = file_search(path+'*.jpg', count=ctjpg)
          for k = 0, ctjpg-1 do if file_test(jpg[k]) then file_delete, jpg[k], /allow_nonexistent
          f = file_search(path+'*.fits.gz', count=ctfits)
          for k = 0, ctfits-1 do begin
             if n_elements(destination) eq 0 then begin
               h = headfits(f[k])
               deststr = rundatestr(sxpar(h,'DATE',/silent))
               if file_test(deststr) eq 0 then print, 'Creating '+deststr
               if file_test(deststr) eq 0 then spawn, 'mkdir '+deststr
             endif
             outfile = deststr+'/'+removepath(f[k])
             if file_test(outfile) eq 0 then begin
                file_move, f[k], outfile
                if keyword_set(rename) then lrisrename, outfile
             endif else begin
                print, outfile+' already exists?'
             endelse
          endfor
          spawn, 'rmdir '+path
       endif
     endfor
     spawn, 'rmdir '+outdir+'/LRIS/raw/'
     spawn, 'rmdir '+outdir+'/LRIS/'
     spawn, 'rmdir '+outdir ; won't succeed if non-LRIS runs present
     if keyword_set(deletetar) then file_delete, files[wtar[i]]
     if n_elements(untarred) eq 0 then untarred = files[wtar[i]] else untarred = [untarred, files[wtar[i]]]
  endfor

  print, 'Unwrapped: '
  for u = 0, n_elements(untarred)-1 do begin
    print, '  ', untarred[u]
  endfor

end


