; autolris.pro - Library file for LPipe.
;
; This file contains a large number of subroutines called by the top-level LPipe code.
; Most of these execute one of the individual 'steps' of the processing sequence, such
; as flat-field division or flux-calibration, although a few are general functions used
; by the software.
;
; In general, none of these should be called except by LPipe itself.
;
; List of subroutines (as of 2022 July 4):
; other routines:
;  autolrissetup - Mostly obsolete; creates a local version of the configuration file
;  readseq() - Get some information out of the header useful to check for component moves
;  updateseq - Used to determine and store the last component move
; primary step functions:
;  autolrisprepare      - prepare step
;  autolrismakebias     - in development
;  autolrisdebias       - in development
;  autolrismakeimflat   - image flat field construction
;  autolrismakespecflat - spectroscopy flat field construction
;  autolrisimflatten    - image flat field correction
;  autolrisspecflatten  - spectroscopy flat field correction
;  autolrisreflatten    - "reflattening" (illumination correction) using e.g. a supersky flat
;  autolrismakefringe   - LRIS-R1 only, very old, needs updating
;  autolrisrmfringe     - LRIS-R1 only, very old, needs updating
;  autolriscrcleanim    - image cosmic ray cleaning
;  autolriscrcleanspec  - spectroscopy cosmic ray cleaning
;  autolrissplit        - split step
;  autolrisskysubtract  - 2D sky subtraction
;  autolriswavcal       - wavelength solution
;  autolrissum          - 2D spectrum coaddition
;  autolristrace        - tracing
;  autolrisextract      - 1D spectrum extraction
;  autolriswavapply     - apply wavelength solution to data
;  autolrisresponse     - response function solution
;  autolrisfluxcal      - flux calibration (apply response function)
;  autolriscombine      - combine 1D spectra
;  autolrisconnect      - join red and blue spectra
;  autolrisastrorometry - solve astrometry for images
;  autolrisphotometry   - determine zeropoints for images
;  autolrisphotcat      - output photometric catalogs (part of phot step)
;  autolrisstack        - coadd reduced images
;  lrisfailuresummary   - output summary of any problems that arose



; -------------------------
pro autolrissetup

   ; No longer normally called.

   ; Copy the default configuration file to the current directory to allow a local configuration.

   ; Formerly this was needed because of the changing default trace position (the region to look for sources
   ; needed to be customized by the user between different nights).  This is no longer needed because standards
   ; are used to find the trace automatically.

   islocalfile = file_test('lpipe.par')

   if islocalfile eq 0 then begin
      print, '  LPIPE - LRIS automatic processing pipeline'
      print, '          version 2022.06-BETA'
      print, '          authored by Daniel Perley (d.a.perley@ljmu.ac.uk)'

      paths = strsplit(!path,':',/extract)
      for p = -1, n_elements(paths)-1 do begin   ; find the directory this file is in
         if p ge 0 then path = paths[p] $
                        else path = ''
         sfile = path+'/lpipe.pro'
         n = countfiles(sfile)
         if n gt 0 then begin
           defaultparfile = path+'/lpipe.par'
           if file_test(defaultparfile) eq 0 then defaultparfile = path+'/defaults/lpipe.par'
           if file_test(defaultparfile) then begin
              spawn, 'cp '+defaultparfile+' ./' 
              break
           endif
         endif
      endfor

      if file_test('lpipe.par') then begin
         print, '  To edit configuration for this reduction, modify ./lpipe.par'
         print, '  To edit default configuration for future reductions, modify '+defaultparfile
         print, '  To run the pipeline, re-run the command.'
      endif else begin
         print, '  ERROR: Unable to find default parameter file.  Installation problem?'
      endelse
      retall
   endif

end



; these seqs have nothing to do with seqfilter

function readseq, h
   
    ; Get some specific information out of the header regarding the
    ; configuration of the image, including when the grating/filter was
    ; last moved.

    seq = {dich:'', rfilt:'', bfilt:'', grat:'', grangle:0., slit:'', $
           dichf:'', rfiltf:'', bfiltf:'', gratf:'', granglef:'', slitf:''}
    seq.dich = sxpar(h, 'DICHNAME')
    seq.rfilt = sxpar(h,'REDFILT')
    seq.bfilt = sxpar(h, 'BLUFILT')
    seq.grat = sxpar(h, 'GRANAME')
    seq.grangle = sxpar(h, 'GRANGLE')
    seq.slit = sxpar(h,'SLITNAME')
    dum = sxpar(h, 'DICHSEQF', count=ct) 
    if ct gt 0 then begin
      seq.dichf = sxpar(h, 'DICHSEQF')
      seq.rfiltf = sxpar(h, 'RFLTSEQF')
      seq.bfiltf = sxpar(h, 'BFLTSEQF')
      seq.gratf = sxpar(h, 'GRATSEQF')
      seq.granglef = sxpar(h, 'GRANSEQF')
      seq.slitf = sxpar(h, 'SLITSEQF')
    endif

    return, seq
end

pro updateseq, seq, prevseq, seqfile

    ; Used by the reflat procedure to look for moves of the gratings, filters, etc.

    if seq.dich eq prevseq.dich then seq.dichf = prevseq.dichf else seq.dichf = seqfile
    if seq.rfilt eq prevseq.rfilt then seq.rfiltf = prevseq.rfiltf else seq.rfiltf = seqfile
    if seq.bfilt eq prevseq.bfilt then seq.bfiltf = prevseq.bfiltf else seq.bfiltf = seqfile
    if seq.grat eq prevseq.grat then seq.gratf = prevseq.gratf else seq.gratf = seqfile
    if seq.grangle eq prevseq.grangle then seq.granglef = prevseq.granglef else seq.granglef = seqfile
    if seq.slit eq prevseq.slit then seq.slitf = prevseq.slitf else seq.slitf = seqfile
end



; -------------------------
pro autolrisprepare, modestr=modestr, camstr=camstr, nmissingkey=nmissingkeys, prepareall=prepareall, skipfocus=skipfocus, setredwav=setredwav
   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

  ; Prepare images for processing - debias, crop, fix some bad pixels, add header information,
  ; and merge the separate extensions into a simple fits file - see lrisprepare.pro for details.

  mosworkingdir = 'mosredux/'
  darkdir = 'dark/'

  if n_elements(modestr) gt 0 then begin
     if modestr eq 'is' or modestr eq 'i,s' or modestr eq 'im,sp'  then mode = ''
     if modestr eq 'si' or modestr eq 's,i' or modestr eq 'sp,im'  then mode = ''
     if n_elements(mode) eq 0 then mode = strmid(modestr,0,1)
  endif else begin
     mode = ''
  endelse

  cam = strmid(camstr,0,1)
  prefchar = strmid(camstr,0,1) ; i.e. camchar
 
  if lpar.koa then begin
      prefchar='L'+strupcase(cam)+'.??'
      if cam eq 'b' then otherprefchar = 'LR.??'
      if cam eq 'r' then otherprefchar = 'LB.??'
  endif else begin
      if cam eq 'r' then otherprefchar = 'b'
      if cam eq 'b' then otherprefchar = 'r'
  endelse

  ;files =        file_search(lpar.datadir+    prefchar+lpar.wildchar+'.fits')
  files = lrisselectfiles('d',prefchar,'','.fits*')
  nfiles = n_elements(files)
  if nfiles eq 1 then if files[0] eq '' then nfiles = 0

  ;if lpar.datadir ne '' then  begin
     logprint, ' Looking for raw data at: ', lpar.datadir+    prefchar+lpar.wildchar+'.fits*'
     if nfiles gt 0 then begin
        logprint, '   Found ', clip(nfiles), ' files'
     endif else begin
        logprint, '   Did not find any files!  Check your data directory path!'
     endelse
  ;endif

  badfiles = ['']
  badfilefile = 'badfiles.txt'  ; ignorefiles.txt would be a better name?
  if file_test(badfilefile) then readcol, badfilefile, badfiles, format='a'
  ;'~/research/redux/allgrbpos.txt' 
  ;'~/research/redux/standardpos.txt'

  namefixfiles = ['']
  catalogfile = 'catalog.txt' ; current directory
  if file_test(catalogfile) then namefixfiles = [namefixfiles, catalogfile]
  catalogfile = 'starlist.txt'
  if file_test(catalogfile) then namefixfiles = [namefixfiles, catalogfile]
  landoltfieldposfile = lpar.refdatapath+'/landoltpos.txt'
  if file_test(landoltfieldposfile) then namefixfiles = [namefixfiles, landoltfieldposfile]
  specstandardposfile = lpar.refdatapath+'/specstandards.dat'
  if file_test(specstandardposfile) then namefixfiles = [namefixfiles, specstandardposfile]
  if n_elements(namefixfiles) gt 1 then namefixfiles = namefixfiles[1:*] else delvarx, namefixfiles

   prevtarg = ''
   prevobj = ''
   prevradeg = 0.
   prevdecdeg = 0.

  for f = 0, n_elements(files)-1 do begin
     if files[f] eq '' then continue

     dum = where(files[f] eq badfiles or removepath(files[f]) eq badfiles, ct)
     if ct gt 0 then continue
    
     fileroot = removepath(files[f])
     outfileroot = repstr(fileroot,'.gz','')
     if lpar.koa then outfileroot = cam+strmid(outfileroot,5)
     outnameim = lpar.imworkingdir + 'p' + outfileroot
     outnamesp = lpar.spworkingdir2d + 'p' + outfileroot
     outnamemos = mosworkingdir + 'p' + outfileroot
     outnamedark = darkdir + 'p' + outfileroot
     matchi = file_test(outnameim)
     matchs = file_test(outnamesp)
     matchm = file_test(outnamemos)
     matchd = file_test(outnamedark)
     ct = matchi+matchs+matchm+matchd
     if ct eq 0 or lpar.overwrite gt 0 then begin
        h = headfits(files[f], /silent)

        ; Store in the header the name of the first file in which the current configuration was set.
        ; This is useful for identifying a block of files in the same config 
        ; (since LRIS doesn't always return to the same position exactly) - currently the reflat
        ; procedure uses this to limit blocks of supersky flats, and the spectroscopic extractions
        ; could also use it to optimize the choice of standard frame.
        ; This should be robust to re-runs and unprepared files, etc.
        ; However, if that first file is deleted in between program runs, this will cause files 
        ; processed later to be assigned different identifiers.
        if n_elements(seq) gt 0 then begin
           prevseq = seq
           prevfilef = where(files eq prevfile)
        endif else begin
           prevfilef = -999
        endelse
        seq = readseq(h)
        if f eq prevfilef + 1 then begin
           ; No breaks since last time seq was populated; just check against last file.
           updateseq, seq, prevseq, files[f]
        endif else begin
           ; break in sequence.  Have to look back at the most recent processed file and then run forward from there.
           ;print, 'Break in sequence before ', files[f]
           oldreffile = ''
           for fn = f-1, 0, -1 do begin  ; Find the most recent processed file.
             prevoutnameim = lpar.imworkingdir + 'p' + repstr(removepath(files[fn]),'.gz','')
             prevoutnamesp = lpar.imworkingdir + 'p' + repstr(removepath(files[fn]),'.gz','')
             prevoutnamemo = mosworkingdir + 'p' + repstr(removepath(files[fn]),'.gz','')
             prevoutnamedk = darkdir + 'p' + repstr(removepath(files[fn]),'.gz','')
             if file_test(prevoutnameim) then oldreffile = prevoutnameim
             if file_test(prevoutnamesp) then oldreffile = prevoutnamesp
             if file_test(prevoutnamemo) then oldreffile = prevoutnamemo
             if file_test(prevoutnamedk) then oldreffile = prevoutnamedk
             if oldreffile ne '' then break
           endfor
           if oldreffile ne '' then begin
             ;print, 'Found an earlier file:', oldreffile
             oldseq = readseq(headfits(oldreffile))  ; Read it in.
             prevoldseq = oldseq
             for fn = fn+1, f-1 do begin     ; Iterate forward, updating oldseq until the previous file is reached.
                ;;print, '   ', files[fn]
                oldh = headfits(files[fn])
                oldseq = readseq(oldh)
                updateseq, oldseq, prevoldseq, files[fn]
             endfor
             prevseq = oldseq
          endif else begin                 ; There are no earlier processed files - read in a blank seq.
             prevseq = {dich:'', rfilt:'', bfilt:'', grat:'', grangle:0., slit:'', $
                        dichf:'', rfiltf:'', bfiltf:'', gratf:'', granglef:'', slitf:''}
          endelse
          updateseq, seq, prevseq, files[f]
        endelse
        prevfile = files[f]

        ; Find the blue match of a red file, or find the red match of a blue file.
        ; This will be used to fill in missing header values if necessary.
        if n_elements(matchdata) eq 0 then begin
           ;if cam eq 'b' then matchfiles = file_search(lpar.datadir+'r'+lpar.wildchar+'.fits')
           ;if cam eq 'r' then matchfiles = file_search(lpar.datadir+'b'+lpar.wildchar+'.fits')
           if cam eq 'b' then matchfiles = lrisselectfiles('d',otherprefchar,'','.fits*', /nofilter)
           if cam eq 'r' then matchfiles = lrisselectfiles('d',otherprefchar,'','.fits*', /nofilter)
           if matchfiles[0] ne '' then begin
             matchdata = lrisheaderinfo(matchfiles) ; in principle only UT/expt is needed now.
           endif
        endif
        if n_elements(matchdata) gt 0 then begin
           hdata = lrisheaderinfo(files[f]) ; could replace with the header.

           ; Find any exposures that overlap.
           wmatch = where((matchdata.tstart gt hdata.tstart-10 and matchdata.tstart lt hdata.tstop+10) or $
                          (matchdata.tstop  gt hdata.tstart-10 and matchdata.tstop  lt hdata.tstop+10) or $ 
                          (hdata.tstart gt matchdata.tstart-10 and hdata.tstart lt matchdata.tstop+10) or $
                          (hdata.tstop  gt matchdata.tstart-10 and hdata.tstop  lt matchdata.tstop+10), ct)
           if ct eq 0 then begin
              fmatch = 'none'
           endif else begin
             if ct gt 1 then wmatch = wmatch[ct/2] else wmatch = wmatch[0] ; lazy
             fmatch = matchfiles[wmatch]
           endelse
        endif else begin
           fmatch = 'unknown'
        endelse

        ; Get some header information that will be useful in sorting the files.
        pane = strtrim(sxpar(h, 'PANE'),2)
        window = strtrim(sxpar(h, 'WINDOW'),2)
        ofile = sxpar(h,'OFNAME',count=ctoname)
        if ctoname gt 0 then isfocus = strpos(ofile,'foc') ge 0
        if ctoname eq 0 then isfocus = strpos(files[f],'foc') ge 0 and strpos(files[f],'foc') le 4
        if isfocus and keyword_set(skipfocus) then begin
           logprint, ' '+files[f], ' is a focus image'
           continue
        endif
        slit =  clip(strlowcase(clip(sxpar(h, 'SLITNAME', count=cts))))
        shutter = clip(sxpar(h,'TRAPDOOR'))
        if cam eq 'r' and lpar.lrisversion eq 4 then shutter = 'unknown' ; provisonal/temporary, hopefully can remove
        if cam eq 'b' then grname = clip(sxpar(h,'GRISNAME'))
        if cam eq 'r' then grname = clip(sxpar(h,'GRANAME'))
        lamps = sxpar(h,'LAMPS',count=ctl)
        if ctl gt 0 then begin
          nlamps = total(fix(strsplit(lamps,',',/extract)))
        endif else begin
          lampkeys = ['MERCURY','NEON','ARGON','CADMIUM','ZINC','HALOGEN','KRYPTON',$
                      'XENON','FEARGON','DEUTERI','FLAMP1','FLAMP2']
          nlamps = 0
          for ki = 0, n_elements(lampkeys)-1 do nlamps += clip(sxpar(h,lampkeys[ki])) eq 'on'
        endelse
        filemode = ''

        if slit eq 'direct' and strpos(grname,'/') eq -1 then filemode = 'i'   ; should probably put into lrisheaderinfo
        if strpos(slit,'long') ge 0  and strpos(grname,'/') ge 0 then filemode = 's' ; longslit and grating
        if strpos(slit,'long') eq -1 and strpos(grname,'/') ge 0 then filemode = 'm' ; non-longslit and grating
        if strpos(slit,'long') eq -1 and strpos(slit,'pol') eq -1 and slit ne 'direct' then filemode = 'm' ; slitmask image
        if shutter eq 'closed' and nlamps eq 0 then filemode = 'd'

        if (slit eq 'goh_lris' or slit eq 'focus_holes') and filemode ne 'd' then begin
           logprint, ' '+files[f], ' is a focus holes image'
           continue
        endif

        if filemode ne 'i' and filemode ne 's' and filemode ne 'm' and filemode ne 'd' then begin
           ; unknown file type
           if cts eq 0 then begin
                 logprint, ' '+files[f], ': SLITNAME keyword absent; cannot distinguish image/spectrum; skipping...' 
           endif else begin
                 if strpos(slit,'pol') ge 0 then logprint, ' '+files[f],': spectropolarimetry not supported'
                 logprint, ' '+files[f], ': unusual combination of slit ('+clip(slit)+') and disperser ('+clip(grname)+'); skipping...'
                 ; These seem to be mask images and in the future should be processed.
           endelse
           ;if slit eq 'direct' and strpos(grname,'/') ge 0 then $
           ;      print, files[f], ' Spectroscopic slitless flats not supported yet, skipping for now...'
           prepfail = [prepfail, files[f]]
           continue
        endif

        ; If the user forgot to change the object name but moved to a clearly different target, update using TARGNAME
        if n_elements(newobj) gt 0 then delvarx, newobj        
        obj = clip(sxpar(h,'OBJECT'))
        targ = clip(sxpar(h,'TARGNAME',count=cttarg))
        radeg = 15*ten(clip(sxpar(h,'RA')))
        decdeg = ten(clip(sxpar(h,'DEC')))
        ltarg = strlowcase(targ)
        ;trapdoor = clip(sxpar(h,'TRAPDOOR'))
        if strpos(ltarg,'flat') eq -1 and strpos(ltarg,'stow') eq -1 and $            ; trapdoor eq 'open' and 
           strpos(ltarg,'dome') eq -1 and strpos(ltarg,'horizon') eq -1 and ltarg ne '' then begin
            if obj eq prevobj and targ ne prevtarg and $
               cttarg gt 0 and strlen(targ) gt 2 and $
               (abs(decdeg-prevdecdeg) gt 0.1 or abs(radeg-prevradeg) gt 0.3) then begin
              ; New target and position, same object as before - set OBJECT to TARGET
              newobj = (strsplit(targ,'_',/extract)) [0]
            endif else begin
              prevtarg = targ
              prevradeg = radeg
              prevdecdeg = decdeg
            endelse
          prevobj = obj
        endif

        if prepareall eq 0 then begin
          if filemode eq 'd' then begin
             exp = sxpar(h,'EXPTIME',count=ctt)
             if ctt eq 0 then exp = sxpar(h,'ELAPTIME',count=ctt)
             if ctt eq 0 then exp = sxpar(h,'TELAPSE')
             if exp lt 1.5 then $
                logprint, ' '+files[f]+' is a bias frame'
             if exp gt 1.5 and exp lt 30 then begin
                logprint, ' '+files[f]+' is a dark image ('+clip(exp)+'s)'
             endif
             if exp ge 30. then begin
                logprint, ' '+files[f]+' claims to have a shut trapdoor but is long ('+clip(exp)+'s)'
                prepfail = [prepfail, files[f]] ; will be filed in dark with no warning if prepareall=1
             endif
             continue
          endif
          if filemode eq 'm' then begin
             logprint, ' '+files[f]+' is a multislit frame; not supported yet'
             continue
          endif
        endif
        if mode eq 'i' and filemode ne 'i' then continue
        if mode eq 's' and filemode ne 's' and filemode ne 'm' then continue
        if filemode eq 'i' then outname = outnameim
        if filemode eq 's' then outname = outnamesp
        if filemode eq 'm' then outname = outnamemos
        if filemode eq 'd' then outname = outnamedark

        if filemode eq 'i' and file_test(lpar.imworkingdir) eq 0 then spawn, 'mkdir '+lpar.imworkingdir
        if filemode eq 's' and file_test(lpar.spworkingdir2d) eq 0 then spawn, 'mkdir '+lpar.spworkingdir2d
        if filemode eq 'm' and file_test(mosworkingdir) eq 0 then spawn, 'mkdir '+mosworkingdir
        if filemode eq 'd' and file_test(darkdir) eq 0 then spawn, 'mkdir '+darkdir


        ;print, 'Preparing ', files[f]
        if lpar.debug eq 0 then catch, errorstatus $
           else errorstatus = 0
        if errorstatus ne 0 then begin
           logprint, '   Could not read in file (unexpected error):'
           logprint, '      '+!error_state.msg
           ;print, '   Ignoring this file, consider adding to badfiles.txt'
           prepfail = [prepfail, files[f]]
           success = -10
        endif else begin
          lrisprepare, files[f], gain=[1,1,1,1], /merge, crop='auto', $
            outname=outname, namefixfiles=namefixfiles, seq=seq, /requirecompleteh, newobj=newobj, $
            reducer=lpar.reducer, version=lpar.lpversion, matchfile=fmatch, errcode=errcode, $
            setredwav=setredwav
          if file_test(outname) eq 0 and errcode ne 1 then prepfail = [prepfail, files[f]]
            ; errcode = 1 is a mira: ignore these silently.  Different from errorstatus (for crashes)
        endelse
;        catch, /cancel 

     endif
  endfor

end


pro autolrismakebias, camera=camera
  
   ; Not complete yet and not used by the pipeline.
   ; Will need to either transpose images too, or not transpose spectra until
   ; the split stage, or make bias construction and removal recognize and de-transpose.

   common lrisauto, lpar

    fchip = '' 

    camchar = strmid(camera,0,1)

    ;files = file_search(lpar.imworkingdir+'p'+camchar+lpar.wildchar+fchip+'.fits') ; camera dealt with here
    ;imh = lrisheaderinfo(files)
    files = lrisselectfiles('s2','p'+camchar,fchip,'.fits', imh=imh)

    wbias = where(imh.exp lt 1 and imh.trapdoor eq 'closed' and imh.nlamp eq 0)
    isbias = intarr(n_elements(imh))
    isbias[wbias] = 1

    ; this is where transposing the data gets us into trouble.
    ; Biases are generic to imaging and spectra.
    ; Need to consider transposing images, too.

    binlist = unique(imh[wbias].binning)
    for b = 0, n_elements(binlist)-1 do begin
      winlist = unique(imh[wbias].win)
    endfor


end

;pro autolrisdebias, camera=camera
;end


; -------------------------
pro autolrismakeimflat, chip=chip, camera=camera, forceflat=forceflat

   ; Construct imaging flatfields.   Looks at what science images actually need flat-fielding, then looks for flats
   ; in the appropriate configurations and combines them, choosing dome or sky based on how many of each or available.

   ; Currently it constructs a flat even if the ability to do so is rather marginal.  Ideally in these cases it should use
   ; a set of archival flats kept in a special directory (in fact it's quite possible that this would be better in general.)

    common lrisauto, lpar

    if strlen(chip) gt 1 then fchip = strmid(chip,0,1) else fchip = chip

    prefchar = strmid(camera,0,1) ; camchar

    ;if prefchar eq 'r' and lpar.lrisversion eq 1 then ynormkey = 'AMP1' else ynormkey = 'AMPR1'
    if prefchar eq 'b' then ynormkey = 'AMPR1'
    if prefchar eq 'r' then begin
      if lpar.lrisversion eq 1 or lpar.lrisversion eq 4 then ynormkey = 'AMP1'
      if lpar.lrisversion eq 2 or lpar.lrisversion eq 3 then ynormkey = ['AMPR1','AMPR2'] ; 2nd is backup
    endif

    ;files = file_search(lpar.imworkingdir+'p'+prefchar+lpar.wildchar+fchip+'.fits') ; camera dealt with here
    ;imh = lrisheaderinfo(files)
    files = lrisselectfiles('i','p'+prefchar,'','.fits', count=ct, imh=imh)
    if ct eq 0 then return

    w = where(imh.slit eq 'direct', ct)
    if ct eq 0 then return
    files = files[w]
    imh   = imh[w]
    isdomeflat = intarr(ct)
    isskyflat  = intarr(ct)
    isscience  = intarr(ct) 

    for f = 0, n_elements(files)-1 do begin
       if files[f] eq '' then continue
       h = headfits(files[f], /silent)

       if imh[f].trapdoor eq 'open' then begin    ;abs(domerelaz-90) lt 1 or abs(domerelaz+90) lt 1))
         if (abs(imh[f].elevation-45) lt 0.5 and abs(abs(imh[f].domerelaz)-90) lt 3) or strpos(imh[f].target,'cass') ge 0 then begin
             isdomeflat[f] = 1  ;if counts[f] gt 7000 and counts[f] lt 40000 then 
         endif else begin
             if ((imh[f].skycounts gt 9000 and camera eq 'blue') or $
                 (imh[f].skycounts gt 3500 and imh[f].filter eq 'U') or $  
                 (imh[f].skycounts gt 3500 and clip(imh[f].filter,2) eq 'NB') or $  ; narrowband
                 (imh[f].skycounts gt 6000 and imh[f].filter eq 'B') or $
                 (imh[f].skycounts gt 6000 and imh[f].skycounts/imh[f].exp gt 450. and camera eq 'red')    $ 
                ) and $
                  imh[f].skycounts lt 43000 and imh[f].exp le 90 and abs(imh[f].domerelaz) lt 3 $ 
                then isskyflat[f] = 1
             if abs(imh[f].domerelaz) lt 5.0 and ((imh[f].skycounts lt 40000) and (imh[f].skycounts le 15000 or (imh[f].filter eq 'I' and imh[f].skycounts le 30000) or (imh[f].filter eq 'RG850') or (imh[f].exp ge 60))) then isscience[f] = 1
         endelse
       endif

      if keyword_set(lpar.debug) then print, files[f], ' ', clip(imh[f].target,12), string(round(imh[f].exp),format='(I4)'), ' ', clip(imh[f].filter,5), ' ', clip(imh[f].dichroic,4), ' ', clip(imh[f].bin,3), '  ', rclip(fix(imh[f].skycounts),5), ' ', clip(imh[f].skycounts/imh[f].exp,7), ' ', fpr(imh[f].domerelaz,3.2), ' ', fpr(imh[f].elevation,2.2), '  ', isdomeflat[f] ? 'dom':'   ',' ', isskyflat[f]?'twi':'   ',' ',isscience[f]?'sci':'   '

    endfor
    
    if keyword_set(forceflat) then isscience = isscience or isdomeflat or isskyflat

    wsci = where(isscience, ct)
    if ct eq 0 then return
    datestr = strmid(repstr(imh[wsci[0]].date,'-',''),2,6) ; single date string
    scifilterlist  = unique(imh[wsci].filter)
    for i = 0, n_elements(scifilterlist)-1 do begin
      filt = scifilterlist[i]

      dichlist = unique(imh[wsci[where(filt eq imh[wsci].filter)]].dichroic)
      for d = 0, n_elements(dichlist)-1 do begin
        dich = dichlist[d]
        if n_elements(dichlist) gt 1 then dichstr = dichlist[d] else dichstr = ''

        binlist = unique(imh[wsci[where(filt eq imh[wsci].filter and dich eq imh[wsci].dichroic)]].bin)
        for b = 0, n_elements(binlist)-1 do begin
          bin = binlist[b]
          if n_elements(binlist) gt 1 then binstr = 'b'+binlist[b] else binstr = ''

        winlist = unique(imh[wsci[where(filt eq imh[wsci].filter and dich eq imh[wsci].dichroic and bin eq imh[wsci].bin)]].win)
        for wi = 0, n_elements(winlist)-1 do begin
          win = winlist[wi]
          if n_elements(winlist) gt 1 then winstr = 'w'+winlist[wi] else winstr = ''        

          ;outflatname = lpar.imworkingdir+'lris'+strmid(camera,0,1)+'flat'+filt+dichstr+binstr+winstr+fchip+'.fits'

          flatsuccess = 0

          print, filt, ' ', dich, ' ', bin
        
         ; Regular sky flat.
         if flatsuccess eq 0 then begin
            flattyp = 'sky'
            outflatname = lpar.flatdir+'l'+strmid(camera,0,1)+datestr+'_'+flattyp+'flat'+filt+dichstr+binstr+winstr+fchip+'.fits'
            if file_test(outflatname) and lpar.overwrite eq 0 then continue

            skyflats = where(isskyflat and imh.filter eq filt and imh.dichroic eq dich and imh.bin eq bin and wincontain(win, imh.win), ctsky)
            
            if ctsky ge 5 then begin
               logprint, ' '+filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctsky),2), ' sky flats.'
               if file_test(lpar.flatdir) eq 0 then spawn, 'mkdir '+lpar.flatdir
               flatcombine, files[skyflats], outflatname, /removeobjects, type='sky', ynormkey=ynormkey ;  normzone=normzone
               flatsuccess = 1
            endif
         endif

        ; Dome flat.
        if flatsuccess eq 0 then begin
           flattyp = 'dome'
           outflatname = lpar.flatdir+'l'+strmid(camera,0,1)+datestr+'_'+flattyp+'flat'+filt+dichstr+binstr+winstr+fchip+'.fits'
           if file_test(outflatname) and lpar.overwrite eq 0 then continue

           if strmid(camera,0,1) eq 'b' then mincounts=900. else mincounts=1000.  ; formerly 5000 and 7000
           domeflats = where(isdomeflat and imh.filter eq filt and imh.dichroic eq dich and imh.bin eq bin and wincontain(win, imh.win) and $
                             imh.skycounts gt mincounts and imh.skycounts lt 44000., ctdome)
           if ctdome gt 5 then domeflats = domeflats[where(imh[domeflats].skycounts gt ( median(imh[domeflats].skycounts, /even)*0.6) < 9000.)]
           if ctdome ge 3 then begin
              logprint, ' '+filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctdome),2), ' dome flats'
              if file_test(lpar.flatdir) eq 0 then spawn, 'mkdir '+lpar.flatdir
              flatcombine, files[domeflats], outflatname, type='dome', ynormkey=ynormkey ; normzone
              flatsuccess = 1
           endif
        endif

         ; "Good" supersky flat [now avoided - supersky correction should be done in "reflat" portion]
        ctsupersky = 0
        if flatsuccess eq 0 and (lpar.lrisversion eq 1 and camera eq 'red' and (filt eq 'I' or filt eq 'RG850' or filt eq 'GG570')) ne 1 then begin ; don't supersky fringed frames
            flattyp = 'sup'
            outflatname = lpar.flatdir+'l'+strmid(camera,0,1)+datestr+'_'+flattyp+'flat'+filt+dichstr+binstr+winstr+fchip+'.fits'
            if file_test(outflatname) and lpar.overwrite eq 0 then continue

            ; First, see if we have enough frames to do a supersky flat.
            poss = where(imh.filter eq filt and imh.dichroic eq dich and imh.bin eq bin and wincontain(win, imh.win) and isdomeflat eq 0 and imh.skycounts gt 300 and imh.exp gt 10., ctposs)
            if ctposs gt 0 and filt ne 'U' then begin ;otherwise usually means some weird situation like the dichroic info is missing, etc.
               if ctposs gt 1 then $
                  medcounts = median(imh[poss].skycounts) $
               else $
                  medcounts = imh[poss].skycounts
               if camera eq 'red' then superskymincount = 1000 > medcounts/3.
               if camera eq 'blue' then superskymincount = 350 > medcounts/3.
               superskymaxcount = (medcounts * 5) < (30000. > medcounts*1.2) ; 
               superskyminexp = 10
               
               ;;;if filt eq 'g' or filt eq 'V' or filt eq 'R' then $
               superskyflats = where(imh.filter eq filt and imh.dichroic eq dich and imh.bin eq bin and wincontain(win, imh.win) and isdomeflat eq 0 and $
                                     imh.skycounts gt superskymincount and imh.skycounts lt superskymaxcount and imh.exp ge superskyminexp, ctsupersky)        ;    else  $
               ;superskyflats = where(imh.filter eq filt and                   imh.bin eq bin and wincontain(win, imh.win) and isdomeflat eq 0 and $
               ;                      imh.skycounts gt superskymincount and imh.skycounts lt superskymaxcount and imh.exp ge superskyminexp, ctsupersky)
               ;print, superskymincount, ' - ', superskymaxcount
               ;for i = 0, ctsupersky-1 do print, '  ', clip(removepath(files[superskyflats[i]]),20), clip(filters[superskyflats[i]],3), clip(imh[superskyflats[i]].dichroic,5), fix(counts[superskyflats[i]]), ' ', targets[superskyflats[i]]
               

 
               ; Exclude extended sources or anything else labeled as a bad flat

               extendedsources = [''] 
               if file_test('extendedsources.txt') then readcol, 'extendedsources.txt', extendedsources, format='a', comment='#', /silent
               badflattargets = [''] 
               if file_test('badflattargets.txt') then  readcol, 'badflattargets.txt',  badflattargets,  format='a', comment='#', /silent
               badflattargets = unique([badflattargets, extendedsources])
               notblank = where(badflattargets ne '', ct)
               if ct gt 0 then badflattargets = badflattargets[notblank]


               for bad = 0, n_elements(badflattargets)-1 do begin
                  if badflattargets[bad] eq '' then continue
                  w = where(imh[superskyflats].target ne badflattargets[bad], ctsupersky)
                  superskyflats = superskyflats[w]
               endfor
  
               if ctsupersky gt 1 then superskyfields = unique(imh[superskyflats].target) else superskyfields = ['']
               nsuperskyfields = n_elements(superskyfields)
               ; this criterion is a bit loose and depends on the number of very bright stars in the image and how big the dither steps were.
               ; eventually, should make flatcombine able to find bright (ultra-saturated) objects and add halo masks, and 
               ; treat big dithers like new fields.
               ;print, ctsupersky, nsuperskyfields
               ;if (ctsupersky ge  9 and nsuperskyfields ge 5) or (ctsupersky ge 17 and nsuperskyfields ge 4) or $
               ;   (ctsupersky gt 17 and nsuperskyfields gt 3) or  ctsupersky ge 21 then begin
               if (ctsupersky ge 21 and nsuperskyfields ge 5) or (ctsupersky ge 25 and nsuperskyfields ge 4) or $
                 (ctsupersky gt 29 and nsuperskyfields gt 3) or ctsupersky ge 33 then begin
                  logprint, ' '+filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctsupersky),2), ' supersky flats (', clip(nsuperskyfields), ' unique fields)'
                  if file_test(lpar.flatdir) eq 0 then spawn, 'mkdir '+lpar.flatdir
                  flatcombine, files[superskyflats], outflatname, /removeobjects, type='supersky', ynormkey=ynormkey
                  flatsuccess = 1
               endif
            endif
         endif

        ; marginal sky flat
        if flatsuccess eq 0 then begin
            flattyp = 'sky'
            outflatname = lpar.flatdir+'l'+strmid(camera,0,1)+datestr+'_'+flattyp+'flat'+filt+dichstr+binstr+winstr+fchip+'.fits'
            if file_test(outflatname) and lpar.overwrite eq 0 then continue
            
            logprint, ' WARNING - insufficient images for a good flat.  ('+clip(ctdome)+' dome, '+clip(ctsky)+' sky, '+clip(ctsupersky)+' science).'
            if ctsky ge 3 then begin
               logprint, ' '+filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctsky),2), ' sky flats (WARNING - marginal!)'
               flatcombine, files[skyflats], outflatname, /removeobjects, type='sky', ynormkey=ynormkey
               flatsuccess = 1
            endif
        endif

        ; marginal supersky flat
        if flatsuccess eq 0 then begin
            flattyp = 'sup'
            outflatname = lpar.flatdir+'l'+strmid(camera,0,1)+datestr+'_'+flattyp+'flat'+filt+dichstr+binstr+winstr+fchip+'.fits'
            if ctsupersky ge 5 then begin
               logprint, ' '+filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctsupersky),2), ' supersky flats  (WARNING - EXTREMELY marginal)'
               logprint, ' Flat field for this filter/dichroic combination is likely to be extremely poor.'
               if file_test(lpar.flatdir) eq 0 then spawn, 'mkdir '+lpar.flatdir
               flatcombine, files[superskyflats], outflatname, /removeobjects, type='supersky', ynormkey=ynormkey
               flatsuccess = 1
            endif
        endif

        if flatsuccess eq 0 then begin
           logprint, ' Unable to produce a flat field for this setting: ', filt, '-band / ', 'D'+dich,' '+binstr+' '+winstr
           noproc = where(isscience and imh.filter eq filt and imh.dichroic eq dich and imh.binning eq bin and wincontain(win, wins), ctnoproc)
           if ctnoproc gt 0 then begin
              logprint, ' Will not be able to further process '+clip(ctnoproc)+' files without a flat from another source.'
              if ctnoproc lt 6 then $
              for j = 0, ctnoproc-1 do logprint, '   ', files[noproc[j]], imh[noproc[j]].exp, '  ', imh[noproc[j]].target
                   ; this seems to actually miss some sometimes, as shown 2013-05-10
           endif
        endif

       endfor
       endfor
     endfor  
  endfor


end






; -------------

pro autolrismakespecflat, chip=chip, camera=camera, forceflat=forceflat, flatrequire=flatrequire, rightchipdead=rightchipdead

   common lrisauto, lpar

    ; Construct spectroscopic flatfields.  Always uses dome flats unless "desparate", doesn't recognize 
    ; the new lamps yet or any sky flats.  Does a y-axis medianing to remove the wavelength response - this
    ; does leave residuals of the telluric lines (from air in the dome) and emission lines in the lamp.

   fchip = strmid(chip,0,1)      ; chipchar
   prefchar = strmid(camera,0,1) ; camchar

   files = lrisselectfiles('s2','p'+prefchar,'','.fits', count=ct, imh=imh)
   if ct eq 0 then return

   w = where(imh.slit ne 'direct', ct)
   if ct eq 0 then return
   files = files[w]
   imh   = imh[w]
   isdomeflat = intarr(ct)
   ishalogen  = intarr(ct)
   isskyflat  = intarr(ct)
   isdeuflat  = intarr(ct)
   isscience  = intarr(ct) 

   ; Distinguish calibration files from science files.

   for f = 0, n_elements(files)-1 do begin
       if files[f] eq '' then continue

       if imh[f].trapdoor eq 'open' then begin 
         if (abs(imh[f].elevation-45) lt 0.5 and $
             (abs(imh[f].domerelaz) gt 3) or strpos(imh[f].target,'cass') ge 0 or $
               strpos(imh[f].target,'DOME_FLAT') ge 0) then begin
             isdomeflat[f] = 1
         endif else begin
             if abs(imh[f].domerelaz) lt 5.0 and imh[f].skycounts ge 1000 and $
                imh[f].skycounts lt 43000 and imh[f].exp lt 60 then isskyflat[f] = 1 ; need a way to exclude objects
             if abs(imh[f].domerelaz) lt 5.0 and (imh[f].skycounts lt 1000) then isscience[f] = 1
         endelse
       endif else begin
         if imh[f].trapdoor eq 'closed' then begin
              if imh[f].hal and (imh[f].neon+imh[f].ar+imh[f].hg+imh[f].cd+imh[f].zn+imh[f].kr+imh[f].xe+imh[f].de eq 0)  then ishalogen[f] = 1 
              if imh[f].de  and (imh[f].neon+imh[f].ar+imh[f].hg+imh[f].cd+imh[f].zn+imh[f].kr+imh[f].xe+imh[f].hal eq 0) then isdeuflat[f] = 1
         endif
       endelse
       ;if isdomeflat[f] then $
       ;print, files[f], ' ', clip(imh[f].target,12), string(imh[f].exp,format='(I4)'), ' ', clip(imh[f].grname,9), ' ', imh[f].skycounts, imh[f].domerelaz, ' ', clip(imh[f].trapdoor,6), ' ', imh[f].lamps, ' ', clip(isdomeflat[f]), ' ', clip(ishalogen[f])
       ;print, files[f], ' ', clip(imh[f].target,12), string(imh[f].exp,format='(I4)'), ' ', clip(imh[f].grname,9), ' ', imh[f].skycounts, imh[f].domerelaz, ' ', clip(imh[f].trapdoor,6), ' ', imh[f].lamps, ' ', clip(isdomeflat[f]), ' ', clip(ishalogen[f])  
   endfor

   if keyword_set(forceflat) eq 0 then begin
     ; By default, flat-fields are only constructed when science files actually need them.
     wsci = where(isscience, ct)
     if ct eq 0 then begin
        if total(isdeuflat)+total(isdomeflat)+total(ishalogen)+total(isskyflat) gt 0 then begin
          logprint, '  No science files have been taken.'
          logprint, '  Set /forceflats to make flat fields regardless, for all available setups.'
        endif
        return
     endif
   endif else begin
     ; If forceflat is set, then flat-field are constructed for any configuration with a flat taken, too.
     wsci = where(isscience or isdomeflat or ishalogen or isskyflat or isdeuflat, ct)
     if ct eq 0 then return
   endelse

   config  = imh.grname + 'D'+imh.dichroic + 'b'+imh.bin
   configlist = unique(config[wsci]) ; every mode we need flats for

   subconfig = config ;+ 's'+imh.slit
   if prefchar eq 'r' then subconfig = subconfig + 'w'+clip(fix(imh.wavelength+0.44))

   outflatfiles = file_search(lpar.flatdir+'l*flat.fits', count=ctoutflat)
   if ctoutflat gt 0 then begin
      outflatimh = lrisheaderinfo(outflatfiles)
      outflatconfig  = outflatimh.grname + 'D'+outflatimh.dichroic + 'b'+outflatimh.bin
   endif

   for i = 0, n_elements(configlist)-1 do begin
     w = where(config eq configlist[i], ct) ; ct will always be positive

     grat = clip(imh[w[0]].grname)
     dich = clip(imh[w[0]].dichroic)
     bin = clip(imh[w[0]].bin)

     flatneeded = 1
     if lpar.overwrite eq 0 then begin ; new addition in 2020-06, not carefully checked
       if ctoutflat gt 0 then begin
          wout = where(outflatconfig eq configlist[i], ct)
          if ct gt 0 then flatneeded = 0  ; We already have a useable (if not necessarily ideal) flat, 
                                          ; so hide printed output.
       endif
     endif

     flatsuccess = 0
     flattypes = ['dom','hal','deu','sky']
     flatnames = ['dome','halogen','deuterium','sky']
     for t = 0, 3 do begin
        if flattypes[t] eq 'deu' and prefchar eq 'r' then continue ; never make deuterium flats for red camera

        if flattypes[t] eq 'dom' then isflat = isdomeflat
        if flattypes[t] eq 'hal' then isflat = ishalogen
        if flattypes[t] eq 'deu' then isflat = isdeuflat
        if flattypes[t] eq 'sky' then isflat = isskyflat

        flats = where(isflat and config eq configlist[i] and $
                        imh.skycounts gt 1000. and imh.skycounts lt 39000., $
                        ctflat)
        weakflats = where(isflat and config eq configlist[i] and $
                        imh.skycounts le 1000., $
                        ctweakflat) ; flats with too little signal
        satflats = where(isflat and config eq configlist[i] and $
                        imh.skycounts ge 39000., $
                        ctsatflat) ; flats with too much signal

        require = 3
        if n_elements(flatrequire) gt 0 then require = flatrequire ; override
        if ctflat lt require and flatneeded then begin
           flats = where(isflat and config eq configlist[i] and $
                           imh.skycounts gt 500. and imh.skycounts lt 39000., $
                           ctflat)
           require = 5 ; += 2?
        endif
        if ctflat lt require and flatneeded then begin
           flats = where(isflat and config eq configlist[i] and $
                           imh.skycounts gt 250. and imh.skycounts lt 39000., $
                           ctflat)
           require = 9 ; += 2?
        endif

        if ctflat lt require and flatneeded then begin
           addstr = ''
           if ctweakflat gt 0 then addstr += ' ['+clip(ctweakflat)+' with insufficient signal]'
           if ctsatflat gt 0  then addstr += ' ['+clip(ctsatflat)+' close to saturation]'
           if ctflat ge 3 and require ge 5 then addstr += '; '+clip(require)+' needed'
           if flatsuccess eq 0 and ctflat eq 0 then logprint, ' '+grat, ', D'+dich, ' ', bin, ': no '+flatnames[t]+' flats'+addstr
           if ctflat gt 0 then logprint, ' '+grat, ', D'+dich, ' ', bin, ': '+clip(ctflat)+' '+flatnames[t]+' flats'+addstr
           continue
           ; not enough flats even when combining slits, etc.
        endif

        subconfiglist = unique(subconfig[flats], count=ctsubconfig)
        for j = 0, n_elements(subconfiglist) do begin
           if j lt n_elements(subconfiglist) then begin
              if ctsubconfig[j] lt 3 then continue
              sflats = flats[where(subconfig[flats] eq subconfiglist[j])]
           endif else begin
              if flatsuccess then continue
              ; "extra loop" if there were not enough individual flats:
              ; ignore subconfiguration (and mix slits, etc.)
              sflats = flats

              ; if mixing slits, don't use the 0.7" slit if possible.
              dslits = imh[sflats].slit
              wnot7 = where(imh[sflats].slit ne 'long_0.7', ctnot7)
              if ctnot7 ge 3 then sflats = sflats[wnot7]
           endelse

           ; make sure the windowing is consistent by only combining the most common windowing
           dwindows = imh[sflats].win
           winlist = unique(dwindows, count=nwin)
           mostcommonwin = winlist[(where(nwin eq max(nwin))) [0]] ; what if this is different from science windowing...?
           sflats = sflats[where(imh[sflats].win eq mostcommonwin, cts)]

           wavinfo = ''
           if j lt n_elements(subconfiglist) then begin
              ;subinfo = '('+strmid(subconfiglist[j],1+strpos(subconfiglist[j],'s'))+' '+mostcommonwin+')'
              if strpos(subconfiglist[j],'w') gt 0 then $
                 wavinfo = clip(strmid(subconfiglist[j],strpos(subconfiglist[j],'w')))
              if wavinfo ne '' then wavinfo += ', '
           endif
           subinfo = '('+wavinfo+mostcommonwin+')'
           subinfo = repstr(subinfo, 'w','w=')
           subinfo = repstr(subinfo, 'long_0.7','0.7"')
           subinfo = repstr(subinfo, 'long_1.0','1.0"')
           subinfo = repstr(subinfo, 'long_1.5','1.5"')
           subinfo = repstr(subinfo, '0,0,4096,4096','full')
           subinfo = repstr(subinfo, '1500,0,1200,4096','slit')
           subinfo = repstr(subinfo, '400,625,3400,2500','img') ; should rarely encounter in this mode

           if cts ge 1 and flatneeded then $
              logprint, ' '+grat, ', D'+dich, ' ', bin, ': ', strtrim(string(cts),2), ' '+flatnames[t]+' flats '+subinfo

           if cts ge 3<require then begin

              ; Flat is worth making - check if it already exists

              datestr = repstr(strmid(imh[sflats[0]].date,2,8),'-','')
              if lpar.koa eq 0 then framestr = string(imh[sflats[0]].framenum,'(I04)') $
                               else framestr = imh[sflats[0]].koaframe
              ctstr = clip(cts)
              outflatname = lpar.flatdir+'l'+strmid(camera,0,1)+datestr+'_'+framestr+'s'+ctstr+'_'+flattypes[t]+'flat.fits'
              if file_test(outflatname) eq 0 or lpar.overwrite then begin

                 ; Actually make the flat

                 if flatneeded eq 0 then $  ; wasn't printed before so print it now
                 logprint, ' '+grat, ', D'+dich, ' ', bin, ': ', strtrim(string(cts),2), ' '+flatnames[t]+' flats '+subinfo

                 ; if n_elements(subconfiglist) gt 0...??
                 ;if j eq n_elements(subconfiglist) then logprint, '  Note: flat combines multiple wavelength configurations'
                                                                                       ;slit and/or 
                 xdisc = -1
                 if lpar.lrisversion eq 1 and camera eq 'red' then begin  ; discontinuity on old LRIS red
                    h = headfits(files[sflats[0]])
                    xdisc = (fix(strsplit(sxpar(h,'AMP2'),',',/extract))) [0]
                 endif
 
                 indresp  = flattypes[t] ne 'dom'
                 slitnorm = flattypes[t] eq 'hal' or flattypes[t] eq 'deu'
                 
                 h = headfits(files[sflats[0]])
                    ; nx = fix(sxpar(h,'NAXIS1'))
                 ampr = sxpar(h,'AMPR1',count=ctk)
                 if ctk eq 0 then ampr = sxpar(h,'AMPR2',count=ctk)
                 gapy = (fix(strsplit(ampr,',',/extract))) [0] - sxpar(h,'CROPY0')

                 ybin = sxpar(h,'YBIN')
                 ; Generally, use the right amplifier for determining the normalization to 1.
                 yresp = gapy+20+indgen(800/ybin)
                 yresp = yresp[where(yresp le imh[sflats[0]].ny-5)]
                 if keyword_set(rightchipdead) then begin
                    ; Use the left chip for normalizing instead if right chip is dead.
                    ampl = sxpar(h,'AMPL2',count=ctk)
                    if ctk eq 0 then ampl = sxpar(h,'AMPL1',count=ctk)
                    ampln = fix(strsplit(ampl,',',/extract))
                    yresp = [(ampln[0]+100)>10, (ampln[1]-100)]
                    ;print, sxpar(h,'AMPL1')
                    ;print, sxpar(h,'AMPL2')
                    ;print, ampln
                    ;print, yresp
                 endif
                 if flattypes[t] eq 'deu' or flattypes[t] eq 'sky' then begin
                    smoothscale = 3 
                 endif else begin
                    ;if n_elements(yresp) gt 0 then delvarx, yresp
                    if n_elements(smoothscale) gt 0 then delvarx, smoothscale
                 endelse

                 slitstr = ''
                 slitsused = unique(imh[sflats].slit) 
                 for s = 0, n_elements(slitsused)-1 do $
                    slitstr += (s > 0 ? '+':'') + repstr(slitsused[s],'long_','')+'"'
                 wavstr = ''
                 if prefchar eq 'r' then begin
                    minwav = min(imh[sflats].wavelength)
                    maxwav = max(imh[sflats].wavelength)
                    if maxwav-minwav gt 5 then wavstr = ' W'+clip(fix(round(minwav)))+'-'+clip(fix(round(maxwav))) $
                                          else wavstr = ' W'+clip(fix(round((minwav+minwav)/2)))
                 endif
                 binstr = imh[sflats[0]].bin
                 outobject = flatnames[t]+' flat: '+grat+' D'+dich+wavstr+' S'+slitstr+' '+binstr

                 if file_test(lpar.flatdir) eq 0 then spawn, 'mkdir '+lpar.flatdir
                 if flattypes[t] eq 'dom' then begin
                   flatcombine, files[sflats], outflatname, type=flatnames[t], mincounts=10, xdisc=xdisc, $
                     /response, indresp=indresp, slitnorm=slitnorm, yresp=yresp, smoothscale=smoothscale, $
                     outobject=outobject
                 endif
                 if flattypes[t] eq 'hal' then begin
                   flatcombine3, files[sflats], outflatname, type=flatnames[t], mincounts=10, $
                     yresp=yresp, outobject=outobject
                 endif
                 if flattypes[t] eq 'sky' or flattypes[t] eq 'deu' then begin
                   lineflatcombine, files[sflats], outflatname, type=flatnames[t], ydisc=gapy[0], $
                     outobject=outobject, slitnorm=slitnorm ;  mincounts=10, 
                 endif
              endif ; produce flat

              if file_test(outflatname) then flatsuccess = 1

            endif ; enough files to produce flat
       
        endfor ; subconfig loop

     endfor ; flat type loop

     if flatneeded eq 1 and flatsuccess eq 0 then begin
        logprint, ' Unable to produce any flat field for this combination: ', grat+' '+'D'+dich+' ('+ bin + ').'
        noproc = where(isscience and imh.grname eq grat and imh.dichroic eq dich and imh.bin eq bin, ctnoproc)
        if ctnoproc gt 0 then begin
           logprint, ' Will not be able to process '+clip(ctnoproc)+' science files without a flat from another source.'
           for j = 0, ((ctnoproc-1)<5) do $
              logprint, '   ', files[noproc[j]], fpr(imh[noproc[j]].exp,5.1)+'s', '  ', imh[noproc[j]].target
           if ctnoproc gt 5 then print, '   + '+clip(ctnoproc-6)+' others'
           ; Note - this message will still appear if rebinnable flats are available.
        endif
     endif

   endfor ; config loop

end





; -------------------------
pro autolrisimflatten, chip=chip, camera=camera, ignoredichroic=ignoredichroic, altflatdir=altflatdir

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   camchar = strmid(camera,0,1)
   fchip = strmid(chip,0,1)

   ; files = file_search(lpar.imworkingdir+'p'+camchar+lpar.wildchar+fchip+'.fits')
   files = lrisselectfiles('i','p'+camchar,fchip,'.fits', count=ct)
   if ct eq 0 then return

   if lpar.overwrite eq 0 then files = unmatched(files, 'f')
   if n_elements(files) eq 1 and files[0] eq '' then return

   ; un-flattened imaging files
   imh = lrisheaderinfo(files)
   w = where(imh.slit eq 'direct' and strmid(imh.camera,0,1) eq camchar and imh.exptime gt 0.5, ct)
   if ct eq 0 then return      ; return silently if nothing to process
   files = files[w]
   imh   = imh[w]

   ; flat field files
   flatfiles = file_search(lpar.flatdir+'*flat*.fits')
   for a = 0, n_elements(altflatdir)-1 do flatfiles = [flatfiles, file_search(altflatdir[a]+'*flat*.fits')] ; blanks are OK
   flatimh   = lrisheaderinfo(flatfiles)
   w = where(flatimh.slit eq 'direct' and strmid(flatimh.camera,0,1) eq camchar, nflats)
   ws = where(abs(flatimh.elevation-45) ge 0.5 or abs(flatimh.domerelaz) gt 3, nsci)
   if nflats eq 0 and nsci gt 0 then begin   ; return loudly if something to process and no flats made yet
     if camchar eq 'r' then cam = 'red'
     if camchar eq 'b' then cam = 'blue'
     logprint, ' No processed '+cam+' image flats found!  Cannot perform imaging flatfielding.'
     for f = 0, n_elements(files)-1 do flatfail = [flatfail, files[f]]
     return
   endif
   flatfiles = flatfiles[w]
   flatimh   = flatimh[w]
   
;   if keyword_set(ignoredichroic) then begin
;      ; w = where(filter ne 'R' and filter ne 'V') could choose to require dichroic in these filters
;      flatimh.dichroic = ''
;      imh.dichroic = ''
;   endif
  
   for f = 0, n_elements(files)-1 do begin
      outfile = fileappend(files[f], 'f')

      if imh[f].trapdoor eq 'closed' then continue    ; skip darks (and internal flats)

      if abs(imh[f].elevation-45) lt 0.5 and abs(imh[f].domerelaz) gt 3 then continue ; silently skip dome flats
      if strpos(imh[f].target,'cass') ge 0 then continue 
      if strpos(imh[f].target,'dome') ge 0 then continue     ; sometimes the telescope is pointed wrong; we still don't want these...
      if strpos(imh[f].target,'horizon') ge 0 then continue 

      if imh[f].skycounts gt 55000 then begin
         logprint,  '            ', removepath(files[f]), ' is saturated.'
         continue
      endif

      ;;print, imh[f].skycounts, imh[f].target, imh[f].exptime, imh[f].skycounts gt 10000., strpos(strlowcase(imh[f].object), 'sky') ge 0

      ; We may actually want to keep these for use in the reflattening procedure 
      ;  (but perhaps mark them for further processing, or make this a function call.)
      if (imh[f].skycounts gt 30000. and imh[f].exptime lt 30) or $
         (imh[f].skycounts gt 20000. and imh[f].exptime lt 15 and imh[f].exptime gt 1.9) or $
         (imh[f].skycounts/imh[f].exptime gt 2000. and imh[f].filter ne 'RG850' and imh[f].exptime gt 1.9) or  $ 
         (imh[f].skycounts/imh[f].exptime gt 6000. and imh[f].filter eq 'RG850' and imh[f].exptime gt 1.9) or  $ 
         (imh[f].skycounts/imh[f].exptime gt 15000. and imh[f].exptime le 1.9) or $
         (imh[f].skycounts gt 10000. and imh[f].exptime gt 1.9 and imh[f].exptime lt 30 and $
          strpos(strlowcase(imh[f].object), 'flat') ge 0 or strpos(strlowcase(imh[f].object), 'sky') ge 0) or $
         (imh[f].skycounts gt 15000. and strpos(strlowcase(imh[f].target), 'flat') ge 0) or $
         (imh[f].skycounts gt 15000. and strpos(strlowcase(imh[f].target), 'sky') ge 0) then begin
            logprint, '   Skipping '+removepath(files[f])+$
                   ' ('+clip(imh[f].target)+', '+clipzero(imh[f].skycounts)+'ct, '+ clip(round(imh[f].exptime)), 's)'+ $
                   ': sky flat?'
            continue
      endif
      if imh[f].skycounts gt 40000. and imh[f].exptime gt 60 then begin
         if imh[f].satval gt 0 then satstr = '('+clip(imh[f].satval)+')' else satstr=''
         logprint, '  WARNING: ', files[f], ' has sky (', clip(round(imh[f].skycounts)), ') close to saturation '+satstr+'!'
      endif

      flatfileno = where(flatimh.filter eq imh[f].filter and flatimh.chip eq imh[f].chip and flatimh.dichroic eq imh[f].dichroic and  $
                         flatimh.binning eq imh[f].binning and flatimh.win eq imh[f].win, ct)
      if ct eq 0 then $
      flatfileno = where(flatimh.filter eq imh[f].filter and flatimh.chip eq imh[f].chip and flatimh.dichroic eq imh[f].dichroic and $
                         flatimh.binning eq imh[f].binning, ct)

      if ct eq 0 then begin
        flatfileno = where(flatimh.filter eq imh[f].filter and flatimh.chip eq imh[f].chip and $ 
                         flatimh.binning eq imh[f].binning, ct)
        if ct gt 0 then begin
          print, ' WARNING: Flat field for  ', imh[f].filter+'+D'+imh[f].dichroic+$
                 ' uses a different dichroic (D'+flatimh[flatfileno].dichroic+')'
        endif
      endif
      ; allow a different dichroic (dangerous if dichroic hits the filter band)

      if ct eq 0 then begin
         flatfilenoothbin = where(flatimh.filter eq imh[f].filter and flatimh.chip eq imh[f].chip, ctwobin) ; and flatwins eq win
            ; Currently doesn't support flat-fielding with a different binning until I can be sure the case where both windowing and binning
            ; are wrong are dealt with

            if ctwobin gt 0 then binstr = imh[f].bin else binstr = '' ; .bin or .binning ?
            logprint, ' Flat field not found for ', removepath(files[f]), ' (', imh[f].filter+', '+'D'+imh[f].dichroic+ binstr,')' ; , ' / ', imh[f].win
            flatfail = [flatfail, files[f]]
            continue
      endif
      flatfile = flatfiles[flatfileno[0]]
      logprint, ' Flattening ', removepath(files[f]), ' using ', removepath(flatfile)

      flatproc, files[f], flatfile, flatminval=0.6, crop='auto', /quiet  ; add autocropping of low-signal zones

    endfor

end

; -------------

pro autolrisspecflatten, chip=chip, camera=camera, altflatdir=altflatdir, norebin=norebin

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail
   common slitwarn, slit7warn

   camchar = strmid(camera,0,1)
   fchip = strmid(chip,0,1)
   if n_elements(slit7warn) eq 0 then slit7warn = 0

   ;files = file_search(lpar.spworkingdir2d+'p'+camchar+lpar.wildchar+fchip+'.fits')
   files = lrisselectfiles('s2','p'+camchar,fchip,'.fits', count=ct)
   if ct eq 0 then return

   if lpar.overwrite eq 0 then files = unmatched(files, 'f')
   if n_elements(files) eq 1 and files[0] eq '' then return

   ; un-flattened data files
   imh = lrisheaderinfo(files)
   w = where(imh.slit ne 'direct' and strmid(imh.camera,0,1) eq camchar and imh.exptime gt 0.5, ct)
   if ct eq 0 then return      ; return silently if nothing to process
   files = files[w]
   imh   = imh[w]

   ; flat field files
   flatfiles = file_search(lpar.flatdir+'*flat*.fits')
   for a = 0, n_elements(altflatdir)-1 do flatfiles = [flatfiles, file_search(altflatdir[a]+'*flat*.fits')] ; blanks are OK
   flatimh   = lrisheaderinfo(flatfiles)
   w = where(flatimh.slit ne 'direct' and strmid(flatimh.camera,0,1) eq camchar, nflats)
   if nflats eq 0 then begin   ; return loudly if something to process and no flats made yet
     logprint, ' No processed spectroscopy flats found!  Cannot perform spectroscopic flatfielding.'
     for f = 0, n_elements(files)-1 do flatfail = [flatfail, files[f]]
     return
   endif
   flatfiles = flatfiles[w]
   flatimh   = flatimh[w]

       ;  Option to remove unused spec columns that seems to be no longer lused
       ;if total(ycrop eq [0,0]) eq 2.0 then begin  ; automatic cropping
       ;   flat = mrdfits(flats[f], /silent)
       ;   s = size(flat)
       ;   nx = s[1]
       ;   yprofile = total(flat[0:nx/5.,*], 1)/median(flat[0:nx/5.,*]) + $
       ;              total(flat[4*nx/5.:nx-1,*], 1)/median(flat[4*nx/5.:nx-1,*])
       ;              ; sum over x-axis, using the edge areas since we are liberally cropping out all rows with some
       ;              ; signal-less pixels and the spectrum is often tilted
       ;   signalzone = where(yprofile gt 0.86*median(yprofile)) ; hopefully there's no slit profile variation.
       ;   flatycrop[f,*] = [min(signalzone)+1,max(signalzone)-2]
       ;
       ;endif
       ;
       ; ycrop = median(flatycrop, dimension=1) ; assume it's the same for all gratings (and slits)
       ; this is probably not a good assumption!

 
   for f = 0, n_elements(files)-1 do begin
      outfile = fileappend(files[f], 'f')
      ; we do flatten arcs, so trapdoor isn't a prerequisite

      ; skip biases
      if imh[f].exptime lt 0.5 then continue
      ; skip darks
      if imh[f].trapdoor eq 'closed' and imh[f].nlamp eq 0 then continue
      ; skip internal flats
      if imh[f].hal or imh[f].de then continue

      ;print, files[f], elevation, domerelaz, target, counts, exptime, counts/exptime
    
      if imh[f].trapdoor eq 'open' then begin 
         ; silently skip dome flats
         if abs(imh[f].elevation-45) lt 0.5 and abs(imh[f].domerelaz) gt 3 then continue ; might also skip some older LRIS data?
         if strpos(imh[f].target,'dome') ge 0 then continue     ; sometimes the telescope is pointed wrong; we still don't want these...
         if strpos(imh[f].target,'horizon') ge 0 then continue  ; hopefully the words dome or horizon don't appear in science
         if imh[f].sunelev gt 0 then continue      ; daytime: dome lights could be on and telescope mis-pointed,

         ; skip sky flats
         targ = clip(strlowcase(imh[f].target))
         if ((imh[f].sunelev gt -6) or $
             (imh[f].sunelev gt -12 and imh[f].sunelev lt 0 and $
               (strpos(targ,'sky') ge 0 or strpos(targ,'ski') ge 0 or $ ; sky target names are often weird...
                strpos(targ,'flat') ge 0 or strpos(targ,'phlat') ge 0 or $
                strpos(targ,'flut') ge 0 or strpos(targ,'blank') ge 0))) then begin
            logprint, '   Skipping '+removepath(files[f])+$
                   ' ('+clip(imh[f].target)+', '+clipzero(imh[f].skycounts)+'ct, '+ clip(round(imh[f].exptime)), 's)'+ $
                   ': sky flat?'
            continue
         endif  

      endif
 
      ; Attempt to find a flat in an identical configuration (does not consider central wavelength yet)
      flatfileno = where(flatimh.grname eq imh[f].grname and flatimh.filter eq imh[f].filter and $
                         flatimh.dichroic eq imh[f].dichroic and $
                         flatimh.binning eq imh[f].binning and flatimh.detector eq imh[f].detector, ct)
                        ;and flatimh.wins eq imh[f].win
      if ct eq 0 then begin
         ; If unsuccessful, allow flats in a different filter
         flatfileno = where(flatimh.grname eq imh[f].grname and $
                            flatimh.dichroic eq imh[f].dichroic and $
                            flatimh.binning eq imh[f].binning and flatimh.detector eq imh[f].detector, ct)
         if ct ge 1 then $
         logprint, '  WARNING - no flat with filter='+clip(imh[f].filter)+'; using a flat with different filter'
      endif

      flatfile = ''
      if ct ge 1 then begin 
         if ct ge 2 then begin ; more than one usable flat exists
           not7 = where(imh[flatfileno].slit ne 'long_0.7', ct)
           if ct gt 0 then flatfileno = flatfileno[not7] 
           ct = n_elements(flatfileno)
         endif
         if ct ge 2 and camchar eq 'r' then begin
            samewav = where(abs(imh[flatfileno].wavelength-imh[f].wavelength) eq $
                            min(abs(imh[flatfileno].wavelength-imh[f].wavelength)), ct)
            if ct gt 0 then flatfileno = flatfileno[samewav]
            ct = n_elements(flatfileno)
         endif
         if ct ge 2 then begin
            nothal = where(imh[flatfileno].hal eq 0, ct)
            if ct gt 0 then flatfileno = flatfileno[nothal]
            ct = n_elements(flatfileno)
         endif
         if ct ge 2 then begin
           mostcomb = where(imh[flatfileno].ncomb eq max(imh[flatfileno].ncomb), ct)
           if ct gt 0 then flatfileno = flatfileno[mostcomb] ; ct cannot be zero 
           ct = n_elements(flatfileno)
         endif
         if ct ge 2 then begin
           sameslit = where(imh[flatfileno].slit eq imh[f].slit, ct)
           if ct gt 0 then flatfileno = flatfileno[sameslit] 
           ct = n_elements(flatfileno)
         endif
         flatfile = flatfiles[flatfileno[0]]
      endif else begin
         ; If above was still unsuccessful, see if it's possible to rebin a flat

         wflatotherbin = where(flatimh.grname eq imh[f].grname and flatimh.filter eq imh[f].filter and $
                         flatimh.dichroic eq imh[f].dichroic and $
                         flatimh.detector eq imh[f].detector, ctother)
         if ctother eq 0 then begin
           flatfileno = where(flatimh.grname eq imh[f].grname and $
                            flatimh.dichroic eq imh[f].dichroic and $
                            flatimh.detector eq imh[f].detector, ctother)
           if ctother ge 1 then $
             logprint, '  WARNING - no flat with filter='+clip(imh[f].filter)+'; using a flat with different filter'
         endif

         if ctother gt 0 and keyword_set(norebin) eq 0 then begin
           logprint, '  WARNING - no flat with same binning as '+clip(removepath(files[f]))+'.  Trying to rebin.'
           ; will only rebin 1x1 to others.  2x1 won't be rebinned to 2x2.  Could change this?  Rare situation.
           fflat = wflatotherbin[0]
;           if imh[f].binning eq '1,1' then begin ; note this is y,x
;             ; rebin data to match flat (should it really be doing this??  Probably not, should change for next release)
;             newbinning = flatimh[fflat].bin
;             logprint, '  Rebinning ', removepath(files[f]), ' from ', imh[f].bin, ' to ', newbinning, ' to match flat-field '+removepath(flatimh[fflat].filename)
;             outbinfile = repstr(files[f],'.fits','.rebin'+newbinning+'.fits')
;             lrisrebin, files[f], newbinning, out=outbinfile, /fluxsum
;             files[f] = outbinfile
;             imh[f].binning = newbinning
;             flatfile = flatimh[fflat].filename
;           endif
;           if flatimh[fflat].binning eq '1,1' then begin ; note this 
             ; rebin flat to match data
             newbinning = imh[f].bin
             logprint, '  Rebinning ', removepath(flatimh[fflat].filename), ' to ', newbinning, ' to match '+removepath(files[f])
             outflatbinfile = repstr(flatimh[fflat].filename,'.fits','.rebin'+newbinning+'.fits')
             lrisrebin, flatimh[fflat].filename, newbinning, out=outflatbinfile ; default is to average...
             flatfiles = file_search(lpar.flatdir+'*flat*.fits')
             flatimh   = lrisheaderinfo(flatfiles)
             w = where(flatimh.slit ne 'direct' and strmid(flatimh.camera,0,1) eq camchar, nflats)
             flatfiles = flatfiles[w]
             flatimh   = flatimh[w]
             flatfile = outflatbinfile
;           endif
         endif
      endelse
      if flatfile eq '' then begin
         logprint, ' Flat field not found for ', removepath(files[f]), $
                   ' (', strtrim(imh[f].grname,2), ', D', strtrim(imh[f].dichroic,2), ', ', imh[f].bin, ')'
         flatfail = [flatfail, files[f]]
      endif else begin
        if imh[f].slit eq 'long_0.7' and slit7warn eq 0 then begin           
           logprint, ' WARNING - 0.7" longslit data was used in this run!'
           logprint, '    Tiny irregularities in the slit make it flat-field poorly.'
           ;logprint, '    It may be possible to manually improve the flat-fielding using reflatslit.'
           slit7warn = 1
        endif

        logprint, ' Flattening ', removepath(files[f]), ' using ', removepath(flatfile)
        flatproc, files[f], flatfile, flatminval=0.1, crop='auto', /quiet, outname=outfile 
            ; note that the autocropping basically is overridden at the split phase.
                   ;crop=[0, ycrop[0], 0, ycrop[1]],  
      endelse
    endfor

end


; ----------------------
pro autolrisreflatten, chip=chip, camera=camera, filters=filters

; Create a special super-sky flat specific to these exact settings.

   common lrisauto, lpar

    if strlen(chip) gt 1 then fchip = strmid(chip,0,1) else fchip = chip
    camchar = strmid(camera,0,1)

    ;files = file_search(lpar.imworkingdir+'fp'+camchar+lpar.wildchar+fchip+'.fits')
    ;imh = lrisheaderinfo(files)
    files = lrisselectfiles('i','fp'+camchar,fchip,'.fits', count=ct)
    if ct eq 0 then return

    w = where(imh.slit eq 'direct', ct)
    if ct eq 0 then return
    files = files[w]
    imh = imh[w]
    reflatfiles = lpar.imworkingdir + 'i' + removepath(files)

               extendedsources = [''] 
               if file_test('extendedsources.txt') then readcol, 'extendedsources.txt', extendedsources, format='a', comment='#', /silent
               badflattargets = [''] 
               if file_test('badflattargets.txt') then  readcol, 'badflattargets.txt',  badflattargets,  format='a', comment='#', /silent
               badflattargets = unique([badflattargets, extendedsources])
               notblank = where(badflattargets ne '', ct)
               if ct gt 0 then badflattargets = badflattargets[notblank]

    ugr = unique(imh.gratf)       
    for g = 0, n_elements(ugr)-1 do begin
      if n_elements(filters) eq 0 then filts = unique(imh[where(imh.gratf eq ugr)].filter) else filts = filters
      for f = 0, n_elements(filts)-1 do begin

          wfg = where(imh.filter eq filts[f] and imh.gratf eq ugr[g], ctfg)
          if ctfg eq 0 then continue
          mednx = median([imh[wfg].nx])
          medny = median([imh[wfg].ny])  ; this is a lazy check, since it doesn't check the x0/y0 parameters.
          wfg = where(imh.filter eq filts[f] and imh.gratf eq ugr[g] and imh.nx eq mednx and imh.ny eq medny, ctfg)

          if ctfg eq 0 then continue

          unreflattened = 0
          for i = 0, ctfg-1 do begin   ; need to implement lpar.fileseq, etc., here
             if file_test(reflatfiles[wfg[i]]) eq 0 then unreflattened = 1
          endfor
          if unreflattened eq 0 and lpar.overwrite eq 0 then continue
          minfiles = 9
          if clip(filts[f]) eq 'G' then minfiles = 15
          if clip(filts[f]) eq 'V' then minfiles = 11  ; how many files before the last grating change needed?
          if clip(filts[f]) eq 'R' then minfiles = 10
          if clip(filts[f]) eq 'I' then minfiles = 8
          if clip(filts[f]) eq 'RG850' then minfiles = 6
          if ctfg lt minfiles then begin
             logprint, ' Not enough files in filter ', filts[f],' to reflat block of imaging starting at ', removepath(files[wfg[0]]), $
                    '  (',clip(ctfg), ' files)'
             continue
          endif
          ufields = unique(imh[wfg].target)  ; need to do a badflattarget/extendedsource check
          nfields = n_elements(ufields)
          goodufields = ['']
          wfggood = [-1]
          for u = 0, n_Elements(ufields)-1 do begin
             if total(ufields[u] eq badflattargets) eq 0 then begin
                goodufields = [goodufields, ufields[u]]
                wfggood = [wfggood, wfg[where(imh[wfg].target eq ufields[u])] ]   
             endif
          endfor

          ngoodfields = n_elements(goodufields)-1
          ngoodimages = n_elements(wfggood)-1
          if ngoodfields gt 0 then goodufields = goodufields[1:n_elements(goodufields)-1]
          skytblock = median(imh[wfggood].exptime, /even) lt 120 and median(imh[wfggood].ctrate, /even) gt 300
             ; seems to be a flag indicating sky flats.  
             ; formerly 'and skytblock eq 0' was required to trigger skipping -
             ;  we sometimes want to reflat twilight images?  deactivated for now, but consider re-adding...
          if (ngoodfields eq 0) or $
             (ngoodfields le 2 and ngoodimages le 13) or $  
             (ngoodfields le 3 and ngoodimages le 9) then begin
             logprint, ' Not enough fields in filter ', filts[f],' to reflat block of imaging starting at ', removepath(files[wfg[0]]), $
                    '  (',clip(ngoodimages), ' files, ', clip(nfields),' fields)'
             ;ugr[g]
             continue
          endif

          wfggood = wfggood[1:n_elements(wfggood)-1]
          wfggood = wfggood[sort(wfggood)]

          logprint, ' Creating special supersky reflats for ', clip(ctfg), ' ', filts[f]+'-band files between ', removepath(files[wfg[0]]), ' and ', removepath(files[wfg[ctfg-1]]), ' (', clip(nfields),' fields)'

          selfsupersky, files[wfg], outfilelist=reflatfiles[wfg]   ; include an "also" option to flat-field extra files
          
          ;lpar.fileseqstr = lrisfilenum(ugr[g], /str)
          ;outflatname = lpar.imworkingdir+'lrisrefl_'+filts[f]+'_'+lpar.fileseqstr+'.fits'
          ;if file_test(outflatname) eq 0 then begin
          ;   print, 'Creating special supersky reflat for ', clip(ctfg), ' files between ', removepath(files[wfg[0]]), ' and ', removepath(files[wfg[ctfg-1]])
          ;   flatcombine, files[wfg], outflatname, /removeobjects, type='supersky', ynormkey='AMPR1'
          ;endif

          ;for i = 0, ctfg-1 do begin    ; i for "improved" flatfield.
          ;   file = files[wfg[i]]
          ;   outfile = lpar.imworkingdir + 'i' + removepath(file)
          ;   if file_test(outfile) then continue
          ;   print, 'Reflattening ', file, ' -> ', outfile
          ;   if n_elements(flatim) le 1 then flatim = readfits(outflatname)
 
          ;   im = readfits(file, h)
          ;   im /= flatim
          ;   sxaddpar, h, 'REFLAT', outflatname, after='FLATTYPE'

          ;   writefits, outfile, im, h
          ;endfor
          ;flatim = 0

       endfor
    endfor


end



; -------------------------
pro autolrismakefringe

   ; currently long deprecated.  Only needed for old LRIS anyway.

  common lrisauto, lpar

  if lpar.lrisversion eq 1 and makefringe ne 0 and camera eq 'red' then begin
    ffiles = file_search('f'+prefchar+lpar.wildchar+'?'+'.fits')
    fileinfo = replicate({filename:'', filter:'', counts:0., exptime:0., airmass:0., object:''}, n_elements(ffiles))
    fileinfo.filename = ffiles
    for f = 0, n_elements(fileinfo)-1 do begin
      if fileinfo[f].filename eq '' then continue
      h = headfits(fileinfo[f].filename, /silent)
      fileinfo[f].filter = strtrim(sxpar(h, filtkey),2)
      fileinfo[f].counts =  sxpar(h, 'COUNTS')
      fileinfo[f].exptime = sxpar(h, 'ELAPTIME')
      fileinfo[f].object  = sxpar(h, 'OBJECT')
      fileinfo[f].airmass = sxpar(h, 'AIRMASS')
      ;print, fileinfo[f].filename, ' ', fileinfo[f].filter, fileinfo[f].counts, fileinfo[f].exptime
    endfor
  
    fringablefilts = ['I','RG850']
    fringeexcludeobj = ['']  ;['GRB070412', '050412', 'GRB051109B', 'GRB070521', 'GRB100205A']
    for ff = 0, n_elements(fringablefilts)-1 do begin
       ffilt = fringablefilts[ff]
       fringefile = ffilt + 'fringe.fits'
       if file_test(fringefile) then continue
       thisfilt = where(fileinfo.filter eq ffilt, ct)
       if ct lt 3 then continue
       ffileinfo = fileinfo[thisfilt]
       thisdark = where(ffileinfo.exptime gt 60 and ffileinfo.counts lt 40000, ct)
       if ct lt 3 then continue
       ffileinfo = ffileinfo[thisdark]
       ctratethreshold = median((ffileinfo.counts/ffileinfo.exptime)/ffileinfo.airmass) * 1.8
       thisdark = where((ffileinfo.counts/ffileinfo.exptime)/ffileinfo.airmass le ctratethreshold)
       if ct lt 3 then continue
       ffileinfo = ffileinfo[thisdark]
       for nf = 0, n_elements(fringeexcludeobj)-1 do begin
          notforbiddenobj = where(ffileinfo.object ne fringeexcludeobj[nf], ct)
          if ct ge 1 then ffileinfo = ffileinfo[notforbiddenobj]
       endfor
       if ct lt 3 then continue

       logprint, ' Making ', ffilt+'-band fringe frame...'

       logprint, ' '+ffileinfo.filename

       mkfringe, ffileinfo.filename, outfile=fringefile, /removeobjects
    endfor
  endif


end




; -------------------------
pro autolrisrmfringe

   ; currently long deprecated.

  common lrisauto, lpar

  if lpar.lrisversion eq 1 and rmfringe ne 0 and camera eq 'red' then begin
    ffiles = file_search('f'+prefchar+lpar.wildchar+'?'+'.fits')
    if lpar.overwrite eq 0 then ffiles = unmatched(ffiles,'i')
    fileinfo = replicate({filename:'', filter:'', counts:0., exptime:0., object:''}, n_elements(ffiles))
    fileinfo.filename = ffiles
    for f = 0, n_elements(fileinfo)-1 do begin
      if fileinfo[f].filename eq '' then continue
      h = headfits(fileinfo[f].filename, /silent)
      fileinfo[f].filter = strtrim(sxpar(h, filtkey),2)
      fileinfo[f].counts =  sxpar(h, 'COUNTS')
      fileinfo[f].exptime = sxpar(h, 'ELAPTIME')
      fileinfo[f].object  = sxpar(h, 'OBJECT')
    endfor

    fringablefilts = ['I','RG850']
    for ff = 0, n_elements(fringablefilts)-1 do begin
       ffilt = fringablefilts[ff]
       fringefile = ffilt + 'fringe.fits'
       if file_test(fringefile) eq 0 then continue
       thisfringe = where(fileinfo.filter eq ffilt and fileinfo.counts/fileinfo.exptime lt 1000., ctfringe)

       if ctfringe ge 1 then begin
         fringablefiles = fileinfo[thisfringe].filename
         if fringablefiles[0] ne '' then $
           fringecorrect, fringablefiles, fringefile=fringefile, prefix='i';, /iterate, itersec=[50,880,200,-200]
         ; alternative method uses weights rather than iteration
         ;fringescale = filesi.counts
         ;continuumimg = where(filesi.counts/(filesi.exptime*filesi.airmass) gt 10, ctcontinuum)
         ;if ctcontinuum gt 0 then fringescale[continuumimg] = (filesi[continuumimg].exptime*filesi[continuumimg].airmass)/0.15
         ;fringecorrect, filesi.name, fringefile='Ifringe.fits', /iterate, itersec=[300,700,400,600]
      endif
    endfor
  endif

end




; -------------------------
pro autolriscrcleanim, chip=chip, camera=camera, setzeal=setzeal

   common lrisauto, lpar

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   ;if chipchar eq 'r' and camchar eq 'r' and (lpar.lrisversion eq 1 or lpar.lrisversion eq 4) then chipchar = ''

   ;files = choosefiles(camchar+lpar.wildchar+chipchar+'.fits',lpar.imworkingdir+'ifp',lpar.imworkingdir+'fp')
   files = lrisselectfiles('i',['ifp','fp']+camchar,chipchar,'.fits', count=ct)

   if ct eq 0 then return

   if lpar.overwrite eq 0 then files = unmatched(files,'z')

   for f = 0, n_elements(files)-1 do begin
      if files[f] eq '' then continue
      h = headfits(files[f])
      counts = sxpar(h,'COUNTS')
      exptime = sxpar(h,'ELAPTIME')
      target = sxpar(h,'TARGNAME')
      ;if counts gt 40000. then continue
      if counts gt 30000. and exptime lt 10. then continue
      if counts gt 20000. and exptime gt 5. and exptime lt 10. then continue
      if target eq '' then continue   ; blank target is certainly not a science object.
      if strpos(target,'flat') ge 0 then continue
      if strpos(target,'sky') ge 0 then continue
      if strpos(target,'twilight') ge 0 then continue

      logprint, ' Cleaning cosmic rays from ', removepath(files[f])
      zeal = 1.1
      if camera eq 'blue' then zeal = 1.0
      if (camera eq 'red' and lpar.lrisversion eq 1) then usamp=1 else usamp=0
      if camera eq 'red' and lpar.lrisversion gt 1 then zeal = 0.8
      if n_elements(setzeal) eq 1 then zeal *= setzeal
      pzap, files[f], zeal=zeal, usamp=usamp, /quiet, /mask   ; /weight,   no longer dealing with weight maps.  but ideally we should.
   endfor

end


; -------------------------
pro autolriscrcleanspec, chip=chip, camera=camera, czeal=czeal

   common lrisauto, lpar

camchar = strmid(camera,0,1)
chipchar = strmid(chip,0,1)

;if chipchar eq 'r' and camchar eq 'r' and (lpar.lrisversion eq 1 or lpar.lrisversion eq 4) then chipchar = ''

;files = choosefiles(lpar.spworkingdir2d+'fp'+camchar+lpar.wildchar+chipchar+'.fits')
files = lrisselectfiles('s2','fp'+camchar,chipchar,'.fits', count=ct)
if ct eq 0 then return

if lpar.overwrite eq 0 then files = unmatched(files,'z')
if n_elements(czeal) eq 0 then czeal = 1.

;if camchar eq 'b' then boxsize = 5
;if camchar eq 'r' then boxsize = 7

for f = 0, n_elements(files)-1 do begin
   if files[f] eq '' then continue

   h = headfits(files[f])
   lampstr = clip(sxpar(h,'LAMPS'))
   lamps = strsplit(lampstr,',',/extract)
   ;;if total(lamps ne '0') gt 0 then continue ;  don't bother with lamps, arcs, etc.

   trapdoor = clip(sxpar(h,'TRAPDOOR'))
   if trapdoor eq 'closed' then continue

   binning = clip(sxpar(h,'BINNING'))
   bins = fix(strsplit(binning,',',/extract))
   ybin = bins[0] gt 1 ; note that these are bool (0=unbinned, 1=binned) 
   xbin = bins[1] gt 1

   ef = 1.0
   exp = sxpar(h,'EXPTIME')
   if exp le 95. then ef = 0.9
   if exp lt 65. then ef = 0.8
   if exp lt 45. then ef = 0.7
   if exp lt 25. then ef = 0.6
   if exp lt 15. then ef = 0.5
   zf = 1.0
   bigcr = 0
   if camchar eq 'r' and lpar.lrisversion eq 3 then begin
      zf = 0.45  ; still gets basically everything
   endif
   if camchar eq 'r' and lpar.lrisversion eq 4 then begin
      zf=1.5
      if (xbin eq 0 or ybin eq 0) then bigcr = 1
      ; performed OK for one example, further tests needed
   endif
   
   ;print, files[f], czeal, ef, zf
   
   ;if (bins[0] ge 2 or bins[1] ge 2) and boxsize ge 5 then boxsize -= 2   ; note the order is y,x for lris-red

   ; should use a faster and less dangerous algorithm for the blue side?

   logprint, ' Cleaning cosmic rays from ', removepath(files[f])
   if camchar eq 'b' then begin
       ; single cleaning
       ; usamp (and bigcr) control the sampling of the CR's
       ; nsigma defaults to 22.5
       pzapspec, files[f], /usamp, ybin=(camchar eq 'r' and ybin gt 1), /mask, sourceblocksize=100/xbin, sfactor=1./sqrt(czeal*ef*zf)
   endif else begin
       ; double cleaning       
       outname = lpar.spworkingdir2d+'z'+removepath(files[f])
       outmaskname = repstr(outname,'.fits','.mask.fits')       
       ; first step is the main cleaning
       pzapspec, files[f], 'temp.fits', 'temp.mask.fits', nsigma=19./(czeal*ef*zf), ybin=ybin, /mask, bigcr=bigcr, sfactor=1./sqrt(czeal*ef*(zf<1))
       ; second step cleans up the fragments left behind (usamp mode)
       pzapspec, 'temp.fits', outname, outmaskname, nsigma=13./(czeal*ef*zf), inmask='temp.mask.fits', /usamp, /mask, sfactor=1./sqrt(czeal*ef*(zf<1))  ;, /debug
       ;spawn, 'rm -f temp.fits temp.mask.fits'
       file_delete, 'temp.fits', 'temp.mask.fits', /quiet
       ; red 2x2 bin files sometimes have the telluric bands get zapped...
   endelse

         ;, /nan  ; the nan is new and temporary

endfor


end


; ------------------
pro autolrissplit, chip=chip, camera=camera, mode=mode, nodistort=nodistort


  common lrisauto, lpar

  camchar = strmid(camera,0,1)
  if n_elements(chip) gt 0 then chipchar = strmid(chip,0,1) else chipchar = ''
  
  if mode eq 'im' then workingdir = lpar.imworkingdir
  if mode eq 'sp' then workingdir = lpar.spworkingdir2d

  if chipchar eq 'r' then rightonly=1
  if chipchar eq 'l' then leftonly=1

  ;files = choosefiles(camchar+lpar.wildchar+'.fits',workingdir+'ifp',workingdir+'fp')
  if mode eq 'im' then files = lrisselectfiles('i',['ifp','fp']+camchar,'','.fits', count=ct)
  if mode eq 'sp' then files = lrisselectfiles('s2',['ifp','fp']+camchar,'','.fits', count=ct)
  if ct eq 0 then return

  for f = 0, n_elements(files)-1 do begin
    if files[f] eq '' then continue

    rfile = repstr(repstr(files[f],'.fits','r.fits'),'if','f')
    lfile = repstr(repstr(files[f],'.fits','l.fits'),'if','f')

    if lpar.overwrite eq 0 then begin
      if file_test(rfile) and file_test(lfile) then continue
      if file_test(rfile) and keyword_set(rightonly) then continue
      if file_test(lfile) and keyword_set(leftonly) then continue
    endif

    lrissplit, files[f], rightonly=rightonly, leftonly=leftonly, renameas=repstr(files[f],'if','f'), nodistort=nodistort

  endfor

end

; -------------------------
pro autolrisskysubtract, camera=camera, chip=chip, nocrclean=nocrclean, skyalg=skyalg

  common lrisauto, lpar

  camchar = strmid(camera,0,1)
  chipchar = strmid(chip,0,1)
  ;if chipchar eq 'r' and camchar eq 'r' and (lpar.lrisversion eq 1 or lpar.lrisversion eq 4) then chipchar = ''

  if n_elements(skyalg) eq 0 then skyalg = 'median'
  if skyalg ne 'median' and skyalg ne 'kelson' then skyalg = 'median'

  if keyword_set(nocrclean) then $
  ;   files = choosefiles(camchar+lpar.wildchar+chipchar+'.fits',lpar.spworkingdir2d+'zfp',lpar.spworkingdir2d+'fp') $
      files = lrisselectfiles('s2',['zfp','fp']+camchar,chipchar,'.fits', count=ct) $
  else $
  ;   files = choosefiles(camchar+lpar.wildchar+chipchar+'.fits',lpar.spworkingdir2d+'zfp')
      files = lrisselectfiles('s2','zfp'+camchar,chipchar,'.fits', count=ct)
  if ct eq 0 then return

  if lpar.overwrite eq 0 then files = unmatched(files,'s')

  for f = 0, n_elements(files)-1 do begin
    if files[f] eq '' then continue

    h = headfits(files[f])
    lampstr = clip(sxpar(h,'LAMPS'))
    lamps = strsplit(lampstr,',',/extract)
    ;;if total(lamps ne '0') gt 0 then continue ;  skip flats/arcs
 
    trapdoor = clip(sxpar(h,'TRAPDOOR'))
    if trapdoor eq 'closed' then continue
 
    if skyalg eq 'median' then dum = subtractsky2(files[f])
    if skyalg eq 'kelson' then subtractsky4, files[f]

  endfor

end

; -------------------------

pro autolriswavcal, camera=camera, chip=chip, forcewavcal=forcewavcal, novalidatewavcal=novalidatewavcal

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

    ; new functionality: order files as to always start with the standards
    ; visually display the extraction
 

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   ;if chipchar eq 'l' then begin
   ;   return
   ;endif

   arclistfilelowres = lpar.refdatapath+'/'+'lrisarcall.list'   
   brightarclistfilelowres = lpar.refdatapath+'/'+'lrisarcbright.list'

   arclistfilehires = lpar.refdatapath+'/'+'lrisarcallhires.list'   
   brightarclistfilehires = lpar.refdatapath+'/'+'lrisarcbrighthires.list'


   ; Find useable 2D arcs.
   
   ; Get list of files which weren't crcleaned/skysubtracted (probable arcs)
   ;files = choosefiles(camchar+lpar.wildchar+chipchar+'.fits',lpar.spworkingdir2d+'fp', ct=ct)
   files = lrisselectfiles('s2','fp'+camchar,chipchar,'.fits', count=ct) ; nofilefilter??
   if ct gt 0 then files = unmatched(files, 'z', ct=ct)
   if ct gt 0 then files = unmatched(files, 's', ct=ct)
   if ct gt 0 then files = unmatched(files, 'sz', ct=ct)

   ;altfiles = choosefiles(camchar+lpar.wildchar+chipchar+'_arc.fits',lpar.spworkingdir1d+'fp', ct=ctalt)
   altfiles = lrisselectfiles('a','fp'+camchar,chipchar+'_arc','.fits', count=ctalt)
   if ct eq 0 and ctalt gt 0 then files = altfiles  
   if ct eq 0 and ctalt gt 0 then ct = ctalt

   ; This is a kludgy way to do it.  The arcs/ 2D files are looked at only if there are no arcs in the 2d/ directory.
   ; This means arcs can only be copied as 1D or as solutions.  Explore a way to fix this.


   ; Read in bad arc list

   if n_elements(badarclist) eq 0 then begin
     badarclist = ['']
     if file_test('badarc.txt') then begin
       nbarc = countlines ('badarc.txt')
       openr, 6, 'badarc.txt'
       ;badarcs = strarr(nbf)
       badarcstr = ''
       inline = ''
       while not eof(6) do begin
          readf, 6, inline
          badarcstr += (inline + ' ')
       endwhile
       close, 6 
       badarclist = strlowcase(strsplit(badarcstr,/extract))
       ;notblank = where(badarcs ne '', ctb) ; this isn't used for anything
     endif
   endif

  if ct gt 0 then begin
   
   ; future: copy these to 1d directory?  and also check there?

   arcp = lrisheaderinfo(files, badarclist=badarclist)

   ; Remove all but the last member of of identical consecutive doubles/multiples (probably unwarmed)   
   ; Change 2018-01: must also be in same position to count as a consecutive double
   keep = bytarr(ct)+1
   stackblocks = intarr(ct)-1
   b = 0 ; block ID
   for f = 1, ct-1 do begin
      if arcp[f].framenum-1 lt arcp[f-1].framenum-5 then begin
          b = b + 1
          continue ; gap / nonconsecutive
      endif
      sameconfig = arcp[f].grname eq arcp[f-1].grname and arcp[f].cwave eq arcp[f-1].cwave and $
         arcp[f].dichroic eq arcp[f-1].dichroic and arcp[f].slit eq arcp[f-1].slit and arcp[f].filter eq arcp[f-1].filter
      sameccd = arcp[f].binning eq arcp[f-1].binning and arcp[f].win eq arcp[f-1].win and $
                arcp[f].nx eq arcp[f-1].nx and arcp[f].ny eq arcp[f-1].ny ; extra safety check: rebinned/cropped frames might not match
      sameposition = $
         ( (abs(arcp[f].azimuth-arcp[f-1].azimuth)*cos(arcp[f].elevation) lt 0.5 and abs(arcp[f].elevation-arcp[f-1].elevation) lt 0.5) $
           or  (abs(arcp[f].radeg-arcp[f-1].radeg)*cos(arcp[f].decdeg) lt 0.5 and abs(arcp[f].decdeg-arcp[f-1].decdeg) lt 0.5) $
         )
      samelamps = arcp[f].lamps eq arcp[f-1].lamps 

      ; print, arcp[f].framenum, sameconfig, sameccd, sameposition, samelamps
      if sameconfig and sameccd and sameposition then begin
         if samelamps then begin
            keep[f-1] = 0
            if stackblocks[f-1] ge 0 then begin
               stackblocks[f-1] = -1
               stackblocks[f] = b  ; stack this one instead
            endif
          endif else begin
            if (arcp[f].hg and arcp[f-1].hg) eq 0 and (arcp[f].cd and arcp[f-1].cd) eq 0 and $
               (arcp[f].zn and arcp[f-1].zn) eq 0 and (arcp[f].kr and arcp[f-1].kr) eq 0 and $
               (arcp[f].xe and arcp[f-1].xe) eq 0 and (arcp[f].ar and arcp[f-1].ar) eq 0 and $
               (arcp[f].neon and arcp[f-1].neon) eq 0 then begin
                 stackblocks[[f-1,f]] = b
                 ; Only stack if fully 'complementary' (none of the same lamps)
                 ; Might be better to just add to block and figure out which specific ones to stack later...
             endif
          endelse
      endif else begin
          b += 1
      endelse
   endfor

   ; stack separately-exposed arcs
   ws = where(stackblocks ge 0, ctstack) 
   if ctstack gt 0 then begin
      ublocks = unique(stackblocks)
      ublocks = ublocks[where(ublocks ge 0)]
      for u = 0, n_elements(ublocks)-1 do begin
          comb = where(stackblocks eq ublocks[u], ct)
          combfiles = arcp[comb].filename
          outfile = lpar.arcdir+removepath(repstr(combfiles[0],'.fits','-sum'+clip(ct)+'_arc.fits'))
          if file_test(outfile) eq 0 or lpar.overwrite then begin
            if strlen(lpar.arcdir) gt 0 and file_test(lpar.arcdir) eq 0 then begin
               spawn, 'mkdir '+lpar.arcdir
               check = file_test(lpar.arcdir)
               if check eq 0 then begin 
                print, 'WARNING - Failed to create arcs directory: '+lpar.arcdir
               endif
            endif

            logprint, ' Summing '+clip(n_elements(combfiles))+' arc exposures:'
            for f = 0, n_elements(combfiles)-1 do begin
                hf = headfits(combfiles[f])
                inlampstr = sxpar(hf,'LAMPS')
                logprint, '    '+removepath(combfiles[f])+' '+inlampstr
                inlamparr = fix(strsplit(inlampstr,',',/extract))
                if f eq 0 then lamparr = inlamparr
                if f gt 0 then lamparr = (lamparr or inlamparr)
            endfor
            logprint, '  -> '+outfile
            outlampstr = clip(lamparr[0])
            for l = 1, n_elements(lamparr)-1 do outlampstr += ','+clip(lamparr[l])
            combinesum, combfiles, outfile, /noweight, /silent, idkey='KOAID'
            lampkey = ['MERCURY','NEON','ARGON','CADMIUM','ZINC','HALOGEN','KRYPTON','XENON','FEARGON','DEUTERI']
            lampon = fix(strsplit(outlampstr,',',/extract))
            lampkey = lampkey[0:(n_elements(lampon)-1)]
            lampval = replicate('off',n_elements(lampkey))
            lampval[where(lampon)] = 'on'
            editheader, outfile, ['LAMPS',lampkey], [outlampstr,lampval], /silent ;, 'Coadded arc'
          endif
          ; keep[comb] = 0 ; ???
      endfor
   endif

   ; officially remove the duplicates
   arcp = arcp[where(keep)]
   files = files[where(keep)]

   ; add the sums
   sumfiles = file_search(lpar.arcdir+'fp'+camchar+'*-sum*_arc.fits', count=ct)
   if ct gt 0 then begin
     sumarcp = lrisheaderinfo(sumfiles)
     files = [files, sumfiles]
     arcp = [arcp, sumarcp]
   endif

   ; Determine what are real arcs, with at least two lamps (one is not sufficient)
   ; Arcs with fewer lamps are ignored.  In the future should use for offsetting.
   if camchar eq 'b' then arcnlamps = arcp.hg + arcp.cd + arcp.zn   ; +arc.fe
   if camchar eq 'r' then arcnlamps = arcp.hg + arcp.neon + arcp.ar  ; + arc.kr + arc.xe
   if keyword_set(forcewavcal) then begin
      isarc = arcnlamps ge 1
   endif else begin
      isarc = (arcnlamps ge 3 and arcp.hal eq 0 and arcp.de eq 0)
      if total(isarc) eq 0 then begin
        wav0 = arcp.cwave - arcp.dispersion*(arcp.nx/2)
        wav1 = arcp.cwave + arcp.dispersion*(arcp.nx/2)
        arcnlamps = arcp.neon*0
        if camchar eq 'r' then arcnlamps += (wav0 gt 5800. or wav0 lt 7600.)*arcp.neon
        if camchar eq 'r' then arcnlamps += (wav0 lt 5800.)*arcp.hg
        if camchar eq 'r' then arcnlamps += (wav1 gt 7600.)*arcp.ar 
        if camchar eq 'b' then arcnlamps = arcp.hg + 0.5*(arcp.cd + arcp.zn + arcp.kr + arcp.xe)
        isarc = (arcnlamps ge 2 and arcp.hal eq 0 and arcp.de eq 0)
      endif
   endelse
   w = where(isarc, narc)
   if narc eq 0 then begin
       logprint, ' No trustworthy '+clip(camera)+' arcs found.'
       npossarc = 0
       for j = 0, n_elements(arcp)-1 do begin
          if arcp[j].nlamp eq 0 then continue
          lampstr = ''
          lampstr += (arcp[j].hg ? 'Hg' : '')
          lampstr += (arcp[j].neon ? 'Ne' : '')
          lampstr += (arcp[j].ar ? 'Ar' : '')
          lampstr += (arcp[j].cd ? 'Cd' : '')
          lampstr += (arcp[j].zn ? 'Zn' : '')
          lampstr += (arcp[j].hal ? 'hal' : '')
          lampstr += (arcp[j].kr ? 'Kr' : '')
          lampstr += (arcp[j].xe ? 'Xe' : '')
          lampstr += (arcp[j].fe ? 'Fe' : '')
          lampstr += (arcp[j].de ? 'De' : '')
          
          print, '   '+clip(removepath(arcp[j].filename),20)+' '+ $
              clip(arcp[j].grname,9)+' D'+clip(arcp[j].dichroic,3)+' '+clip(arcp[j].filter,5)+' '+$
              clip(arcp[j].slit,9)+' cw='+clip(fix(arcp[j].cwave))+' '+clip(lampstr,16)+' elev='+clip(fpr(arcp[j].elevation,2.1))
       endfor
       return ; no arcs good enough to solve with
   endif
   arcfiles = files[w]
   arc = arcp[w]

   ;outarcfiles = lpar.spworkingdir1d+repstr(removepath(arcfiles),'.fits','_arc.spec')

   ; Which are already extracted?
   if lpar.overwrite eq 0 then begin
     outspec = lpar.arcdir+removepath(repstr(repstr(arcfiles,'_arc',''),'.fits','_arc.spec'))
     w = where(file_test(outspec) eq 0, narc) ; unsolved arcs
     if narc gt 0 then begin
       arcfiles = arcfiles[w]
       arc = arc[w]
     endif
   endif
   ; maybe sort by central wavelength...?
 

   ; Extract arcs

   if narc gt 0 then begin
      ; Sort to enable easier comparisons
       s = sort(arc.cwave)
      arc = arc[s]
      arcfiles = arcfiles[s]

      ; We now have our list arcs to solve.  Proceed to solve them.


      ; Extract the arc spectrum at the standard position (still comes from the old defaults)

      if strlen(lpar.arcdir) gt 0 and file_test(lpar.arcdir) eq 0 then begin
        spawn, 'mkdir '+lpar.arcdir
        check = file_test(lpar.arcdir)
        if check eq 0 then begin 
          print, 'WARNING - Failed to create arcs directory: '+lpar.arcdir
        endif
      endif
 
      for f = 0, narc-1 do begin

          ; Extract an arc spectrum at the "normal" position and save the solution.
          ; The normal position is hard coded (but can be changed in the .par file).
          ; Spatial dependence is modeled later as a linear correction using sky lines.

          arcfile = arcfiles[f]
          ; outarcfile = outarcfiles[f]
          outarcfile = lpar.arcdir+repstr(removepath(repstr(arcfile,'_arc','')),'.fits','_arc.spec')
          if file_test(outarcfile) and (lpar.overwrite eq 0) then continue

          logprint, ' Extracting arc '+removepath(arcfile)
          arch = headfits(arcfile)
          arcdata = mrdfits(arcfile, 0, /silent)

          s = size(arcdata)
          ny = s[2]        
          if n_elements(normaltracepos) eq 0 then begin
             if camchar eq 'b'  then normaltracepos = lpar.defaulttraceyblue
             if camchar eq 'r'  then normaltracepos = lpar.defaulttraceyred
             if normaltracepos le 0 then begin
               if camchar eq 'b' or lpar.lrisversion eq 2 or lpar.lrisversion eq 3 then begin
                 logprint, '  Standard trace position unknown; using 1/4 point on CCD.'
                 normaltracepos = (ny/4);*arc[f].spatbin
               endif else begin
                 logprint, '  Standard trace position unknown; using 5/8 point on CCD.'
                 normaltracepos = (ny*5/8)
               endelse
             endif
          endif
          y0 = (normaltracepos/arc[f].spatbin-16/arc[f].spatbin) > 2
          y1 = (normaltracepos/arc[f].spatbin+16/arc[f].spatbin) < ny-2
          arcdata = arcdata[*,y0:y1]
                  ;nans = where(finite(arcdata) eq 0, ct)
                  ;if ct gt 0 then arcdata[nans] = 0. ; could interpolate, but usually nans are just bad rows
                  ;farc = total(arcdata,2)  ; could use a median to exclude cosmics, but tilted lines will vanish...
          farc = median(arcdata, dimension=2) ;*(y1-y0)
          writespec, outarcfile, farc, arch

          ; save a copy in the arcs directory (for successful arcs only)
          outfits1d = repstr(outarcfile,'.spec','.fits')
          if outfits1d ne arcfile then begin
            if file_test(outfits1d) then file_delete, outfits1d
            spawn, 'cp -pf '+arcfile+' '+outfits1d
          endif
              ; in the future, at this point will write out file with an astrometrically accurate header
      endfor
  endif

   endif

   ;earcfiles = file_search(lpar.spworkingdir1d+'fp'+camchar+'*_arc.spec', count=nearc) ; identical to outarcfiles on first run
   earcfiles = lrisselectfiles('a','fp'+camchar,chipchar,'*_arc.spec', count=nearc)
   updated = 0  ; flag to indicate whether or not a new solution has been determined and recorded

   for f = 0, nearc-1 do begin
       earcfile = earcfiles[f]
       arcsp = readspec(earcfile, h)
       ; print, earcfile
       ; print, sxpar(h,'UTC')
       arcinfo = lrisheaderinfo(earcfile)
       solfile = repstr(repstr(earcfile,'.fits','.sol'),'.spec','.sol')
       fitsfile = repstr(solfile,'.sol','.fits') ; for copying the fits file over to the 1d/ directory on success
       if file_test(solfile) and lpar.overwrite eq 0 then continue
                                      ;'~/research/redux/arc/lrisall.list'

       lampstr = arcinfo.lamps
       fixlampstr, lampstr, badarclist
       lamps = strsplit(lampstr,',',/extract)
       while n_elements(lamps) lt 8 do lamps = [lamps, '']
       lampnames = ['']
       if fix(lamps[0]) then lampnames = [lampnames, 'Hg'] ; need to update this method...
       if fix(lamps[1]) then lampnames = [lampnames, 'Ne']
       if fix(lamps[2]) then lampnames = [lampnames, 'Ar']
       if fix(lamps[3]) then lampnames = [lampnames, 'Cd']
       if fix(lamps[4]) then lampnames = [lampnames, 'Zn']
       if fix(lamps[6]) then lampnames = [lampnames, 'Kr']
       if fix(lamps[7]) then lampnames = [lampnames, 'Xe']
       if n_elements(lampnames) gt 1 then lampnames = lampnames[1:*] $
          else continue ; shouldn't happen unless every arc is broken

       ; reduce arc lists to just the arcs in the file
       if arcinfo.grname eq '1200/3400' or arcinfo.grname eq '1200/7500' or arcinfo.grname eq '1200/9000' then begin
         arclistfile = arclistfilehires
         brightarclistfile = brightarclistfilehires
         ; print, 'Using high-res linelist'
       endif else begin
         arclistfile = arclistfilelowres
         brightarclistfile = brightarclistfilelowres
       endelse
 

       readcol, arclistfile, wav, name, format='f,a'
       arclist = filterarclist(wav, name, lampnames, id=arcid)  ; returns a filtered list of wavs
       readcol, brightarclistfile, wav, name, format='f,a'
       brightarclist = filterarclist(wav, name, lampnames, id=brightid)

       nx = n_elements(arcsp)
       cwave = sxpar(h,'CWAVE') ; should be the same as arcinfo.cwave
       if cwave eq 0 then begin
          ;guesswav = !values.f_nan
          ;if clip(sxpar(h,'GRNAME')) eq '400/8500' then begin
          ;  guesswav = 7828.  ; this is a temporary block that hopefully we can delete later
          ;endif
          ;if clip(sxpar(h,'GRNAME')) eq '600/7500' then begin
          ;  guesswav = 7270.
          ;endif 
          ;if finite(guesswav) eq 0 then begin
            logprint, '  ERROR: Central wavelength value (CWAVE) is absent or set to zero!'
            logprint, '         You may need to edit the file headers directly using editheader.'
            continue
          ;endif
          ;logprint, ' cwave = 0?!?!  changing to '+clip(guesswav) ; would be better to use GRANGLE back in setup
          ;arcinfo.cwav = guesswav
          ;arcinfo.cwave = guesswav
          ;arcinfo.wavelength = guesswav
          ;cwave = guesswav
       endif
       ;print, ' Dispersion = ', sxpar(h,'DISPERSN')
       ;  logprint,   ' Changing dispersion from ', arcinfo.dispersion, ' to ', 1.193*arcinfo.specbin
       ; sxaddpar, h, 'DISPERSN', 1.193*arcinfo.specbin
       
       guessminwav = cwave - 300. - sxpar(h,'DISPERSN')*nx/2
       guessmaxwav = cwave + 300. + sxpar(h,'DISPERSN')*nx/2
       ; print, sxpar(h,'DISPERSN'), nx, guessminwav, guessmaxwav
       brightinrange = where(brightarclist gt guessminwav and brightarclist lt guessmaxwav, ct)
       if ct gt 0 then brightarclist = brightarclist[brightinrange]
       if ct gt 0 then brightid      = brightid[brightinrange]
       if ct le 3 then logprint, '   WARNING - not enough bright lines expected to be present to solve this spectrum'
       inrange = where(arclist gt guessminwav and arclist lt guessmaxwav, ct)
       if ct gt 0 then arclist = arclist[inrange]
       if ct gt 0 then arcid = arcid[inrange]

       ;if clip(sxpar(h,'DICHNAME')) eq '560' then begin
       ;     if camchar eq 'b' then arclist = arclist[where(arclist lt 6000.)]
       ;     if camchar eq 'b' then brightarclist = brightarclist[where(brightarclist lt 6000.)]
       ;endif
       ;if clip(sxpar(h,'DICHNAME')) eq '680' then begin
       ;     if camchar eq 'r' then arclist = arclist[where(arclist gt 6350.)]
       ;     if camchar eq 'r' then brightarclist = brightarclist[where(arclist gt 6350.)]
       ;endif
       if strpos(clip(sxpar(h,'DICHNAME')),'0') gt 0 then begin
          dichwav = float(clip(sxpar(h,'DICHNAME'))) * 10.
          dichbluelimit = dichwav + 450.
          dichredlimit = dichwav - 450.
          if camchar eq 'b' then wd = where(arclist lt dichbluelimit, ctd)
          if camchar eq 'r' then wd = where(arclist gt dichredlimit, ctd)
          if ctd eq 0 then logprint, '  ERROR - no lines???'
          arcid = arcid[wd]
          arclist = arclist[wd]

          if camchar eq 'b' and total(brightarclist ge dichbluelimit) gt 0 then begin
             arclist = [arclist, brightarclist[where(brightarclist ge dichbluelimit)]]
             arcid =   [arcid,   brightid[where(brightarclist ge dichbluelimit)]]
          endif
          if camchar eq 'r' and total(brightarclist le dichredlimit) gt 0 then begin
             arclist = [brightarclist[where(brightarclist le dichredlimit)], arclist]
             arcid   = [brightid[where(brightarclist le dichredlimit)], arcid]
          endif
          ;if camchar eq 'b' then brightarclist = brightarclist[where(brightarclist lt dichbluelimit)]
          ;if camchar eq 'r' then brightarclist = brightarclist[where(brightarclist gt dichredlimit)]
       endif
  

       cwavstr = ''
       if camchar eq 'r' then cwavstr = ', wv='+clip(fpr(arcinfo.wavelength,5.1))
       lampstr = ''
       for l = 0, n_elements(lampnames)-1 do lampstr += lampnames[l]
       needed = ''
       if camchar eq 'r' and total(lampstr eq 'Ar') eq 0 then needed = 'Ar' ; only if >8000 A ?
       if camchar eq 'r' and total(lampstr eq 'Ne') eq 0 then needed = 'Ne' 
       logprint, ' Wavelength-calibrating '+ removepath(earcfile)+ ' ('+lampstr+', '+arcinfo.grname+cwavstr+')'


       nx = n_elements(arcsp)

       dispersion = arcinfo.dispersion
       cwav = arcinfo.cwave
       ; could do a try loop here
       
       if n_elements(initsol) gt 0 then delvarx, initsol

       uncrefwav = 0.3   ; systematic uncertainty floor on wavelength measurement
       if arcinfo.grname eq '300/5000' then uncrefwav = 0.6 ; severe blending makes this solution always bad...

       grn = arcinfo.grname
       if camchar eq 'r' then nbrightline = 20
       if camchar eq 'b' then nbrightline = 16
       if camchar eq 'r' then order = 7
       if camchar eq 'r' and (grn eq '400/8500') then nbrightline = 22
       if camchar eq 'r' and (grn eq '831/8200' or grn eq '900/5500') then order = 6
       if camchar eq 'r' and (grn eq '1200/7500' or grn eq '1200/9000') then order = 5
       if camchar eq 'b' then order = 5 ; order has been increased.  Will this be safe?
       if camchar eq 'b' and grn eq '1200/3400' then order = 3 ; so few lines
       ; if camchar eq 'r' and lpar.lrisversion eq 4 then order = 5

       ;;;if camchar eq 'r' and chipchar eq 'l' then dispersion *= 1.01

       errorstatus = 0
       if lpar.debug eq 0 then catch, errorstatus
       if errorstatus ne 0 then begin
          errmsg = clip(!error_state.msg)
          if errmsg ne 'SOLVEARC: No line matches found' then begin
            logprint, '     Error during wavelength calibration:'; of '+removepath(earcfile)+':'
            logprint, '     '+errmsg ;!error_state.msg
          endif
          catch, /cancel 
          if !d.name eq 'PS' then psclose ; need to close the plot if an exception was thrown
          !p.multi = 0
          if file_test(checkplotfile) then file_delete, checkplotfile
          continue
       endif

       centroidrad = 4
       if arcinfo.specbin ge 2 then centroidrad = 2
       
       ; need to set x0 for the unlikely situation that the CCD was windowed in x directon

       ; fractional uncertainty/variation in dispersion per pixel
       ;df = 0.075 old
       df = 0.075 ; lots of variation on the blue, though less so on the red.
       ;if grn eq '600/4000' then df = 0.075
       if grn eq '1200/3400' or grn eq '900/5500' then df = 0.025 ; new

       if grn eq '400/3400' then xmin = 1400/arcinfo.specbin else delvarx, xmin ; ignore lines in FUV

       unccwav = (100 > 100*dispersion/arcinfo.specbin) + cwav*0.005 ;cwav*0.05
       ; print, '  unccwav: ', unccwav
       checkplotfile = repstr(earcfile,'.spec','.wavsol.ps') ; formerly from .fits
       outarclistfile = repstr(earcfile,'.spec','.list')     ;          from .fits to arc.list
       checkwav = [2900,10500.]
       if camchar eq 'b' then begin
           if arcinfo.dichroic eq '460' then checkwav[1] = 4900.
           if arcinfo.dichroic eq '500' then checkwav[1] = 5300.
           if arcinfo.dichroic eq '560' then checkwav[1] = 5900.
           if arcinfo.dichroic eq '680' then checkwav[1] = 7100.
       endif
       if camchar eq 'r' then begin
           if arcinfo.dichroic eq '460' then checkwav[0] = 4000.
           if arcinfo.dichroic eq '500' then checkwav[0] = 4700.
           if arcinfo.dichroic eq '560' then checkwav[0] = 5300.
           if arcinfo.dichroic eq '680' then checkwav[0] = 6500.
       endif

       if n_elements(initsol) gt 0 then print, '    Solution:', initsol
       nretry = 5
       for t = 0, nretry do begin
         if t gt 0 then begin
           print, ' Trying again...'
           if t eq 1 then order -= 1
           if t eq 2 then dispersion *= 0.95
           if t eq 3 then dispersion *= 1.10
           if t eq 4 then cwav -= cwav*0.05 
           if t eq 5 then cwav += cwav*0.10 
         endif
       
         ;print, ' Solver inputs: '
         ;print, '    cwav = ', cwav
         ;print, '    disp = ', dispersion
         ;print, '    maxorder = ', order
         arcsol = solvearcspec(arcsp.flux, wav, refwav=arclist, brightrefwav=brightarclist, refid=arcid, brightrefid=brightid, $
                  cwav=cwav*1., unccwav=unccwav, dispersion=dispersion, uncdisp=df*dispersion, $
                  maxsep=nx*dispersion*0.5, uncrefwav=uncrefwav, initsol=initsol, nbrightline=nbrightline, $
                  minorder=order, maxorder=order, xmin=xmin, $ 
                  centroidrad=centroidrad, $
                  ndet=ndet, nmatch=nmatch, ngoodmatch=ngoodmatch, nbrightunmatched=nbrightunmatched,$
                  checkplotfile=checkplotfile, medabsresidual=medabsresidual, checkwav=checkwav, $
                  errorcode=errorcode)
                      ; , outarclistfile=outarclistfile ; for user solutions?

         if errorcode eq -1 then break ; don't retry if arc simply has no lilnes

         validated = 0
         if n_elements(arcsol) ge 2 and keyword_set(novalidatewavcal) eq 0 then begin
           validated = 1
           ; Validate the wavelength solution by making sure that it has reasonable properties.
           if nbrightunmatched[0] gt 2+fix(nbrightunmatched[1]*0.2) then validated = 0
           checkcwav = arcsol[0]
           checkdisp = arcsol[1]
           if abs(checkcwav - cwav) gt cwav*0.05 then valerrstr = 0
           if abs(checkdisp - dispersion) gt dispersion*0.075 then validated = 0
           outwav = poly(findgen(nx)-nx/2., arcsol)
           uv = where(outwav lt 2900, ctuv)
           if max(uv) gt n_elements(outwav)-nx/4 then validated = 0
           if validated then begin
             if ctuv gt 0 then outwav = outwav[max(uv):*] ; this will remove the UV and any inflected wavelengths
              ;outwav = outwav[where(outwav gt 2900. and outwav lt 11000.)] ; the lt 2900 is unnecessary now
             ;if camchar eq 'r' and arcinfo.dichroic eq '560' then outwav = outwav[where(outwav gt 5400.)]
             ;if camchar eq 'r' and arcinfo.dichroic eq '680' then outwav = outwav[where(outwav gt 6500.)]
             outwav = outwav[where(outwav gt checkwav[0] and outwav lt checkwav[1])]
             dispx = outwav-shift(outwav,1)
             dispx[0] = dispx[1]
             avdisp = median(dispx)
             if abs(avdisp-dispersion) gt 0.075*dispersion then validated = 0
             dum = where(abs(avdisp-dispersion) gt 0.15*dispersion, ctbaddisp)
             if ctbaddisp gt 10 then validated = 0
           endif
           ;okdisp = (dispx - avdisp)/avdisp lt 0.2*dispersion
           ;wokdisp = where(okdisp eq 0, ct)
           ;if ct gt 1 then validated = 0
           if validated eq 0 then begin
              logprint, '     Automatic wavelength solution failed validation check.'
              ;  Not using this arc.  Were lamps not warm?'   Duplicated by below message.
           endif
         endif
         if keyword_set(novalidatewavcal) and n_elements(arcsol) ge 2 then validated = 1

         if camchar eq 'r' then residualthresh = 0.5
         if camchar eq 'b' then residualthresh = 0.8 ; formerly 1.7
         if camchar eq 'b' and arcinfo.grname eq '300/5000' then residualthresh=2.0
         if n_elements(arcsol) gt 2 and validated and nmatch gt 8 and ngoodmatch gt nmatch*0.25 and medabsresidual lt residualthresh then begin
           logprint, '   -> ', removepath(solfile)
           if t gt 0 then print, ' Wavelength calibration succeeded on retry #'+clip(t)
           openw, 9, solfile
           for i = 0, n_elements(h)-1 do printf, 9, '#'+h[i]
           for i = 0, n_elements(arcsol)-1 do printf, 9, arcsol[i]
           close, 9  ; could print out a header (and even search it), but it's not really necessary now
           updated = 1
         endif else begin
           logprint, '     FAILED to solve arc.   Were the lamps warmed up, and is the central wavelength correct?' 
           ; could now just redo it until it works...?
           if t eq nretry then wavsolfail = [wavsolfail, earcfile]
           continue ; retry
         endelse
         break ; don't retry
       endfor
       
       checkplotfile = ''
       catch, /cancel 
       
   endfor

   ; Finally, extract a "generic" solution that averages all the non-constant terms together to ensure consistency between frames.
   ; Also produce a summary of all the individual wavelength solutions.

   if updated then begin
    psopen, lpar.arcdir+'wavsol'+camchar+'.ps', xsize=8, ysize=5.8, /inches
    device, /color
    !p.font=0
    device, /times, font_index=17                                  
    colors = transpose([[0,0,0],$
                      [255,50,50],$ 
                      [120,120,40],$ 
                      [50,200,50],$ 
                      [50,190,190],$
                      [60,60,255],$
                      [140,40,140]])
    tvlct, colors
    cols = indgen(6)
    !p.multi = [0,1,3]

    solfiles = findfile(lpar.arcdir+'*fp'+camchar+'*_arc.sol')
    sh = lrisheaderinfo(solfiles)
    gr = unique(sh.grname) ; should also consider the possibility of highly different CWAV values
    for g = 0, n_elements(gr)-1 do begin
      ; print, gr[g]
      grstr = repstr(gr[g],'/','-')
      ;avfile = lpar.spworkingdir1d+'lris'+camchar+grstr+'_average.sol'
      ;print, avfile
      w = where(sh.grname eq gr[g], nsol)
      csolfiles = solfiles[w]
      csh = sh[w]
      for s = 0, nsol-1 do begin
         insol = grabcolumn(csolfiles[s],0)      
         order = n_elements(insol)-1
         if s eq 0 then solmatrix = fltarr(order+1, nsol) + !values.f_nan
         if n_elements(insol) ne (size(solmatrix)) [1] then logprint, 'Order mismatch in trying to measure a standard wavelength solution' ; this happens when order is changed and old solution files exist
         for p = 0, order do insol[p] /= csh[s].specbin^p
         solmatrix[0:(n_elements(insol)-1),s] = insol  
      endfor
      avsol = fltarr(order+1)
      rangecwav = max(solmatrix[0,*])-min(solmatrix[0,*])  ; make sure this is not too high
      if rangecwav gt 20 then logprint, ' Warning - large offsets present in wavelength solutions for '+clip(gr[g])
      for p = 0, order do begin
         avsol[p] = mean(solmatrix[p,*])
      endfor
      ;if file_test(avfile) eq 0 or lpar.overwrite then begin
      ;  openw, 5, avfile
      ;  printf, 5, '# '+clip(nsol)+' solutions'
      ;  printf, 5, '# GRNAME = '+clip(gr[g])
      ;  for p = 0, order do begin
      ;     printf, 5, avsol[p]
      ;  endfor
      ;  close, 5
      ;endif

      if camchar eq 'r' and lpar.lrisversion eq 1 then begin
         nx = 2048
         x0 = 1024
      endif else begin 
         nx = 4096
         x0 = 2048
      endelse
      ; print, nx, x0, camchar, lpar.lrisversion

      ybars = [0.08, 0.38, 0.68, 0.98]
      pix = findgen(nx)
      if camchar eq 'b' and grstr eq '400-3400' then pix = pix[1300:*]
      avdisp = median( poly(pix-x0,avsol) - shift(poly(pix-x0, avsol),1) )
      for b = 0, 2 do begin
        !p.position = [0.08, ybars[2-b], 0.98, ybars[3-b]]
        if b eq 0 then yrange=[min(poly(pix-x0, avsol)), max(poly(pix-x0, avsol))]
        if b eq 1 then yrange=[-100,100]
        if b eq 2 then yrange=[-4 < min(poly(pix-x0, avsol)-(b gt 0)*poly(pix-x0, [avsol[0], avdisp])), $
                               +4 > max(poly(pix-x0, avsol)-(b gt 0)*poly(pix-x0, [avsol[0], avdisp]))]
        ;plot, pix, poly(pix-x0, avsol)-(b gt 0)*poly(pix-x0, [avsol[0], avdisp]), $
        ;              thick=4, xrange=[0,nx], /xsty, yrange=yrange, /ynoz, ytickformat='(I5)', charsize=2.0
        if b eq 2 then xtitle = 'Pixel (physical)' else xtitle = ''
        if b eq 1 or b eq 2 then ytitle = 'Wavelength residual (A)' else ytitle = 'Wavelength (A)'
        plot, [0], [0], xrange=[0,nx], /xsty, yrange=yrange, /ynoz, ytickformat='(I5)', charsize=2.0, xtitle=xtitle, ytitle=ytitle
        if b eq 0 then xyouts, 0.1, 0.93, /norm, gr[g], charsize=1.5
        if b eq 0 then xyouts, 0.2, ybars[2-b]+0.04, /norm, 'Wavelength curve'
        if b eq 1 then xyouts, 0.1, ybars[2-b]+0.04, /norm, 'Linear differential'
        if b eq 2 then xyouts, 0.1, ybars[2-b]+0.04, /norm, 'Linear differential after linear correction'
        ;if b eq 0 then print, avsol
        for s = 0, nsol-1 do begin
           ;if b eq 0 then print, solmatrix[*,s]
           oplot, pix, poly(pix-x0, solmatrix[*,s])-(b gt 0)*poly(pix-x0, [avsol[0], avdisp]) $
                       - (b eq 2)*(solmatrix[0,s]-avsol[0]), $
                       color=(s mod 5)+1
        endfor
      endfor
    endfor 

    psclose
    retitleplot, lpar.arcdir+'wavsol'+camchar+'.ps', (camchar eq 'b' ? 'Blue' : 'Red')+' Wavelength solutions'
    !p.multi = 0
   endif

end

; -----------

pro autolrissum, camera=camera, chip=chip, nosum=nosum

   ; Combine 2D spectra taken at a common orientation.

   ; Should probably only stack sequences - try looping through files, and with each file look ahead to see if the next one
   ; is of the same target and PA and configuration and similar exposure and no significant x-dither.

   ; Very rarely encountered bug - /nosum after regular run deletes summed files and redoes

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)
   ;if chipchar eq 'r' and camchar eq 'r' and (lpar.lrisversion eq 1 or lpar.lrisversion eq 4) then chipchar = ''
   
   ;ulpar.wildchar = lpar.wildchar  ; this is a temporary measure for now until generalized date filtering is implemented
   ;if n_elements(date) gt 0 then ulpar.wildchar = clip(date)+strmid(ulpar.wildchar,6)
   ;files = choosefiles(camchar+ulpar.wildchar+chipchar+'.fits',lpar.spworkingdir2d+'szfp',lpar.spworkingdir2d+'sfp', ct=ct)
   files = lrisselectfiles('s2',['szfp','sfp']+camchar,chipchar,'.fits', count=ct)
   if ct eq 0 then return
 
   if strlen(lpar.spworkingdir1d) gt 0 and file_test(lpar.spworkingdir1d) eq 0 then begin
     print, 'Creating spectroscopy 1D working directory ', lpar.spworkingdir1d
     spawn, 'mkdir '+lpar.spworkingdir1d
     check = file_test(lpar.spworkingdir1d)
     if check eq 0 then begin 
       print, 'WARNING - Failed to create 1D spectrum working directory: '+lpar.spworkingdir1d
     endif
   endif

   fileh = lrisheaderinfo(files) ; should eventually put this in one of the conditionals below

   i = 0
   while i le n_elements(files)-1 do begin
      combfiles = [files[i]]
      ; look ahead, unless this is a short exposure
      ; print, fileh[i].filename, ' ', rclip(fix(fileh[i].exptime),6), '  ', fileh[i].target
      if fileh[i].exptime gt 100. and keyword_set(nosum) eq 0 then begin
        j = i+1
        while j le n_elements(files)-1 do begin
          ; print, '   ', fileh[j].filename, fileh[j].target, fileh[j].slitpa, fileh[j].grname, fileh[j].dichroic, fileh[j].cwave, fileh[j].gratf
          ; Note that files are being compared to the FIRST file, since j increments until the loop breaks but i remains fixed.
          if fileh[j].target ne fileh[i].target then break
          if abs((fileh[j].slitpa - fileh[i].slitpa)+360. mod 360) gt 1.0 then break
          if fileh[j].specbin ne fileh[i].specbin or fileh[j].spatbin ne fileh[i].spatbin then break
          if fileh[j].grname ne fileh[i].grname then break
          if fileh[j].dichroic ne fileh[i].dichroic then break
          if abs(fileh[j].cwave - fileh[i].cwave) gt 3. then break
          if fileh[j].gratf ne fileh[i].gratf then break  ; configuration changed, presumably a file sequence gap
          if fileh[j].nx ne fileh[i].nx or fileh[j].specbin ne fileh[i].specbin or fileh[j].x0 ne fileh[i].x0 or $
             fileh[j].ny ne fileh[i].ny or fileh[j].spatbin ne fileh[i].spatbin or fileh[j].y0 ne fileh[i].y0 then break  
          if (ten(fileh[j].ut) - ( ten(fileh[j-1].ut)+fileh[j-1].exptime/3600. )) gt 0.25 then break ; long gap - probably guiding was lost,
                                                                                               ; or the intent was to get spaced spectra
          slitparad = fileh[i].slitpa*!pi/180.
          cosdec = cos(fileh[i].decdeg*!pi/180.)
          raoff = ((fileh[j].radeg-fileh[i].radeg)*cosdec)*3600.
          decoff = (fileh[j].decdeg-fileh[i].decdeg)*3600.
          xoff = raoff*cos(slitparad)-decoff*sin(slitparad)  ; should be zero if slit remains on source
          yoff = raoff*sin(slitparad)+decoff*cos(slitparad)  ; dither step
          ;print, removepath(files[i]), ' ', removepath(files[j]), raoff, decoff, ' | ', xoff, yoff
          if abs(xoff) gt 1.5 or abs(yoff) gt 30. then break     ; off-slit or huge dither
          ;print, fileh[j].exptime, fileh[i].exptime
          ; could also check filter and some other things
          if fileh[j].exptime gt 60. and fileh[j].exptime gt fileh[i].exptime*0.2 then begin
             combfiles = [combfiles, files[j]]  
          endif
          j += 1
        endwhile
      endif else begin
        ; maybe consider doing something special if the next exposure is long on the same 'target'?
        j = i+1
      endelse

      nsum = n_elements(combfiles)
      outtarg = repstr(fileh[i].target,' ','')
      outtarg = repstr(fileh[i].target,'-','') ; ?
      outtarg = repstr(outtarg,'_S','|')
      outtarg = (strsplit(outtarg,'|',/extract)) [0]
      outtargstr = repstr(legalfilestr(outtarg),'_','')
      outtargstr = clip(clip(outtargstr,12)) ; formerly 10.  Reluctantly changing to 12 to accommodate ZTF names better.
      ;if strpos(strupcase(outtargstr),'GRB') eq -1 and $
      ;   strpos(strupcase(outtargstr),'PTFS') eq -1 then outtargstr = clip(clip(outtargstr,8))
                                                        
      datestr = strmid(repstr(fileh[i].date,'-',''),2,6)
      ; outpastr = string(upa[p],'(I03)')
      ; grnamestr = repstr(fileh[w[0]].grname,'/','.')
      if lpar.koa then begin
         filen0str = fileh[i].koaframe
         sepstr = '.'
      endif else begin
         filen0str = string(fileh[i].framenum,'(I04)')
         sepstr = '_'
      endelse
      outfile = lpar.spworkingdir1d+'ls'+camchar+datestr+sepstr+filen0str+chipchar+'_'+outtargstr+'_sum'+clip(nsum)+'.fits'
      outfile = repstr(outfile,'_sum1.','.')
      outfile = repstr(outfile,'_sum','_s')
      ; print, outfile, file_test(outfile)
      if file_test(outfile) and lpar.overwrite eq 0 then begin
         i = j
         continue
      endif
      competingoutfile = lpar.spworkingdir1d+'ls'+camchar+datestr+sepstr+filen0str+chipchar+'_'+outtargstr+'*'+'.fits'
      checkcompeting = file_search(competingoutfile,count=ctcompete)
      if ctcompete gt 0 then begin
         for k = 0, ctcompete-1 do begin
            if checkcompeting[k] ne outfile then logprint, ' Deleting '+checkcompeting[k]
            if checkcompeting[k] ne outfile then file_delete, checkcompeting[k]
         endfor
      endif

      outmaskfile = repstr(outfile,'.fits','.mask.fits')

      if nsum eq 1 then begin
         if file_test(outfile) then file_delete, outfile
         logprint, ' '+removepath(combfiles[0])+' '+clip('('+clipzero(round(fileh[i].exptime))+'s)',7)+'-> '+removepath(outfile)
         ;spawn, 'cp '+combfiles[0]+' '+outfile
         combmask = repstr(repstr(combfiles[0],'.fits','.mask.fits'),'sz','z')
         combinesum, combfiles[0], outfile, /noweight, /silent, idkey='KOAID', addext=combmask   ; purely for header management
         ;combinesum, combmasks[0], outmaskfile, /noweight, /silent, idkey='KOAID'
      endif else begin
         combmasks = repstr(repstr(combfiles,'.fits','.mask.fits'),'sz','z')
         if total(file_test(combmasks)) lt n_elements(combmasks) then delvarx, combmasks ; if no mask files, skip masking...
         pixoffsets = intarr(nsum)
         xoffsets = fltarr(nsum) ; in arcsec
         ra0 = fileh[i].radeg
         dec0 = fileh[i].decdeg
         cosdec0 = cos(dec0*!pi/180.)
         slitparad = fileh[i].slitpa*!pi/180.
         for k = 1, nsum-1 do begin
           wk = where(files eq combfiles[k])
           ra = fileh[wk].radeg
           dec = fileh[wk].decdeg
           raoff = ((ra-ra0)*cosdec0)*3600.
           decoff = (dec-dec0)*3600.
           xoff = raoff*cos(slitparad)-decoff*sin(slitparad)  ; should be zero if slit remains on source
           yoff = raoff*sin(slitparad)+decoff*cos(slitparad)  ; dither step
           if abs(yoff) lt 0.01 then yoff = 0.0
           yoffpix = yoff/(0.135*fileh[wk].spatbin)
           pixoffsets[k] = fix(round(yoffpix))
           xoffsets[k] = xoff
         endfor
         logprint, ' Combining 2D frames: '
         for k = 0, nsum-1 do begin
           if abs(xoffsets[k]) gt 0.01 then sstr = ' lateral offset: '+clip(fpr(xoffsets[k],4.2))+'"' else sstr = ''
           logprint, '   '+removepath(combfiles[k])+' '+rclip(fix(round(fileh[wk].exptime)),5)+'s'+' '+clip(pixoffsets[k])+'px'+sstr
         endfor
         logprint, clip('',20)+'   --> '+removepath(outfile)


         ; default idkeys are EXPTIME, AIRMASS, UT

         combinesum, combfiles, outfile, /nomask, yshift=pixoffsets, /noweight, /silent, idkey='KOAID', addext=combmasks  ; previously, mask=combmasks. But masking creates problems if conditions change
        ; combinesum, combmasks, outmaskfile, /nomask, yshift=pixoffsets, /noweight, /silent, idkey='KOAID'
      endelse


      i = j
   endwhile

   return
end


; -----------

pro autolristrace, camera=camera, chip=chip, skiprbalign=skiprbalign

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   ; This function does several things:
   ;  (1) a) Uses "trace2d" to measure the trace of every frame using ALL sources in each image, and writes out
   ;      .trace and .profile files containing the trace function and summed spatial profile.
   ;      b) Failing that, it uses the trace offsets from a similar observation to still write out the profile.
   ;  (2) Looks through profile files corresponding to 'bright' stars to measure where the trace normally resides.
   ;      These are saved to brightobjectpos.txt
   ;  (3) Finally, goes through all profile files to find the best guess of the science target position and aperture width.
   ;      These are saved to objectpos.txt

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)
   ;ulpar.wildchar = lpar.wildchar  ; this is a temporary measure for now until generalized date filtering is implemented
   ;if n_elements(date) gt 0 then ulpar.wildchar = clip(date)+strmid(ulpar.wildchar,6)

   ;if chipchar eq 'l' then return

   brightobjfile = lpar.spworkingdir1d+'brightobjectpos.txt'
   objfile = lpar.spworkingdir1d+'objectpos.txt' ; consider adding camera...
   userobjfile = 'userobjectpos.txt'

   ; files = choosefiles(camchar+ulpar.wildchar+chipchar+'.fits',spworkingdir+'szfp',spworkingdir+'sfp', ct=ct)
   ;files = file_search(lpar.spworkingdir1d+'ls'+camchar+'*.fits', count=ct)
   files = lrisselectfiles('s1','ls'+camchar,'*','.fits', count=ct, /nofilter)
   if ct eq 0 then return

   tracefiles = strarr(n_elements(files))
   fileh = lrisheaderinfo(files)


   dum = file_search(lpar.spworkingdir1d+'ls'+camchar+'*.profile', count=ntrfilebefore)

   ; Trace spectra: look for a moderately strong source and determine its trace curve

   for f = 0, n_elements(files)-1 do begin
     if seqfilter(fileh[f]) then continue

     outfile = repstr(files[f], '.fits', '.trace')
     profilefile = repstr(files[f], '.fits', '.profile')
     plotfile = repstr(files[f], '.fits', '.trace.eps')
     ;;;print, clip(file_test(profilefile)), ' ', clip(file_test(outfile)), '   ', files[f]
     ; uncomment below to retry failed traces
     ;if (file_test(profilefile) and file_test(outfile)) and keyword_set(lpar.overwrite) eq 0 then continue
     if (file_test(profilefile) or file_test(outfile)) and keyword_set(lpar.overwrite) eq 0 then continue
     h = headfits(files[f])
     delvarx, xstart
     delvarx, xrange
     delvarx, yrange
     if camchar eq 'b' and sxpar(h,'GRNAME') eq '400/3400' then xstart=sxpar(h,'NAXIS1')*mean([0.65,0.9])
     if camchar eq 'b' and sxpar(h,'GRNAME') eq '400/3400' then xrange=sxpar(h,'NAXIS1')*[0.35,1.0]
     if camchar eq 'r' and lpar.lrisversion eq 4 then yrange = [fileh[f].ny/2,fileh[f].ny-30]
     ; if n_elements(yrange) gt 0 then print, fileh[f].ny, ' | ', yrange[0], yrange[1]
     logprint, ' Measuring spatial profile and trace function for '+removepath(files[f])
     errorstatus = 0
     if lpar.debug eq 0 then catch, errorstatus
     if errorstatus ne 0 then begin
         logprint, ' Unexpected error during tracing:'
         logprint, !error_state.msg
     endif else begin
        if file_test(outfile) then file_delete,outfile ; these are not guaranteed to be rewritten, so keeping them could induce a mixup
        if file_test(profilefile) then file_delete,profilefile
        if file_test(plotfile) then file_delete,plotfile
        trace2d, files[f], xstart=xstart, xrange=xrange, tracefile=outfile, profilefile=profilefile, plotfile=plotfile, yrange=yrange
     endelse
     catch, /cancel
   endfor


   ; For any failed traces, find another observation with same config, similar airmass/angle
   ; and use that trace file as a substitute for this object

   success = intarr(n_elements(files))
   for f = 0, n_elements(files)-1 do begin
      if file_test(repstr(files[f], '.fits', '.trace')) then success[f] = 1
   endfor
   fut = fltarr(n_elements(files))
   for f = 0, n_elements(files)-1 do fut[f] = ten(fileh[f].ut)
   for f = 0, n_elements(files)-1 do begin
     if seqfilter(fileh[f]) then continue
     if success[f] eq 0 then begin
        profilefile = repstr(files[f], '.fits', '.profile')
        if (file_test(profilefile)) and keyword_set(lpar.overwrite) eq 0 then continue
           ; if profile exists but not in objectpos.txt, then objectpos.txt won't be written.  maybe don't want that?

        hf = fileh[f]
        wsimilar = where(success and hf.spatbin eq fileh.spatbin and hf.specbin eq fileh.specbin and $
                         hf.grname eq fileh.grname and abs(hf.cwave-fileh.cwave) lt 10., ct)
        if ct eq 0 then begin 
           logprint, '  '+removepath(files[f])+' has no strong traces and no similar observations found - no way to determine profile or trace.'
           continue
        endif
        ranksimilar = (hf.airmass-fileh[wsimilar].airmass)^2 + abs(fut[f]-fut[wsimilar])/24.
        s = sort(ranksimilar)
        tracefiles[f] = repstr(files[wsimilar[s[0]]],'.fits','.trace') ; includes the directory path
        logprint, '  Using trace file ', removepath(tracefiles[f]), ' for ', removepath(files[f])

        plotfile = repstr(files[f], '.fits', '.trace.eps')
        trace2d, files[f], xstart=xstart, settracefile=tracefiles[f], profilefile=profilefile, plotfile=plotfile
     endif
   endfor

   ; Output at this point:  profile file and trace file for each 2D spectrum.






   ; Get locations of bright objects (trace stars, usually standards)
   ; and write their centroid locations to brightobjects.txt

   tfile = ['']
   if file_test(brightobjfile) eq 0 then begin
       openw, 10, brightobjfile
       printf, 10, '#LPipe-v'+lpar.lpversion+' bright object positions'
       printf, 10, '#coord=array'
       printf, 10, '#filename                           ycen    width'
       close, 10
   endif else begin
      readcol, brightobjfile, tfile, tcenter, format='a,f', comment='#' ; crashes later if file is blank
      for i = 0, n_elements(tfile)-1 do if strpos(tfile[i],'.fits') le 0 then tfile[i] = tfile[i]+'.fits'
   endelse

      ; Identify bright sources (probably but not necessarily standards)
   isstd = fileh.exptime le 91. and fileh.maxsource gt 10000. 
   if total(isstd) lt 3 then isstd = fileh.exptime le 181. and fileh.maxsource gt 10000.
   if total(isstd) lt 3 then isstd = fileh.exptime le 181. and fileh.maxsource gt 7000.
   if total(isstd) lt 2 then isstd = fileh.exptime le 181. and fileh.maxsource gt 4000.
   if total(isstd) lt 1 then isstd = fileh.exptime le 181. and fileh.maxsource gt 2000.
   if total(isstd) lt 1 then isstd =(fileh.exptime le 181. and fileh.maxsource gt 1000.) or $
                                    (fileh.exptime le 900. and fileh.maxsource gt 4000.) ; no standard?  Do what we can...
          ; note: maxcounts is not always a reliable indicator...
          ; note: objects are never removed from the brightobjpos file even with /redo mode, unless deleted

  ; for j = 0, n_elements(fileh)-1 do print, ' '+clip(removepath(fileh[j].filename),36),' ',fpr(fileh[j].exptime,5.1),' ', fpr(fileh[j].maxsource,6.1), isstd[j] ? ' * ' : '  '

   wstd = where(isstd, ct) ; probable standards (or other bright tracing objects)
   for i = 0, ct-1 do begin
       if total(files[wstd[i]] eq tfile) gt 0 then continue  ; already measured the center
       ; measure a new center
       profilefile = repstr(files[wstd[i]],'.fits','.profile')
       if file_test(profilefile) eq 0 then continue ; no profile
       readcol, profilefile, y, prof, format='f,f', comment='#'
       if n_elements(prof) lt 25 then begin
          logprint, '  Error: something wrong with '+clip(profilefile)+'?  Only '+clip(n_elements(prof))+'elements.'
          continue
       endif
       maxprof = max(prof[10:n_elements(prof)-10])
       wmax = where(prof eq max(prof))
       maxpix = median(y[wmax], /even)
       if (chipchar ne 'l' and (maxpix lt 20 or maxpix gt n_elements(y)-50)) or $
          (chipchar eq 'l' and (maxpix lt 50 or maxpix gt n_elements(y)-20)) then begin
          ;print, '  Error: brightest object in '+clip(profilefile)+' too close to edge; not tracing.'
          continue
       endif
        ; continuing will re-open profile files lots of times unnecessarily, but no big deal
       centroidpix = total(y[maxpix-5:maxpix+5]*prof[maxpix-5:maxpix+5])/total(prof[maxpix-5:maxpix+5])
       fitcenter = getlinecenter(y, prof, centroidpix, fitrad=10., width=width)
       if width gt 1.0 and width lt 20.0 then begin
          objectpos = fitcenter
       endif else begin
          objectpos = centroidpix
       endelse
       ; this should really use trsoutstring in saveobjectposfile.pro
       ; y position is in FITS pixel coordinates
       outstr = clip(repstr(removepath(files[wstd[i]]),'.fits',''),32)+' '+fpr(objectpos,5.2)+' '+fpr(width,3.2) $
                       +'  # '+clip(fileh[wstd[i]].grname,10)  $
                       +rclip(long(fileh[wstd[i]].exptime),5)+'s' $
                       +rclip(long(fileh[wstd[i]].maxsource),7)+'cts' ; needs to be updated if nchar changes
       updatefile, brightobjfile, outstr
   endfor


   ; Read in the list of bright standards and centroid locations, and mark the (likely) standards.

   readcol, brightobjfile, tfile, tcenter, format='a,f', comment='#'
   for i = 0, n_elements(tfile)-1 do if strpos(tfile[i],'.fits') le 0 then tfile[i] = tfile[i]+'.fits'

   isstd = intarr(n_elements(files))  ; this will generally be the same as isstd above, but need to recalc if loading from disk
   centerpos = fltarr(n_elements(files)) + !values.f_nan
   for i = 0, n_elements(tfile)-1 do begin
       match = where(tfile[i] eq removepath(fileh.filename), ctmatch)
       if ctmatch gt 0 then begin
         isstd[match] = 1   ; mark the bright standards
         centerpos[match] = tcenter[i]
       endif
   endfor



   ; Create or update the object apertures file

   ; Create the file if it does not exist
   ofile = ['']
   if file_test(objfile) eq 0 then begin
       openw, 10, objfile
       printf, 10, '#LPipe-v'+lpar.lpversion+' default object positions'
       printf, 10, '#coord=array'
       printf, 10, '#filename                            ycen    rad   background  tracefile # (alternates) '
       close, 10
   endif else begin
       readcol, objfile, ofile, ocenter, orad, format='a,f,f', comment='#' ; crashes later if file blank
       for i = 0, n_elements(ofile)-1 do if strpos(ofile[i],'.fits') le 0 then ofile[i] = ofile[i]+'.fits'
   endelse

   ; Update it, object by object
   objfilechanged = 0
   for i = 0, n_elements(files)-1 do begin
      filei = removepath(files[i])
      if total(filei eq ofile) gt 0 and lpar.overwrite eq 0 then continue  ; already measured the center

      profilefile = repstr(files[i], '.fits', '.profile')
      if file_test(profilefile) eq 0 then continue ; no profile, so there is no way to measure the center

      hi = fileh[i]
      ybin = hi.spatbin

      if seqfilter(hi) then continue

      ; Find the trace position of a standard star taken in same configuration
      wcam = where(fileh.camera eq hi.camera and isstd, ctcam)
      if ctcam eq 0 then begin
         avstdpos = 50./hi.spatbin
         if lpar.lrisversion eq 4 or lpar.lrisversion eq 1 then avstdpos = (hi.ny/2+50 < hi.ny-25)
         logprint, '  No bright objects observed; not sure where to look for the trace (guessing '+clipzero(avstdpos)+')'
      endif else begin
        avstdpos = median(centerpos[wcam]*fileh[wcam].spatbin, /even)/hi.spatbin  ; default to search around the global average standard position
        wsimstd = where(fileh.specbin eq hi.specbin and fileh.spatbin eq hi.spatbin and fileh.grname eq hi.grname and isstd, ctsimstd)
        if ctsimstd ge 1 then avstdpos = median(centerpos[wsimstd], /even)
      endelse

      ; Find possible sources
      readcol, profilefile, y, prof, format='f,f', comment='#'
      nm = n_elements(prof)
      mprof = fltarr(nm)
      for k = 0, n_elements(mprof)-1 do mprof[k] = prof[k] - median([prof[(k-10)>0:(k-2)>0],prof[(k+2)<(nm-1):(k+10)<(nm-1)]])
      rms = median(abs(mprof-shift(mprof,1)),/even) ; some outlier rejection?
      if finite(rms) eq 0 or rms le 0. then begin
         logprint, ' WARNING: had difficulty estimating profile noise level.'
         rms = 1.
      endif
      smoothprof = convol(mprof,[0.5,1,1,1,0.5]/4.)
       
      ismax4 = smoothprof gt shift(smoothprof,-1) and smoothprof gt shift(smoothprof,1) and $
               smoothprof gt shift(smoothprof,-2) and smoothprof gt shift(smoothprof,2) and $
               smoothprof gt shift(smoothprof,-3) and smoothprof gt shift(smoothprof,3) and $
               smoothprof gt shift(smoothprof,-4) and smoothprof gt shift(smoothprof,4)
      if chipchar eq 'r' then begin
        ytracemin = 20/ybin
        ytracemax = n_elements(y)-140/ybin ; n_elements(y)-40/ybin-100/ybin
      endif
      if chipchar eq 'l' then begin
        ytracemin = 140/ybin
        ytracemax = n_elements(y)-80/ybin
      endif
      if chipchar eq '' then begin
        ytracemin = 800/ybin               ; very temporary
        ytracemax = n_elements(y)-140/ybin
      endif
      peaks = where(ismax4 and $
                    smoothprof gt 3*rms and y gt ytracemin and y lt ytracemax, ctpeak)
      if ctpeak eq 0 then begin
         if chipchar eq 'r' then begin
            ytracemin = 40/ybin
            ytracemax = n_elements(y)-180/ybin ; n_elements(y)-80/ybin-100/ybin
         endif
         if chipchar eq 'l' then begin
            ytracemin = 180/ybin
            ytracemax = n_elements(y)-90/ybin
         endif
         peaks = where(ismax4 and $
                    smoothprof gt 2.5*rms and y gt ytracemin and y lt ytracemax, ctweakpeak)
         if ctweakpeak eq 0 then begin
            logprint, ' '+clip(removepath(profilefile))+': found no sources.'
            continue
         endif
      endif
      peakcen   = fltarr(n_elements(peaks))
      peakwidth = fltarr(n_elements(peaks))  ; this is sigma.  fwhm = 2*sqrt(2*alog(2.))*sigma = 2.35482*sigma
      for p = 0, n_elements(peaks)-1 do begin
        maxpix = peaks[p]
        ; try different peak-measurement input parameters until a reasonable result is hopefully obtained
        fitcenter = getlinecenter(y, prof, maxpix,      fitrad=16./ybin, peak=peakf, width=width, bkpoly=1, pix=pix,flux=flux,ffit=ffit)   
        if abs(fitcenter-maxpix) gt 10./ybin then $
          fitcenter = getlinecenter(y, prof, maxpix,      fitrad=30./ybin, peak=peakf, width=width, bkpoly=2, pix=pix,flux=flux,ffit=ffit)   
        if abs(fitcenter-maxpix) gt 6./ybin or peakf lt 0 or width gt 20./ybin then $
          fitcenter = getlinecenter(y, mprof, maxpix,      fitrad=20./ybin, peak=peakf, width=width, bkpoly=1, pix=pix,flux=flux,ffit=ffit)   
        if abs(fitcenter-maxpix) gt 6./ybin or peakf lt 0 or (width lt 2./ybin) then $
          fitcenter = getlinecenter(y, mprof, maxpix,      fitrad=10./ybin, peak=peakf, width=width, bkpoly=1, pix=pix,flux=flux,ffit=ffit)   
        if abs(fitcenter-maxpix) lt 6./ybin and (width gt 1.5/ybin and width le 30.0/ybin) then begin          
          ; use the Gaussian fit if the results are reasonable
          peakcen[p] = fitcenter
          peakwidth[p] = width
        endif else begin
          ; use a centroid (or even just the max value) if results are questionable
          centroidpix = total(y[maxpix-5:maxpix+5]*(prof[maxpix-5:maxpix+5] > 0))/total(prof[maxpix-5:maxpix+5] > 0)
          if abs(centroidpix-maxpix) le 2./ybin then peakcen[p] = centroidpix else $
                                                     peakcen[p] = maxpix
          peakwidth[p] = 6.0/ybin
        endelse

      endfor
      peaksig = (smoothprof[peaks]/rms) ; significance, not sigma
      peakdist = abs(peaks-avstdpos) 
      ; Determine which source is most likely to be the target object
      peakrank = (1./sqrt(abs(peaksig-2))) * ((peakdist*4/ybin) > 5.0)^2  ; lower=more likely to be peak, could explore optimal reweighting here
      ; peakrank = (1./abs(peaksig-2)) * (peakdist > 1.0)^2  ; older weighting
      ss = sort(peakrank) 
      slitwidtharcsec = strnum(strmid(hi.slit,1+strpos(hi.slit,'_')))
      slitwidth = slitwidtharcsec/(0.135*hi.spatbin)
      if slitwidth gt 0. then slitfrac = erf(slitwidth/2./peakwidth[ss[0]]/sqrt(2.))  else slitfrac='?' ; assumes a gaussian PSF
      if slitwidth gt 0. then slitloss = 1./slitfrac else slitloss = '?'
      updatefile, profilefile, ['#SIGMA   = '+clip(peakwidth[ss[0]]),'#FWHM    = '+clip(peakwidth[ss[0]]*2.35),$
                                '#SLITFRAC= '+clip(slitfrac),'#SLITLOSS= '+clip(slitloss)]
      objectpos = peakcen[ss[0]]
      objectrad = peakwidth[ss[0]] * 1.8 ; the 'optimum' 1D value is ~1.40 sigma.  ; FWHM would be 2.35 sigma. (full not half!)
      objectsig = peaksig[ss[0]] ; only for contamination checking
      if isstd[i] then objectrad += 2/ybin
      ; bakrad = [0.,0.]
      baksize = 28 ; unbinned pixels (0.135"/pix) in each background window- formely 18
      bakrad = objectrad + 6/ybin + [0,baksize/ybin] ; expanded from [0,12] pix
        ; test to check for contaminators in the background 
      upeaks = where(peakcen gt objectpos+bakrad[0] and peakcen lt objectpos+bakrad[1] and peaksig gt 0.1*objectsig, ctup)
      lpeaks = where(peakcen lt objectpos-bakrad[0] and peakcen gt objectpos-bakrad[1] and peaksig gt 0.1*objectsig, ctlp)
      bakradu = bakrad
      bakradl = bakrad
        ; if present, try a new background region once.
      if ctlp gt 0 then begin
         logprint, '  Background source in lower aperture', format='($,A)'
         bakradl = abs(min(peakcen[lpeaks])-objectpos)+10/ybin+[0,baksize/ybin]
         lpeaks2 = where(peakcen gt objectpos+bakradl[0] and peakcen lt objectpos+bakradl[1] and peaksig gt 0.1*objectsig, ctlp2)
         if objectpos-bakradl[1] le 5 then bakradl[1] = objectpos-5
         ; revert if it's no good
         if objectpos-bakradl[0] le 5 or bakradl[1]-bakradl[0] le 3 then bakradl = bakrad
         if ctlp2 ge ctlp then bakradl = bakrad
         if ctlp2 eq ctlp then if max(peaksig[lpeaks2]) gt 0.5*max(peaksig[lpeaks]) then bakradl = bakrad
         if total(bakradl-bakrad) eq 0 then logprint else logprint, ', adjusting'
      endif
      if ctup gt 0 then begin
         logprint, '  Background source in upper aperture', format='($,A)'
         bakradu = abs(max(peakcen[upeaks])-objectpos)+10/ybin+[0,baksize/ybin]
         upeaks2 = where(peakcen gt objectpos+bakradu[0] and peakcen lt objectpos+bakradu[1] and peaksig gt 0.1*objectsig, ctup2)
         ; if max(bakradu) ge ny-5 then bakradu = bakrad would need to reload the profile file to determine ny for this check
         if ctup2 ge ctup then bakradu = bakrad
         if ctup2 eq ctup then if max(peaksig[upeaks2]) gt 0.5*max(peaksig[upeaks]) then bakradu = bakrad
         if total(bakradu-bakrad) eq 0 then logprint else logprint, ', adjusting'
      endif

      ;window, 18
      ;!p.multi = [0,1,2]
      ;plot, mprof
      ;oplot, prof*max(mprof)/max(prof), color=rgb(100,100,255)
      ;oplot, smoothprof, color=rgb(255,255,0)
      ;for c = 0, n_elements(peakcen)-1 do oplot, peaks[[c,c]], [-1000,10000], color=rgb(0,128,128)
      ;for c = 0, n_elements(peakcen)-1 do oplot, peakcen[[c,c]], [-1000,10000], color=rgb(0,255,0)

      localmed = median([prof[(objectpos-objectrad-26/ybin)>0:(objectpos-objectrad-6/ybin)],$
                         prof[(objectpos+objectrad+6/ybin)<(nm-1):(objectpos+objectrad+26/ybin)<(nm-1)]])
      isnotclean = prof gt (localmed+5*rms) ; profile pixels well above the local median
      nLwinnotclean = total(isnotclean[((objectpos-bakrad[1])>0)     :((objectpos-bakrad[0])>0)     ]) ; #pix contaminated
      nRwinnotclean = total(isnotclean[((objectpos+bakrad[0])<(nm-1)):((objectpos+bakrad[1])<(nm-1))]) ; #pix contaminated
      if nLwinnotclean gt 0.4*(baksize/ybin) then lcont = 1 else lcont = 0
      if nRwinnotclean gt 0.4*(baksize/ybin) then rcont = 1 else rcont = 0
      ;if lcont or rcont then logprint, '   Warning: background contamination for '+repstr(filei,'.fits','')
      ;outstr = clip(repstr(filei,'.fits',''),32)+' '+fpr(objectpos,5.2)+' '+fpr(objectrad,3.2) + '  ' $
      ;        +clip(clip(fpr(bakrad[0],2.1))+'-'+clip(fpr(bakrad[1],2.1)),12) $
      ;        +removepath(tracefiles[i])  $
      ;        +'  # '
      ; convention for bakmin/bakmax is lower, upper in specextract and elsewhere
      tr = {file:filei, ymin:0., ymax:0., yfix:filei, aprad:objectrad, $
            bakmin:[bakradl[0],bakradu[0]], bakmax:[bakradl[1],bakradu[1]], $ 
            tracefile:tracefiles[i], tracefileypos:0., apname:'', addsky:0, comment:'', coord:''}
      tr.file = filei
      tr.yfix = objectpos
      for p = 1, (n_elements(peaks)-1) < 5 do tr.comment += clip(fpr(peakcen[ss[p]],4.1))+' '
      outstr = trsoutstring(tr)
      if objfilechanged eq 0 then logprint, ' Updating object position file...'
      logprint, '  '+outstr
      updatefile, objfile, outstr
      objfilechanged += 1
   endfor

  if (objfilechanged gt 0 or lpar.overwrite) then begin
    if keyword_set(skiprbalign) eq 0 and file_test(brightobjfile) then begin
       lrisalignapertures, camchar=camchar, objfilechanged=objfilechanged
    endif else begin
       print, '  Not aligning red/blue apertures.'
    endelse
  endif

   ; slit capture fraction is:  S[-w to +w]             e^(-x^2/2*sigma^2) dx  / S[-infinity to infinity] e^(-x^2/2*sigma)
   ;   z = (x/sigma)/sqrt(2)    S[-w/sigma/sqrt(2), +]  e^(-z^2)               / S[] e^(-z^2)
   ;                            erf(x/sigma/sqrt(2))  ??
   ; erf([slit half-width]/sigma/sqrt2) * erf([aperture radius]/sigma/sqrt2)
   ; slitloss = 1./(erf(slitwidth/2./sigma/sqrt(2.)) * erf(rad/sigma/sqrt(2.))

   dum = file_search(lpar.spworkingdir1d+'ls'+camchar+'*.profile', count=ntrfileafter) 

   profileplotfile = 'profiles'+camchar+'.ps' ; at top level
   if ntrfileafter gt ntrfilebefore or file_test(profileplotfile) eq 0 or objfilechanged gt 0 then begin
      ; print, ntrfileafter, ntrfilebefore, objfilechanged
      lrisplotprofiles, camera=camchar
      if file_test('profilesr.ps') and file_test('profilesb.ps') then lrisplotprofiles2
   endif

end


; -------------------------

pro autolrisextract, camera=camera, chip=chip, $
   date=date, yblue=yblue, yred=yred, alwaysreaddskyin=alwaysreaddskyin  ; for now these are hidden options

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

    ; visually display the extraction
 
   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   ;if chipchar eq 'l' then begin
   ;   logprint, ' Automatic extraction of left chip not yet supported.'
   ;   return
   ;endif
 
   nsuccess = 0

   alwaysreaddsky = keyword_set(alwaysreaddskyin)

   ; search for science files

   ;ulpar.wildchar = lpar.wildchar  ; this is a temporary measure for now until generalized date filtering is implemented
   ;if n_elements(date) gt 0 then ulpar.wildchar = clip(date)+strmid(ulpar.wildchar,6)

   ;files = choosefiles(camchar+ulpar.wildchar+chipchar+'.fits',spworkingdir+'szfp',spworkingdir+'sfp', ct=ct)
   ;files = file_search(lpar.spworkingdir1d+'ls'+camchar+'*.fits', count=ct)
   files = lrisselectfiles('s1','ls'+camchar,'*','.fits', count=ct)
   if ct eq 0 then return
     ; check for unmatched?  but there can be mutiple traces per file

   fileh = lrisheaderinfo(files)


   ; ----- Default trace position -----

   if n_elements(yblue) gt 0 then lpar.defaulttraceyblue = yblue 
   if n_elements(yred) gt 0 then  lpar.defaulttraceyred  = yred 
   if camchar eq 'b'  then normaltracepos = lpar.defaulttraceyblue
   if camchar eq 'r'  then normaltracepos = lpar.defaulttraceyred

  
   ; ----- Get custom trace positions --------

   ; old method
   ; if file_test('tracepos.dat') $
   ;    tr = loadtraceposfile('tracepos.dat', ct=ntr)

   objfile = lpar.spworkingdir1d+'objectpos.txt'
   userobjfile = 'userobjectpos.txt'

   if file_test(objfile) eq 0 then begin
      logprint, ' No object position file found: ', clip(objfile)
      logprint, ' This is required to extract the spectra.  Run trace step.'
      return
   endif

   tr = loadobjectposfile(objfile, ct=ntr)
   if ntr eq 0 then begin
      logprint, ' Object position file is empty; cannot perform any extractions.  Rerun trace step?'
      return
   endif
   userset = intarr(ntr)
   usertr = loadobjectposfile(userobjfile, ct=nutr)
   for u = 0, nutr-1 do begin
      w = where(usertr[u].file eq tr.file, ctmatch) 
      if ctmatch gt 0 then begin
      if usertr[u].apname eq '' then begin
         ; regular aperture, override what's already in the file
         tr[w[0]] = usertr[u]
         userset[w[0]] = 1
      endif else begin
         ; special (named) aperture
         tr = [tr, usertr[u]]
         userset = [userset, 1]
      endelse
      endif
   endfor

   ; ------ Other final setups ------

   if camchar eq 'r' then grkey = 'GRANAME' else grkey = 'GRISNAME'
   if camchar eq 'r' then filtkey = 'REDFILT' else filtkey = 'BLUFILT'
   if camchar eq 'r' then fltseqkey = 'RFLTSEQF' else fltseqkey = 'BFILTSEQF'

   keywords = ['FILTER','EXPTIME','OBJECT','TARGNAME','DICHNAME','GRNAME',grkey,filtkey,'SLITNAME',$
               'INSTRUME','TELESCOP','DETECTOR',$
               'PROGID','PROGINST','PROGTL1','PROGTL2','PROGTL3','SEMESTER','PROGPI','OBSERVER','OA',$
               'REDUCER','LPIPEVER','HOSTNAME','IDLVER',$
               'ELAPTIME','DATE','UTC','MJD-OBS',$
               'JD','JDMID','JDEND','MJD','MJDMID','MJDEND','DATE_BEG','DATE_END','OBSDATE',$
               'AZ','EL','RA','DEC','HA','ST','EQUINOX','AIRMASS', $
               'SUNELEV','MOONELEV','MOONDIST','MOONPHAS',$
               'PARANGLE','OFFPA','SLITPA','ROTPOSN','ROTPPOSN','BINNING','GAIN', $
               'CAMERA','FLATFILE','FLATTYPE','DICHSEQF',fltseqkey, $
               'FRAMENO','DOMEPOSN','TRAPDOOR','LAMPS','WINDOW','PANE','WINDET', $
               'CURRINST','ADCWAVE0','ADCWAVE1', $ ; ADC keywords
               'WXDOMHUM','WXDOMTMP','WXDWPT','WXOUTHUM','WXOUTTMP','WXPRESS','GUIDFWHM',$
               'NAXIS1','NAXIS2','LTV1','LTV2','XBIN','YBIN','PXSCALE','DISPERSN','SATURATE', $
               'SFILE'+['0','1','2','3','4','5','6','7','8','9'], $
               'COMB'+['0','1','2','3','4','5','6','7','8','9'], $
               'EXPT'+['0','1','2','3','4','5','6','7','8','9'], $
               'UT'+['0','1','2','3','4','5','6','7','8','9'], $
               'AIRMASS'+['0','1','2','3','4','5','6','7','8','9'], $
               'YSHIFT'+['0','1','2','3','4','5','6','7','8','9'], $
               'KOAID'+['0','1','2','3','4','5','6','7','8','9'], $
               'PREPARED','TOTALEXP','NSUM']
   if camchar eq 'r' then keywords  = [keywords,'GRATSEQF','GRANSEQF','WAVELEN']
   keywords = [keywords, 'CWAVE'] 
   ; could do a uniqueness check to avoid duplicates...
   ;stds = where(fileh.exptime le 60. and fileh.maxcounts-fileh.skycounts gt 6000., complement=notstds, ctstd)
   ;if ctstd gt 0 then files = files[[stds, notstds]]   ; do the obvious standards first if possible


   ; -------- Now do the extraction ------------

   unextracted = ['']
   nunextracted = 0

   for f = 0, n_elements(files)-1 do begin
      filename = files[f]
      if filename eq '' then continue

      baseoutfile = fileappend(repstr(filename,'.fits', '.spec'),'x')
      ; normally we would check here if the outfile already exists, but there could be new apertures
      ; specified.  We have to load the objectpos file every time.

      ; theoretically this is inefficient
      hi = lrisheaderinfo(filename)

      skyfile = 1 ; extension number
      crfile = 2

      spatbin = hi.spatbin

    ;   if normaltracepos le 0 then begin
    ;     print, ' Standard trace position unknown; using 1/4 point on CCD.'
    ;     normaltracepos = (hi.ny/4)*hi.spatbin
    ;   endif
    ;   defaultysearch = normaltracepos/spatbin + 22./spatbin*[-1,1]   ; doesn't make much difference when tracing is actually performed.
    ;  if n_elements(defaultysearch) gt 0 then ysearch = defaultysearch ; otherwise specextract default, i.e. the whole chip
      if ntr gt 0 then begin
         ta = where(tr.file eq removepath(files[f]), ct)
         ;if ct ge 2 then begin
         ;   print, 'WARNING:  Multiple entries for ', files[f], ' in tracepos.dat'
         ;   ; in the future, this is OK - write multiple spectra out to extract multiple sources.  ; need to create a second loop
         ;endif
         if ct eq 0 then begin
            ; print, 'error message'
            nunextracted += 1
            unextracted = [unextracted, files[f]]
            extractfail = [extractfail, files[f]]
            continue
            ;print, '         Need to rerun trace step, or specify manually?'
         endif
      endif else begin
         logprint, ' No trace data... the pipeline theoretically supports continuing but please run the trace step to enable good behavior.'
         continue
      endelse

      if hi.grname eq '400/3400' then xsearch=hi.nx*[0.65,0.9] else xsearch=-1
      ; this is critical as the midpoint of xsearch defines the zero position of the trace.
      ; really should be more explicit about this in the future and make xcent a parameter in specextract.
      ; express object.pos positions as x,y values instead of just y.

      for tt = 0, ct-1 do begin

          if n_elements(ycent) gt 0 then delvarx, ycent
          if n_elements(bakdist) gt 0 then delvarx, bakdist
          if n_elements(bakwidth) gt 0 then delvarx, bakwidth
          if n_elements(tracefile) gt 0 then delvarx, tracefile
          if n_elements(traceycent) gt 0 then delvarx, traceycent
          if n_elements(aprad) gt 0 then delvarx, aprad
          if n_elements(fixtrace) gt 0 then delvarx, fixtrace
          if n_elements(ysearch) gt 0 then delvarx, ysearch
          if n_elements(readdsky) gt 0 then delvarx, readdsky

          ; print, '  User-specified trace constraints for ', tr[t].file
          ;  help, tr[t], /str
          t = ta[tt]
          if tr[t].apname ne '' and tr[t].apname ne '0' then outfile = repstr(baseoutfile,'.spec','-'+tr[t].apname+'.spec') else outfile = baseoutfile
          if lpar.overwrite eq 0 and file_test(outfile) then continue

          if tr[t].yfix gt 0 and tr[t].yfix le hi.ny-1 then begin
             ycent = tr[t].yfix
             fixtrace = 1
          endif else begin
             ; formerly could specify a position of 0 (or -1)? to search within specextract.
             ; This is no longer an option.
             print, '  ERROR - extraction location for '+$
                    tr[t].file+((tr[t].apname eq '') ? '' : '|')+tr[t].apname+$
                   ' (y='+clipzero(tr[t].yfix)+')'+$
                   ' is not in the file!'
             continue
          endelse
          if tr[t].ymin ge 0 and tr[t].ymax gt 0 then begin
             ysearch = [tr[t].ymin, tr[t].ymax] 
          endif 
          if max(tr[t].bakmin) ge 0 and max(tr[t].bakmax) ge 0 then begin 
             bakdist = tr[t].bakmin            ;(tr[t].bakmin+tr[t].bakmax)/2.  ; I must have redefined this...?
             bakwidth = (tr[t].bakmax-tr[t].bakmin)
          endif
          if tr[t].tracefile ne '' then    tracefile  = lpar.spworkingdir1d+tr[t].tracefile 
          if tr[t].tracefileypos gt 0 then traceycent = tr[t].tracefileypos
          if tr[t].aprad gt 0 then         aprad      = tr[t].aprad
          if tr[t].addsky gt 0 or alwaysreaddsky then        readdsky    = 1

          logprint, ' Extracting ', removepath(filename), ' ('+clip(hi.target)+')'
          if userset[t] then logprint, '    User-specified trace constraints: ' + $ 
                                   ' y='+clipzero(ycent) + $
                                   ' rad='+clipzero(aprad) + $
                                   ((total(bakdist) gt 0 or total(bakwidth) gt 0) ? ' bak='+$
                                       clip(fpr(tr[t].bakmin[0],2.1))+'-'+clip(fpr(tr[t].bakmax[0],2.1))+'|'+$
                                       clip(fpr(tr[t].bakmin[1],2.1))+'-'+clip(fpr(tr[t].bakmax[1],2.1)) : '') + $
                                   (tr[t].apname eq '' ? '' : '  ['+tr[t].apname+']')
                                  ; ' bak='+clip(traceycent) + $



          errorstatus = 0
          if lpar.debug eq 0 then catch, errorstatus
          if errorstatus ne 0 then begin
             logprint, ' Unexpected error during spectrum extraction:'
             logprint, !error_state.msg
             success = -10
          endif else begin
             if n_elements(tracefile) gt 0 then logprint, '    Trace file: ', removepath(tracefile)

             specextract, filename, ycent, xsearch=xsearch, $ 
                ysearch=ysearch, aprad=aprad, bakdist=bakdist, bakwidth=bakwidth, $
                tracefile=tracefile, traceycent=traceycent, fixtrace=fixtrace, $
                skyfile=skyfile, crfile=crfile, outfile=outfile, keywords=keywords, success=success, $
                setkey=setkey, setval=setval, readdsky=readdsky, maxsaturated=100, /quiet

          endelse
          catch, /cancel 

          ; success:  1=good, 2=questionablewavsol, -1=filenotfound, -2=badfile, -3=tracefail, -4=wavsolfail
          if success lt 0 and success ne -4 then extractfail = [extractfail, filename]
          if success gt 0 then nsuccess += 1
      endfor  ; sub-apertures
   endfor   ; fits files


   catch, errorstatus
   if errorstatus ne 0 then begin
        logprint, '   Error opening display window.'
        logprint, '   Turning off window display.'
        lpar.displayoff = 1
   endif

   if nunextracted gt 0 then logprint, '  ', clip(nunextracted)+' '+strupcase(clip(camera))+' spectra have no trace data and cannot be extracted.'

   if nsuccess gt 0 then begin
      ; update the plots if any new extractions performed
      if lpar.displayoff eq 0 then lrisplottraces, cam=camera   
      lrisplotprofiles, camera=camchar  ; user could have changed the object positions.
   endif

   catch, /cancel 

end

; -------------------------

pro autolriswavapply, camera=camera, chip=chip, altarcsoldir=altarcsoldir

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   nsuccess = 0

   ;if chipchar eq 'l' then return

   ; search for science files

   ;ulpar.wildchar = lpar.wildchar  ; this is a temporary measure for now until generalized date filtering is implemented
   ;if n_elements(date) gt 0 then ulpar.wildchar = clip(date)+strmid(ulpar.wildchar,6)
   ;files = choosefiles(camchar+ulpar.wildchar+chipchar+'.fits',spworkingdir+'szfp',spworkingdir+'sfp', ct=ct)
   ;files = file_search(lpar.spworkingdir1d+'xls'+camchar+'*.spec', count=ct)
   files = lrisselectfiles('s1','xls'+camchar,'*','.spec', count=ct)
   if ct eq 0 then return
   ; could also add a fast-exit here if all outfiles exist.

   fileh = lrisheaderinfo(files)

   ; search for arc solutions
   
   ; We want both a 'nightly' file (common to the configuration) and a 'local' file (possibly taken during the night)
   ; with the latter only used for the linear adjustment.

   ;arcsolfiles = choosefiles(camchar+ulpar.wildchar+chipchar+'_arc.sol',lpar.spworkingdir1d+'fp', ct=ct)
   arcsolfiles = lrisselectfiles('a','fp'+camchar,chipchar,'*_arc.sol', count=ct, /nofilter) ; note the star
   for a = 0, n_elements(altarcsoldir)-1 do begin
     more = file_search(altarcsoldir[a]+'fp'+camchar+'*'+chipchar+'*_arc.sol', count=ctmore) ; a more primitive method
     ct = ct+ctmore
     if ctmore gt 0 then arcsolfiles = [arcsolfiles, more]
   endfor
   if ct eq 0 then begin
      ;logprint, ' '+lpar.spworkingdir1d+'fp', camchar+ulpar.wildchar+chipchar+'.sol'
      logprint, ' No '+camera+' arc solutions found.  Cannot extract wavelength-calibrated spectra.'
      print
      wavsolfail = [wavsolfail, files]

      return
   endif
   narcsol = n_elements(arcsolfiles)
   arcfitsfiles = lpar.arcdir+removepath(repstr(arcsolfiles,'.sol','.fits'))
   arcfitsfilesexist = bytarr(narcsol)
;   for f = 0, narcsol-1 do begin
;      if file_test(arcfitsfiles[f]) then arcfitsfilesexist[f] = 1 
;      if arcfitsfilesexist[f] eq 0 then begin
;         logprint, '    Data file ', arcfitsfiles[f], ' corresponding to ', removepath(arcsolfiles[f]), ' is missing - cannot use this solution.'
;         ; Is it really necessary anymore?
;      endif
;   endfor
;   if total(arcfitsfilesexist eq 0) gt 0 then $
;         logprint, '  Must rerun wavcal step to regenerate these files.'


   arch = lrisheaderinfo(arcsolfiles, badarclist=badarclist) ; arc metadata
   if camchar eq 'b' then arcnlamps = arch.hg + arch.cd + arch.zn    ; +arc.fe
   if camchar eq 'r' then arcnlamps = arch.hg + arch.neon + arch.ar  ; + arc.kr + arc.xe


           ;'~/research/Useful/skyspec/gident_most.dat'
   skylist = lpar.refdatapath + '/' + 'skylines.dat'
   telspec = lpar.refdatapath + '/' + 'telluricabs.dat'

   updated = 0
   for f = 0, n_elements(files)-1 do begin
      filename = files[f]
      if filename eq '' then continue
      outfile = lpar.spworkingdir1d + 'e'+strmid(removepath(files[f]),1)  ; e replaces x.  Different convention from normal for back-compatibility.
      if lpar.overwrite eq 0 and file_test(outfile) then continue

      hi = fileh[f]

      if n_elements(arcfile) gt 0 then delvarx, arcfile
      if n_elements(skyfile) gt 0 then delvarx, skyfile
      if n_elements(initsol) gt 0 then delvarx, initsol
      ; if n_elements(extractarcfile) gt 0 then delvarx, extractarcfile ; deprecated?
      skyrefine = 1 ; on by default
      telrefine = 0

      shiftwarn = 0
      ; find matching arcs
      matches = where(arch.grname eq hi.grname and $  ; and arch.binning eq hi.binning and arch.window eq hi.window 
                    (abs(arch.cwave-hi.cwave) lt 10.) and $  
                    (arch.dichroic eq hi.dichroic or camchar eq 'r'), ct)

      if ct eq 0 then begin
         ; embark on the dangerous game of using an arc taken in a different wavelength setting.
         matches = where(arch.grname eq hi.grname and $  ; and arch.binning eq hi.binning and arch.window eq hi.window 
                       (abs(arch.cwave-hi.cwave) lt 150.) and $  
                       (arch.dichroic eq hi.dichroic or camchar eq 'r'), ct)
         shiftwarn = 1
      endif

      if ct eq 0 then begin
             ; logprint, 'WARNING: Found no useable arcs in the same configuration as ', removepath(filename), ' ('+hi.target+')'
             ; logprint, '     gr='+hi.grname, ' bin='+repstr(hi.binning,',','x'), ' win='+hi.win, ' cwav='+clip(hi.cwave), ' dich='+clip(hi.dichroic)
             ;   print, filename, ' ', gr, ' ', binning, ' ', window, ' ', cwav, ' ', dich
             printarcs = 0
             if printarcs then for a = 0, n_elements(arch)-1 do begin
                logprint, removepath(arcsolfiles[a]), ' gr=', arch[a].grname, ' bin=', arch[a].binning, ' win=', arch[a].win, ' cwav=', clip(arch[a].cwave), ' dich=', clip(arch[a].dichroic)
             endfor
             logprint, '  No solved arcs matching '+removepath(filename)+'!  Skipping...'
             wavsolfail = [wavsolfail, filename]
             continue
      endif 

      logprint, ' Wavelength-calibrating ', removepath(files[f])

      if shiftwarn then begin
         logprint, '  WARNING: arc matching '+removepath(filename)+' has a different central wavelength ('+clip(fix(arch[matches[0]].cwave))+' vs '+clip(fix(hi.cwave))+')'
      endif

      ; get the 'initial' arc from the start of the night
      ; for now just look at the first with all lamps, but could see if there is a way to decide on the best one

      w = where(arcnlamps[matches] ge min([max(arcnlamps[matches]),5]) and abs(arch[matches].cwave-hi.cwave) lt 0.1, ct) 
             ; a better check might be arch[matches].granglef eq hi.granglef
      if ct eq 0 then begin
          logprint, '  WARNING: no arc was taken at this precise grating angle.  This can lead to imperfect wavelength/flux calibration.'
          w = where(arcnlamps[matches] eq max(arcnlamps[matches]), ct) 
      endif
      winit = matches[w[0]]
      timediffinit = arch[winit].mjd - hi.mjd

      ; get the 'local' arc, possibly taken during the night
      ; if there is no local arc it ends up being the master arc
 
      archere = 0
      w = where(abs(arch[matches].radeg-hi.radeg) lt 0.1 and abs(arch[matches].decdeg-hi.decdeg) lt 0.1 and $
                abs(arch[matches].mjd-hi.mjd) lt 10./60./24., ct) 
                 ; first see if an arc was taken here and now
      if ct gt 0 then begin
          wlocal = max(matches[w]) ; major bug fixes
          archere = 1
      endif else begin
          ; same filter, if possible (does the filter affect the wavelength solution???)
          ; w = where(arch[matches].filter eq hi.filter, ct)
          ; if ct gt 0 then matches = matches[w]

          ; get an arc not near horizon, if we can.
          w = where(arch[matches].airmass lt 3.0, ct)
          if ct gt 0 then matches = matches[w]

          ; as many lamps as possible
          w = where(arcnlamps[matches] eq max(arcnlamps[matches]), ct) 
          matches = matches[w]

          ; at this exact central wavelength, if possible
          w = where(abs(arch[matches].cwave-hi.cwave) lt 0.1, ct)
          if ct gt 0 then matches = matches[w]

          ; this configuration-move, if possible
          w = where(arch[matches].gratf eq hi.gratf, ct)
          if ct gt 0 then matches = matches[w]
          w = where(arch[matches].granglef eq hi.granglef, ct)
          if ct gt 0 then matches = matches[w]

          ; within 10 Ang if possible.
          w = where(abs(arch[matches].cwave-hi.cwave) lt 10, ct)
          if ct gt 0 then matches = matches[w]

          ; within 100 Ang if possible.
          w = where(abs(arch[matches].cwave-hi.cwave) lt 100, ct)
          if ct gt 0 then matches = matches[w]

          ; as close to this elevation as possible.  [ really should look at drive angle as well if not zenith ]
          w = where(abs(arch[matches].airmass - hi.airmass) eq min(abs(arch[matches].airmass-hi.airmass)))
          if ct gt 0 then matches = matches[w]

          ; as close to this time as possible.
          w = where(abs(arch[matches].mjd - hi.mjd) eq min(abs(arch[matches].mjd-hi.mjd)))
          if ct gt 0 then matches = matches[w]

          wlocal = matches[0] ; previously [max(w)] which was surely wrong.

             ; could also write info about the fit to the solfile and choose the best-fitting arc.

      endelse
      delvarx, w

      ; Need a tool for wavelength refinement using only the mercury lamp (for blue morning twilight standards).
      ; Probably best done within autowavcal

      arcinitfitsfile = arcfitsfiles[winit]
      arcinithi = arch[winit]
      solinitfile = arcsolfiles[winit]
      
      arcfitsfile = arcfitsfiles[wlocal]
      archi = arch[wlocal]
      solfile = arcsolfiles[wlocal]

      arcsol = grabcolumn(solfile,0)
      for o = 0, n_elements(arcsol)-1 do arcsol[o] *= (hi.specbin*1./archi.specbin)^o
      if solinitfile ne solfile then begin
             ; use a common set of higher-order terms.  Only the first term (c.wav.) is from the local arc.
             logprint, '  Loading arc solutions from '+removepath(solinitfile)+' and '+removepath(solfile)
             arcinitsol = grabcolumn(solinitfile,0)
             for o = 0, n_elements(arcinitsol)-1 do arcinitsol[o] *= (hi.specbin*1./arcinithi.specbin)^o
             initsol = [arcsol[0], arcinitsol[1:*]]
             ; print, 'Reference cwave='+clip(arcinitsol[0])+',  local cwave='+clip(arcsol[0])+', diff = '+clip(arcsol[0]-arcinitsol[0])
      endif else begin
             initsol = arcsol
             logprint, '  Loading arc solution from '+removepath(solfile)
      endelse

      setkey = ['ARCFILE0', 'SOLFILE0', 'ARCFILE', 'SOLFILE']
      setval = removepath([arcinitfitsfile, solinitfile, arcfitsfile, solfile])
      setcom = [' Primary wave calibration arc file',' Primary wave calibration solution',$ 
                ' Secondary wave calibration arc file',' Secondary wave calibration solution']
          ; previously these appended onto a blank string.  Unsure why - only one operation.


      dispersion = hi.dispersion
      cwav = hi.cwave

      if camchar eq 'r' and hi.exptime le 18.0 then  skyrefine = 0  ; don't sky-refine short exposures  (formerly 9R, 16B)
      if camchar eq 'b' and hi.exptime le 18.0 then skyrefine = 0  
      if camchar eq 'b' and hi.sunelev gt -11 then skyrefine = 0   ; don't sky-refine twilight exposures.  
      if camchar eq 'r' and hi.sunelev gt -11 then skyrefine = 0 ; formerly -6, then -9.  Not sure how good the sunelev estimate is.
      if cwav lt 6600. and hi.exptime lt 20. and hi.sunelev gt -13 then skyrefine = 0
                                                                    ; a frame with sunlv=-9.6 (airmass 2) is definitely not usable.
                                                                    ; probably should also consider airmass.
      ; arc refinement based on limited set of lines?
           ; if skylist eq '' and archere then begin
               ; in development
           ;     if file_test(arcfile) then skyfile = arcfitsfile 
           ;     skylist = lines
           ; endif
                                   ;This is conservative, better solar-subtraction will enable improvement.

      if camchar eq 'r' then maxnudge = 220.  ; maximum flexure adjustments in Ang
      if camchar eq 'b' then maxnudge = 15.  

      uncrefwav = 0.3   ; still needed?

      success = 0
      if skyrefine then begin
          logprint, '  Correcting wavelength solution using sky lines...'
          specwavcal, files[f], outfile=outfile, /skyrefine, skylist=skylist, $
                    initsol=initsol, uncrefwav=uncrefwav, $
                    maxnudge=maxnudge, success=success, /quiet, $
                    setkey=setkey, setval=setval, setcom=setcom
                    ;unccwav=cwav*0.04, $ minorder=7, maxorder=7, $
                    ;cwav=cwav, dispersion=dispersion, uncdisp=0.075*dispersion, 
      endif 
      if success le 0 then begin ; sky solution not attempted or failed such that nothing was written
         ; press on anyway with the uncorrected arc
         specwavcal, files[f], outfile=outfile, initsol=initsol, success=success, $
                    setkey=setkey, setval=setval, setcom=setcom
      endif

      ; success code is not yet documented, but 1 seems to indicate successful sky solution. A positive value
      ; that is not 1 indicates success but either no sky solution or a problematic one.

      if success ne 1 and camchar eq 'r' then begin 
          ; sky line correction was not attempted or is questionable.
          ; is this spectrum suitable for telluric correction?
          temp = readspec(outfile)
          ;aband = where(temp.wav gt 7600 and temp.wav lt 7700, cta)
          ;bband = where(temp.wav gt 6840 and temp.wav lt 6910, ctb)
          ;if cta gt 10 then if max(temp[aband].wav)-min(temp[aband].wav) gt 75 then if median(temp[aband].flux) gt 500 then telrefine = 1
          ;if ctb gt 10 then if max(temp[bband].wav)-min(temp[bband].wav) gt 50 then if median(temp[bband].flux) gt 500 then telrefine = 1
           if ((min(temp.wav) le 7530-100 and max(temp.wav) ge 7740+100) or $
               (min(temp.wav) le 6840-100 and max(temp.wav) ge 6910+100)) then telrefine = 1
          if telrefine then begin     
            logprint, '  Correcting wavelength solution using telluric lines...'
            specwavcal, outfile, outfile=outfile, /telrefine, telspec=telspec, $
                    initsol=initsol, maxnudge=maxnudge, success=success, $
                    setkey=setkey, setval=setval, setcom=setcom
            ; This overwrites the file which we normally avoid, since an extremely
            ; well-timed interrupt could prevent this connection on next run.
          endif
      endif
      if success eq -4 then wavsolfail = [wavsolfail, filename]
      if success eq 2 then wavsolwarn = [wavsolwarn, filename]
      if success eq 0 then wavsolfail = [wavsolfail, filename] ; formerly warn??
      if success gt 0 then updated = 1
   endfor


   ; Produce plots of all the sky spectra.

   if updated then begin
    plotfilename = 'skyplots'+camchar+'.ps'
    psopen, plotfilename, xsize=8, ysize=11, /inches
    device, /color
    !p.font=0
    device, /times, font_index=17                                  
    colors = transpose([[0,0,0],$
                      [255,50,50],$ 
                      [150,50,50],$ 
                      [170,170,170]])
    tvlct, colors
    nperpage = 8
    !p.multi = [0,1,nperpage]

    escifiles = file_search(lpar.spworkingdir1d+'els'+camchar+'*.spec')
    sh = lrisheaderinfo(escifiles)
    gr = unique(sh.grname) ; should also consider the possibility of highly different CWAV values
    readcol, skylist, idn, allreflinewav, allreflinestr, format='i,f,f', /silent
   
    j = 0
    for g = 0, n_elements(gr)-1 do begin
      grstr = gr[g] ;repstr(gr[g],'/','-')
      w = where(sh.grname eq gr[g], nscig)

      for i = 0, nscig-1 do begin
        sp = readspec(escifiles[w[i]])  ;, /quiet)
        minwav = min(sp.wav)
        maxwav = max(sp.wav)

        ; mimic the line selection in specwavcal
        wlr = where(allreflinestr gt 100. and allreflinewav gt minwav and allreflinewav lt maxwav)
        skylinewav = allreflinewav[wlr]
        skylinestr = allreflinestr[wlr]
        nline = 100
        if maxwav lt 7000. then nline = 20
        if maxwav lt 6000. then nline = 5
        brightskywav = findbrightest(skylinewav, skylinestr, (nline/5)>3)   
        moreskywav = findbrightest(skylinewav, skylinestr, nline)
 
        !p.position = [0.04, 0.88-0.12*j, 0.99, 0.98-0.12*j]  
        
        z = where(sp.wav gt minwav+100 and sp.wav lt maxwav-200)
        yrange = [(-100 > min(sp[z].sky)) < 0,max(sp[z].sky)*1.05]
        plot, [0], [0], xrange=[minwav > 2900,maxwav], /xsty, yrange=yrange, /ysty
        xyouts, 0.07, 0.96-0.12*j, /norm, repstr(removepath(escifiles[w[i]]),'.spec',''), charsize=0.6
        xyouts, 0.07, 0.945-0.12*j, /norm, grstr, charsize=0.55

        for l = 0, n_elements(moreskywav)-1 do begin
           oplot, moreskywav[l]*[1,1],[-100,1e6], color=3
        endfor       
        for l = 0, n_elements(brightskywav)-1 do begin
           oplot, brightskywav[l]*[1,1],[-100,1e6], color=2, thick=2
        endfor       
        oplot, sp.wav, sp.sky
        j = (j + 1) mod nperpage
      endfor
    endfor 

    psclose
    retitleplot, plotfilename, (camchar eq 'b' ? 'Blue' : 'Red')+' sky spectra'
    !p.multi = 0
   endif

end


; -------------------------

pro autolrisresponse, camera=camera, chip=chip

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   if chipchar eq 'l' then return

   ;if n_elements(files) eq 0 then $
   ;    files = file_search(lpar.spworkingdir1d+'els'+camchar+'*.spec', count=ct)
   files = lrisselectfiles('s1','els'+camchar,'*','.spec', count=ct)
   if ct eq 0 then return

   imh = lrisheaderinfo(files)

   ; identify flux calibrators.

   updated = 0
   nstdobs = 0
   for f = 0, n_elements(files)-1 do begin

      readstandard, imh[f].target, pos=[imh[f].radeg, imh[f].decdeg], stdname=stdname, /check, result=result, path=lpar.refdatapath+'/'
      if result eq 0 then continue

      nstdobs = nstdobs + 1

      ;outrespfile = lpar.responsedir+removepath(repstr(files[f],'.spec','.response'))
      datestr = strmid(repstr(imh[f].date,'-',''),2,6)
      if lpar.koa then divpos = '.' else $
                       divpos = '_'
      if lpar.koa then filestr = imh[f].koaframe else $ 
                       filestr = string(imh[f].framenum,'(I04)')
      outrespfile = lpar.responsedir+'els'+camchar+datestr+divpos+filestr+'_'+legalfilestr(stdname)+'.response'

      if file_test(outrespfile) and lpar.overwrite eq 0 then begin 
         nstdobs = nstdobs + 1
         continue
      endif

      if file_test(lpar.responsedir) eq 0 then begin 
         spawn, 'mkdir '+lpar.responsedir
         if file_test(lpar.responsedir) eq 0 then begin
            logprint, ' ERROR: Cannot create '+lpar.responsedir
         endif
      endif
      lrisresponse, files[f], imh=imh[f], updated=updated, outfile=outrespfile
   endfor


   if nstdobs eq 0 then begin  
      print, ' Cannot find any '+camera+' standard-star observations.'
   endif

   ; make a summary plot

   if updated then begin
      lrisplotresponses, camchar
   endif


end

; ------------------------

; -------------------------

pro autolrisfluxcal, camera=camera, chip=chip, notelluric=notelluric, reducestandards=reducestandards, altresponsedir=altresponsedir

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   standardpath = lpar.refdatapath
   missingstd = 0

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   if chipchar eq 'l' then return

   files = lrisselectfiles('s1','els'+camchar,'*','.spec', count=nf)
   if nf eq 0 then return

   imh = lrisheaderinfo(files)
   issci = intarr(nf)

   for f = 0, nf-1 do begin
      h = headspec(files[f])

      readstandard, imh[f].target, pos=[imh[f].radeg, imh[f].decdeg], stdname=stdname, /check, result=result, path=standardpath+'/'
      ;print, imh[f].target, imh[f].radeg, imh[f].decdeg, result

      if (~result) or keyword_set(reducestandards) then issci[f] = 1
   endfor

   wscience = where(issci, nsci)
   if nsci eq 0 then begin
      logprint, ' Found no science observations!' ; only standard observations
      return
   endif
   files = files[wscience]
   imh = imh[wscience]

   
   respfiles = lrisselectfiles('r','els'+camchar,'*','.response', count=ctfluxcal, /nofilter, imh=respimh)
   for d = 0, n_elements(altresponsedir)-1 do begin
     addfiles = file_search(altresponsedir[d]+'els'+camchar+'*'+'.response', count=ctadd)
     if ctadd gt 0 then respfiles = [respfiles, addfiles]
   endfor
              

   if ctfluxcal eq 0 then begin
      logprint, ' Found no LRIS-'+strupcase(camchar)+' response files.  Run response step or acquire standard-star observations.'
      fluxcalfail = [fluxcalfail, files] ; add all of them to the failure list
      return
   endif

   stdnames = strarr(ctfluxcal)
   greatstandardlist = ['G191','BD+28','Feige34','Feige67','Feige110']
   greatstandard = bytarr(ctfluxcal)
   for rf = 0, ctfluxcal-1 do begin
      readstandard, respimh[rf].target, pos=[respimh[rf].radeg, respimh[rf].decdeg], stdname=stdname, /check, result=result, path=standardpath+'/'
      stdnames[rf] = stdname 
      for j = 0, n_elements(greatstandardlist)-1 do $
         if strpos(stdname,greatstandardlist[j]) ge 0 then greatstandard[rf] = 1
      h = headspec(respfiles[rf])
   endfor

   for s = 0, nsci-1 do begin
      filename = files[s]

      outfile = fileappend(filename,'c')
      if file_test(outfile) and lpar.overwrite eq 0 then continue

      sameconfig = respimh.filter eq imh[s].filter and respimh.grname eq imh[s].grname and $
                   respimh.slit eq imh[s].slit and respimh.dichroic eq imh[s].dichroic
      similarconfig = respimh.filter eq imh[s].filter and respimh.grname eq imh[s].grname and $
                      respimh.dichroic eq imh[s].dichroic  ; allow different slits
                ; always required
      airdiff = abs(respimh.airmass-imh[s].airmass)
      timediff = abs(respimh.mjd - imh[s].mjd)/24. ; hours
      cwavdiff = abs(respimh.cwav - imh[s].cwav)

       ; Try to determine appropriate standards to use.
       ; other things to check would be whether or not the wavelength solution is precise, how bright twilight was,
       ; and the actual wavelength solution 
       ; Using hard cuts like this is non-optimal - should instead try a goodness metric.
       ; Also, should try to choose the same standard for both observations of the same target...

         matchstandards = where(sameconfig and cwavdiff lt 10. and $
                                timediff lt 3. and airdiff lt 0.35 and greatstandard, nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(sameconfig and cwavdiff lt 15. and $
                                timediff lt 5. and airdiff lt 0.35 and greatstandard, nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(sameconfig and cwavdiff lt 15. and $
                                timediff lt 5. and airdiff lt 0.6 and greatstandard, nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(sameconfig and cwavdiff lt 15., nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(similarconfig and cwavdiff lt 15., nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(similarconfig and cwavdiff lt 50., nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(similarconfig and cwavdiff lt 250., nmatchstd) ; formely 200
      ; progressively search for standards with less and less similar central wavelength (since note that LRIS shifts)

      if nmatchstd eq 0 then begin
         logprint, ' No matching standard for ', removepath(files[s]), ' ('+imh[s].grname+', D'+clip(imh[s].dichroic)+', '+imh[s].filter+', '+clip(imh[s].slit)+', cw='+clip(fix(imh[s].cwav))+')'
         fluxcalfail = [fluxcalfail, files[s]]
         missingstd = missingstd + 1 ; not used right now
         continue
      endif

      lrisfluxcal, files[s], respfiles[matchstandards], stdname=stdnames[matchstandards], outfile=outfile

   endfor

end

; -----------------------------

pro autolriscombine, camera=camera, live=live

   common lrisauto, lpar


   standardpath = lpar.refdatapath

   ; automatic combination of 1D spectra of a common source in a common setting.

   ; this function is largely unnecessary now that things are summed in 2D space.
   ; However, it's still useful if observations aren't summed and you want them to be summed.

   ; assumes the central wavelength wasn't wildly adjusted between exposures.
 
   ; live mode doesn't do anything yet, and hasn't been tested at any rate.

   ; not used in a long time and keyword management may be obsolete / may not work

   camchar = strmid(camera,0,1)
   chipchar = 'r'

   ;files = choosefiles(camchar+lpar.wildchar+chipchar+'.spec',lpar.spworkingdir1d+'ceszfp',lpar.spworkingdir1d+'cesfp')
   ;files = file_search(lpar.spworkingdir1d+'cels'+camchar+'*.spec', count=ct)
   files = lrisselectfiles('s1','cels'+camchar,'*','.spec', count=ct)

   nf = n_elements(files)

   if nf eq 1 and files[0] eq '' then return

   targets = strarr(nf)
   igrs = strarr(nf)
   ; worry about different cwavs?  probably not since it can shift slightly.

   for f = 0, n_elements(files)-1 do begin
     filename = files[f]
     h = headspec(filename)
     target = clip(sxpar(h,'TARGNAME'))
     if strpos(target,'_S') gt 0 then begin  ; strip the offset star from the target info
        ustriptarget = strmid(target,0,strpos(target,'_', /reverse_search))
        ;if ustriptarget ne 'HD' and ustriptarget ne 'BD' and ustriptarget ne 'PTF' then 
        target = ustriptarget
     endif
     targets[f] = target
     inst = clip(sxpar(h,'INSTRUME'))
     gr = clip(sxpar(h,'GRNAME'))
     incamchar = 'u' ; unknown
     if inst eq 'LRIS' then incamchar = 'r'
     if inst eq 'LRISBLUE' then incamchar = 'b'
     igrs[f] = incamchar + gr ; this is sort of obsolete now, but allows this to work with camera='?'
   endfor

   targetlist = unique(targets)
   ntargets = n_elements(targetlist)

   for t = 0, ntargets-1 do begin 
      target = targetlist[t]
      readstandard, target, /check, result=isstandard, path=standardpath+'/'
      if isstandard then continue ; don't stack observations of flux standards

      wtarget = where(targets eq target, nexp)
      
      ; determine intra-color normalization

      igrlist = unique(igrs[wtarget])
      ngr = n_elements(igrlist) 

      minwav = fltarr(ngr)
      maxwav = fltarr(ngr)
      for g = 0, ngr-1 do begin
         igr = igrlist[g]
         cam = strmid(igr,0,1)
         gr  = strmid(igr,1)
               ; make this a legal filename.  should really create a funciton to do this.
         ;outfile =  lpar.spworkingdir1d+'vc'+repstr(repstr(repstr(target,'_',''),'*',''),'?','')+'_'+repstr(igr,'/','-')+'.spec'

         wtargigr = where(targets eq target and igrs eq igr, ngexp)
         if ngexp le 1 then continue
         outfile = files[wtargigr[0]]
         for si = 2, 10 do outfile = repstr(outfile,'_s'+clip(si)+'.spec','.spec')
         outfile = repstr(outfile,'.spec','_sa.spec')
         if file_test(outfile) and lpar.overwrite eq 0 then continue
         logprint, outfile
            

         fluxnormval = fltarr(ngexp)
         totexp = 0.
         logprint, ' Combining ', format='($,A)'
         c = 0
         for i = 0, ngexp-1 do begin
            f = wtargigr[i]
            logprint, removepath(files[f])+' ', format='($,A)'
            s = readspec(files[f], h)
            inexp = sxpar(h,'ELAPTIME')

            if i eq 0 then sums = s
            if i gt 0 then begin
               sums = addspec(sums, s, w1=totexp, w2=inexp)
            endif
            totexp += inexp

            ; header info 
            if i eq 0 then begin
               outh = h  
            endif

            sxaddpar, outh, 'CFILE'+clip(i+1), files[f], i eq 0 ? 'summed 1D spectrum file name' : ''

            ;if ngexp gt 1 then begin
            sxaddpar, outh, 'APRAD'+clip(i+1), sxpar(h,'APRAD')
            sxaddpar, outh, 'YEXTRAC'+clip(i+1), sxpar(h,'YEXTRACT')
            sxaddpar, outh, 'EXPTIME'+clip(i+1), inexp
            ;endif

            for cc = 0, 9 do begin
                combcc = sxpar(h, 'COMB'+clip(cc+1), count=ctcc)
                ; if cc eq 0 then print, cc+1, combcc, ctcc
                if ctcc gt 0 then begin
                   sxaddpar, outh, 'COMB'+clip(c+1), combcc, removepath(files[f]), after='COMB'+clip(c)
                   sxaddpar, outh, 'EXPT'+clip(c+1), sxpar(h,'EXPT'+clip(cc+1)), after='EXPT'+clip(c)
                   sxaddpar, outh, 'AIRMASS'+clip(c+1), sxpar(h,'AIRMASS'+clip(cc+1)), after='AIRMASS'+clip(c)
                   sxaddpar, outh, 'UT'+clip(c+1), sxpar(h,'UT'+clip(cc+1)), after='UT'+clip(c)
                   c += 1
                endif
            endfor


         endfor
         sxaddpar, outh, 'NSUM', ngexp
         sxaddpar, outh, 'DATE_END', sxpar(h, 'DATE_END')
         sxaddpar, outh, 'AIRMASS', mean([sxpar(h, 'AIRMASS') , sxpar(outh,'AIRMASS')])
         sxaddpar, outh, 'AZ', mean([sxpar(h, 'AZ') , sxpar(outh,'AZ')])
         sxaddpar, outh, 'EL', mean([sxpar(h, 'EL') , sxpar(outh,'EL')])
         sxaddpar, outh, 'EXPTIME',  totexp
         sxaddpar, outh, 'TOTALEXP', totexp
         logprint
         logprint, '           -> ', removepath(outfile)

         writespec, outfile, sums, outh

      endfor

   endfor

end



pro autolrisconnect, files=files, noscale=noscale, live=live, datenames=datenames, requirepair=requirepair

   common lrisauto, lpar

   ; This routine needs to be redesigned.
   ; Probably should scan through the blues in reverse order of exposure time and identify overlapping red matches, 
   ; checking them off.

;   if n_elements(files) eq 0 then $
;      files = file_search(lpar.spworkingdir1d+'vc*.spec')
;       files = file_search(lpar.spworkingdir1d+'cels*.spec', count=ct)
   files = lrisselectfiles('s1','cels?','*','.spec', count=ct)

   files = files[sort(files)]
   nf = n_elements(files)

   if nf eq 1 and files[0] eq '' then return

   targets = strarr(nf)
   dates = strarr(nf)

   for f = 0, n_elements(files)-1 do begin
     filename = files[f]
     h = headspec(filename)
     target = clip(sxpar(h,'TARGNAME'))
     if strpos(target,'_S') gt 0 then begin
        posstarget = strmid(target,0,strpos(target,'_S'))  ; offset stars
        if posstarget ne 'HD' and posstarget ne 'BD' and posstarget ne 'PTF' then target = posstarget
     endif
     targets[f] = target 
     dates[f] = strmid(repstr(clip(sxpar(h,'DATE')),'-',''),0,8)
   endfor

   targetlist = unique(targets)
   ntargets = n_elements(targetlist)
   nconnected = 0

   for t = 0, ntargets-1 do begin 
      target = targetlist[t]

      wtarget = where(targets eq target, nexp)
      vcfiles = files[wtarget] ; all observations of this target
      if keyword_set(datenames) then datestr = dates[wtarget[0]]+'_' else datestr = ''

      filesuff = datestr+legalfilestr(target,sp='_')+'.spec'
      outfile = lpar.spfinaldir+'lris'+filesuff
      outbfile = lpar.spfinaldir+'lrisb'+filesuff
      outrfile = lpar.spfinaldir+'lrisr'+filesuff

      if file_test(outfile) and lpar.overwrite eq 0 then continue
      plotfile = lpar.spworkingdir1d+'rescale_'+legalfilestr(target,sp='_')+'.eps'

      if nexp gt 2 then begin 
          ; Check for 1D sums (_sa means sum of everything)
          ; old capability, only crudely supported (not tested, may not work)
          hinfo = lrisheaderinfo(vcfiles) ; unnecessarily does this twice
          cams = hinfo.camera
          wb = where(cams eq 'blue' and strpos(vcfiles,'_sa.spec') gt 0, ctb)
          wr = where(cams eq 'red' and strpos(vcfiles,'_sa.spec') gt 0, ctr)
          if ctb eq 1 and ctr eq 1 then begin
              vcfiles = vcfiles[[wb[0], wr[0]]] 
              nexp = 2
              ; just use the all-summed blue and all-summed red if 1D summed.
          endif
      endif

      hinfo = lrisheaderinfo(vcfiles)
      exptimes = float(hinfo.exptime)
      cams = hinfo.camera

      wb = where(cams eq 'blue', ctb)
      wr = where(cams eq 'red', ctr)

      if ctb eq 0 or ctr eq 0 then begin
         ; either all red or all blue files are missing for this target
         if ctb gt 0 then outfile = outbfile
         if ctr gt 0 then outfile = outrfile
         if file_test(outfile) and lpar.overwrite eq 0 then continue
         logprint, ' No match for: '
         for j = 0, nexp-1 do logprint, '   '+removepath(vcfiles[j])+' '+fpr(hinfo[j].exptime,5.1)+'s'+'  '+hinfo[j].grname+'  '+clip(fix(hinfo[j].cwave))
         if keyword_set(requirepair) then continue
          
         ; trivially 'connect' unpaired files
         if nexp eq 1 then begin
            if file_test(outfile) and lpar.overwrite then file_delete, outfile
            lrisconnect, vcfiles[0], out=outfile
            nconnected += 1
         endif else begin
           for i = 0, nexp-1 do begin
             if cams[i] eq 'blue' then eoutfile = repstr(outbfile,'.spec','.'+clip(i)+'.spec')
             if cams[i] eq 'red'  then eoutfile = repstr(outrfile,'.spec','.'+clip(i)+'.spec')
             if file_test(eoutfile) and lpar.overwrite eq 0 then continue
             lrisconnect, vcfiles[i], out=eoutfile
             nconnected += 1
           endfor
         endelse
      endif else begin
        ; both red and blue files exist for this target

        if ctb eq 1 and ctr eq 1 then begin   
           ; the simplest and most common case: one red and one blue file
           lrisconnect, vcfiles[wb[0]], vcfiles[wr[0]], out=outfile, plot=plotfile, displayoff=lpar.displayoff
           nconnected += 1
           if outfile ne outbfile and file_test(outbfile) then file_delete, outbfile
           if outfile ne outrfile and file_test(outrfile) then file_delete, outrfile
        endif else begin

          logprint, ' Multiple summed exposures exist for '+target+' ('+clip(ctb)+' blue, '+clip(ctr)+' red)'
          if ctb ne ctr then begin
             logprint, '  Warning:  Some exposures will not be connected.'
             logprint, '            Sum in 1D space, or delete unwanted frames?'
          endif
          for i = 0, (ctb < ctr)-1 do begin
             ; Things get complicated if there are different numbers of spectra even at this stage.
             ; Loop over the shorter list and match each with the longest overlapping exposure on the longer list.

             eoutfile = repstr(outfile,'.spec','.'+clip(i)+'.spec')
             eoutbfile = repstr(outbfile,'.spec','.'+clip(i)+'.spec')
             eoutrfile = repstr(outrfile,'.spec','.'+clip(i)+'.spec')
             eplotfile = lpar.spworkingdir1d+'rescale_'+legalfilestr(target)+clip(i)+'.eps'
             if file_test(eoutfile) and lpar.overwrite eq 0 then continue
   
             if file_test(eoutbfile) then file_delete, eoutbfile
             if file_test(eoutrfile) then file_delete, eoutrfile

             if ctb le ctr then ii = wb[i] else ii = wr[i]
             if ctb le ctr then wc = wr    else wc = wb
             ;print, removepath(vcfiles[ii]), '   ', utstart[ii], utstop[ii]
             ;for j = 0, n_Elements(wc)-1 do print, '   ', clip(removepath(vcfiles[wc[j]]),30), utstart[wc[j]], utstop[wc[j]]
             overlap = where(toverlap(hinfo[wc].mjd, hinfo[wc].mjdstop, hinfo[ii].mjd, hinfo[ii].mjdstop), ctover)
             if ctover eq 0 then begin
                logprint, '  Having difficulty matching multiple spectra of '+target
                continue
             endif 
             wcoverlap = wc[overlap]
             wbest = wcoverlap[ (where(exptimes[wcoverlap] eq max(exptimes[wcoverlap]))) [0] ]

             if (ctb le ctr) then $
                lrisconnect, vcfiles[ii], vcfiles[wbest], out=eoutfile, plot=eplotfile, displayoff=lpar.displayoff $
                             else $
                lrisconnect, vcfiles[wbest], vcfiles[ii], out=eoutfile, plot=eplotfile, displayoff=lpar.displayoff
             nconnected += 1
          endfor
          continue
        endelse
      endelse
   endfor  ; target loop

   if nconnected gt 0 then begin
;     print, ' Updating lrisspectra.ps:'
;     lrisplotall
   endif

end


; -------------------------
pro autolrisastrometry, camera=camera, chip=chip, nocrclean=nocrclean, nodistort=nodistort

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail
   
    achip = strmid(chip,0,1) ; chipchar
    prefchar = strmid(camera,0,1) ; camchar
    
    if keyword_set(nocrclean) eq 0 then begin
       ;zffiles = choosefiles(prefchar+lpar.wildchar+achip+'.fits',lpar.imworkingdir+'zifp',lpar.imworkingdir+'zfp')
       zffiles = lrisselectfiles('i',['zifp','zfp']+prefchar,achip,'.fits', count=ct)
       if (lpar.lrisversion eq 1 or lpar.lrisversion eq 4) and camera eq 'red' then $
         allzffiles = choosefiles('?'+lpar.wildchar+'.fits',lpar.imworkingdir+'zifp',lpar.imworkingdir+'zfp') $
       else $
         allzffiles = choosefiles('?'+lpar.wildchar+'?'+'.fits',lpar.imworkingdir+'zifp',lpar.imworkingdir+'zfp')
    endif else begin
       ;zffiles = choosefiles(prefchar+lpar.wildchar+achip+'.fits',lpar.imworkingdir+'ifp',lpar.imworkingdir+'fp')
       zffiles = lrisselectfiles('i',['ifp','fp']+prefchar,achip,'.fits', count=ct)
       allzffiles = choosefiles('?'+lpar.wildchar+'?'+'.fits',lpar.imworkingdir+'ifp',lpar.imworkingdir+'fp')
    endelse

    ; "all" files include the other camera and chips, and is for the initial catalog lookup.

    if n_elements(allzffiles) eq 1 and removepath(allzffiles[0]) eq '' then return

    allhinfo = lrisheaderinfo(allzffiles)
    allfiletargets = strarr(n_elements(allhinfo))
    for f = 0, n_elements(allzffiles)-1 do begin
       if allzffiles[f] eq '' then continue
       h = headfits(allzffiles[f], /silent)
       allfiletargets[f] = legalfilestr(strtrim(allhinfo[f].target,2),space='_')
    endfor

    alltargets = unique(allfiletargets)
    catalogs = ['sdss','ps1','usno','2mass'] ; note: PS1 has been very slow lately
   
    ; Get catalog reference data for every field (avoids having to do repeated lookups later)
    ;   and reduces risk that SDSS will start rejecting the catalog queries)
    ;   Note that this is chip/camera-independent    

    for t = 0, n_elements(alltargets)-1 do begin
      if lpar.targetseq[0] ne '' then if total(alltargets[t] eq lpar.targetseq) eq 0 then continue ; user filtering
      wthistarget = where(allfiletargets eq alltargets[t], ntargfiles)
      if ntargfiles eq 0 then continue ; shouldn't happen
      for c = 0, n_elements(catalogs)-1 do begin
         cat = catalogs[c]
         ; saves the number of images in the filename in case more appear later
         targname = alltargets[t]  ; name already repaired
         outcatfile = lpar.imworkingdir+targname+'.'+string(ntargfiles,'(I03)')+'.'+cat+'.ccat'
         if file_test(outcatfile) eq 0 then begin
            medra = median([allhinfo[wthistarget].radeg], /even)
            meddec = median([allhinfo[wthistarget].decdeg], /even)
            if meddec lt -25 and cat eq 'sdss' then continue ; not actually sure where the cut is
            if meddec lt -32 and cat eq 'ps1' then continue
            ; allow for a pointing error of up to 2'.  (another ~1-2' also OK but only partial overlap)
            logprint, '  Looking up '+strupcase(cat)+' catalog file for '+alltargets[t]+$
                   ' ('+clip(ntargfiles)+' files)' ;, format='($,A)'
            ; print, allzffiles[wthistarget] ; lpar.debug
            ; continue                       ; lpar.debug
            sepra = (max(allhinfo[wthistarget].radeg)-min(allhinfo[wthistarget].radeg))*cos(meddec*!pi/180.)
            sepdec = max(allhinfo[wthistarget].decdeg)-min(allhinfo[wthistarget].decdeg)
            if sepra gt 60*20 or sepdec gt 60*20 then begin
                 logprint, '   WARNING - Files of target '+alltargets[t]+' are quite widely spaced'
            endif
            if sepra gt 60*60 or sepdec gt 60*60 then begin
                 logprint, '   ERROR   - Files of target '+alltargets[t]+' are more than a degree apart!'
                 continue
            endif
            tstart = systime(1)
            printcatalog, allzffiles[wthistarget], cat, edgebuffer=-889, /unc, out=outcatfile
            tstop = systime(1)
            dt = tstop - tstart
            if dt gt 10. then begin
               logprint, '  Sorry, catalog lookup for '+cat+' is very slow right now (took '+clip(fpr(dt,5.1))+'sec)'
            endif
            ; Sometimes the catalog is on the edge of an SDSS field - trash it if nothing within 1' of the object.
            if file_test(outcatfile) and cat eq 'sdss' then begin
               readcol, outcatfile, ccra, ccdec, /silent, format='d,d'
               cdist  = 3600 * sqrt( ((ccra-medra)*cos(meddec*!pi/180.))^2 + (ccdec-meddec)^2 )
               mindist = min(cdist)
               if mindist gt 60 then begin
                  logprint, '  SDSS appears to only partially cover this field; not saving the file.'
                  file_delete, outcatfile
               endif
            endif
            if file_test(outcatfile) then begin
               ; delete any earlier versions with fewer images included
               others = file_search(lpar.imworkingdir+targname+'.*.'+cat+'.ccat')
               for j = 0, n_elements(others)-1 do if others[j] ne outcatfile then file_delete, others[j]

               ; also write out a region file (doesn't need the number)
               readcol, outcatfile, ccra, ccdec, /silent, format='d,d'
               outcatregfile = lpar.imworkingdir+targname+'.'+cat+'.ccat.reg'
               ;print, '  Converting '+clip(outcatfile)+' to region file '+clip(outcatregfile)
               openw, 1, outcatregfile
               printf, 1, 'global color=green width=1'
               printf, 1, 'fk5'
               for i = 0, n_elements(ccra)-1 do begin
                  printf, 1, 'point('+clip(ccra[i])+','+clip(ccdec[i])+') # point=boxcircle'
               endfor
               close, 1
            endif else begin
               if cat eq 'sdss' then begin
                   openw, 1, outcatfile
                   printf, 1, '# SDSS query unsuccessful - probably not an SDSS field'
                   close, 1
                   ;print, ' (unsuccessful)'
               endif else begin
                   logprint, '  WARNING - unsuccessful - problem with internet connection?' ;, format='($,A)'
               endelse
            endelse
         endif
      endfor
    endfor



    ; Solve every image (of this chip and camera) directly against SDSS/USNO

    hinfo = lrisheaderinfo(zffiles)

    if lpar.overwrite eq 0 then zffiles = unmatched(zffiles,'a')  ; catalogs should always be the same; only do this now

    for f = 0, n_elements(zffiles)-1 do begin
       if zffiles[f] eq '' then continue
       outfile = fileappend(zffiles[f],'a')
       if file_test(outfile) and lpar.overwrite eq 0 then continue
       h = headfits(zffiles[f], /silent)
       pa = strtrim(string((float(sxpar(h, 'ROTPOSN'))+0) MOD 360),2)
       exptime = sxpar(h,'ELAPTIME')
       counts = sxpar(h,'COUNTS')
       if counts gt 30000 and exptime le 10 or counts gt 20000 and exptime le 5 or counts gt 15000 and exptime le 4 then begin
          logprint, ' '+zffiles[f], ' is in bright twilight; not solving astrometry.'
          continue
       endif

       targname = legalfilestr(strtrim(sxpar(h,'TARGNAME'),2),space='_')
       if targname eq '' then continue
       if strpos(targname,'flat') ge 0 then continue

       ; Find the "best" catalog
       refcatfile = ''
       backuprefcatfile = ''
       for c = 0, n_elements(catalogs)-1 do begin
          ; in case there are multiple versions, find the newest
          searchstr = lpar.imworkingdir+targname+'.*.'+catalogs[c]+'.ccat'
          refcatfiles = file_search(searchstr)
          tryrefcatfile = refcatfiles[n_elements(refcatfiles)-1]
          if tryrefcatfile eq '' then logprint, "  ERROR - can't find the catalog file "+searchstr+"...?"
          nline = countlines(tryrefcatfile)
          ;print, '  ', tryrefcatfile, nline
          if nline gt 4+1 then begin
             if refcatfile ne '' and backuprefcatfile eq '' then backuprefcatfile = tryrefcatfile
             if refcatfile eq '' then refcatfile = tryrefcatfile
          endif
       endfor

       sat = sxpar(h,'SATURATE')
       if sat eq 0 then sat = 60000
       for i = 0, 2 do begin
          if i eq 0 and strpos(refcatfile,'sdss') eq -1 then continue  ; skip the run without saturated stars on USNO/2MASS
          if i eq 2 then refcatfile = backuprefcatfile
          if refcatfile eq '' then catstr = '' else catstr = ' -c '+refcatfile

          if i eq 0 then $
             acommand =  lpar.autoastrocommand+' '+zffiles[f] + catstr+' -upa 1.0' + ' -x '+clip(sat)+' -q'
          if i eq 1 then $
             acommand =  lpar.autoastrocommand+' '+zffiles[f] + catstr+' -upa 1.0' + ' -q'
          if i eq 2 then $
             acommand =  lpar.autoastrocommand+' '+zffiles[f] + catstr+' -upa 1.0' + ' -q'
          ;logprint, ' '+acommand
          logprint, ' Solving '+removepath(zffiles[f])+ ((refcatfile eq '') ? '' : (' against catalog '+removepath(refcatfile)))
          spawn, acommand
          if file_test(outfile) then begin
              sip2pv, outfile, /modify  ; could skip this if not SIP
              break  ; don't do the run with saturated stars if the first one succeeds
          endif
       endfor
       if file_test(outfile) eq 0 then begin
          fullastrofail = [fullastrofail, zffiles[f]]
          continue
       endif
 
    endfor

    ; future improvement: should go through fullastrofail list and align against neighboring images.






    ; Align images against each other, including distortion terms up to 4th order.
    ; 4th does not actually appear to be quite high enough, but the SIP conversion code only works to 4th.
    ; This is essentially a separate step, but still within lrisautoastrometry for now.

    ; For non-mosaic fields (the only capability at present), start with a central image 
    ; and work outward.


    ;afiles = choosefiles(prefchar+lpar.wildchar+achip+'.fits',lpar.imworkingdir+'azifp',lpar.imworkingdir+'azfp') 
    ;afiles = lrisselectfiles('i',['azifp','azfp']+camchar,achip,'.fits', count=ct)


    if keyword_set(nocrclean) eq 0 then begin
       ;afiles = choosefiles(prefchar+lpar.wildchar+achip+'.fits',lpar.imworkingdir+'azifp',lpar.imworkingdir+'azfp')
       afiles = lrisselectfiles('i',['azifp','azfp']+prefchar,achip,'.fits', count=ct)
    endif else begin
       ;afiles = choosefiles(prefchar+lpar.wildchar+achip+'.fits',lpar.imworkingdir+'azifp',lpar.imworkingdir+'azfp')
       afiles = lrisselectfiles('i',['aifp','afp']+prefchar,achip,'.fits', count=ct)
    endelse
    if ct eq 0 then return
    if lpar.overwrite eq 0 then uafiles = unmatched(afiles,'e') else uafiles = afiles
    if n_elements(uafiles) eq 1 and uafiles[0] eq '' then return

    ahinfo = lrisheaderinfo(afiles, /crval)     ; all the headers get read in every time, I guess this is OK.


    targetstorealign = unique(ahinfo.target) ; for now
    distort = keyword_set(nodistort) eq 0
    
    for t = 0, n_elements(targetstorealign)-1 do begin
       if targetstorealign[t] eq '' then continue

       w = where(ahinfo.target eq targetstorealign[t], ct)
       if ct eq 0 then logprint, '????'
       tahinfo = ahinfo[w]
       tafiles = afiles[w]
       if lpar.overwrite eq 0 then uatfiles = unmatched(tafiles,'e') else uatfiles = tafiles
       if n_elements(uatfiles) eq 1 and uatfiles[0] eq '' then continue
       if max(tahinfo.exp) lt 60. then continue  ; don't bother to realign e.g. standards
       logprint
       logprint, ' Aligning images of '+targetstorealign[t]

       cosdec = cos(median([tahinfo[w].decdeg])*!pi/180.)    
       dr = (tahinfo[w].radeg - median([tahinfo[w].radeg],/even))*cosdec*3600.
       dd = tahinfo[w].decdeg - median([tahinfo[w].decdeg],/even)*3600.
       dc = sqrt(dr^2 + dd^2)
       score = dc ;+ sqrt(((ahinfo[w].exptime < 100) > 20.) - 100)

       firstimagei = (where(score eq min(score))) [0] ; index
       firstimagename = tafiles[ firstimagei ]
       alignstatus = intarr(ct)
       alignstatus[firstimagei] = 1

       hfirst = headfits(firstimagename)
       acatfile = sxpar(hfirst,'ASTR_CAT')
       logprint, '  Starting realignment with '+removepath(firstimagename)+' vs. catalog '+removepath(acatfile)
       if file_test(acatfile) eq 0 then acatfile = lpar.imworkingdir+acatfile
       acatnum  = sxpar(hfirst,'ASTR_NUM') ; if this is high enough could re-distort here too?
       outfile = fileappend(firstimagename,'e')
       if file_test(acatfile) then begin
          if lpar.debug eq 0 then catch, errorstatus else errorstatus = 0
          if errorstatus ne 0 then begin
             logprint, '  Error during re-alignment of first image:'
             logprint, '   '+!error_state.msg
             logprint, '  Cannot realign this target: '+targetstorealign[t]
             continue
             ; in principle we could align against the a* instead of the ea* below and it might still work.
          endif else begin
            realignimage, firstimagename, acatfile, mindist=0.5, redistort=0  ; consider turning on redistort for SDSS with 30+ matches
          endelse
          catch, /cancel
       endif
       if file_test(outfile) eq 0 then begin
          logprint, ' Not re-aligning the first image.'
          spawn, 'cp -p '+firstimagename+' '+outfile
       endif else begin
          if distort then sip2pv, outfile, /modify
       endelse

       for i = 1, ct-1 do begin ; 0 is already done
          aligned = where(alignstatus eq 1, naligned, complement=unaligned, ncomplement=nunaligned)
          if naligned eq 0 or nunaligned eq 0 then begin
            logprint, ' Everything is aligned?'
            continue
          endif
          ;print, 'unaligned: ', tafiles[unaligned]
          ;print, 'aligned: ', tafiles[aligned]
          mind = 60.    ; don't trust this process for large offsets
          bestd = 180.*3600.
          bestj = -1
          bestk = -1
          for j = 0, n_elements(unaligned)-1 do begin
             for k = 0, n_elements(aligned)-1 do begin
                d = sqrt( ((tahinfo[unaligned[j]].radeg -tahinfo[aligned[k]].radeg)*cosdec)^2 + $  ; meridian cross?
                           (tahinfo[unaligned[j]].decdeg-tahinfo[aligned[k]].decdeg)^2 ) * 3600.
                ;print, removepath(tafiles[unaligned[j]]), ' - ', removepath(tafiles[aligned[k]]), ' ', d
                if d lt mind then begin
                   bestj = j
                   bestk = k
                   mind = d
                endif
                if d lt bestd then bestd = d
             endfor
          endfor

          if bestj ge 0 then begin
             infile =  tafiles[unaligned[bestj]]
             reffile = fileappend(tafiles[aligned[bestk]],'e')
             outfile = fileappend(infile,'e')
             logprint, '  Aligning '+removepath(infile)+' against '+removepath(reffile)+' (sep = '+clip(bestd)+')'
             if lpar.debug eq 0 then catch, errorstatus else errorstatus = 0
             if errorstatus ne 0 then begin
                logprint, '  Error during re-alignment:'
                logprint, '   '+!error_state.msg
             endif else begin
                realignimage, infile, reffile, redistort=distort
             endelse
             catch, /cancel
             ;;spawn, 'cp -f '+fileappend(tafiles[unaligned[bestj]],'e')+' '+tafiles[unaligned[bestj]]
             if file_test(outfile) and distort then sip2pv, outfile, /modify
             alignstatus[unaligned[bestj]] = 1
          endif else begin
             logprint, '  Remaining images of this field are too far apart for relative alignment (minsep = '+clip(bestd)+')'
          endelse
       endfor

    endfor



end



; -------------------------
pro autolrisphotometry, camera=camera, chip=chip, photometric=photometric

   common lrisauto, lpar

   if n_elements(chip) eq 0 then chipchar = '?' else chipchar = strmid(chip,0,1)
   camchar = strmid(camera,0,1)

   ;files = choosefiles(camchar+lpar.wildchar+chip+'.fits',  lpar.imworkingdir+'eazfp', lpar.imworkingdir+'azfp',lpar.imworkingdir+'eafp',lpar.imworkingdir+'afp')
   ;if n_elements(files) eq 1 and files[0] eq '' then return
   files = lrisselectfiles('i',['eazfp','azfp','eafp','afp']+camchar,chipchar,'.fits', count=ct)
   if ct eq 0 then return

   if camera eq 'blue' then istr = 'lrisb'
   if camera eq 'red'  then istr = 'lrisr'

   calibratenight, instrument='LRIS', istr=istr, files, workingdir=lpar.imworkingdir, photometric=photometric, archivezpt=lpar.refdatapath+'/zeropoints.dat'

end




; -------------------------
pro autolrisphotcat

   common lrisauto, lpar

   ; Last step of the photo pipeline combines red/blue left/right and needs to be run separately.

   ; merge all tcal files (red AND blue, left AND right) together!

   ; can't use lrisselectfiles here.
   tcalfiles = [file_search(lpar.imworkingdir+'ea*f*'+'?'+lpar.wildchar+'?'+'.tcal'), $
                file_search(lpar.imworkingdir+'ea*f*'+'?'+lpar.wildchar+'?'+'.tcaz')]
   w = where(tcalfiles ne '', ct)
   if ct eq 0 then return

   tcalfiles = tcalfiles[w]
   ;tcalinfo = readcatheaders(tcalfiles, count=ct) ; seems to be the only use of this routine, so deprecating it
   tcalinfo = replicate({target:''}, ct)
   for f = 0, ct-1 do begin
       cat = readcat(tcalfiles[f], h)
       tcalinfo[f].target = clip(sxpar(h,'TARGNAME'))
   endfor
   
   printed = 0
   targets = unique(tcalinfo.target)
   for t = 0, n_elements(targets)-1 do begin
       if lpar.targetseq[0] ne '' then if total(targets[t] eq lpar.targetseq) eq 0 then continue
       wt = where(tcalinfo.target eq targets[t])
       photcaldir = lpar.imfinaldir+'photcal/'
       checkdir = file_test(photcaldir)
       if checkdir eq 0 then spawn, 'mkdir '+photcaldir
       checkdir = file_test(photcaldir)
       if checkdir eq 0 then logprint, ' ERROR: cannot create photometric calibrations directory?'
       targetoutcat = photcaldir+'stars'+repstr(targets[t],' ','_')+'.cat'
       if file_test(targetoutcat) and keyword_set(overwrite) eq 0 then continue  ; could check contents as well.
       if printed eq 0 then logprint, ' Combining photometric catalogs...'
       printed = 1
       logprint, '     ', targets[t]
       mergecat, tcalfiles[wt], out=targetoutcat  ; would also be nice to write out the origin of the calib, etc.
   endfor
 

end


; -------------------------
pro autolrisstack, camera=camera, chip=chip, live=live, final=final, nophotometry=nophotometry, datenames=datenames

   common lrisauto, lpar
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

  if file_test('default.swarp') eq 0 then $
     spawn, lpar.swarpcommand+' -d > default.swarp'

  if n_elements(chip) eq '' then chip = '*'
  chipchar = strmid(chip,0,1)
  prefchar = strmid(camera,0,1)

  latesttarget = ''
  if keyword_set(live) then begin   ; if "live" reduction, find what target was most recently read out
    rawfiles = file_search(lpar.datadir+prefchar+lpar.wildchar+chipchar+'.fits')
    if n_elements(rawfiles) gt 1 then begin
       rawfiles = rawfiles[sort(rawfiles)]                ; this is not a perfect solution for truly continuous running.
       latestfile = rawfiles[n_elements(rawfiles)-1]      ; the appearance of raw-data may outrun the data processing.
       h = headfits(latestfile)                          
       if sxpar(h,'SLITNAME') eq 'DIRECT' then $
         latesttarget = repstr(strtrim(sxpar(h,'TARGNAME'),2), ' ', '_')
    endif
  endif

  ; azffiles = file_search(lpar.imworkingdir+'ea*'+prefchar+lpar.wildchar+chipchar+'.fits')
  ; azffiles = choosefiles(prefchar+lpar.wildchar+chipchar+'.fits',lpar.imworkingdir+'eazfp',lpar.imworkingdir+'azfp')
  azffiles = lrisselectfiles('i',['eazfp','azfp']+prefchar,chipchar,'.fits', count=ct)  
  if ct eq 0 then return

  realfiles = where(azffiles ne '', ct)

  nfiles = ct
  if nfiles eq 0 then return ; can't stack if there are no astrometry files
  if nfiles ge 1 then azffiles = azffiles[realfiles] ; unnecessary now

  ; make a custom structure for this (but, in principle could update lrisheaderinfo)
  ; this predates lrisheaderinfo and should probably be replaced with calls to that
  info = replicate({filename:'', target:'', exp:0., filter:'', satval:0., skyval:0., skyrate:0., airmass:0., seepix:0., zpt:0., $
                    rzpt:0., trans:0., reltrans:0., utstart:'', utend:'', jdstart:0.d, jdend:0.d, mjdstart:0.d, mjdend:0.d, $
                    sunelev:0., moonelev:0., datestart:'', dateend:'', koaid:''}, nfiles)

  for i = 0, nfiles-1 do begin
     h = headfits(azffiles[i], /silent)
     info[i].filename = removepath(azffiles[i])
     info[i].target = repstr(strtrim(sxpar(h,'TARGNAME'),2), ' ', '_')
     info[i].exp    = round(sxpar(h,'EXPTIME')*10.)/10.  ; round to 0.1 sec
     info[i].filter = sxpar(h,'FILTER')
     info[i].satval = sxpar(h,'SATURATE')
     info[i].skyval = sxpar(h,'SKYCTS')
     info[i].skyrate= sxpar(h,'CTRATE')
     info[i].airmass= sxpar(h,'AIRMASS')
     info[i].seepix = sxpar(h,'SEEING')
     info[i].zpt    = sxpar(h,'ZEROPT')
     info[i].rzpt   = sxpar(h,'RZPT')
     info[i].trans  = sxpar(h,'TRANS')
     info[i].reltrans=sxpar(h,'RELTRANS')
     info[i].utstart =sxpar(h,'UTC') ; does not include the date
     info[i].utend   =sxpar(h,'UTC-END')
     info[i].jdstart =sxpar(h,'JD')
     info[i].jdend   =info[i].jdstart+info[i].exp/(24*3600.)
     info[i].mjdstart=double(sxpar(h,'MJD'))
     info[i].mjdend  =info[i].mjdstart+info[i].exp/(24*3600.)
     info[i].datestart=sxpar(h,'UTSTART')
     info[i].dateend  =sxpar(h,'UTEND')
     info[i].sunelev =sxpar(h,'SUNELEV')
     info[i].moonelev=sxpar(h,'MOONELEV')
     info[i].koaid   =sxpar(h,'KOAID', count=checkkoaid)
   endfor  

   if keyword_set(nophotometry) then info.zpt = 0
   if keyword_set(nophotometry) then info.rzpt = 0
   if keyword_set(nophotometry) then info.trans = 1
   if keyword_set(nophotometry) then info.reltrans = 1
 
   bkey = 'GRISNAME,'
   rkey = 'GRANAME,GRANGLE,'
   if prefchar eq 'r' then skey = rkey 
   if prefchar eq 'b' then skey = bkey 
   keepkeystr = 'OBJECT,TARGNAME,TELESCOP,INSTRUME,CAMERA,DETECTOR,CURRINST,FILTER,DICHNAME,SLITNAME,'+$
                skey+'UT,UTC,JD,MJD_OBS,MJD-OBS,ST,'+$
                'DATE,DATE-OBS,AIRMASS,AZ,RA,DEC,EL,HA,TELFOCUS,ROTPOSN,DOMEPOSN,CURRINST,PARANGLE,OFFPA,'+$
                'PROGID,PROGINST,PROGTL1,PROGTL2,PROGTL3,SEMESTER,PROGPI,OBSERVER,OA,'+$
                'REDUCER,LPIPEVER,HOSTNAME,USERNAME,IDLVER,'+$
                'DOMEPOSN,ADCWAVE0,ADCWAVE1,AMPMODE,CCDGAIN,CCDSPEED,NUMAMPS,'+$
                'DATE_BEG,BINNING,TRAPDOOR,LAMPS,FLIMAGIN,FLSPECTR,'+$
                'FLATFILE,FLATTYPE,NSUPFLT,CALFILT,MOONDIST,SUNELEV,MOONELEV,GUIDFWHM,'+$
                'WXDOMHUM,WXDOMTMP,WXDWPT,WXOUTHUM,WXOUTTMP,WXPRESS'
   keepkeys = strsplit(keepkeystr, ',', /extract)
   ; some of these are overwritten by med, min, max, etc. values, but they are here also to preserve order.


   if file_test('extendedsources.txt') eq 0 then begin
      extendedsources = [''] 
   endif else begin
      readcol, 'extendedsources.txt', extendedsources, format='a', /silent
      notblank = where(extendedsources ne '', ct)
      if ct gt 0 then extendedsources = extendedsources[notblank]
   endelse


   targets = unique(info.target)
   w = where(targets ne latesttarget, ct)  ; if we are still integrating on this target, don't stack it yet
   if ct eq 0 then return
   targets = targets[w]

   for t = 0, n_elements(targets)-1 do begin
     target = targets[t]
     thistarget = where(info.target eq target, cttarg)
     if cttarg eq 0 then continue
     thistargetfilts = unique(info[thistarget].filter)
     for l = 0, n_elements(thistargetfilts)-1 do begin
        filter = thistargetfilts[l]
        thistargfilt = where(info.target eq target and info.filter eq filter, cttargfilt)
        if cttargfilt eq 0 then continue

        outfile       = lpar.imworkingdir + 'coadd' + strtrim(target,2) +'_'+ strtrim(filter,2) + '.' + chipchar + '.fits'
        outweightfile = lpar.imworkingdir + 'coadd' + strtrim(target,2) +'_'+ strtrim(filter,2) + '.' + chipchar + '.weight.fits'
        if file_test(outfile) and lpar.overwrite eq 0 then continue

        if max(info[thistargfilt].exp) lt 59 then continue
        ;expthresh = (median([info[thistargfilt].exp], /even) * 0.5) > (max(info[thistargfilt].exp) * 0.19) ; formerly median*0.8
        ;ctratethresh = 2 * median([info[thistargfilt].skyrate], /even)
        ;ww = where(info[thistargfilt].reltrans gt 0.2)
        ;;reltransthresh = 0.5*median([info[thistargfilt[ww]].reltrans], /even)
        ;expxtransthresh =  0.3*max([info[thistargfilt[ww]].reltrans * sqrt(info[thistargfilt[ww]].exp)]) > $
        ;                   0.7*median([info[thistargfilt[ww]].reltrans * sqrt(info[thistargfilt[ww]].exp)], /even)
        ; should do a full sort by contribution and determine what actually helps/hurts the S/N...
        ;print, info[thistargfilt[ww]].reltrans * sqrt(info[thistargfilt[ww]].exp)
        ;print, expxtransthresh

        tinfo = info[thistargfilt]
        relvar = tinfo.skyrate / (tinfo.exp * tinfo.reltrans^2)  ; scaled variance of the sky in this image
        badphot = where(finite(relvar) eq 0, ct)
        if ct gt 0 then logprint, ' WARNING: ',clip(ct),' files do not have usable photometric header information.'
        if ct gt 0 then relvar[badphot] = tinfo[badphot].skyrate / tinfo[badphot].exp   ; images for which photometry failed
        
        ord = sort(relvar)                                    ; lowest to highest variance (best to worst)
        prevcumnoise = !values.f_infinity
        for j = 0, cttargfilt-1 do begin                         ; could also incorporate seeing into this...
           cumnoise = sqrt( total( relvar[ord[0:j]]) ) / (j+1)
           ;;print, tinfo[ord[j]].filename, relvar[ord[j]], cumnoise
           if cumnoise gt prevcumnoise then break
           prevcumnoise = cumnoise
        endfor
        nstack = j
        stacki = thistargfilt[ord[0:j-1]]
        stacki = stacki[sort(stacki)]
        nexclude = cttargfilt-nstack
        if nexclude gt 0 then excludei = thistargfilt[ord[j:cttargfilt-1]]

        ;stacki = where(info.target eq target and info.filter eq filter and $
        ;               info.exp gt expthresh and info.skyrate le ctratethresh and $
        ;               info.reltrans*sqrt(info.exp) ge expxtransthresh, nstack)
        ;excludei = where((info.target eq target and info.filter eq filter) and $
        ;               (info.exp le expthresh or info.skyrate gt ctratethresh or $
        ;               info.reltrans*sqrt(info.exp) lt expxtransthresh), nexclude)
        ;if nstack lt 2 then continue
        if nstack eq 0 then continue
        ;;if nstack eq 1 then print, 'Only one deep exposure:  trivial stack'
        stacklist = azffiles[stacki]
        if nexclude gt 0 then excludelist = azffiles[excludei]
        totalexp = total(info[stacki].exp)
        medianexp = median([info[stacki].exp], /even)
        medair = median([info[stacki].airmass], /even)
        minair = min(info[stacki].airmass)
        maxair = max(info[stacki].airmass)
        minseeingpix = min(info[stacki].seepix)
        maxseeingpix = max(info[stacki].seepix)
        medseeingpix = median([info[stacki].seepix], /even)
        mintrans = min(info[stacki].trans)
        maxtrans = max(info[stacki].trans)
        medtrans = median([info[stacki].trans], /even)
        minreltrans = min(info[stacki].reltrans)
        maxreltrans = max(info[stacki].reltrans)
        medreltrans = median([info[stacki].reltrans], /even)        
        if mintrans gt 0 then transratio = maxreltrans/minreltrans else transratio = !values.f_nan

        if nstack gt 1 then stdtrans = stdev(info[stacki].reltrans) else stdtrans = 0

        stackcmd = lpar.swarpcommand+' '    ;'swarp '
        for s = 0, nstack-1 do begin
           if s eq 0 then stackcmd = stackcmd + stacklist[s] 
           if s gt 0 then stackcmd = stackcmd + ',' + stacklist[s] 
        endfor
        for s = 0, nstack-1 do begin  ; this looks like an opaque way to convert zfpXXX.weight.fits to azfpXXX.wei...
           weightfilename = strmid(stacklist[s],0,strlen(stacklist[s])-5-0)  + '.weight.fits'
           weightexists = file_test(weightfilename)
           if weightexists eq 0 then begin
              slashpos = strpos(weightfilename,'/',/reverse_search)
              weightfilenameinit = lpar.imworkingdir + strmid(weightfilename,slashpos+2)  
              weightexists = file_test(weightfilenameinit)
              if weightexists then begin
                  logprint, ' mv '+weightfilenameinit+' '+weightfilename
                  spawn, 'mv '+weightfilenameinit+' '+weightfilename
              endif
           endif
        endfor

        stackcmd = stackcmd + ' -IMAGEOUT_NAME ' + outfile + ' -WEIGHTOUT_NAME ' + outweightfile
        stackcmd = stackcmd + ' -WEIGHT_TYPE NONE'  ; weighting is now turned off.
        ;if nstack gt 1 then begin ; used to be 3...
        ;  stackcmd = stackcmd + ' -WEIGHT_TYPE MAP_WEIGHT'
        ;endif else begin
        ;  stackcmd = stackcmd + ' -WEIGHT_TYPE BACKGROUND'
        ;endelse

        ;stackcmd = stackcmd + ' -VERBOSE_TYPE QUIET'

        dum = where(target eq extendedsources, extct)
        if extct gt 0 then stackcmd += ' -BACK_SIZE 2048'

        if nstack gt 1 then begin
          if transratio le 1.07 or finite(transratio) eq 0 then fluxscales = medianexp/info[stacki].exp $
                                                           else fluxscales = medianexp/(info[stacki].exp*info[stacki].reltrans)
        endif else begin
          fluxscales = 1
        endelse

        if nstack gt 1 then begin
          stackcmd = stackcmd + ' -FSCALE_DEFAULT '
          for s = 0, nstack-1 do stackcmd = stackcmd + clip(fluxscales[s]) + ','   ; multiplicative factor...
        endif
 
        indmodzeropts = info[stacki].zpt + 2.5*alog10(fluxscales*(info[stacki].exp/medianexp))  ; divide the exptime ratios back out since zeropt includes this...
        outzeropt = median([indmodzeropts], /even)

          ; Too many keywords causes pyfits to be unable to load the mosaic.  It's rather freakish.
          ; Also, these all have the same irritating conversion to scientific notation with absurd sigdigits.
          ; make this a string and add to the swarpscikey iter below.
        ;stackcmd = stackcmd + ' -BACK_SIZE 256'
        stackcmd = stackcmd + ' -VERBOSE_TYPE QUIET'

        if (file_test(outfile) eq 0) or lpar.overwrite then begin
           logprint, ' Stacking ', target, ', ', clip(filter),'-band:'
           if nstack eq 1 then $ ; used to be 3 for some bizarre reason?
             logprint, ' Warning - only ', clip(nstack), ' exposure; not flagging bad pixels.'
           for s = 0, nstack-1 do logprint, '  ', removepath(stacklist[s]), ' x ', clip(fluxscales[s],5), $
                                         ' (exp =', fpr(info[stacki[s]].exp,4.1) + 's', $
                                         ', rtrans = ', clip(info[stacki[s]].reltrans,4), $
                                         ', zpt = ', fpr(info[stacki[s]].zpt,2.3), $
                                         ', skyr = ', clip(info[stacki[s]].skyrate,4) + 'ct/s)'
           logprint, ' --> '+removepath(outfile)
           if nexclude gt 0 then logprint, '   Excluding:'
           for x = 0, nexclude-1 do logprint, '  ', removepath(excludelist[x]), '   ', clip('',5), $
                                         ' (exp =', fpr(info[excludei[x]].exp,4.1) + 's', $
                                         ', rtrans = ', clip(info[excludei[x]].reltrans,4), $
                                         ', zpt = ', fpr(info[excludei[x]].zpt,2.3), $
                                         ', skyr = ', clip(info[excludei[x]].skyrate,4) + 'ct/s)'
           ;logprint, ' '+stackcmd
           spawn, stackcmd  ; do the stack
           data = mrdfits(outfile, 0, h, /silent)
           weightdata = readfits(outweightfile, /silent)
           wnodata = where(weightdata eq 0, ct)
           if ct gt 0 then data[wnodata] = !values.f_nan  ; dangerous to keep these at zero like swarp prefers
           swarpscikeys = ['CRPIX1','CRPIX2','EXPTIME','GAIN']
           h0 = headfits(stacklist[0])
           sxrenamepar, h, 'DATE', 'DATESWAR'
           for k = 0, n_elements(swarpscikeys)-1 do sxaddpar, h, swarpscikeys[k], sxpar(h, swarpscikeys[k], comment=cmt), cmt 
           for k = 0, n_elements(keepkeys)-1 do     sxaddpar, h, keepkeys[k],     sxpar(h0, keepkeys[k],     comment=cmt), cmt 

           ;dates = unique(info[stacki].date, ctd)
           ;if ctd gt 0 then for d = 0, ctd-1 do sxaddpar, h, 'DATE'+clip(d), dates[d]  ; in case imaged on multiple nights
           sxaddpar, h, 'DATE_BEG', info[stacki[0]].datestart, ' UTC datetime at start'
           sxaddpar, h, 'DATE_END',  info[stacki[nstack-1]].dateend, ' UTC datetime at end'
           sxaddpar, header, 'UTSTART', info[stacki[0]].datestart, ' UTC datetime at start'
           ;sxaddpar, header, 'UTMID', utmidstr, ' UTC datetime at midpoint'
           sxaddpar, header, 'UTEND', info[stacki[nstack-1]].dateend, ' UTC datetime at end'
           sxaddpar, h, 'UTC', info[stacki[0]].utstart, ' UTC time at start'
           sxaddpar, h, 'UTC-END',  info[stacki[nstack-1]].utend, ' UTC time at end'
           sxaddpar, h, 'MJDSTART', info[stacki[0]].mjdstart, ' Mean julian date at start'
           sxaddpar, h, 'MJDMID', info[stacki[0]].mjdstart+0.5*(info[stacki[nstack-1]].mjdend-info[stacki[0]].mjdstart), ' Mean julian date at midpoint'
           sxaddpar, h, 'MJDSTOP',  info[stacki[nstack-1]].mjdend, ' Mean julian date at end'
           sxaddpar, h, 'JDSTART', info[stacki[0]].jdstart, ' Julian date at start'
           sxaddpar, h, 'JDMID', info[stacki[0]].jdstart+0.5*(info[stacki[nstack-1]].jdend-info[stacki[nstack-1]].jdstart), ' Julian date at midpoint'
           sxaddpar, h, 'JDSTOP',  info[stacki[nstack-1]].jdend,' Julian date at start'
           ;sxaddpar, h, 'TELAPSED', 
           sxaddpar, h, 'NSTACK', nstack, ' # of exposures'
           sxaddpar, h, 'AIRMASS', medair, ' Median exposure airmass', after='NSTACK'
           sxaddpar, h, 'AIRMIN', minair, ' Minimum exposure airmass'
           sxaddpar, h, 'AIRMAX', maxair, ' Maximum exposure airmass'
           sxaddpar, h, 'EXPOSURE', medianexp, ' Effective rescaled exposure time (sec)'
           sxaddpar, h, 'SWARPEXP', sxpar(h, 'EXPTIME'), ' Maximum equivalent exposure time (sec)'
           sxaddpar, h, 'EXPTIME', medianexp, ' Effective rescaled exposure time (sec)', after='EXPOSURE' ; this is relevant for ZEROPT - lpar.overwrites SWARP value
           sxaddpar, h, 'TOTALEXP', totalexp, ' Total summed integration time (sec)'
           sxaddpar, h, 'MAXEXP', max(info[stacki].exp), ' Length of longest exposure (sec)'
           sxaddpar, h, 'MINEXP', min(info[stacki].exp), ' Length of shortest exposure (sec)'
           sxaddpar, h, 'SATURATE', min(info[stacki].satval-info[stacki].skyval)
           sxaddpar, h, 'MEDSKY', median([info[stacki].skyval], /even)
           sxaddpar, h, 'SUNELEV', max(info[stacki].sunelev), ' Highest sun elevation (deg)'
           sxaddpar, h, 'MOONELEV', max(info[stacki].moonelev), ' Highest moon elevation (deg)'
           sxaddpar, h, 'ZEROPT', outzeropt, ' Effective image zeropoint of stack'
           sxaddpar, h, 'RZPT', median([info[stacki].rzpt], /even), ' Aperture radius for zeropoint (pixels)'
           if minseeingpix gt 0 then begin
              sxaddpar, h, 'SEEING', medseeingpix, ' Median seeing in pixels'
              sxaddpar, h, 'SEEMIN', minseeingpix, ' Best exposure seeing in pixels'
              sxaddpar, h, 'SEEMAX', maxseeingpix, ' Worst exposure seeing in pixels'
              if camera eq 'red' and lpar.lrisversion eq 1 then medseeingarcsec = medseeingpix*0.210 $
                                                      else medseeingarcsec = medseeingpix*0.135
              sxaddpar, h, 'SEEARCSC', medseeingarcsec, ' Median seeing (arcsec)'
           endif
           if finite(transratio) then begin
              sxaddpar, h, 'TRANSRAT', maxreltrans/minreltrans, ' Transmission ratio of max/min images'
              sxaddpar, h, 'TRANSSTD', stdtrans, ' Std. dev. of relative transmission'
           endif
           for s = 0, nstack-1 do  sxaddpar, h, 'IMAGE'+clip(s), info[stacki[s]].filename
           for s = 0, nstack-1 do  sxaddpar, h, 'IMEXP'+clip(s), info[stacki[s]].exp
           for s = 0, nstack-1 do $
               if info[stacki[s]].zpt gt 0 then sxaddpar, h, 'ZEROPT'+clip(s), info[stacki[s]].zpt
           for s = 0, nstack-1 do $
               if info[stacki[s]].seepix gt 0 then sxaddpar, h, 'SEEING'+clip(s), info[stacki[s]].seepix
           for s = 0, nstack-1 do $
               if info[stacki[s]].trans gt 0 then sxaddpar, h, 'TRANS'+clip(s), info[stacki[s]].trans
           for s = 0, nstack-1 do $
               if info[stacki[s]].reltrans gt 0 then sxaddpar, h, 'RTRANS'+clip(s), info[stacki[s]].reltrans
           if nstack gt 1 then for s = 0, nstack-1 do sxaddpar, h, 'FSCALE'+clip(s), fluxscales[s]
           dum = sxpar(h0,'KOAID')
           for s = 0, nstack-1 do $
               if checkkoaid then sxaddpar, h, 'KOAID'+clip(s), info[stacki[s]].koaid ; this info 
           get_date, now
           sxdelpar, h, ['SOFTNAME','SOFTVERS','SOFTDATE','SOFTAUTH','SOFTINST','AUTHOR']
           sxaddpar, h, 'REDUCED',  repstr(tstamp(/ut),' ', 'T'), ' Timestamp of stack reduction step'
           hist = 'Processed by lpipe '+now
           sxaddhist, hist, h, /COMMENT
           mwrfits, data, outfile, h, /create ; add this header info

           ;;print, lpar.autoastrocommand+' '+outfile+' -q'
           ;;spawn, lpar.autoastrocommand+' '+outfile+' -q'
           ;;                 ; do another round of precise astrometry on the stack (probably unnecessary now)

           ; In the future, there will be a final alignment step.

        endif


     endfor
  endfor


     ; Combine left and right coadded images to produce final l+r images

  if keyword_set(final) then begin   ; only run this for whatever chip comes next (otherwise if r comes first it will assume l doesn't exist)

     ; Honestly this is a little dangerous as the photometric scaling may be slightly different especially on cloudy nights where the "reference"
     ; image is not guaranteed to be the same for the right and left chips.
     ; Would probably be best to actually open up the coadd files and get the list of images and weights and completely re-stack using that
     ; information.
     ; However, we are sticking with this for now because it is faster.  Just be careful on nonphotometric nights.

     coaddfilesr = file_search(lpar.imworkingdir+'coadd*r.fits')
     coaddfilesl = file_search(lpar.imworkingdir+'coadd*l.fits')
     if coaddfilesr ne [''] then coaddfilesr = coaddfilesr[where(coaddfilesr ne '')] ;else continue
     if coaddfilesl ne [''] then coaddfilesl = coaddfilesl[where(coaddfilesl ne '')] ;else continue
     ; filenames no longer indicate what camera; check to avoid duplication
     keep = bytarr(n_elements(coaddfilesr))
     for f = 0, n_elements(coaddfilesr)-1 do begin
        if coaddfilesr[f] eq '' then continue
        h = headfits(coaddfilesr[f])
        if strlowcase(strmid(sxpar(h,'CAMERA'),0,1)) ne prefchar then continue
        keep[f] = 1
     endfor
     if total(keep) eq 0 then coaddfilesr = '' else coaddfilesr = coaddfilesr[where(keep)]
     keep = bytarr(n_elements(coaddfilesl))
     for f = 0, n_elements(coaddfilesl)-1 do begin
        if coaddfilesl[f] eq '' then continue
        h = headfits(coaddfilesl[f])
        if strlowcase(strmid(sxpar(h,'CAMERA'),0,1)) ne prefchar then continue
        keep[f] = 1
     endfor
     if total(keep) eq 0 then coaddfilesl = '' else coaddfilesl = coaddfilesl[where(keep)]


     ; match every r with an l and coadd
     for f = 0, n_elements(coaddfilesr)-1 do begin
        rfilename = coaddfilesr[f]
        if rfilename eq '' then continue
        if lpar.targetseq[0] ne '' then begin
           h = headfits(rfilename)
           htarg = clip(sxpar(h,'TARGNAME'))
           passtarget = 0
           for t = 0, n_elements(lpar.targetseq)-1 do if htarg eq lpar.targetseq[t] then passtarget = 1
           if passtarget eq 0 then continue
        endif
        datestr = ''
        if keyword_set(datenames) then begin
           h = headfits(rfilename)
           datestr = strmid(repstr(sxpar(h,'DATE'),'-',''),0,8)+'_'
        endif
        lfilename = repstr(rfilename, 'r.fits','l.fits') ; assumes target name doesn't include these strings
        outfileidir = repstr(repstr(rfilename, '.r.fits', '.fits'),'coadd','lris'+prefchar+datestr)
        outfilefdir = lpar.imfinaldir+repstr(repstr(removepath(outfileidir),'_RG850','_z'),'_I','_i')
        if file_test(outfilefdir) eq 0 or lpar.overwrite then begin
           match = where(coaddfilesl eq lfilename, ct)
           ;if ct eq 0 then logprint, 'No match for ', rfilename, ' among ', coaddfilesl, ' (looking for ', lfilename, ')'
           if ct ge 1 then begin
              stackcmd = lpar.swarpcommand + ' ' + rfilename + ',' + lfilename
              rweightname = repstr(rfilename, '.fits','.weight.fits') 
              lweightname = repstr(lfilename, '.fits','.weight.fits') 
              ;if file_test(rweightname) eq 0 then spawn, 'mv ' + strmid(rweightname,1) + ' ' + rweightname
              ;if file_test(lweightname) eq 0 then spawn, 'mv ' + strmid(lweightname,1) + ' ' + lweightname
              logprint
              logprint, ' Combining ', rfilename, ' and ', lfilename
              stackcmd = stackcmd + ' -IMAGEOUT_NAME ' + outfileidir
              stackcmd = stackcmd + ' -WEIGHTOUT_NAME ' + 'weight.temp.fits'
              stackcmd = stackcmd + ' -WEIGHT_TYPE MAP_WEIGHT'
              stackcmd = stackcmd + ' -VERBOSE_TYPE QUIET'

              logprint, ' '+stackcmd
              spawn, stackcmd
              data = mrdfits(outfileidir, 0, h, /silent)
              weightdata = readfits('weight.temp.fits', /silent)
              file_delete, 'weight.temp.fits' ; don't need it anymore
              wnodata = where(weightdata eq 0, ct)
              if ct gt 0 then data[wnodata] = !values.f_nan
              swarpscikeys = ['EXPTIME','GAIN']  ; formerly also included 'CRPIX1','CRPIX2', which was insane
              hr = headfits(rfilename)
              hl = headfits(lfilename)
              for k = 0, n_elements(swarpscikeys)-1 do sxaddpar, h, swarpscikeys[k], sxpar(hr, swarpscikeys[k], comment=cmt), cmt 
              for k = 0, n_elements(keepkeys)-1 do     sxaddpar, h, keepkeys[k],     sxpar(hr, keepkeys[k],     comment=cmt), cmt 
              for i = 0, 100 do begin
                key = 'IMAGE'
                rimagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, 'RIMAGE'+clip(i), clip(rimagekeyn)
              endfor
              for i = 0, 100 do begin
                key = 'IMAGE'
                limagekeyn = sxpar(hl, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, 'LIMAGE'+clip(i), clip(limagekeyn)
              endfor
              for i = 0, 100 do begin
                key = 'IMEXP'
                imagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, key+clip(i), imagekeyn
              endfor
              for i = 0, 100 do begin
                key = 'SEEING'
                imagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, key+clip(i), imagekeyn
              endfor
              for i = 0, 100 do begin
                key = 'ZEROPT'
                imagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, key+clip(i), imagekeyn
              endfor
              for i = 0, 100 do begin ; the above four loops could be merged
                key = 'KOAID'
                imagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, key+clip(i), imagekeyn
              endfor
              sxdelpar, h, ['SOFTNAME','SOFTVERS','SOFTDATE','SOFTAUTH','SOFTINST','AUTHOR']
              sxaddpar, h, 'REDUCED',  repstr(tstamp(/ut),' ', 'T'), ' Timestamp of stack reduction step'
              get_date, now
              hist = 'Processed by lpipe '+now
              sxaddhist, hist, h, /COMMENT
              mwrfits, data, outfileidir, h, /create
              spawn, 'mv '+outfileidir+' '+outfilefdir
           endif else begin

            ; spawn, 'cp -f '+rfilename+' '+outfilefdir
              file_copy, rfilename, outfilefdir, /overwrite
              logprint, ' '+rfilename+' -> '+outfilefdir

           endelse
        endif
     endfor
  endif

end


pro lrisfailuresummary, dirs, reset=reset, altflatdir=altflatdir;, path=path

  common lrisauto, lpar
  common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail


  if keyword_set(reset) then begin
     prepfail = ['']
     flatfail = ['']
     catastrofail = ['']
     relastrofail = ['']
     fullastrofail = ['']
     extractfail = ['']
     wavsolfail = ['']
     wavsolwarn = ['']
     fluxcalfail = ['']
     return
  endif

  ; dealing with multiple directories here is a bit of a mess.
  ; would be better to include subdirectories in the failure logs.

  nprepfail = n_elements(prepfail)-1
  nredfail = 0
  if nprepfail ge 1 then begin ;element 0 is blank
    logprint
    logprint, 'Unable to prepare the following files:'
    for f = 1, n_elements(prepfail)-1 do begin 
      w = where(file_test(dirs+prepfail[f]), ct)
      if ct eq 0 then continue
      h = headfits(dirs[w[0]]+prepfail[f])
      slitstr = clip(sxpar(h,'SLITNAME',count=ctsl),11)
      targstr = clip(sxpar(h,'TARGNAME',count=ctta),16)
      if ctsl eq 0 then slitstr = clip('no SLITNAME',11)
      if ctta eq 0 then targstr = clip('[no TARGNAME]',16)
      trapstr = clip(sxpar(h,'TRAPDOOR'))
      logprint, ' '+clip(removepath(prepfail[f]),25)+' '+slitstr+' '+string(sxpar(h,'ELAPTIME'),format='(I5)')+'s'+ $
                ' '+targstr+' '+clip(sxpar(h,'OBJECT'),16)+' '+repstr(repstr(clip(sxpar(h,'BINNING')),',','x'),' ','')+' '+trapstr
      if clip(sxpar(h,'INSTRUME')) eq 'LRIS' then nredfail += 1
    endfor
  endif

  nflatfail = n_elements(flatfail)-1
  if nflatfail ge 1 then begin ;element 0 is blank
    logprint
    logprint, 'Unable to flat-field the following images:'
    for f = 1, n_elements(flatfail)-1 do begin
      w = where(file_test(dirs+flatfail[f]), ct)
      if ct eq 0 then continue
      h = headfits(dirs[w[0]]+flatfail[f])
      gg = sxpar(h,'GRNAME')
      if clip(gg) eq '0' then gg = sxpar(h,'SLITNAME')
      logprint, ' '+clip(removepath(flatfail[f]),30)+string(sxpar(h,'ELAPTIME'),format='(I5)')+'s '+clip(sxpar(h,'FILTER'),5)+$
                ' D'+clip(sxpar(h,'DICHNAME'),7)+' '+clip(gg,12)+' '+clip(sxpar(h,'TARGNAME'),16)+' '+repstr(clip(sxpar(h,'BINNING')),',','x')
     ;, clip(sxpar(h,'WINDOW')), ' ', clip(sxpar(h,'PANE'))
    endfor
    flats = file_search(lpar.flatdir+'*flat*.fits')
    flats = flats[sort(strmid(removepath(flats),2))]
      ; formerly needed to look at two directories
    for a = 0, n_elements(altflatdir)-1 do flats = [flats, file_search(altflatdir[a]+'*flat*.fits')]
    nflats = total(flats ne '')
    if nflats gt 0 then begin
      flats = flats[where(flats ne '')]
      logprint, 'Available flats are:'
      for f = 0, nflats-1 do begin
         h = headfits(flats[f], /silent)
         gg = sxpar(h,'GRNAME')
         if clip(gg) eq '0' then gg = sxpar(h,'SLITNAME')
         logprint, ' '+clip(removepath(flats[f]),30)+string(sxpar(h,'ELAPTIME'),format='(I5)')+'s '+clip(sxpar(h,'FILTER'),5)+$
            ' D'+clip(sxpar(h,'DICHNAME'),7)+' '+clip(gg,12)+' '+clip(sxpar(h,'TARGNAME'),16)+' '+repstr(clip(sxpar(h,'BINNING')),',','x')
     ;, clip(sxpar(h,'WINDOW')), ' ', clip(sxpar(h,'PANE'))
       endfor
     endif else begin
       if dirs[0] eq '' then $
       logprint, 'No processed flat-fields exist!  Check inputs or copy raw flat-field files to data directory.'
     endelse
  endif
  
  nefail = n_elements(extractfail)-1
  if nefail ge 1 then begin
    logprint, 'Unable to model trace and/or extract 1D spectrum for the following 2D spectra:'
    for f = 1, n_elements(extractfail)-1 do begin
      w = where(file_test(dirs+extractfail[f]), ct)
      if ct eq 0 then continue
      h = headfits(dirs[w[0]]+extractfail[f])
      logprint, ' '+clip(removepath(extractfail[f]),36)+rclip(sxpar(h,'ELAPTIME'),5)+'s '+clip(sxpar(h,'GRNAME'),9)+$
                ' '+clip(sxpar(h,'TARGNAME'),16)
    endfor
  endif
  nwfail = n_elements(wavsolfail)-1
  if nwfail ge 1 then begin
     logprint, 'Unable to wavelength-calibrate the following spectra:'
     for f = 0, n_elements(wavsolfail)-1 do begin
        w = where(file_test(dirs+wavsolfail[f]), ct)
        if ct eq 0 then continue
        ff = dirs[w[0]]+wavsolfail[f]
        h = ['']
        if strpos(wavsolfail[f],'.spec') gt 0 then h = headspec(ff)
        if strpos(wavsolfail[f],'.fits') gt 0 then h = headfits(ff)
        logprint, ' '+clip(removepath(wavsolfail[f]),36)+rclip(sxpar(h,'ELAPTIME'),5)+'s'+' '+clip(sxpar(h,'GRNAME'),9)+$
                  ' ', 'D'+clip(sxpar(h,'DICHNAME'),3)+' cw=',clip(sxpar(h,'CWAVE'),7)+' '+clip(sxpar(h,'TARGNAME'),16)
     endfor
     arcsolfiles = findfile('arcs/*.sol'); this is really crude, should use LRIS finder
     ;arcfitsfiles = lpar.arcdir+removepath(repstr(arcsolfiles,'.sol','.fits'))
     ;arcfitsfilesexist = bytarr(narcsol)
     ;arch = lrisheaderinfo(arcsolfiles, badarclist=badarclist) ; arc metadata
     logprint, 'Available solved arcs:'
     for f = 0, n_elements(arcsolfiles)-1 do begin
       if file_test(arcsolfiles[f]) eq 0 then continue
       h = headspec(arcsolfiles[f])
       logprint, ' '+clip(removepath(arcsolfiles[f]),36)+rclip(sxpar(h,'ELAPTIME'),5)+'s'+' '+clip(sxpar(h,'GRNAME'),9)+$
                  ' ', 'D'+clip(sxpar(h,'DICHNAME'),3)+' cw=',clip(sxpar(h,'CWAVE'),7)+' '+clip(sxpar(h,'TARGNAME'),16)
     endfor
  endif
  nwwarn = n_elements(wavsolwarn)-1
  if nwwarn ge 1 then begin
     logprint, 'Wavelength solutions for the following spectra are questionable; check sky-line verification plots:'
     for f = 1, n_elements(wavsolwarn)-1 do begin
        w = where(file_test(dirs+wavsolwarn[f]), ct)
        if ct eq 0 then continue
        h = headspec(dirs[w[0]]+wavsolwarn[f])
        logprint, ' '+clip(removepath(wavsolwarn[f]),36)+rclip(sxpar(h,'ELAPTIME'),5)+'s'+' '+clip(sxpar(h,'GRNAME'),9)+$
                  ' '+clip(sxpar(h,'TARGNAME'),16)
     endfor
  endif
  nfluxfail = n_elements(fluxcalfail)-1
  if nfluxfail ge 1 then begin
     logprint, 'Unable to flux-calibrate the following spectra:'
     for f = 1, n_elements(fluxcalfail)-1 do begin
        w = where(file_test(dirs+fluxcalfail[f]), ct)
        if ct eq 0 then continue
        h = headspec(dirs[w[0]]+fluxcalfail[f])
        logprint, ' '+clip(removepath(fluxcalfail[f]),36)+rclip(sxpar(h,'ELAPTIME'),5)+'s'+' '+clip(sxpar(h,'GRNAME'),9)+$
                  ' D'+clip(sxpar(h,'DICHNAME'),3)+' '+clip(sxpar(h,'FILTER'),5)+' '+clip(sxpar(h,'SLITNAME'),9)+$
                  ' '+clip(sxpar(h,'TARGNAME'),16)+' cw='+clip(fix(sxpar(h,'CWAV')))
     endfor
     respfiles = file_search(lpar.responsedir+'*.response', count=nf)
     if nf gt 0 then logprint, ' Available solved standards:'
     for f = 0, nf-1 do begin
        ;spfile = repstr(repstr(respfiles[f],'.response','.spec') ,lpar.responsedir,lpar.spworkingdir1d)
        ;if file_test(spfile) eq 0 then begin
        ;   print, ' '+spfile
        ;   continue
        ;endif
        spfile = respfiles[f]
        h = headspec(spfile)
        cwav = sxpar(h,'CWAV')
        gr = clip(sxpar(h,'GRNAME'))
        dich = clip(sxpar(h,'DICHNAME'))
        filt = clip(sxpar(h,'FILTER'))
        slit = clip(sxpar(h,'SLITNAME'))
        targ = clip(sxpar(h,'TARGNAME'))
        logprint, '   '+clip(removepath(spfile),34)+rclip(sxpar(h,'ELAPTIME'),5)+'s'+' '+clip(sxpar(h,'GRNAME'),9)+$
                  ' D'+clip(sxpar(h,'DICHNAME'),3)+' '+clip(sxpar(h,'FILTER'),5)+' '+clip(sxpar(h,'SLITNAME'),9)+$
                  ' '+clip(sxpar(h,'TARGNAME'),16)+' cw='+clip(fix(sxpar(h,'CWAV')))
        
        ;+' '+clip(gr,9)+', D'+clip(dich,3)+', '+clip(filt,5)+', '+clip(slit,9)+', cw='+clip(fix(cwav))
     endfor
  endif

  nafail = n_elements(relastrofail)-1 + n_elements(fullastrofail)-1 + n_elements(catastrofail)-1
  if nafail ge 1 then begin
    logprint
    if n_elements(catastrofail) gt 1 then logprint, 'Unable to produce astrometric catalogs for the following reference images:'
    for f = 1, n_elements(catastrofail)-1 do begin
      w = where(file_test(dirs+catastrofail[f]), ct)
      if ct eq 0 then continue
      h = headfits(dirs[w[0]]+catastrofail[f])
      logprint, ' '+clip(removepath(catastrofail[f]),25)+string(sxpar(h,'ELAPTIME'),format='(I5)')+' '+clip(sxpar(h,'FILTER'),4)+' '+ clip(sxpar(h,'TARGNAME'),16)+' '+clip(sxpar(h,'COUNTS'))
    endfor
    if n_elements(relastrofail) gt 1 then logprint, 'Relative astrometry failed for the following images, but absolute was successful:'
    for f = 1, n_elements(relastrofail)-1 do begin
      w = where(file_test(dirs+relastrofail[f]), ct)
      if ct eq 0 then continue
      h = headfits(dirs[w[0]]+relastrofail[f])
      logprint, ' '+clip(removepath(relastrofail[f]),25)+string(sxpar(h,'ELAPTIME'),format='(I5)')+' '+clip(sxpar(h,'FILTER'),4)+' '+ clip(sxpar(h,'TARGNAME'),16)+' '+clip(sxpar(h,'COUNTS'))
    endfor
    if n_elements(fullastrofail) gt 1 then logprint, 'All astrometry failed for the following images (not stacked):'
    for f = 1, n_elements(fullastrofail)-1 do begin
      w = where(file_test(dirs+fullastrofail[f]), ct)
      if ct eq 0 then continue
      h = headfits(dirs[w[0]]+fullastrofail[f])
      logprint, ' '+clip(removepath(fullastrofail[f]),25)+string(sxpar(h,'ELAPTIME'),format='(I5)')+' '+clip(sxpar(h,'FILTER'),4)+' '+ clip(sxpar(h,'TARGNAME'),16)+' '+clip(sxpar(h,'COUNTS'))
    endfor
  endif; else begin
  ;print, 'Finished with no astrometric problems.'
  ;endelse

  if nredfail gt 3 and lpar.lrisversion eq 4 then begin
      print
      print, '  Multiple red images could not be prepared.'
      print, '  This probably indicates bad header information.'
      print, '  You may need to correct the header information by editing'
      print, '  the data files to run LPipe on red-side data from this run.  
      print, '  See README.lrisred4.txt for more information.'
  endif

end

